/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 -2016 Samsung Electronics Co., Ltd All Rights Reserved.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <vconf.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"
#include "bt-hal-adapter-le.h"

#include <bt-hal-adapter-dbus-handler.h>
#include <bt-hal-dbus-common-utils.h>

#define BT_ENABLE_TIMEOUT 20000 /* 20 seconds */
#define BT_CORE_NAME "org.projectx.bt_core"
#define BT_CORE_PATH "/org/projectx/bt_core"
#define BT_CORE_INTERFACE "org.projectx.btcore"

static GDBusProxy *core_proxy = NULL;
static GDBusConnection *system_conn = NULL;
static handle_stack_msg event_cb = NULL;

handle_stack_msg _bt_get_adapter_event_cb(void)
{
	if (!event_cb)
		return event_cb;
	else
		return NULL;
}

GDBusConnection *__bt_hal_get_system_gconn(void)
{
	DBG("+");
	if (system_conn == NULL)
		system_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);

	DBG("-");
	return system_conn;
}

GDBusProxy *_bt_init_core_proxy(void)
{	GDBusProxy *proxy;
	GDBusConnection *conn;

	DBG("+");
	conn = __bt_hal_get_system_gconn();
	if (!conn)
		return NULL;

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
			NULL,
			BT_CORE_NAME,
			BT_CORE_PATH,
			BT_CORE_INTERFACE,
			NULL, NULL);

	if (!proxy)
		return NULL;

	core_proxy = proxy;

	DBG("-");
	return proxy;
}

static GDBusProxy *__bt_get_core_proxy(void)
{
	return (core_proxy) ? core_proxy : _bt_init_core_proxy();
}

/* To send stack event to hal-bluetooth handler */
void _bt_hal_dbus_store_stack_msg_cb(handle_stack_msg cb)
{
	event_cb = cb;
}

int _bt_hal_enable_core(void)
{
	DBG("");
	GDBusProxy *proxy;
	GVariant *result;
	GError *error = NULL;

	proxy = __bt_get_core_proxy();
	if(!proxy)
		return BT_HAL_ERROR_INTERNAL;

	/* Clean up the process */
	result = g_dbus_proxy_call_sync(proxy,
			       "EnableCore",
			       NULL,
			       G_DBUS_CALL_FLAGS_NONE,
			      -1,
			      NULL,
			       &error);

	if (!result) {
		if (error != NULL) {
			ERR("Bt core call failed(Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Bt core call failed");
		return BT_HAL_ERROR_INTERNAL;
	}

	g_variant_unref(result);
	return BT_HAL_ERROR_NONE;
}

static gboolean __bt_hal_send_adapter_event(gpointer user_data)
{
	struct hal_ev_adapter_state_changed ev;
	gboolean *enable = (gboolean *)user_data;

	if (*enable == TRUE)
		ev.state = HAL_POWER_ON;
	else
		ev.state = HAL_POWER_OFF;

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();

	if (event_cb)
		event_cb(HAL_EV_ADAPTER_STATE_CHANGED, &ev, sizeof(ev));

	g_free(enable);
	return false;
}

static gboolean __bt_hal_send_le_event(gpointer user_data)
{
	struct hal_ev_le_state_changed ev;
	gboolean *enable = (gboolean *)user_data;

	if (*enable == TRUE)
		ev.state = HAL_POWER_ON;
	else
		ev.state = HAL_POWER_OFF;

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();

	if (event_cb)
		event_cb(HAL_EV_LE_STATE_CHANGED, &ev, sizeof(ev));

	g_free(enable);
	return false;

}

int _bt_hal_dbus_enable_adapter(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result = NULL;
	unsigned char powered = 0;

	DBG("+");

	int le_value = VCONFKEY_BT_LE_STATUS_OFF;
	if (vconf_get_int(VCONFKEY_BT_LE_STATUS, &le_value))
		ERR("fail to get vconf key!");

	if (_bt_hal_get_adapter_powered_state(&powered) == BT_STATUS_FAIL)
		ERR("_bt_hal_get_adapter_powered_state failed");

	/*if LE is enabled and bluez is UP , send event from here and return.*/
	if ((le_value == VCONFKEY_BT_LE_STATUS_ON) && (powered == 1)) {
		gboolean *user_data = g_new0(int, 1);
		*user_data = TRUE;
		g_idle_add(__bt_hal_send_adapter_event, (gpointer)user_data);
		return BT_STATUS_SUCCESS;
	}
	proxy = __bt_get_core_proxy();

	if (!proxy) {
		DBG("_bt_hal_dbus_enable_adapter: Core proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	_bt_hal_set_adapter_request_state(TRUE);
	_bt_hal_set_le_request_state(TRUE);
	result = g_dbus_proxy_call_sync(proxy, "EnableAdapter",
			NULL,
			G_DBUS_CALL_FLAGS_NONE, BT_ENABLE_TIMEOUT,
			NULL, &error);
	if (error) {
		DBG("EnableAdapter failed: %s", error->message);
		g_clear_error(&error);
		error = NULL;
		_bt_hal_set_adapter_request_state(FALSE);
		_bt_hal_set_le_request_state(FALSE);
		result = g_dbus_proxy_call_sync(proxy,
				"DisableAdapter",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);

		if (error != NULL) {
			DBG("Bt core call failed(Error: %s)", error->message);
			g_clear_error(&error);
		}
		g_variant_unref(result);
		/*Terminate myself*/
		/* TODO: Terminate bluetooth service or not, need to check */
		/*g_idle_add((GSourceFunc)_bt_terminate_service, NULL);*/
		return BT_STATUS_FAIL;
	}

	DBG("-");
	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}

int _bt_hal_dbus_disable_adapter(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result = NULL;

	DBG("+");
	proxy = __bt_get_core_proxy();

	if (!proxy) {
		DBG("_bt_hal_dbus_enable_adapter: Core proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy, "DisableAdapter",
			NULL,
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	if (error) {
		DBG("DisableAdapter failed: %s", error->message);
		g_clear_error(&error);
		error = NULL;
		g_variant_unref(result);
		/*Terminate myself*/
		/* TODO: Terminate bluetooth service or not, need to check */
		/*g_idle_add((GSourceFunc)_bt_terminate_service, NULL);*/
		return BT_STATUS_FAIL;
	}

	DBG("-");
	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}

#ifdef TIZEN_BT_HAL
int _bt_hal_dbus_enable_le(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result = NULL;
	int adapter_value = VCONFKEY_BT_STATUS_OFF;
	unsigned char powered = 0;

	DBG("+");

	if (vconf_get_int(VCONFKEY_BT_STATUS, &adapter_value) != 0)
		ERR("fail to get vconf key!");

	if (_bt_hal_get_adapter_powered_state(&powered) == BT_STATUS_FAIL)
		ERR("_bt_hal_get_adapter_powered_state failed");

	if ((adapter_value == VCONFKEY_BT_STATUS_ON) && (powered == 1)) {
		gboolean *user_data = g_new0(int, 1);
		*user_data = TRUE;
		g_idle_add(__bt_hal_send_le_event, (gpointer)user_data);
		return BT_STATUS_SUCCESS;
	}

	proxy = __bt_get_core_proxy();

	if (!proxy) {
		DBG("_bt_hal_dbus_enable_le: Core proxy get failed!!!");
		return BT_STATUS_FAIL;
	}
	_bt_hal_set_le_request_state(TRUE);

	result = g_dbus_proxy_call_sync(proxy, "EnableAdapterLe",
			NULL,
			G_DBUS_CALL_FLAGS_NONE, BT_ENABLE_TIMEOUT,
			NULL, &error);
	if (error) {
		DBG("EnableAdapter failed: %s", error->message);
		g_clear_error(&error);
		error = NULL;
		result = g_dbus_proxy_call_sync(proxy,
				"DisableAdapter",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);

		_bt_hal_set_le_request_state(FALSE);
		if (error != NULL) {
			DBG("Bt core call failed(Error: %s)", error->message);
			g_clear_error(&error);
		}
		g_variant_unref(result);

		return BT_STATUS_FAIL;
	}

	DBG("-");
	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}

int _bt_hal_dbus_disable_le(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result = NULL;
	int adapter_value = VCONFKEY_BT_STATUS_OFF;

	DBG("+");
	if (vconf_get_int(VCONFKEY_BT_STATUS, &adapter_value) != 0)
		ERR("fail to get vconf key!");

	if (adapter_value == VCONFKEY_BT_STATUS_ON) {
		gboolean *user_data = g_new0(int, 1);
		*user_data = FALSE;
		g_idle_add(__bt_hal_send_le_event, (gpointer)user_data);
		return BT_STATUS_SUCCESS;
	}

	proxy = __bt_get_core_proxy();

	if (!proxy) {
		DBG("_bt_hal_dbus_disable_le: Core proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy, "DisableAdapterLe",
			NULL,
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	if (error) {
		DBG("DisableAdapter failed: %s", error->message);
		g_clear_error(&error);
		error = NULL;
		g_variant_unref(result);
		return BT_STATUS_FAIL;
	}



	DBG("-");
	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}
#endif

#ifdef TIZEN_BT_HAL
int _bt_hal_dbus_reset_adapter(void)
{
	GDBusProxy *proxy;
	GVariant *result;
	GError *error = NULL;
	INFO("+");

	proxy = __bt_get_core_proxy();
	if (!proxy)
		return BT_STATUS_FAIL;

	result = g_dbus_proxy_call_sync(proxy,
			"ResetAdapter",
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to ResetAdapter (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to ResetAdapter");
		return BT_STATUS_FAIL;
	}

	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}
#endif

int _bt_hal_dbus_start_discovery(gboolean is_custom,
					bt_disc_role_type_t role)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result;
	const gchar *disc_type;
	DBG("Discovery Type: Is Custom : [%s]", is_custom ? "TRUE" : "FALSE");

	proxy = _bt_hal_get_adapter_proxy();
	if (!proxy) {
		DBG("Adapter proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	if (is_custom) {
		if (role == BT_DISC_ROLE_BREDR)
			disc_type = "BREDR";
		else if (role == BT_DISC_ROLE_LE)
			disc_type = "LE";
		else if (role == BT_DISC_ROLE_DUAL)
			disc_type = "LE_BREDR";
		else
			return BT_STATUS_PARM_INVALID;

		result = g_dbus_proxy_call_sync(proxy,
				"StartCustomDiscovery",
				g_variant_new("s", disc_type),
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);
	} else
		result = g_dbus_proxy_call_sync(proxy,
				"StartDiscovery",
				NULL,
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);

	if (!result) {
		if (error != NULL) {
			ERR("StartDiscovery failed (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("StartDiscovery failed");
		return BT_STATUS_FAIL;
	}

	/* discovery status will be change in event */
	DBG("-");
	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}

int _bt_hal_dbus_stop_discovery(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result;
	DBG("+");

	proxy = _bt_hal_get_adapter_proxy();
	if (!proxy) {
		DBG("_bt_hal_dbus_stop_discovery: Adapter proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"StopDiscovery",
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("StopDiscovery failed (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("StopDiscovery failed");
		return BT_STATUS_FAIL;
	}

	/* discovery status will be change in event */
	DBG("-");
	g_variant_unref(result);
	return BT_STATUS_SUCCESS;
}

static gboolean __bt_adapter_all_properties_cb(gpointer user_data)
{
	GVariant *result = user_data;
	GVariantIter *property_iter;
	const gchar *key;
	GVariant *value;

	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;
	size_t size = 0;
	gchar *address = NULL;
	gchar *name = NULL;
	unsigned int cod = 0;
	gboolean discoverable;
	gboolean connectable;
	unsigned int scan_mode = BT_SCAN_MODE_NONE;
	unsigned int disc_timeout;
	gchar *version;
	gboolean is_discovering;
	gboolean is_le_discovering;
	uint8_t ipsp_initialized;
	gboolean powered;
	uint8_t pairable;
	unsigned int pairable_timeout;
	gboolean scan_mode_property_update = FALSE;

	DBG("+");

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	DBG("@@Start parsing properties");
	g_variant_get(result, "(a{sv})", &property_iter);
	while (g_variant_iter_loop(property_iter, "{sv}", &key, &value)) {
		if (!g_strcmp0(key, "Address")) {
			uint8_t bdaddr[6];

			address = (gchar *) g_variant_get_string(value, NULL);
			DBG("Address [%s]", address);
			_bt_hal_convert_addr_string_to_type(bdaddr, address);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_ADDR, sizeof(bdaddr), bdaddr);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Alias")) {
			name = (gchar *) g_variant_get_string(value, NULL);
			DBG("Alias [%s]", name);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_NAME, strlen(name) + 1, name);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Class")) {
			cod = g_variant_get_uint32(value);
			DBG("Class [%d]", cod);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_CLASS, sizeof(unsigned int), &cod);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Discoverable")) {
			discoverable = g_variant_get_boolean(value);
			DBG("Discoverable [%d]", discoverable);
			if (discoverable)
				scan_mode = BT_SCAN_MODE_CONNECTABLE_DISCOVERABLE;
			scan_mode_property_update = TRUE;
		} else if (!g_strcmp0(key, "DiscoverableTimeout")) {
			disc_timeout = g_variant_get_uint32(value);
			DBG("Discoverable Timeout [%d]", disc_timeout);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_DISC_TIMEOUT, sizeof(unsigned int), &disc_timeout);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Connectable")) {
			connectable = g_variant_get_boolean(value);
			DBG("Connectable [%d]", connectable);
			if (scan_mode == BT_SCAN_MODE_NONE)
				scan_mode = BT_SCAN_MODE_CONNECTABLE;
			scan_mode_property_update = TRUE;
		} else if (!g_strcmp0(key, "Version")) {
			version = (gchar *) g_variant_get_string(value, NULL);
			DBG("Version [%s]", version);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_VERSION, strlen(version) + 1, version);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Name")) {
			name = (gchar *) g_variant_get_string(value, NULL);
			DBG("Name [%s]", name);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_NAME, strlen(name) + 1, name);
			ev->num_props++;
		} else if (!g_strcmp0(key, "Powered")) {
			powered = g_variant_get_boolean(value);
			DBG("Powered = [%d]", powered);
		} else if (!g_strcmp0(key, "Pairable")) {
			pairable = (g_variant_get_boolean(value) ? 1 : 0);
			DBG("Pairable [%d]", pairable);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_PAIRABLE, sizeof(uint8_t), &pairable);
			ev->num_props++;
		} else if (!g_strcmp0(key, "PairableTimeout")) {
			pairable_timeout = g_variant_get_uint32(value);
			DBG("Pairable Timeout = [%d]", pairable_timeout);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_PAIRABLE_TIMEOUT, sizeof(unsigned int), &pairable_timeout);
			ev->num_props++;
		} else if (!g_strcmp0(key, "UUIDs")) {
			char **uuid_value;
			int uuid_count = 0;
			gsize size1 = 0;
			int i = 0;
			size1 = g_variant_get_size(value);
			int num_props_tmp = ev->num_props;
			if (size1 > 0) {
				uuid_value = (char **)g_variant_get_strv(value, &size1);
				for (i = 0; uuid_value[i] != NULL; i++)
					uuid_count++;
				/* UUID collection */
				uint8_t uuids[HAL_UUID_LEN * uuid_count];
				for (i = 0; uuid_value[i] != NULL; i++) {
					char *uuid_str = NULL;
					uint8_t uuid[HAL_UUID_LEN];
					uuid_str = g_strdup(uuid_value[i]);
					DBG("UUID string [%s]\n", uuid_str);
					_bt_hal_convert_uuid_string_to_type(uuid, uuid_str);
					memcpy(uuids + i * HAL_UUID_LEN, uuid, HAL_UUID_LEN);
					g_free(uuid_str);
				}
				size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_UUIDS,
						(HAL_UUID_LEN * uuid_count),
						uuids);
				ev->num_props = num_props_tmp + 1;
				g_free(uuid_value);
			}
		} else if (!g_strcmp0(key, "Discovering")) {
			is_discovering = g_variant_get_boolean(value);
			DBG("Discovering = [%d]", is_discovering);
		} else if (!g_strcmp0(key, "LEDiscovering")) {
			is_le_discovering = g_variant_get_boolean(value);
			DBG("LE Discovering = [%d]", is_le_discovering);
		} else if (!g_strcmp0(key, "Modalias")) {
			char *modalias = NULL;
			g_variant_get(value, "s", &modalias);
			DBG("Adapter ModAlias [%s]", modalias);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_MODALIAS, strlen(modalias) + 1, modalias);
			ev->num_props++;
			g_free(modalias);
		} else if (!g_strcmp0(key, "SupportedLEFeatures")) {
			DBG("LE Supported features");
			char *name = NULL;
			char *val = NULL;
			GVariantIter *iter = NULL;
			bt_local_le_features_t le_features;
			gboolean le_features_present = FALSE;

			memset(&le_features, 0x00, sizeof(le_features));

			g_variant_get(value, "as", &iter);
			if (iter) {
				while (g_variant_iter_loop(iter, "s", &name)) {
					DBG("name = %s", name);
					g_variant_iter_loop(iter, "s", &val);
					DBG("Value = %s", val);
					_bt_hal_update_le_feature_support(name, val, &le_features);
					le_features_present = TRUE;
				}
				g_variant_iter_free(iter);

				if (le_features_present) {
					size += __bt_insert_hal_properties(buf + size,
							HAL_PROP_ADAPTER_LOCAL_LE_FEAT, sizeof(le_features), &le_features);
					ev->num_props++;
				} else {
					DBG("le supported features values are NOT provided by Stack");
				}
			}
		} else if (!g_strcmp0(key, "IpspInitStateChanged")) {
			ipsp_initialized = (g_variant_get_boolean(value) ? 1 : 0);
			DBG("IPSP Initialized = %d", ipsp_initialized);
			size += __bt_insert_hal_properties(buf + size,
					HAL_PROP_ADAPTER_IPSP_INITIALIZED, sizeof(uint8_t), &ipsp_initialized);
			ev->num_props++;
		} else {
			ERR("Unhandled Property:[%s]", key);
		}
	}

	if (scan_mode_property_update) {
		size += __bt_insert_hal_properties(buf + size,
				HAL_PROP_ADAPTER_SCAN_MODE, sizeof(int), &scan_mode);
		ev->num_props++;
	}

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter properties changed event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_variant_unref(result);
	return FALSE;
}

static int __bt_hal_dbus_get_all_adapter_properties(void)
{
	GDBusProxy *proxy;
	GVariant *result;
	GError *error = NULL;

	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"GetAll",
			g_variant_new("(s)", BT_HAL_ADAPTER_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to get all adapter properties (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to get all adapter properties");
		return BT_STATUS_FAIL;
	}

	DBG("Got All properties from Bluez Stack!!, time to start parsing");
	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_all_properties_cb, (gpointer)result);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

int _bt_hal_dbus_get_adapter_properties(void)
{
	DBG("+");

	return __bt_hal_dbus_get_all_adapter_properties();
}

int _bt_hal_get_adapter_powered_state(uint8_t *state)
{
	char *adapter_path = NULL;
	gboolean powered = FALSE;
	int ret = BT_STATUS_FAIL;

	/* Check Valid Adapter Path, followed by Adapter Powered state.
	   If any of these fails, adapter will be considered Disabled */
	adapter_path = _bt_hal_get_adapter_path();

	if (adapter_path == NULL) {
		INFO("Adapter is not powered");
		g_free(adapter_path);
		*state = 0;
		return ret;
	}

	/* Check Adapter Powered Status */
	ret = _bt_hal_is_adapter_powered(&powered);
	if (BT_STATUS_FAIL == ret) {
		INFO("Adapter is not powered");
		g_free(adapter_path);
		*state = 0;
		return ret;
	}

	if (powered)
		*state = 1;
	else
		*state = 0;

	return ret;
}

/* Get Discoverable timeout API and callback */
static gboolean __bt_adapter_discovery_timeout_cb(gpointer user_data)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;;
	size_t size = 0;
	unsigned int *timeout = user_data;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	DBG("Discovery timeout in callback: [%d]", *timeout);

	size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_DISC_TIMEOUT,
			sizeof(unsigned int), timeout);

	ev->num_props++;
	DBG("Timeout value [%d] property Num [%d]", *timeout, ev->num_props);

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter Properties changed event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_free(timeout);
	return FALSE;
}

int _bt_hal_dbus_get_discovery_timeout(void)
{
	GDBusProxy *proxy;
	GVariant *result;
	GVariant *temp;
	GError *error = NULL;
	unsigned int *timeout;
	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Get",
			g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
				"DiscoverableTimeout"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to get local version (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to get local version");
		return BT_STATUS_FAIL;
	}

	timeout = g_malloc0(sizeof(int));
	if (!timeout) {
		ERR("Memory allocation failed");
		g_variant_unref(result);
		return BT_STATUS_FAIL;
	}

	g_variant_get(result, "(v)", &temp);
	*timeout = g_variant_get_uint32(temp);
	DBG("Timeout value: [%d]", *timeout);

	g_variant_unref(result);
	g_variant_unref(temp);

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_discovery_timeout_cb, (gpointer) timeout);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* Get Discoverable Mode API and callback */
static gboolean __bt_adapter_scan_mode_cb(gpointer user_data)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;;
	size_t size = 0;
	int *mode = user_data;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	DBG("Scan mode callback: [%d]", *mode);

	size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_SCAN_MODE,
			sizeof(int), mode);

	ev->num_props++;
	DBG("Scan mode [%d] property Num [%d]", *mode, ev->num_props);

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter Properties changed event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_free(mode);
	return FALSE;
}

int _bt_hal_dbus_get_scan_mode(void)
{
	GDBusProxy *proxy;
	gboolean discoverable;
	gboolean connectable;
	GVariant *result;
	GVariant *temp;
	GError *error = NULL;
	int *scan_mode;
		;
	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Get",
			g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
				"Discoverable"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to get discoverable mode (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to get discoverable mode");
		return BT_STATUS_FAIL;
	}

	g_variant_get(result, "(v)", &temp);
	discoverable = g_variant_get_boolean(temp);
	DBG("discoverable:%d", discoverable);

	g_variant_unref(result);
	g_variant_unref(temp);

	if (!discoverable) {
		result = g_dbus_proxy_call_sync(proxy,
				"Get",
				g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
					"Connectable"),
				G_DBUS_CALL_FLAGS_NONE,
				-1,
				NULL,
				&error);
		if (!result) {
			if (error != NULL) {
				ERR("Failed to get connectable mode (Error: %s)", error->message);
				g_clear_error(&error);
			} else
				ERR("Failed to get connectable mode");
			return BT_STATUS_FAIL;
		}

		g_variant_get(result, "(v)", &temp);
		connectable = g_variant_get_boolean(temp);
		DBG("connectable:%d", connectable);

		g_variant_unref(result);
		g_variant_unref(temp);
	}

	scan_mode = g_malloc0(sizeof(int));
	if (!scan_mode) {
		ERR("Memory allocation failed");
		return BT_STATUS_FAIL;
	}

	if (discoverable)
		*scan_mode = BT_SCAN_MODE_CONNECTABLE_DISCOVERABLE;
	else if (connectable)
		*scan_mode = BT_SCAN_MODE_CONNECTABLE;
	else
		*scan_mode = BT_SCAN_MODE_NONE;

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_scan_mode_cb, (gpointer) scan_mode);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* Get Local Version API and callback */
static gboolean __bt_adapter_local_version_cb(gpointer user_data)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;;
	size_t size = 0;
	char *version = NULL;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	version = (char*) user_data;
	DBG("Local Version in callback: [%s]", version);

	size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_VERSION,
			(strlen(version) + 1), version);

	ev->num_props++;
	DBG("Device version [%s] property Num [%d]", version, ev->num_props);

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter Properties changed event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_free(version);
	return FALSE;
}

int _bt_hal_dbus_get_local_version(void)
{
	GDBusProxy *proxy;
	const char *version = NULL;
	GVariant *result;
	GVariant *temp;
	GError *error = NULL;
	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Get",
			g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
				"Version"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to get local version (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to get local version");
		return BT_STATUS_FAIL;
	}

	g_variant_get(result, "(v)", &temp);
	version = g_variant_dup_string(temp, NULL);
	DBG("Local Version: [%s]", version);

	g_variant_unref(result);
	g_variant_unref(temp);

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_local_version_cb, (gpointer) version);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* Get Local Name API and callback */
static gboolean __bt_adapter_local_name_cb(gpointer user_data)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;;
	size_t size = 0;
	char *name = NULL;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	name = (char*) user_data;
	DBG("Local Name in callback: [%s]", name);

	size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_NAME,
			strlen(name) + 1, name);

	ev->num_props++;
	DBG("Device name [%s] property Num [%d]", name, ev->num_props);

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter Properties changed event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_free(name);
	return FALSE;
}

int _bt_hal_dbus_get_local_name(void)
{
	GDBusProxy *proxy;
	const char *name = NULL;
	GVariant *result;
	GVariant *temp;
	GError *error = NULL;
	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("_bt_hal_dbus_get_local_name: Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Get",
			g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
				"Alias"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to get local name (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to get local name");
		return BT_STATUS_FAIL;
	}

	g_variant_get(result, "(v)", &temp);
	name = g_variant_dup_string(temp, NULL);
	DBG("Local Name: [%s]", name);

	g_variant_unref(result);
	g_variant_unref(temp);

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_local_name_cb, (gpointer) name);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* Get Local Address API and callback */
static gboolean __bt_adapter_local_address_cb(gpointer user_data)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;
	size_t size = 0;
	char * address = NULL;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	address = (char*) user_data;

	uint8_t bdaddr[6];
	_bt_hal_convert_addr_string_to_type(bdaddr, address);

	size += __bt_insert_hal_properties(buf + size, HAL_PROP_DEVICE_ADDR,
			sizeof(bdaddr), bdaddr);

	ev->num_props++;
	DBG("Device address [%s] property Num [%d]", address, ev->num_props);

	size += __bt_insert_hal_properties(buf + size, HAL_PROP_ADAPTER_ADDR,
			sizeof(bdaddr), bdaddr);

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 1) {
		DBG("Send Device found event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_free(address);
	return FALSE;
}

int _bt_hal_dbus_get_local_address(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	const char *address;
	GVariant *result;
	GVariant *temp;

	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("_bt_hal_dbus_get_local_address: Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Get",
			g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
				"Address"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		ERR("Failed to get local address");
		if (error != NULL) {
			ERR("Failed to get local address (Error: %s)", error->message);
			g_clear_error(&error);
		}
		return BT_STATUS_FAIL;
	}

	g_variant_get(result, "(v)", &temp);
	address = g_variant_dup_string(temp, NULL);

	if (address)
		DBG("Address:%s", address);
	else
		return BT_STATUS_FAIL;

	g_variant_unref(result);
	g_variant_unref(temp);

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_local_address_cb, (gpointer) address);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* Get Local services API and callback */
static gboolean __bt_adapter_service_uuids_cb(gpointer user_data)
{
	GVariant *result = user_data;
	GVariant *temp;
	GVariantIter *iter = NULL;
	gchar *uuid_str;

	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;
	size_t size = 0;

	/* UUID collection */
	uint8_t uuids[HAL_UUID_LEN * MAX_UUID_COUNT];
	int uuid_count = 0;

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	g_variant_get(result, "(v)", &temp);
	g_variant_get(temp, "as", &iter);
	if (iter == NULL) {
		ERR("Failed to get UUIDs");
		goto fail;
	}

	while (g_variant_iter_loop(iter, "s", &uuid_str)) {
		uint8_t uuid[HAL_UUID_LEN];

		DBG("UUID string [%s]\n", uuid_str);
		_bt_hal_convert_uuid_string_to_type(uuid, uuid_str);
		memcpy(uuids + uuid_count * HAL_UUID_LEN, uuid, HAL_UUID_LEN);
		uuid_count++;
	}

	size += __bt_insert_hal_properties(buf + size,
			HAL_PROP_ADAPTER_UUIDS,
			(HAL_UUID_LEN * uuid_count),
			uuids);
	ev->num_props++;

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter properties changed event to HAL user,"
				" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	g_variant_iter_free(iter);
	g_variant_unref(result);
	g_variant_unref(temp);
	return FALSE;
fail:
	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (event_cb) {
		ev->status = BT_STATUS_FAIL;
		ev->num_props = 0;
		DBG("Send Adapter properties changed event to HAL user,"
				" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}

	g_variant_unref(result);
	return FALSE;
}

int _bt_hal_dbus_get_adapter_supported_uuids(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *result;

	DBG("+");

	proxy = _bt_hal_get_adapter_properties_proxy();

	if (!proxy) {
		DBG("_bt_hal_dbus_get_local_name: Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Get",
			g_variant_new("(ss)", BT_HAL_ADAPTER_INTERFACE,
				"UUIDs"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (!result) {
		if (error != NULL) {
			ERR("Failed to get UUIDs (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to get UUIDs");
		return BT_STATUS_FAIL;
	}


	DBG("Got All Adaptr service UUID's from Bluez Stack!!, time to start parsing");

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_service_uuids_cb, (gpointer)result);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

static gboolean __is_device_paired(GVariantIter *item_iter)
{
	gboolean paired = FALSE;
	GVariant *value;
	gchar *key;

	while (g_variant_iter_loop(item_iter, "{sv}", &key, &value)) {
		if (NULL == key || g_strcmp0(key, "Paired"))
			continue;

		paired = g_variant_get_boolean(value);
		g_variant_unref(value);
		g_free(key);
		break;
	}

	return paired;
}

static gboolean __bt_adapter_bonded_devices_cb(gpointer user_data)
{
	/* Buffer and propety count management */
	uint8_t buf[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	uint8_t addresses[BT_HAL_MAX_PROPERTY_BUF_SIZE];
	struct hal_ev_adapter_props_changed *ev = (void*) buf;;
	size_t size = 0;
	size_t count = 0;

	GVariant *result = user_data;
	GVariantIter *iter;
	GVariantIter *interface_iter;
	GVariantIter *svc_iter;
	char *object_path = NULL;
	char *interface_str = NULL;

	char device_address[BT_HAL_ADDRESS_STRING_SIZE];
	uint8_t bdaddr[BT_HAL_ADDRESS_LENGTH_MAX];

	memset(buf, 0, sizeof(buf));
	size = sizeof(*ev);
	ev->num_props = 0;
	ev->status = BT_STATUS_SUCCESS;

	INFO("Size [%zd]", size);

	/* signature of GetManagedObjects:  a{oa{sa{sv}}} */
	g_variant_get(result, "(a{oa{sa{sv}}})", &iter);

	/* Parse the signature: {oa{sa{sv}}} */
	while (g_variant_iter_loop(iter, "{&oa{sa{sv}}}", &object_path, &interface_iter)) {
		if (object_path == NULL)
			continue;

		while (g_variant_iter_loop(interface_iter, "{sa{sv}}",
					&interface_str, &svc_iter)) {
			if (g_strcmp0(interface_str, "org.bluez.Device1") == 0) {
				DBG("Found a device: %s", object_path);
				if (__is_device_paired(svc_iter)) {
					_bt_hal_convert_device_path_to_address(object_path, device_address);
					DBG("Paired Device Address: [%s]", device_address);

					_bt_hal_convert_addr_string_to_type(bdaddr, device_address);
					memcpy((addresses + (count * BT_HAL_ADDRESS_LENGTH_MAX)),
							bdaddr, BT_HAL_ADDRESS_LENGTH_MAX);
					count++;
				}
				g_variant_iter_free(svc_iter);
				g_free(interface_str);
				interface_str = NULL;
				break;
			}
		}
	}

	g_variant_iter_free(iter);
	g_variant_unref(result);

	size += __bt_insert_hal_properties((buf + size),
			HAL_PROP_ADAPTER_BONDED_DEVICES,
			(count * BT_HAL_ADDRESS_LENGTH_MAX), addresses);
	ev->num_props++;

	if (!event_cb)
		event_cb = _bt_get_adapter_event_cb();
	if (!event_cb) {
		ERR("event_cb is NULL");
		goto done;
	}

	if (size > 2) {
		DBG("Send Adapter properties changed event to HAL user,"
			" Num Prop [%d] total size [%zd]", ev->num_props, size);
		event_cb(HAL_EV_ADAPTER_PROPS_CHANGED, (void*) buf, size);
	}
done:
	return FALSE;
}

int _bt_hal_dbus_get_bonded_devices(void)
{
	GDBusProxy *manager_proxy;
	GVariant *result = NULL;

	DBG("+");

	manager_proxy = _bt_hal_get_manager_proxy();
	if (manager_proxy == NULL)
		return  BT_STATUS_FAIL;

	result = g_dbus_proxy_call_sync(manager_proxy, "GetManagedObjects",
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			NULL);
	if (!result) {
		ERR("Can't get managed objects");
		return BT_STATUS_FAIL;
	}

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	g_idle_add(__bt_adapter_bonded_devices_cb, (gpointer) result);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

int _bt_hal_dbus_get_adapter_property(bt_property_type_t property_type)
{
	DBG("+");

	INFO("property_type: %d", property_type);

	switch (property_type) {
	case BT_PROPERTY_BDADDR:
		return _bt_hal_dbus_get_local_address();
	case BT_PROPERTY_BDNAME:
		return _bt_hal_dbus_get_local_name();
	case BT_PROPERTY_VERSION:
		return _bt_hal_dbus_get_local_version();
	case BT_PROPERTY_ADAPTER_DISCOVERY_TIMEOUT:
		return _bt_hal_dbus_get_discovery_timeout();
	case BT_PROPERTY_ADAPTER_SCAN_MODE:
		return _bt_hal_dbus_get_scan_mode();
	case BT_PROPERTY_CLASS_OF_DEVICE:
		return BT_STATUS_UNSUPPORTED;
	case BT_PROPERTY_UUIDS:
		return _bt_hal_dbus_get_adapter_supported_uuids();
	case BT_PROPERTY_ADAPTER_BONDED_DEVICES:
		return _bt_hal_dbus_get_bonded_devices();
	default:
		return BT_STATUS_UNSUPPORTED;
	}

	DBG("-");
}

static int __bt_hal_dbus_set_local_name(void *data)
{
	GDBusProxy *proxy;
	const char *name = data;
	GVariant *result;
	GError *error = NULL;

	DBG("+");
	DBG("Local Name: [%s]", name);
	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("_bt_hal_dbus_get_local_name: Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Set",
			g_variant_new("(ssv)", BT_HAL_ADAPTER_INTERFACE,
				"Alias", g_variant_new("s", name)),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (!result) {
		if (error != NULL) {
			ERR("Failed to set local name (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to set local name");
		return BT_STATUS_FAIL;
	}
	g_variant_unref(result);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

static int __bt_hal_dbus_set_scan_mode(void *data)
{
	GDBusProxy *proxy;
	int *mode = data;
	GVariant *result;
	GError *error = NULL;
	gboolean pg_scan;
	gboolean inq_scan;

	DBG("+");
	DBG("Scan mode: [%d]", *mode);
	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	switch (*mode) {
	case BT_SCAN_MODE_NONE:
		pg_scan = FALSE;
		inq_scan = FALSE;
		break;
	case BT_SCAN_MODE_CONNECTABLE:
		pg_scan = TRUE;
		inq_scan = FALSE;
		break;
	case BT_SCAN_MODE_CONNECTABLE_DISCOVERABLE:
		pg_scan = TRUE;
		inq_scan = TRUE;
		break;
	default:
		ERR("Invalid scan mode");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Set",
			g_variant_new("(ssv)", BT_HAL_ADAPTER_INTERFACE,
				"Connectable", g_variant_new("b", pg_scan)),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (!result) {
		if (error != NULL) {
			ERR("Failed to set connectable property (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to set connectable property");
		return BT_STATUS_FAIL;
	}
	g_variant_unref(result);

	result = g_dbus_proxy_call_sync(proxy,
			"Set",
			g_variant_new("(ssv)", BT_HAL_ADAPTER_INTERFACE, "Discoverable",
				g_variant_new("b", inq_scan)),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (!result) {
		if (error != NULL) {
			ERR("Failed to set Discoverable property (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to set Discoverable property");
		return BT_STATUS_FAIL;
	}
	g_variant_unref(result);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

static int __bt_hal_dbus_set_discovery_timeout(void *data)
{
	GDBusProxy *proxy;
	unsigned int *timeout = data;
	GVariant *result;
	GError *error = NULL;


	DBG("+");

	DBG("Discovery Timeout: [%d]", *timeout);
	proxy = _bt_hal_get_adapter_properties_proxy();
	if (!proxy) {
		DBG("Adapter Properties proxy get failed!!!");
		return BT_STATUS_FAIL;
	}

	result = g_dbus_proxy_call_sync(proxy,
			"Set",
			g_variant_new("(ssv)", BT_HAL_ADAPTER_INTERFACE,
				"DiscoverableTimeout", g_variant_new("u", *timeout)),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);
	if (!result) {
		if (error != NULL) {
			ERR("Failed to set DiscoverableTimeout property (Error: %s)", error->message);
			g_clear_error(&error);
		} else
			ERR("Failed to set DiscoverableTimeout property");
		return BT_STATUS_FAIL;
	}
	g_variant_unref(result);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* Set Adapter Properties */
int _bt_hal_dbus_set_adapter_property(const bt_property_t *property)
{
	int result;

	DBG("");

	if (property == NULL || property->val == NULL) {
		ERR("Invalid parameters received");
		return BT_STATUS_FAIL;
	}

	switch (property->type) {
	case BT_PROPERTY_BDNAME:
		result =  __bt_hal_dbus_set_local_name(property->val);
		break;
	case BT_PROPERTY_ADAPTER_SCAN_MODE:
		result =  __bt_hal_dbus_set_scan_mode(property->val);
		break;
	case BT_PROPERTY_ADAPTER_DISCOVERY_TIMEOUT:
		result =  __bt_hal_dbus_set_discovery_timeout(property->val);
		break;
	default:
		result = BT_STATUS_UNSUPPORTED;
	}

	DBG("Result= [%d]", result);
	return result;
}
