/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 -2016 Samsung Electronics Co., Ltd All Rights Reserved.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>

#include <hardware/bluetooth.h>
#include <hardware/bt_gatt.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <gio/gio.h>
#include <glib.h>
#include <dlog.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <vconf-internal-radio-keys.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <fcntl.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"

#include <bt-hal-adapter-dbus-handler.h>
#include <bt-hal-dbus-common-utils.h>

#include "bt-hal-gatt-server.h"

typedef struct {
	int adv_inst_max;
	int rpa_offloading;
	int max_filter;
	int le_2m_phy;
	int le_coded_phy;
} bt_adapter_le_feature_info_t;

typedef struct {
	int initialized;
	gboolean is_multi_adv; /* To be removed once we complete descope Legacy Adv */
	int adv_handle;
	gboolean is_advertising;
	guint hold_timer_id;
	bt_uuid_t app_uuid;
} bt_adapter_le_adv_slot_t;

static bt_adapter_le_feature_info_t le_feature_info = { 1, 0, 0 };
static bt_adapter_le_adv_slot_t *le_adv_slot = NULL;

typedef struct {
	uint8_t event;
	int server_if;
	uint8_t status;
	uint8_t data[31];
} bt_hal_adv_event_data_t;

/* Macros */
#define BT_HAL_ADV_CONNECTABLE  0x00 /* ADV_IND */
#define BT_HAL_ADV_CONNECTABLE_DIRECT_HIGH  0x01 /* ADV_DIRECT_IND, high duty cycle */
#define BT_HAL_ADV_SCANNABLE 0x02 /* ADV_SCAN_IND */
#define BT_HAL_ADV_NON_CONNECTABLE  0x03 /* ADV_NONCOND_IND */
#define BT_HAL_ADV_CONNECTABLE_DIRECT_LOW  0x04 /* ADV_DIRECT_IND, low duty cycle */

#define BT_HAL_ADV_INTERVAL_MIN 20 /* msec */
#define BT_HAL_ADV_INTERVAL_MAX 10240
#define BT_HAL_ADV_INTERVAL_SPLIT 0.625
#define BT_HAL_DEFAULT_ADV_MIN_INTERVAL 500
#define BT_HAL_DEFAULT_ADV_MAX_INTERVAL 500
#define BT_HAL_ADV_FILTER_POLICY_DEFAULT    0x00
#define BT_HAL_ADV_TYPE_DEFAULT     0x00
#define BT_HAL_ADV_FILTER_POLICY_ALLOW_SCAN_CONN_WL_ONLY    0x03

/* Multi Advertisement callback event */
#define BT_HAL_MULTI_ADV_ENB_EVT           1
#define BT_HAL_MULTI_ADV_DISABLE_EVT       2
#define BT_HAL_MULTI_ADV_PARAM_EVT         3
#define BT_HAL_MULTI_ADV_DATA_EVT          4
#define BT_HAL_MULTI_ADV_UPDATE_EVT        5
#define BT_HAL_LEGACY_ADV_STATUS           6

handle_stack_msg gatt_le_event_cb;

/* Forward declarations */
static gboolean __bt_hal_is_factory_test_mode(void);
static void __bt_hal_free_le_adv_slot(void);
static gboolean __bt_hal_adv_event_cb(gpointer param);

#ifdef TIZEN_BT_HAL
/* Enable LE Adapter */
int _bt_hal_le_enable(void)
{
	return _bt_hal_dbus_enable_le();
}

/* Disable LE Adapter */
int _bt_hal_le_disable(void)
{
	return _bt_hal_dbus_disable_le();
}

int _bt_hal_set_le_static_random_address(uint8_t enable)
{
	DBG("+");
	GError *error = NULL;
	GVariant *ret;
	GDBusProxy *proxy;

	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	ret = g_dbus_proxy_call_sync(proxy, "SetLeStaticRandomAddress",
			g_variant_new("(b)", (enable ? TRUE : FALSE)),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (error) {
		ERR("Set static address Fail: %s", error->message);
		g_clear_error(&error);
		return BT_STATUS_FAIL;
	}

	INFO("Set le static address [%d]", enable);
	if (ret)
		g_variant_unref(ret);

	return BT_STATUS_SUCCESS;
}
#endif

static void __bt_hal_free_le_adv_slot(void)
{
	int i;

	if (le_adv_slot == NULL)
		return;

	for (i = 0; i < le_feature_info.adv_inst_max; i++)
		memset(&le_adv_slot[i], 0x00, sizeof(bt_adapter_le_adv_slot_t));

	g_free(le_adv_slot);
	le_adv_slot = NULL;
}

void _bt_hal_unregister_adv_slot_owner(int slot_id)
{
	if (le_adv_slot == NULL)
		return;
	INFO("Unregister Adv Slot [%d]", slot_id);
	memset(&le_adv_slot[slot_id], 0x00, sizeof(bt_adapter_le_adv_slot_t));
}

int _bt_hal_get_adv_slot_adv_handle(int slot_id)
{
	if (le_adv_slot == NULL)
		return 0;

	return le_adv_slot[slot_id].adv_handle;
}

gboolean _bt_hal_is_advertising_in_slot(int slot)
{
	return le_adv_slot[slot].is_advertising;
}

void _bt_hal_set_advertising_status(int slot_id, gboolean mode)
{

	DBG("Advertising enabled [%s] slot [%d]",  mode ? "TRUE" : "FALSE", slot_id);
	bt_hal_adv_event_data_t *event;

	if (le_adv_slot == NULL)
		return;

	le_adv_slot[slot_id].is_advertising = mode;

	event = g_malloc0(sizeof(bt_hal_adv_event_data_t));
	event->event = mode ? BT_HAL_MULTI_ADV_ENB_EVT : BT_HAL_MULTI_ADV_DISABLE_EVT;
	event->server_if = slot_id;
	event->status = BT_STATUS_SUCCESS;
	/* To be removed later when we completely descope Legacy Adv concept */
	DBG("Is multi ? [%d]", le_adv_slot[slot_id].is_multi_adv);
	if (le_adv_slot[slot_id].is_multi_adv == FALSE)
		event->event = BT_HAL_LEGACY_ADV_STATUS;

	__bt_hal_adv_event_cb((gpointer)event);
}

gboolean _bt_hal_is_advertising(void)
{
	int i;

	for (i = 0; i < le_feature_info.adv_inst_max; i++) {
		if (le_adv_slot[i].is_advertising == TRUE)
			return TRUE;
	}

	return FALSE;
}

int _bt_hal_le_init(void)
{
	le_adv_slot = g_malloc0(sizeof(bt_adapter_le_adv_slot_t) * le_feature_info.adv_inst_max);
	return BT_STATUS_SUCCESS;
}
void _bt_hal_le_deinit(void)
{
	__bt_hal_free_le_adv_slot();
}

gboolean _bt_hal_update_le_feature_support(const char *item, const char *value,
			bt_local_le_features_t *le_features)
{
	if (item == NULL || value == NULL)
		return FALSE;
	if (!le_adv_slot)
		_bt_hal_le_init();

	if (g_strcmp0(item, "adv_inst_max") == 0) {
		int slot_num;

		slot_num = atoi(value);
		if (slot_num < 0) {
			ERR("ERR:Advertising MAX instance [%d]", slot_num);
			return FALSE;
		}

		if (slot_num != le_feature_info.adv_inst_max) {
			__bt_hal_free_le_adv_slot();
			le_feature_info.adv_inst_max = slot_num;
			INFO("Advertising instance max : %d", le_feature_info.adv_inst_max);
			le_adv_slot = g_malloc0(sizeof(bt_adapter_le_adv_slot_t) * le_feature_info.adv_inst_max);
		}
		/* Fill LE feature bytes */
		le_features->max_adv_instance = atoi(value);

	} else if (g_strcmp0(item, "rpa_offloading") == 0) {
		le_feature_info.rpa_offloading = atoi(value);
		INFO("RPA offloading : %d", le_feature_info.rpa_offloading);

		/* Fill LE feature bytes */
		le_features->rpa_offload_supported = atoi(value);

	} else if (g_strcmp0(item, "max_filter") == 0) {
		le_feature_info.max_filter = atoi(value);
		INFO("BLE Scan max filter : %d", le_feature_info.max_filter);

		/* Fill LE feature bytes */
		le_features->max_adv_filter_supported = atoi(value);

	} else if (g_strcmp0(item, "2m_phy") == 0) {
		if (g_strcmp0(value, "true") == 0) {
			le_feature_info.le_2m_phy = TRUE;
			/* Fill LE feature bytes */
			le_features->le_2m_phy_supported = 0x1;
		} else {
			le_feature_info.le_2m_phy = FALSE;
			/* Fill LE feature bytes */
			le_features->le_2m_phy_supported = 0x0;
		}
		INFO("2M PHY Supported [%s]",  le_feature_info.le_2m_phy ? "TRUE" : "FALSE");
	} else if (g_strcmp0(item, "coded_phy") == 0) {
		if (g_strcmp0(value, "true") == 0) {
			le_feature_info.le_coded_phy = TRUE;
			/* Fill LE feature bytes */
			le_features->le_coded_phy_supported = 0x1;
		} else {
			le_feature_info.le_coded_phy = FALSE;
			/* Fill LE feature bytes */
			le_features->le_coded_phy_supported = 0x0;
		}
		INFO("CODED PHY Supported [%s]",  le_feature_info.le_coded_phy ? "TRUE" : "FALSE");
	} else {
		DBG("No registered item");
		return FALSE;
	}

	return TRUE;
}

void _bt_hal_free_server_slot(int slot_id)
{
	if (le_adv_slot == NULL)
		return;
	memset(&le_adv_slot[slot_id], 0x00, sizeof(bt_adapter_le_adv_slot_t));
}

int _bt_hal_get_available_adv_slot_id(bt_uuid_t *uuid, gboolean use_reserved_slot)
{
	int i;

	if (le_adv_slot == NULL) {
		ERR("le_adv_slot is NULL");
		return -1;
	}

	DBG("adv_inst_max : %d", le_feature_info.adv_inst_max);

	for (i = 0; i < le_feature_info.adv_inst_max; i++) {
		if (le_adv_slot[i].initialized == 0)
			continue;
		if (memcmp(uuid->uu, le_adv_slot[i].app_uuid.uu, sizeof(bt_uuid_t)) == 0) {
			DBG("UUID [%s] matched, return slot [%d]", btuuid2str(uuid->uu), i);
			return i;
		}
	}

	/* We should consider 'use_reverved_slot' in later */
	if (le_feature_info.adv_inst_max <= 1)
		i = 0;
	else if (use_reserved_slot == TRUE)
		i = 1;
	else
		i = 2;

	for (i = 0; i < le_feature_info.adv_inst_max; i++) {
		if (le_adv_slot[i].initialized == 0) {
			DBG("Slot to be allocated [%d] UUID to be registered [%s]",
				i, btuuid2str(uuid->uu));
			le_adv_slot[i].initialized = 1;
			memcpy(&le_adv_slot[i].app_uuid.uu, &uuid->uu, sizeof(bt_uuid_t));
			return i;
		}
	}

	return -1;
}

void _bt_hal_get_gatt_server_instance_initialized(int *instance)
{
	int i;
	if (le_adv_slot == NULL) {
		ERR("le_adv_slot is NULL");
		return;
	}

	DBG("adv_inst_max : %d", le_feature_info.adv_inst_max);

	for (i = 0; i < le_feature_info.adv_inst_max; i++) {
		if (le_adv_slot[i].initialized == 0)
			continue;
		DBG("Initialized Slot found: UUID [%s] slot [%d]", le_adv_slot[i].app_uuid.uu, i);
		*instance = i;
		break;
	}
}

gboolean _bt_is_advertising(void)
{
	return FALSE;
}

int _bt_set_advertising(const char *sender, int adv_handle,
			gboolean enable, gboolean use_reserved_slot)
{
	DBG("+");
	return BT_STATUS_UNSUPPORTED;
}

static int __bt_hal_uuid_type(uint8_t* p_uuid)
{
	int i = 0;
	int match = 0;
	int all_zero = 1;

	for (i = 0; i != 16; ++i) {
		if (i == 12 || i == 13)
			continue;

		if (p_uuid[i] == BASE_UUID_CONVERTED[i])
			++match;

		if (p_uuid[i] != 0)
			all_zero = 0;
	}
	if (all_zero)
		return 0;
	if (match == 12)
		return BT_HAL_UUID_32;
	if (match == 14)
		return BT_HAL_UUID_16;
	return BT_HAL_UUID_128;
}

static void __bt_hal_parse_uuid(int len, char *src, uint8_t *dest, int *length, int is_solicit)
{
	DBG("+");
	int prev_byte_len = 0;
	/* dest[index] will contain the length followed by AD Type
	   Move length only when different byte_len is found Ex) 2->4, 2->16, 4->16 etc */
	int index = 0;

	while (len >= 16) {

		/* Create Local buffer & copy source 16 bytes in sequence */
		int byte_len;
		bt_uuid_t uuid;
		memset(&uuid, 0, sizeof(bt_uuid_t));
		memcpy(&uuid.uu, src, 16);

		/* Compute current UUID's byte length */
		byte_len = __bt_hal_uuid_type(uuid.uu);

		switch (byte_len) {
		case 2: {
			if (prev_byte_len == byte_len) {
				memcpy(&(dest[dest[index] + 1]), &src[12], byte_len);
				*length += 2;
				dest[index] += 2;
			} else {
				if (dest[index] != 0)
					index = dest[index] +1;
				dest[index] = byte_len + 1;
				if (is_solicit)
					dest[index+1] = 0x14; /* AD Type */
				else
					dest[index+1] = 0x02; /* AD Type */
				memcpy(&(dest[index + 2]), &src[12], byte_len);
				*length += 4;
			}

			/* Update current type */
			prev_byte_len = byte_len;
			break;
		}
		case 4: {
			if (prev_byte_len == byte_len) {
				memcpy(&(dest[dest[index] + 1]), &src[12], byte_len);;
				*length += 4;
				dest[index] += 4;
			} else {
				if (dest[index] != 0)
					index = dest[index] +1;
				dest[index] = byte_len + 1;
				if (is_solicit)
					dest[index+1] = 0x1F; /* AD Type */
				else
					dest[index+1] = 0x04; /* AD Type */
				memcpy(&(dest[index + 2]), &src[12], byte_len);
				*length += 6;
			}
			/* Update current type */
			prev_byte_len = byte_len;
			break;
		}
		case 16: {
			if (dest[index] != 0)
				index = dest[index] +1;
			dest[index] = byte_len + 1;
			if (is_solicit)
				dest[index+1] = 0x15; /* AD Type */
			else
				dest[index+1] = 0x06; /* AD Type */
			memcpy(&(dest[index + 2]), &src[12], byte_len);
			/* Update current type */
			prev_byte_len = byte_len;
			*length += 18;
			break;
		}
		default: {
			ERR("Abnormal Byte len [%d]", byte_len);
			break;
		}
		}

		/* Process Next 16 bytes of MW UUID */
		src += 16;
		len -= 16;
	}
	DBG("-");
}

static gboolean __bt_hal_is_factory_test_mode(void)
{
	int mode = 0;

	if (vconf_get_bool(VCONFKEY_BT_DUT_MODE, &mode)) {
		ERR("Get the DUT Mode fail");
		return TRUE;
	}

	if (mode != FALSE) {
		INFO("DUT Test Mode !!");
		return TRUE;
	}

	return FALSE;
}

static gboolean __bt_hal_adv_event_cb(gpointer param)
{
	bt_hal_adv_event_data_t *event = (bt_hal_adv_event_data_t*)param;
	DBG("+");

	if (!event)
		return FALSE;

	if (!gatt_le_event_cb) {
		ERR("GATT event callback not registered!!!");
		return FALSE;
	}

	switch (event->event) {
	case BT_HAL_MULTI_ADV_ENB_EVT: {
		INFO("BLE Advertising enabled slot [%d]", event->server_if);
		struct hal_ev_multi_adv_enable ev;
		memset(&ev, 0, sizeof(struct hal_ev_multi_adv_enable));
		ev.status = event->status;
		ev.server_instance = event->server_if;
		gatt_le_event_cb(HAL_EV_MULTI_ADV_ENABLE, (void *)&ev, sizeof(ev));
		break;
	}
	case BT_HAL_MULTI_ADV_DISABLE_EVT: {
		INFO("BLE Advertising disabled slot [%d]", event->server_if);
		struct hal_ev_multi_adv_disable ev;
		memset(&ev, 0, sizeof(struct hal_ev_multi_adv_disable));
		ev.status = event->status;
		ev.server_instance = event->server_if;
		gatt_le_event_cb(HAL_EV_MULTI_ADV_DISABLE, (void *)&ev, sizeof(ev));
		break;
	}
	case BT_HAL_MULTI_ADV_PARAM_EVT: {
		INFO("Unhandled event slot [%d]", event->server_if);
		break;
	}
	case BT_HAL_MULTI_ADV_UPDATE_EVT: {
		INFO("BLE Advertising Param update slot [%d]", event->server_if);
		struct hal_ev_multi_adv_update ev;
		memset(&ev, 0, sizeof(struct hal_ev_multi_adv_update));
		ev.status = event->status;
		ev.server_instance = event->server_if;
		gatt_le_event_cb(HAL_EV_MULTI_ADV_UPDATE, (void *)&ev, sizeof(ev));
		break;
	}
	case BT_HAL_MULTI_ADV_DATA_EVT: {
		INFO("BLE Advertising data set slot [%d]", event->server_if);
		struct hal_ev_multi_adv_data_set ev_data_set;
		/* Copy data */
		memset(&ev_data_set, 0, sizeof(struct hal_ev_multi_adv_data_set));
		ev_data_set.status = event->status;
		ev_data_set.server_instance = event->server_if;
		gatt_le_event_cb(HAL_EV_MULTI_ADV_DATA_SET, (void *)&ev_data_set, sizeof(ev_data_set));
		break;
	}
	case BT_HAL_LEGACY_ADV_STATUS: {
		INFO("BLE Legacy Advertising [%d]", event->server_if);
		struct hal_ev_legacy_adv_status ev;
		/* Copy data */
		memset(&ev, 0, sizeof(struct hal_ev_legacy_adv_status));
		ev.status = event->status;
		ev.server_instance = event->server_if;
		gatt_le_event_cb(HAL_EV_LEGACY_ADV_ENABLE, (void *)&ev, sizeof(ev));
		break;
	}
	default:
		ERR("Unknown event");
		break;
	}

	g_free(event);

	DBG("-");
	return FALSE;
}

int _bt_hal_enable_advertising(int server_if, bool enable, bool is_multi_adv)
{
	DBG("+");
	GError *error = NULL;
	GVariant *ret;
	GDBusProxy *proxy;

	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	if (le_adv_slot[server_if].is_advertising == TRUE && enable == TRUE)
		return BT_STATUS_BUSY;

	if (le_adv_slot[server_if].initialized == TRUE &&
			le_adv_slot[server_if].is_advertising == FALSE &&
			enable == FALSE)
		return BT_STATUS_DONE;

	if (le_adv_slot[server_if].hold_timer_id > 0) {
		g_source_remove(le_adv_slot[server_if].hold_timer_id);
		le_adv_slot[server_if].hold_timer_id = 0;
	}

	ret = g_dbus_proxy_call_sync(proxy, "SetAdvertising",
			g_variant_new("(bi)", enable, server_if),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&error);

	if (error) {
		ERR("SetAdvertising Fail: %s", error->message);
		g_clear_error(&error);
		return BT_STATUS_FAIL;
	}

	INFO("Enable advertising [%d] SLot Id [%d] Is Multi? [%d]", enable, server_if, is_multi_adv);
	if (ret)
		g_variant_unref(ret);

	le_adv_slot[server_if].is_multi_adv = is_multi_adv;
	le_adv_slot[server_if].is_advertising = enable;
	return BT_STATUS_SUCCESS;
}

int _bt_hal_set_advertising_params(int server_if, int min_interval,
		int max_interval, int adv_type,
		int chnl_map, int tx_power, int timeout_s)
{
	DBG("+");
	GDBusProxy *proxy;
	GVariant *ret;
	GError *error = NULL;
	guint32 min = 0;
	guint32 max = 0;
	bt_hal_adv_event_data_t *event;


	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	if (min_interval > max_interval ||
			min_interval < BT_HAL_ADV_INTERVAL_MIN ||
			max_interval > BT_HAL_ADV_INTERVAL_MAX)
		return BT_STATUS_PARM_INVALID;


	if (adv_type  == BT_HAL_ADV_CONNECTABLE_DIRECT_HIGH ||
			adv_type == BT_HAL_ADV_CONNECTABLE_DIRECT_LOW ||
			adv_type == BT_HAL_ADV_NON_CONNECTABLE)
		return BT_STATUS_UNSUPPORTED;


	min = min_interval / BT_HAL_ADV_INTERVAL_SPLIT;
	max = max_interval / BT_HAL_ADV_INTERVAL_SPLIT;

	ret = g_dbus_proxy_call_sync(proxy, "SetAdvertisingParameters",
			g_variant_new("(uuuuii)", min, max,
			BT_HAL_ADV_FILTER_POLICY_DEFAULT, adv_type,
			tx_power, server_if), G_DBUS_CALL_FLAGS_NONE,
			-1, NULL, &error);
	if (error) {
		ERR("SetAdvertisingParameters Fail: %s", error->message);
		g_clear_error(&error);
		return BT_STATUS_FAIL;
	}

	INFO("Set advertising data");
	if (ret)
		g_variant_unref(ret);

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	event = g_malloc0(sizeof(bt_hal_adv_event_data_t));
	event->event  = BT_HAL_MULTI_ADV_UPDATE_EVT;
	event->server_if = server_if;
	event->status = BT_STATUS_SUCCESS;
	g_idle_add(__bt_hal_adv_event_cb, (gpointer)event);

	return BT_STATUS_SUCCESS;
}

/* Takes care of both Scan Response and Advertising data */
int _bt_hal_set_advertising_data(btgatt_adv_param_setup_t adv_param_setup)
{
	DBG("+");
	uint8_t adv_data[31];
	int index = 0;
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *ret = NULL;
	GVariant *temp = NULL;
	GVariantBuilder *builder;
	bt_hal_adv_event_data_t *event;
	int length = 0;
	int i;

	/* Parse data according to Bluez Interface */
	if (__bt_hal_is_factory_test_mode()) {
		ERR("Unable to set advertising data in factory binary !!");
		return BT_STATUS_UNSUPPORTED;
	}

	/* TODO: Check adapter and LE adapter status */
	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	memset(&adv_data, 0, 31);

	/* Service UUID */
	DBG("Service UUID length [%d]", adv_param_setup.service_uuid_len);
	if (adv_param_setup.service_uuid_len > 0) {
		__bt_hal_parse_uuid(adv_param_setup.service_uuid_len,
				adv_param_setup.service_uuid, &adv_data[index], &length, FALSE);
		index = length;
		DBG("After Service UUID:Index [%d]", index);
	}

	/* Solicit UUID */
	DBG("Solicit UUID length [%d]", adv_param_setup.solicit_uuid_len);
	if (adv_param_setup.solicit_uuid_len > 0) {
		__bt_hal_parse_uuid(adv_param_setup.solicit_uuid_len,
				adv_param_setup.solicit_uuid, &adv_data[index], &length, TRUE);
		index = length;
		DBG("After Solicit UUID: Index [%d]", index);
	}

	/* Service Data  UUID*/
	DBG("Service Data length [%d]", adv_param_setup.service_data_len);
	if (adv_param_setup.service_data_len > 0) {
		adv_data[index] = 1 + adv_param_setup.service_data_len;
		adv_data[index+1] = 0x16; /* Fixed */
		memcpy(&adv_data[index+2], adv_param_setup.service_data, adv_param_setup.service_data_len);
		index += (2 + adv_param_setup.service_data_len);
		length += (2 + adv_param_setup.service_data_len);
		DBG("After Service data: Index [%d]", index);
	}

	/* Set Apperance */
	if (adv_param_setup.appearance > 0) {
		adv_data[index] = 0x03;
		adv_data[index+1] = 0x19;
		adv_data[index+2] = (uint8_t) (adv_param_setup.appearance & 0xFF);
		adv_data[index+3] = (uint8_t) ((adv_param_setup.appearance  >> 8) & 0xFF);
		index += 4;
		length += 4;
		DBG("After Apperance: Index [%d]", index);
	}

	/* TX Power */
	if (adv_param_setup.include_txpower != 0) {
		adv_data[index] = 0x01;
		adv_data[index+1] = 0x0A;
		index += 2;
		length += 2;
		DBG("After TX Power: Index [%d]", index);
	}

	/* Device Name */
	if (adv_param_setup.include_name != 0) {
		adv_data[index] = 0x01;
		adv_data[index+1] = 0x09;
		index += 2;
		length += 2;
		DBG("After Name: Index [%d]", index);
	}

	/* Manufacturer data */
	if (adv_param_setup.manufacturer_data_len > 0) {
		adv_data[index] = 1 + adv_param_setup.manufacturer_data_len;
		adv_data[index+1] = 0xFF;
		memcpy(&adv_data[index+2], adv_param_setup.manufacturer_data, adv_param_setup.manufacturer_data_len);
		index += (2 + adv_param_setup.manufacturer_data_len);
		length += (2 + adv_param_setup.manufacturer_data_len);
		DBG("After Manuf Data: Index [%d]", index);
	}

	/* Create Builder */
	builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for (i = 0; i < length; i++)
		g_variant_builder_add(builder, "y", adv_data[i]);

	temp = g_variant_new("ay", builder);
	g_variant_builder_unref(builder);

	DBG("####Adv data length [%d] Index [%d]", length, index);
	for (i = 0; i < length; i++)
		DBG("##Data[%d] [0x%x]", i, adv_data[i]);

	if (adv_param_setup.set_scan_rsp == 0) {
		/* Set Advertising data to stack */
		ret = g_dbus_proxy_call_sync(proxy, "SetAdvertisingData",
				g_variant_new("(@ayi)", temp, adv_param_setup.server_if),
				G_DBUS_CALL_FLAGS_NONE,
				-1, NULL, &error);
	} else {
		/* Set Scan response data to stack */
		ret = g_dbus_proxy_call_sync(proxy, "SetScanRespData",
				g_variant_new("(@ayi)", temp, adv_param_setup.server_if),
				G_DBUS_CALL_FLAGS_NONE,
				-1, NULL, &error);
	}

	if (error) {
		ERR("SetAdvertisingData Fail: %s", error->message);
		g_clear_error(&error);
		return BT_STATUS_FAIL;
	}

	INFO("Set advertising data");
	if (ret)
		g_variant_unref(ret);

	/*
	 * As we need to provide async callback to user from HAL, simply schedule a
	 * callback method which will carry actual result
	 */
	event = g_malloc0(sizeof(bt_hal_adv_event_data_t));
	event->event  = BT_HAL_MULTI_ADV_DATA_EVT;
	event->server_if = adv_param_setup.server_if;
	event->status = BT_STATUS_SUCCESS;
	memcpy(&event->data, adv_data, 31);
	g_idle_add(__bt_hal_adv_event_cb, (gpointer)event);

	return BT_STATUS_SUCCESS;
}

int _bt_hal_adapter_le_start_scan(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *ret;

	DBG("+");

	/* TODO: Check adapter and LE adapter status */
	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	ret = g_dbus_proxy_call_sync(proxy, "StartLEDiscovery",
			NULL, G_DBUS_CALL_FLAGS_NONE,
			-1, NULL, &error);
	if (ret == NULL) {
		if (error) {
			ERR("StartLEDiscovery Fail: %s", error->message);
			g_clear_error(&error);
		}

		return BT_STATUS_FAIL;
	}

	g_variant_unref(ret);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

int _bt_hal_adapter_le_stop_scan(void)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *ret;

	DBG("+");

	/* TODO: Check adapter and LE adapter status */
	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	ret = g_dbus_proxy_call_sync(proxy, "StopLEDiscovery",
			NULL, G_DBUS_CALL_FLAGS_NONE,
			-1, NULL, &error);
	if (ret == NULL) {
		if (error) {
			g_dbus_error_strip_remote_error(error);
			ERR("StopLEDiscovery Fail: %s", error->message);

			/* Abnormal case for ARTIK530 */
			if (g_strrstr(error->message, "No discovery started") ||
				 g_strrstr(error->message, "Operation already in progress")) {
				g_clear_error(&error);
				return BT_STATUS_SUCCESS;
			}

			g_clear_error(&error);
			return BT_STATUS_FAIL;
		}
	}

	g_variant_unref(ret);

	DBG("-");
	return BT_STATUS_SUCCESS;
}


int _bt_hal_adapter_le_set_scan_parameters(
		int scan_type, int scan_interval, int scan_window)
{
	GDBusProxy *proxy;
	GError *error = NULL;
	GVariant *ret;

	DBG("+");

	/* TODO: Check adapter and LE adapter status */
	proxy = _bt_hal_get_adapter_proxy();
	if (proxy == NULL)
		return BT_STATUS_FAIL;

	ret = g_dbus_proxy_call_sync(proxy, "SetScanParameters",
			g_variant_new("(uuu)", scan_type, scan_interval, scan_window),
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
	if (ret == NULL) {
		if (error) {
			ERR("SetScanParameters Fail: %s", error->message);
			g_clear_error(&error);
		}

		return BT_STATUS_FAIL;
	}

	g_variant_unref(ret);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

/* To send stack event to hal-av handler */
void _bt_hal_register_gatt_le_dbus_handler_cb(handle_stack_msg cb)
{
	gatt_le_event_cb = cb;
}

void _bt_hal_unregister_gatt_le_dbus_handler_cb(void)
{
	gatt_le_event_cb = NULL;
}
