/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <dbus/dbus.h>
#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <vconf.h>

#include "bt-hal-dbus-common-utils.h"
#include "bt-hal-internal.h"
#include "bt-hal-avrcp-tg-dbus-handler.h"

#define HAL_ERROR_INTERNAL "InternalError"
#define HAL_ERROR_INVALID_PARAM "InvalidParameters"
#define HAL_ERROR_INVALID_INTERFACE "InvalidInterface"

#define BT_HAL_MEDIA_OBJECT_PATH "/Musicplayer"
#define BT_HAL_MEDIA_PLAYER_INTERFACE "org.mpris.MediaPlayer2.Player"

typedef enum {
	HAL_AVRCP_ERROR_NONE,
	HAL_AVRCP_ERROR_INTERNAL,
	HAL_AVRCP_ERROR_INVALID_PARAM,
	HAL_AVRCP_ERROR_NOT_SUPPORTED,
	HAL_AVRCP_ERROR_INVALID_INTERFACE
} hal_avrcp_error_t;

static handle_stack_msg event_cb = NULL;

static guint avrcp_reg_id = 0;

/* Introspection data exposed from bt-service */
static const gchar hal_avrcp_tg_introspection_xml[] =
"<node name='/'>"
" <interface name='org.freedesktop.DBus.Properties'>"
"     <method name='Set'>"
"          <arg type='s' name='interface' direction='in'/>"
"          <arg type='s' name='property' direction='in'/>"
"          <arg type='v' name='value' direction='in'/>"
"     </method>"
" </interface>"
"</node>";

static const char *__repeat_val_to_str(btrc_player_repeat_val_t repeat_val)
{
	switch (repeat_val) {
	case BTRC_PLAYER_VAL_OFF_REPEAT:
		return "None";
	case BTRC_PLAYER_VAL_SINGLE_REPEAT:
		return "Track";
	case BTRC_PLAYER_VAL_ALL_REPEAT:
		return "Playlist";
	case BTRC_PLAYER_VAL_GROUP_REPEAT:
		return "Playlist";
	default:
		ERR("Invalid repeat value: %d", repeat_val);
		return "";
	}
}

#if 0
static const char *__shuffle_val_to_str(btrc_player_shuffle_val_t shuffle_val)
{
	switch (shuffle_val) {
	case BTRC_PLAYER_VAL_OFF_SHUFFLE:
		return "off";
	case BTRC_PLAYER_VAL_ALL_SHUFFLE:
		return "alltracks";
	case BTRC_PLAYER_VAL_GROUP_SHUFFLE:
		return "group";
	default:
		ERR("Invalid shuffle value: %d", shuffle_val);
		return "";
	}
}
#endif

static const char *__play_status_to_str(btrc_play_status_t play_status)
{
	switch (play_status) {
	case BTRC_PLAYSTATE_STOPPED:
		return "stopped";
	case BTRC_PLAYSTATE_PLAYING:
		return "playing";
	case BTRC_PLAYSTATE_PAUSED:
		return "paused";
	case BTRC_PLAYSTATE_FWD_SEEK:
		return "forward-seek";
	case BTRC_PLAYSTATE_REV_SEEK:
		return "reverse-seek";
	case BTRC_PLAYSTATE_ERROR:
		return "error";
	default:
		ERR("Invalid play status: %d", play_status);
		return "";
	}
}

static GQuark __hal_avrcp_error_quark(void)
{
	static GQuark quark = 0;

	if (!quark)
		quark = g_quark_from_static_string("bt-avrcp");

	return quark;
}

static GError *__hal_avrcp_set_error(hal_avrcp_error_t error)
{
	ERR("error[%d]\n", error);

	switch (error) {
	case HAL_AVRCP_ERROR_INVALID_PARAM:
		return g_error_new(__hal_avrcp_error_quark(),
				error, HAL_ERROR_INVALID_PARAM);
	case HAL_AVRCP_ERROR_INVALID_INTERFACE:
		return g_error_new(__hal_avrcp_error_quark(),
				error, HAL_ERROR_INVALID_INTERFACE);
	case HAL_AVRCP_ERROR_INTERNAL:
	default:
		return g_error_new(__hal_avrcp_error_quark(),
				error, HAL_ERROR_INTERNAL);
	}
}

static void __hal_avrcp_agent_method(GDBusConnection *connection,
		const gchar *sender,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *method_name,
		GVariant *parameters,
		GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	int ret = HAL_AVRCP_ERROR_NONE;
	GError *err = NULL;
	gboolean shuffle_status;
	gchar *interface = NULL;
	gchar *property = NULL;
	gchar *loop_status = NULL;
	GVariant *value = NULL;

	DBG("+");
	INFO("method %s", method_name);
	INFO("object_path %s", object_path);

	if (g_strcmp0(method_name, "Set") == 0) {
		g_variant_get(parameters, "(&s&sv)",
				&interface, &property, &value);

		if (g_strcmp0(interface, BT_HAL_MEDIA_PLAYER_INTERFACE) != 0) {
			ret = HAL_AVRCP_ERROR_INVALID_INTERFACE;
			goto fail;
		}
	}

	if (value == NULL) {
		ERR("value is NULL");
		ret = HAL_AVRCP_ERROR_INVALID_PARAM;
		goto fail;
	}

	DBG("Property: %s\n", property);
	if (g_strcmp0(property, "Shuffle") == 0) {
		struct hal_ev_avrcp_tg_player_property ev;

		if (!g_variant_is_of_type(value, G_VARIANT_TYPE_BOOLEAN)) {
			ERR("Error");
			ret = HAL_AVRCP_ERROR_INVALID_PARAM;
			goto fail;
		}

		/* Prepare to send AVRCP player property event */
		memset(&ev, 0, sizeof(ev));
		shuffle_status = g_variant_get_boolean(value);
		DBG("Value: %s\n", shuffle_status ? "TRUE" : "FALSE");

		ev.prop_type = HAL_AVRCP_TG_PLAYER_PROP_SHUFFLE;
		if (shuffle_status)
			ev.value = BTRC_PLAYER_VAL_ALL_SHUFFLE;
		else
			ev.value = BTRC_PLAYER_VAL_OFF_SHUFFLE;

		if (!event_cb)
			ERR("AVRCP target DBUS handler callback not registered");
		else
			event_cb(HAL_EV_AVRCP_TG_SET_PLAYER_PROPERTY, (void *)&ev, sizeof(ev));
	} else if (g_strcmp0(property, "LoopStatus") == 0) {
		struct hal_ev_avrcp_tg_player_property ev;

		if (!g_variant_is_of_type(value, G_VARIANT_TYPE_STRING)) {
			ERR("Error");
			ret = HAL_AVRCP_ERROR_INVALID_PARAM;
			goto fail;
		}

		loop_status = (gchar *)g_variant_get_string(value, NULL);
		DBG("Value: %s\n", loop_status);

		/* Prepare to send AVRCP player property event */
		memset(&ev, 0, sizeof(ev));
		ev.prop_type = HAL_AVRCP_TG_PLAYER_PROP_REPEAT;
		if (g_strcmp0(loop_status, "Track") == 0)
			ev.value = BTRC_PLAYER_VAL_SINGLE_REPEAT;
		else if (g_strcmp0(loop_status, "Playlist") == 0)
			ev.value = BTRC_PLAYER_VAL_ALL_REPEAT;
		else
			ev.value = BTRC_PLAYER_VAL_OFF_REPEAT;

		if (!event_cb)
			ERR("AVRCP target DBUS handler callback not registered");
		else
			event_cb(HAL_EV_AVRCP_TG_SET_PLAYER_PROPERTY, (void *)&ev, sizeof(ev));

	}

	DBG("-");
	return;

fail:
	if (value)
		g_variant_unref(value);
	err = __hal_avrcp_set_error(ret);
	g_dbus_method_invocation_return_gerror(invocation, err);
	g_clear_error(&err);
}

static const GDBusInterfaceVTable method_table = {
	__hal_avrcp_agent_method,
	NULL,
	NULL,
};

static GDBusNodeInfo *__hal_avrcp_create_method_node_info
				(const gchar *introspection_data)
{
	GError *err = NULL;
	GDBusNodeInfo *node_info = NULL;

	if (introspection_data == NULL)
		return NULL;

	node_info = g_dbus_node_info_new_for_xml(introspection_data, &err);

	if (err) {
		ERR("Unable to create node: %s", err->message);
		g_clear_error(&err);
	}

	return node_info;
}

static void __bt_hal_avrcp_unregister_object_path(void)
{
	GDBusConnection *conn;

	conn = _bt_hal_get_system_gconn();
	if (conn == NULL) {
		ERR("conn == NULL, return");
		return;
	}

	if (avrcp_reg_id > 0) {
		g_dbus_connection_unregister_object(conn, avrcp_reg_id);
		avrcp_reg_id = 0;
	}
}

bt_status_t _bt_hal_dbus_handler_register_media_player(void)
{
	GDBusConnection *conn;
	gchar *adapter_path;
	gchar *path;
	GDBusNodeInfo *node_info;
	GDBusProxy *proxy;
	GVariantBuilder *builder;
	GVariant *ret;
	GError *error = NULL;

	DBG("+");

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	node_info = __hal_avrcp_create_method_node_info(
			hal_avrcp_tg_introspection_xml);
	if (node_info == NULL) {
		ERR("__hal_avrcp_create_method_node_info failed");
		return BT_STATUS_FAIL;
	}

	avrcp_reg_id = g_dbus_connection_register_object(conn,
			BT_HAL_MEDIA_OBJECT_PATH,
			node_info->interfaces[0],
			&method_table,
			NULL, NULL, &error);
	g_dbus_node_info_unref(node_info);

	if (avrcp_reg_id == 0) {
		ERR("Failed to register: %s", error->message);
		g_clear_error(&error);
		return BT_STATUS_FAIL;
	}

	adapter_path = _bt_hal_get_adapter_path();
	if (!adapter_path) {
		ERR("Could not get adapter path");
		return BT_STATUS_FAIL;
	}

	proxy =  g_dbus_proxy_new_sync(conn,
			G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_HAL_BLUEZ_NAME, adapter_path,
			BT_HAL_MEDIA_INTERFACE, NULL, &error);
	g_free(adapter_path);
	if (proxy == NULL) {
		ERR("Unable to create proxy");
		if (error) {
			ERR("Error: %s", error->message);
			g_clear_error(&error);
		}

		return BT_STATUS_FAIL;
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

	g_variant_builder_add(builder, "{sv}", "LoopStatus",
			g_variant_new("s", __repeat_val_to_str(BTRC_PLAYER_VAL_OFF_REPEAT)));
	g_variant_builder_add(builder, "{sv}", "Shuffle", g_variant_new("b", FALSE));
	g_variant_builder_add(builder, "{sv}", "PlaybackStatus",
			g_variant_new("s", __play_status_to_str(BTRC_PLAYSTATE_STOPPED)));
	g_variant_builder_add(builder, "{sv}", "Position", g_variant_new("u", 0));

	path = g_strdup(BT_HAL_MEDIA_OBJECT_PATH);
	ret = g_dbus_proxy_call_sync(proxy, "RegisterPlayer",
			g_variant_new("(oa{sv})", path, builder),
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	g_variant_builder_unref(builder);
	g_object_unref(proxy);
	g_free(path);
	if (ret == NULL) {
		ERR("Call RegisterPlayer Failed");
		if (error) {
			ERR("errCode[%x], message[%s]",
					error->code, error->message);
			g_clear_error(&error);
		}
		return BT_STATUS_FAIL;
	}

	g_variant_unref(ret);
	DBG("-");

	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_unregister_media_player(void)
{
	GDBusConnection *conn;
	GDBusProxy *proxy;
	gchar *adapter_path;
	GVariant *ret;
	GError *error = NULL;
	gchar *path;
	int result = BT_STATUS_SUCCESS;

	DBG("+");

	adapter_path = _bt_hal_get_adapter_path();
	if (adapter_path == NULL) {
		result = BT_STATUS_FAIL;
		goto FAIL;
	}

	conn = _bt_hal_get_system_gconn();
	if (conn == NULL) {
		ERR("conn is NULL");
		g_free(adapter_path);
		result = BT_STATUS_FAIL;
		goto FAIL;
	}

	proxy =  g_dbus_proxy_new_sync(conn,
			G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_HAL_BLUEZ_NAME, adapter_path,
			BT_HAL_MEDIA_INTERFACE, NULL, &error);
	g_free(adapter_path);
	if (proxy == NULL) {
		ERR("Unable to create proxy");
		if (error) {
			ERR("Error: %s", error->message);
			g_clear_error(&error);
		}
		result = BT_STATUS_FAIL;
		goto FAIL;
	}

	path = g_strdup(BT_HAL_MEDIA_OBJECT_PATH);
	DBG("path is [%s]", path);
	ret = g_dbus_proxy_call_sync(proxy, "UnregisterPlayer",
			g_variant_new("(o)", path),
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &error);
	g_free(path);
	g_object_unref(proxy);
	if (ret == NULL) {
		ERR("UnregisterPlayer failed");
		if (error) {
			ERR("D-Bus API failure: errCode[%x], message[%s]",
					error->code, error->message);
			g_clear_error(&error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(ret);

FAIL:
	__bt_hal_avrcp_unregister_object_path();

	DBG("-");
	return result;
}

#ifdef TIZEN_BT_HAL
static void __bt_hal_avrcp_tg_connect_cb(
		GDBusProxy *proxy, GAsyncResult *res, gpointer user_data)
{
	struct hal_ev_avrcp_tg_conn_state ev;
	char *address = user_data;
	int result = BT_STATUS_SUCCESS;
	GVariant *reply = NULL;
	GError *g_error = NULL;

	DBG("+");

	reply = g_dbus_proxy_call_finish(proxy, res, &g_error);
	g_object_unref(proxy);
	if (reply == NULL) {
		ERR("AVRCP Target Connect Dbus Call Error");
		if (g_error) {
			ERR("Error: %s\n", g_error->message);
			g_clear_error(&g_error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(reply);

	DBG("Address: %s", address);

	/*
	 * If result is success, Remote device connected event will be triggered
	 * automatically from stack, so return from here.
	 */
	if (result == BT_STATUS_SUCCESS)
		goto done;

	/* Prepare to send AVRCP Target connection state event */
	memset(&ev, 0, sizeof(ev));
	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
	ev.state = HAL_AVRCP_TG_STATE_DISCONNECTED;
	if (!event_cb)
		ERR("AVRCP target DBUS handler callback not registered");
	else
		event_cb(HAL_EV_AVRCP_TG_CONN_STATE, (void *)&ev, sizeof(ev));
done:
	g_free(address);
	DBG("-");
}

bt_status_t _bt_hal_dbus_handler_avrcp_tg_connect(bt_bdaddr_t *bd_addr)
{
	GDBusConnection *conn;
	char *address;
	int ret;

	if (!bd_addr) {
		ERR("bd_addr is NULL, return");
		return BT_STATUS_PARM_INVALID;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE * sizeof(char));
	if (!address) {
		ERR("Memory allocation failed");
		return BT_STATUS_NOMEM;
	}
	_bt_hal_convert_addr_type_to_string(address, bd_addr->address);

	ret = _bt_hal_connect_profile(address, AVRCP_CTRL_UUID,
			__bt_hal_avrcp_tg_connect_cb, address);
	if (ret != BT_HAL_ERROR_NONE) {
		ERR("_bt_hal_connect_profile(AVRCP Controller) Error");
		g_free(address);
		return BT_STATUS_FAIL;
	}

	return BT_STATUS_SUCCESS;
}

static void __bt_avrcp_tg_disconnect_cb(
		GDBusProxy *proxy, GAsyncResult *res, gpointer user_data)
{
	struct hal_ev_avrcp_tg_conn_state ev;
	char *address = user_data;
	int result = BT_STATUS_SUCCESS;
	GError *g_error = NULL;
	GVariant *reply = NULL;

	DBG("+");

	reply = g_dbus_proxy_call_finish(proxy, res, &g_error);
	g_object_unref(proxy);
	if (reply == NULL) {
		ERR("AVRCP Target Disconnect Dbus Call Error");
		if (g_error) {
			ERR("Error: %s\n", g_error->message);
			g_clear_error(&g_error);
		}
		result = BT_STATUS_FAIL;
	}
	g_variant_unref(reply);

	/*
	 * If result is success, Remote device disconnected event will be triggered
	 * automatically from stack, so return from here.
	 */
	if (BT_STATUS_FAIL != result) {
		DBG("AVRCP TG Disconnected for Device: %s", address);
		g_free(address);
		return;
	}

	DBG("AVRCP TG Disconnect failed for Device: %s", address);
	/* Prepare to send AVRCP Target disconnection state event */
	memset(&ev, 0, sizeof(ev));
	_bt_hal_convert_addr_string_to_type(ev.bdaddr, address);
	ev.state = HAL_AVRCP_TG_STATE_CONNECTED;
	if (!event_cb)
		ERR("AVRCP Target DBUS handler callback not registered");
	else
		event_cb(HAL_EV_AVRCP_TG_CONN_STATE, (void *)&ev, sizeof(ev));

	g_free(address);
	DBG("-");
}

bt_status_t _bt_hal_dbus_handler_avrcp_tg_disconnect(bt_bdaddr_t *bd_addr)
{
	GDBusConnection *conn;
	char *address;
	int ret;

	if (!bd_addr) {
		ERR("bd_addr is NULL, return");
		return BT_STATUS_PARM_INVALID;
	}

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	address = g_malloc0(BT_HAL_ADDRESS_STRING_SIZE * sizeof(char));
	if (!address) {
		ERR("Memory allocation failed");
		return BT_STATUS_NOMEM;
	}
	_bt_hal_convert_addr_type_to_string(address, bd_addr->address);

	ret = _bt_hal_disconnect_profile(address, AVRCP_TARGET_UUID,
			__bt_avrcp_tg_disconnect_cb, address);
	if (ret != BT_HAL_ERROR_NONE) {
		ERR("_bt_hal_connect_profile(AVRCP Target) Error");
		return BT_STATUS_FAIL;
	}

	return BT_STATUS_SUCCESS;
}
#endif

static gboolean __hal_media_emit_property_changed(GDBusConnection *connection,
		const char *path, const char *interface, const char *name,
		const GVariant *variant)
{
	GVariantBuilder *builder = NULL;
	GVariantBuilder *invalid_builder = NULL;
	GError *error = NULL;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	g_variant_builder_add(builder, "{sv}", name, variant);

	invalid_builder = g_variant_builder_new(G_VARIANT_TYPE("as"));

	g_dbus_connection_emit_signal(connection, NULL, path,
				DBUS_INTERFACE_PROPERTIES,
				"PropertiesChanged",
				g_variant_new("(sa{sv}as)",
				interface, builder, invalid_builder),
				&error);
	g_variant_builder_unref(builder);
	g_variant_builder_unref(invalid_builder);
	if (error) {
		ERR("Could not Emit PropertiesChanged Signal: errCode[%x], message[%s]",
			error->code, error->message);
		g_clear_error(&error);
		return FALSE;
	}

	return TRUE;
}

bt_status_t _bt_hal_dbus_handler_avrcp_tg_play_status_changed(btrc_play_status_t play_status)
{
	GDBusConnection *conn;

	DBG("+");

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	if (!__hal_media_emit_property_changed(conn, BT_HAL_MEDIA_OBJECT_PATH,
			BT_HAL_MEDIA_PLAYER_INTERFACE, "PlaybackStatus",
			g_variant_new_string(__play_status_to_str(play_status)))) {
		ERR("Error sending the PropertyChanged signal");
		return BT_STATUS_FAIL;
	}

	DBG("-");
	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_avrcp_tg_play_pos_changed(uint32_t song_pos)
{
	GDBusConnection *conn;

	DBG("+");

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	if (!__hal_media_emit_property_changed(conn, BT_HAL_MEDIA_OBJECT_PATH,
			BT_HAL_MEDIA_PLAYER_INTERFACE, "Position",
			g_variant_new_uint32(song_pos))) {
		ERR("Error sending the PropertyChanged signal");
		return BT_STATUS_FAIL;
	}

	DBG("-");
	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_avrcp_tg_app_setting_changed(btrc_player_settings_t *player_setting)
{
	GDBusConnection *conn;
	char *property_name = NULL;
	GVariant *variant;
	int i;

	DBG("+");

	conn = _bt_hal_get_system_gconn();
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	for (i = 0; i < player_setting->num_attr; i++) {
		property_name = NULL;
		variant = NULL;

		switch (player_setting->attr_ids[i]) {
		case BTRC_PLAYER_ATTR_REPEAT:
			property_name = "LoopStatus";
			variant = g_variant_new_string(
					__repeat_val_to_str(player_setting->attr_values[i]));
			break;
		case BTRC_PLAYER_ATTR_SHUFFLE:
			property_name = "Shuffle";
			if (BTRC_PLAYER_VAL_OFF_SHUFFLE == player_setting->attr_values[i])
				variant = g_variant_new_boolean(FALSE);
			else
				variant = g_variant_new_boolean(TRUE);
			break;
		case BTRC_PLAYER_ATTR_SCAN:
		case BTRC_PLAYER_ATTR_EQUALIZER:
		default:
			ERR("Unsupported attr type: %d", player_setting->attr_ids[i]);
		}

		if (!property_name)
			continue;

		if (!__hal_media_emit_property_changed(conn, BT_HAL_MEDIA_OBJECT_PATH,
					BT_HAL_MEDIA_PLAYER_INTERFACE, property_name, variant)) {
			ERR("Error sending the PropertyChanged signal");
			return BT_STATUS_FAIL;
		}
	}

	DBG("-");
	return BT_STATUS_SUCCESS;
}

bt_status_t _bt_hal_dbus_handler_avrcp_tg_set_track_info(
		uint8_t num_attr, btrc_element_attr_val_t *p_attrs)
{
	GDBusConnection *conn;
	GError *error = NULL;
	GVariantBuilder *invalid_builder = NULL;
	GVariantBuilder *builder = NULL;
	GVariantBuilder *inner_builder = NULL;
	char *interface = BT_HAL_MEDIA_PLAYER_INTERFACE;
	gboolean ret;
	int i;

	DBG("+");

	if (!p_attrs) {
		ERR("p_attrs is NULL");
		return BT_STATUS_FAIL;
	}

	conn = _bt_hal_get_system_gconn();;
	if (!conn) {
		ERR("_bt_hal_get_system_gconn returned NULL, return");
		return BT_STATUS_FAIL;
	}

	invalid_builder = g_variant_builder_new(G_VARIANT_TYPE("as"));
	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	inner_builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

	for (i = 0; i < num_attr; i++) {
		switch (p_attrs[i].attr_id) {
		case BTRC_MEDIA_ATTR_ID_TITLE:
			INFO("Title: %s", p_attrs[i].text);
			g_variant_builder_add(inner_builder, "{sv}",
					"xesam:title", g_variant_new_string((const char *)p_attrs[i].text));
			break;
		case BTRC_MEDIA_ATTR_ID_ARTIST: {
			GVariant *children[1];

			INFO("Artist: %s", p_attrs[i].text);
			children[0] = g_variant_new_string((const char *)p_attrs[i].text);
			g_variant_builder_add(inner_builder, "{sv}", "xesam:artist",
					g_variant_new_array(G_VARIANT_TYPE_STRING, children, 1));
			break;
		}
		case BTRC_MEDIA_ATTR_ID_ALBUM:
			INFO("Album: %s", p_attrs[i].text);
			g_variant_builder_add(inner_builder, "{sv}",
					"xesam:album", g_variant_new_string((const char *)p_attrs[i].text));
			break;
		case BTRC_MEDIA_ATTR_ID_TRACK_NUM: {
			uint32_t cur_track = 0;

			sscanf((char *)p_attrs[i].text, "%u", &cur_track);
			INFO("Current track: %u", cur_track);
			g_variant_builder_add(inner_builder, "{sv}",
					"xesam:trackNumber", g_variant_new_int32(cur_track));
			break;
		}
		case BTRC_MEDIA_ATTR_ID_NUM_TRACKS: {
			uint32_t num_tracks = 0;

			sscanf((char *)p_attrs[i].text, "%u", &num_tracks);
			INFO("Total tracks: %u", num_tracks);
			g_variant_builder_add(inner_builder, "{sv}",
					"xesam:totalTracks", g_variant_new_int32(num_tracks));
			break;
		}
		case BTRC_MEDIA_ATTR_ID_GENRE: {
			GVariant *children[1];

			INFO("Genre: %s", p_attrs[i].text);
			children[0] = g_variant_new_string((const char *)p_attrs[i].text);
			g_variant_builder_add(inner_builder, "{sv}", "xesam:genre",
					g_variant_new_array(G_VARIANT_TYPE_STRING, children, 1));
			break;
		}
		case BTRC_MEDIA_ATTR_ID_PLAYING_TIME: {
			uint32_t duration = 0;

			sscanf((char *)p_attrs[i].text, "%u", &duration);
			INFO("Song duration: %u", duration);
			g_variant_builder_add(inner_builder, "{sv}",
					"mpris:length", g_variant_new_int64(duration));
			break;
		}
		default:
			INFO("Unknown attribute Id: %d", p_attrs[i].attr_id);
		}
	}

	g_variant_builder_add(builder, "{sv}",
			"Metadata", g_variant_new("a{sv}", inner_builder));
	ret = g_dbus_connection_emit_signal(conn, NULL, BT_HAL_MEDIA_OBJECT_PATH,
			DBUS_INTERFACE_PROPERTIES,
			"PropertiesChanged",
			g_variant_new("(sa{sv}as)",
				interface, builder, invalid_builder),
			&error);
	g_variant_builder_unref(inner_builder);
	g_variant_builder_unref(builder);
	g_variant_builder_unref(invalid_builder);
	if (!ret) {
		if (error != NULL) {
			ERR("D-Bus API failure: errCode[%x], message[%s]",
					error->code, error->message);
			g_clear_error(&error);
		}
		return BT_STATUS_FAIL;
	}

	DBG("-");

	return BT_STATUS_SUCCESS;
}

/* To send stack event to hal-avrcp-tg handler */
void _bt_hal_register_avrcp_tg_dbus_handler_cb(handle_stack_msg cb)
{
	event_cb = cb;
}

/* To send stack event to hal-avrcp-tg handler */
void _bt_hal_unregister_avrcp_tg_dbus_handler_cb()
{
	event_cb = NULL;
}
