/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 -2016 Samsung Electronics Co., Ltd All Rights Reserved.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"

#include <bt-hal-adapter-dbus-handler.h>
#include <bt-hal-device-dbus-handler.h>
#include <bt-hal-hidhost.h>
#include <bt-hal-socket.h>
#include <bt-hal-av.h>
#include <bt-hal-hf.h>
#include <bt-hal-hdp.h>
#include <bt-hal-a2dp-sink.h>
#include <bt-hal-avrcp-tg.h>
#include <bt-hal-avrcp-ctrl.h>
#include <bt-hal-gatt.h>
#ifdef TIZEN_BT_HAL
#include <bt-hal-adapter-le.h>
#include <bt-hal-agent.h>
#endif

#define enum_prop_to_hal(prop, hal_prop, type) do { \
	static type e; \
	prop.val = &e; \
	prop.len = sizeof(e); \
	e = *((uint8_t *) (hal_prop->val)); \
} while (0)

static const bt_callbacks_t *bt_hal_cbacks = NULL;


/* Forward declarations */
static void __bt_adapter_props_to_hal(bt_property_t *send_props, struct hal_property *prop, uint8_t num_props, uint16_t len);
static void __bt_device_props_to_hal(bt_property_t *send_props,
		struct hal_property *prop, uint8_t num_props,
		uint16_t len);
static void __bt_hal_handle_adapter_state_changed(void *buf, uint16_t len);
static void __bt_hal_handle_adapter_property_changed(void *buf, uint16_t len);
static void __bt_hal_handle_stack_messages(int message, void *buf, uint16_t len);
static void __bt_hal_handle_adapter_discovery_state_changed(void *buf, uint16_t len);
static void __bt_hal_handle_device_found_event(void *buf, uint16_t len);
static void __bt_hal_handle_bond_state_changed_event(void *buf, uint16_t len);
static void __bt_hal_handle_device_acl_state_changed_event(void *buf, uint16_t len);
static void __bt_hal_handle_ssp_request_event(void *buf, uint16_t len);
static void __bt_hal_handle_pin_request_event(void *buf, uint16_t len);

static void __bt_hal_handle_device_trust_state_changed_event(void *buf, uint16_t len);

static bool interface_ready(void)
{
	return bt_hal_cbacks != NULL;
}

static int init(bt_callbacks_t *callbacks)
{
	int ret;
	DBG("HAL library Initialization..");

	if (interface_ready())
		return BT_STATUS_DONE;
	else {
		bt_hal_cbacks = callbacks;
		DBG("Store HAL stack msg handler callback");
		_bt_hal_dbus_store_stack_msg_cb(__bt_hal_handle_stack_messages);
		ret = _bt_hal_initialize_event_receiver(__bt_hal_handle_stack_messages);

		if (ret == BT_STATUS_SUCCESS)
			return BT_STATUS_SUCCESS;
		else
			return BT_STATUS_FAIL;

	}
	return BT_STATUS_SUCCESS;
}

/* Enable Adapter */
static int enable(void)
{
	return _bt_hal_dbus_enable_adapter();
}

/* Disable Adapter */
static int disable(void)
{
	return _bt_hal_dbus_disable_adapter();
}

#ifdef  TIZEN_BT_HAL
/* Tizen Specific Reset Adapter */
static int reset(void)
{
	return _bt_hal_dbus_reset_adapter();
}
#endif

static void cleanup(void)
{
	return;
}

#ifdef  TIZEN_BT_HAL
/* Tizen Specific Get Powered state */
static int get_adapter_powered_status(uint8_t *state)
{
	if (!state) {
		ERR("Invalid param");
		return BT_STATUS_PARM_INVALID;
	}

	return _bt_hal_get_adapter_powered_state(state);
}
#endif

static int get_adapter_properties(void)
{
	return _bt_hal_dbus_get_adapter_properties();
}

static int get_adapter_property(bt_property_type_t type)
{
	return _bt_hal_dbus_get_adapter_property(type);
}

static int set_adapter_property(const bt_property_t *property)
{
	if (!property) {
		ERR("Invalid param");
		return BT_STATUS_PARM_INVALID;
	}

	return _bt_hal_dbus_set_adapter_property(property);
}

static int get_remote_device_properties(bt_bdaddr_t *remote_addr)
{
	if (!remote_addr) {
		ERR("Invalid param");
		return BT_STATUS_PARM_INVALID;
	}

	return _bt_hal_dbus_get_remote_device_properties(remote_addr);
}

static int get_remote_device_property(bt_bdaddr_t *remote_addr,
		bt_property_type_t type)
{
	return BT_STATUS_UNSUPPORTED;
}

static int set_remote_device_property(bt_bdaddr_t *remote_addr,
		const bt_property_t *property)
{
	return _bt_hal_dbus_set_remote_device_property(remote_addr, property);
}

static int get_remote_service_record(bt_bdaddr_t *remote_addr, bt_uuid_t *uuid)
{
	return BT_STATUS_UNSUPPORTED;
}

static int get_remote_services(bt_bdaddr_t *remote_addr)
{
	if (!remote_addr) {
		ERR("Invalid param");
		return BT_STATUS_PARM_INVALID;
	}
	return _bt_hal_dbus_get_remote_device_services(remote_addr);
}

static int start_discovery(void)
{
	return _bt_hal_dbus_start_discovery(false, 0x00);
}

#ifdef  TIZEN_BT_HAL
static int start_custom_discovery(bt_disc_role_type_t disc_type)
{
	return _bt_hal_dbus_start_discovery(true, disc_type);
}
#endif

static int cancel_discovery(void)
{
	return _bt_hal_dbus_stop_discovery();
}

static int create_bond(const bt_bdaddr_t *bd_addr, int transport)
{
	DBG("+");
	return _bt_hal_device_create_bond(bd_addr, transport);
}

static int cancel_bond(const bt_bdaddr_t *bd_addr)
{
	return _bt_hal_device_cancel_bond(bd_addr);
}

static int remove_bond(const bt_bdaddr_t *bd_addr)
{
	DBG("+");
	return _bt_hal_device_remove_bond(bd_addr);
}

static int pin_reply(const bt_bdaddr_t *bd_addr, uint8_t accept,
		uint8_t pin_len, bt_pin_code_t *pin_code)
{
	return _bt_hal_device_legacy_pin_reply(bd_addr, accept, pin_len, (char*)pin_code);
}

static int ssp_reply(const bt_bdaddr_t *bd_addr, bt_ssp_variant_t variant,
		uint8_t accept, uint32_t passkey)
{
	return _bt_hal_device_ssp_reply(bd_addr, variant, accept, passkey);
}

static const void *get_profile_interface(const char *profile_id)
{
	/*TODO: Profile interfaces to be included later*/
	DBG("%s", profile_id);

	if (!interface_ready())
		return NULL;

	if (!strncmp(profile_id, BT_PROFILE_HIDHOST_ID, strlen(profile_id)))
		return bt_get_hidhost_interface();

	if (!strcmp(profile_id, BT_PROFILE_SOCKETS_ID))
		return bt_get_socket_interface();

	if (!strcmp(profile_id, BT_PROFILE_PAN_ID))
		return NULL;

	if (!strcmp(profile_id, BT_PROFILE_ADVANCED_AUDIO_ID))
		return bt_get_av_interface();

	if (!strcmp(profile_id, BT_PROFILE_AV_RC_ID))
		return bt_get_avrcp_tg_interface();

	if (!strcmp(profile_id, BT_PROFILE_HANDSFREE_ID))
		return bt_get_hf_interface();

	if (!strcmp(profile_id, BT_PROFILE_GATT_ID))
		return bt_get_gatt_interface();

	if (!strcmp(profile_id, BT_PROFILE_HEALTH_ID))
		return bt_get_hl_interface();

	if (!strcmp(profile_id, BT_PROFILE_AV_RC_CTRL_ID))
		return bt_get_avrcp_ctrl_interface();

	if (!strcmp(profile_id, BT_PROFILE_HANDSFREE_CLIENT_ID))
		return NULL;

	if (!strcmp(profile_id, BT_PROFILE_MAP_CLIENT_ID))
		return NULL;

	if (!strcmp(profile_id, BT_PROFILE_ADVANCED_AUDIO_SINK_ID))
		return bt_get_a2dp_sink_interface();

	return NULL;
}

static int dut_mode_configure(uint8_t enable)
{
	return BT_STATUS_UNSUPPORTED;
}

static int dut_mode_send(uint16_t opcode, uint8_t *buf, uint8_t buf_len)
{
	return BT_STATUS_UNSUPPORTED;
}

static int le_test_mode(uint16_t opcode, uint8_t *buf, uint8_t buf_len)
{
	return BT_STATUS_UNSUPPORTED;
}

static int config_hci_snoop_log(uint8_t enable)
{
	return BT_STATUS_UNSUPPORTED;
}

static int get_connection_state(const bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_device_get_connection_state(bd_addr);
}

static int set_os_callouts(bt_os_callouts_t *callouts)
{
	DBG("callouts: %p", callouts);

	/* TODO: implement */

	return BT_STATUS_UNSUPPORTED;
}

static int read_energy_info(void)
{
	return BT_STATUS_UNSUPPORTED;
}

static int authorize_response(const bt_bdaddr_t *bd_addr, bt_service_id_t service_id,
		uint8_t authorize, uint8_t save_settings)
{
	DBG("+");
	return _bt_hal_device_authorize_response(bd_addr, service_id, authorize, save_settings);
}

int set_authorization(bt_bdaddr_t *bd_addr, uint8_t auth)
{
	DBG("+");
	return _bt_hal_device_set_trust(bd_addr, auth);
}

#ifdef TIZEN_BT_HAL
static int get_service_connection_state(const bt_bdaddr_t *bd_addr, bt_service_id_t rem_svc_id)
{
	DBG("");
	return _bt_hal_device_get_service_connection_state(bd_addr, rem_svc_id);
}

static int set_agent_osp_server(uint32_t type, uint8_t enable)
{
	DBG("");
	return _bt_hal_device_set_osp_server(type, enable);
}

static int set_trusted_profile(bt_bdaddr_t *bd_addr, bt_trusted_profile_t profile, uint8_t trust)
{
	DBG("");
	return _bt_hal_device_set_trusted_profile(bd_addr, profile, trust);
}

static int get_trusted_profile(bt_bdaddr_t *bd_addr, bt_trusted_profile_t profile, uint32_t *trusted)
{
	DBG("");
	return _bt_hal_device_get_trusted_profile(bd_addr, profile, trusted);
}

static int get_connected_link_rssi_strength(bt_bdaddr_t *bd_addr, uint32_t conn_link_type)
{
	DBG("");
	return _bt_hal_device_get_connected_link_rssi_strength(bd_addr, conn_link_type);
}

static int enable_rssi_monitoring(bt_bdaddr_t *bd_addr, uint32_t conn_link_type,
		int low_threshold, int in_range_threshold, int high_threshold)
{
	DBG("");
	return _bt_hal_device_enable_rssi_monitoring(bd_addr, conn_link_type,
			low_threshold, in_range_threshold, high_threshold);
}

static int enable_gap_auth_notifications(uint32_t type, uint8_t enable)
{
	DBG("");
	_bt_hal_enable_gap_auth_notifications(type, ((enable == 0) ? FALSE : TRUE));
	return BT_STATUS_SUCCESS;
}

static int set_hal_adapter_request_state(int enable)
{
	DBG("");
	_bt_hal_set_adapter_request_state(enable);
	return BT_STATUS_SUCCESS;
}

static int set_hal_le_request_state(int enable)
{
	DBG("");
	_bt_hal_set_le_request_state(enable);
	return BT_STATUS_SUCCESS;
}
#endif

static const bt_interface_t bluetooth_if = {
	.size = sizeof(bt_interface_t),
	.init = init,
	.enable = enable,
	.disable = disable,
#ifdef TIZEN_BT_HAL
	.reset = reset,
#endif

#ifdef TIZEN_BT_HAL
	.le_enable = _bt_hal_le_enable,
	.le_disable = _bt_hal_le_disable,
	.le_init = _bt_hal_le_init,
	.le_deinit = _bt_hal_le_deinit,
	.is_advertising = _bt_hal_is_advertising,
#endif
	.cleanup = cleanup,
#ifdef TIZEN_BT_HAL
	.get_adapter_powered_status = get_adapter_powered_status,
#endif
	.get_adapter_properties = get_adapter_properties,
	.get_adapter_property = get_adapter_property,
	.set_adapter_property = set_adapter_property,
	.get_remote_device_properties = get_remote_device_properties,
	.get_remote_device_property = get_remote_device_property,
	.set_remote_device_property = set_remote_device_property,
	.get_remote_service_record = get_remote_service_record,
	.get_remote_services = get_remote_services,
	.start_discovery = start_discovery,
#ifdef TIZEN_BT_HAL
	.start_custom_discovery = start_custom_discovery,
#endif
	.cancel_discovery = cancel_discovery,
	.create_bond = create_bond,
	.remove_bond = remove_bond,
	.cancel_bond = cancel_bond,
	.pin_reply = pin_reply,
	.ssp_reply = ssp_reply,
	.get_profile_interface = get_profile_interface,
	.dut_mode_configure = dut_mode_configure,
	.dut_mode_send = dut_mode_send,
	.le_test_mode = le_test_mode,
	.config_hci_snoop_log = config_hci_snoop_log,
	.get_connection_state = get_connection_state,
	.set_os_callouts = set_os_callouts,
	.read_energy_info = read_energy_info,
	/* Tizen Specific API's*/
	.authorize_response = authorize_response,
	.set_authorization = set_authorization,
#ifdef TIZEN_BT_HAL
	.get_service_connection_state = get_service_connection_state,
	.set_agent_osp_server = set_agent_osp_server,
	.set_trusted_profile = set_trusted_profile,
	.get_trusted_profile = get_trusted_profile,
	.enable_rssi_monitoring = enable_rssi_monitoring,
	.get_connected_link_rssi_strength = get_connected_link_rssi_strength,
	.enable_gap_auth_notifications = enable_gap_auth_notifications,
	.set_le_static_random_address = _bt_hal_set_le_static_random_address,
	.set_hal_adapter_request_state = set_hal_adapter_request_state,
	.set_hal_le_request_state = set_hal_le_request_state,
#endif
};

static const bt_interface_t *get_bluetooth_interface(void)
{
	DBG("");
	return &bluetooth_if;
}

static int close_bluetooth(struct hw_device_t *device)
{
	DBG("");
	cleanup();
	free(device);
	return 0;
}

static int open_bluetooth(const struct hw_module_t *module, char const *name,
		struct hw_device_t **device)
{
	bluetooth_device_t *dev = g_malloc0(sizeof(bluetooth_device_t));

	DBG("");

	memset(dev, 0, sizeof(bluetooth_device_t));
	dev->common.tag = HARDWARE_DEVICE_TAG;
	dev->common.version = 0;
	dev->common.module = (struct hw_module_t *) module;
	dev->common.close = close_bluetooth;
	dev->get_bluetooth_interface = get_bluetooth_interface;

	*device = (struct hw_device_t *) dev;

	return 0;
}

static struct hw_module_methods_t bluetooth_module_methods = {
	.open = open_bluetooth,
};

struct hw_module_t HAL_MODULE_INFO_SYM = {
	.tag = HARDWARE_MODULE_TAG,
	.version_major = 1,
	.version_minor = 0,
	.id = BT_HARDWARE_MODULE_ID,
	.name = "Bluetooth stack",
	.author = "Intel Corporation",
	.methods = &bluetooth_module_methods
};

static void __bt_hal_handle_adapter_state_changed(void *buf, uint16_t len)
{
	struct hal_ev_adapter_state_changed *ev = buf;

	DBG("Adapter State: %d", ev->state);

	if (bt_hal_cbacks->adapter_state_changed_cb)
		bt_hal_cbacks->adapter_state_changed_cb(ev->state);
}

#ifdef TIZEN_BT_HAL
static void __bt_hal_handle_le_state_changed(void *buf, uint16_t len)
{
	struct hal_ev_le_state_changed *ev = buf;

	DBG("Adapter State: %d", ev->state);

	if (bt_hal_cbacks->le_state_changed_cb)
		bt_hal_cbacks->le_state_changed_cb(ev->state);
}
#endif

static void __bt_adapter_props_to_hal(bt_property_t *send_props, struct hal_property *prop,
		uint8_t num_props, uint16_t len)
{
	void *buf = prop;
	uint8_t i;

	for (i = 0; i < num_props; i++) {
		if (sizeof(*prop) + prop->len > len) {
			ERR("invalid adapter properties(%zu > %u), cant process further properties!!!",
					sizeof(*prop) + prop->len, len);
			return;
		}

		send_props[i].type = prop->type;

		switch (prop->type) {
			/* TODO: Add Adapter Properties */
		default:
			send_props[i].len = prop->len;
			send_props[i].val = prop->val;
			break;
		}

		DBG("prop[%d]: %s", i, btproperty2str(&send_props[i]));

		len -= sizeof(*prop) + prop->len;
		buf += sizeof(*prop) + prop->len;
		prop = buf;
	}

	if (!len)
		return;
}

static void __bt_device_props_to_hal(bt_property_t *send_props,
		struct hal_property *prop, uint8_t num_props,
		uint16_t len)
{
	void *buf = prop;
	uint8_t i;

	DBG("+");

	for (i = 0; i < num_props; i++) {

		if (sizeof(*prop) + prop->len > len) {
			ERR("invalid device properties (%zu > %u), cant process further properties!!!",
					sizeof(*prop) + prop->len, len);
			return;
		}

		send_props[i].type = prop->type;

		DBG("HAL prop Type [%d]", prop->type);

		switch (prop->type) {
		case HAL_PROP_DEVICE_TYPE:
		{
			DBG("Device property:HAL_PROP_DEVICE_TYPE:");
			enum_prop_to_hal(send_props[i], prop,
					bt_device_type_t);
			break;
		}
		case HAL_PROP_DEVICE_VERSION_INFO:
		{
			DBG("Device property: HAL_PROP_DEVICE_VERSION_INFO");
			static bt_remote_version_t e;
			const struct hal_prop_device_info *p;
			send_props[i].val = &e;
			send_props[i].len = sizeof(e);
				p = (struct hal_prop_device_info *) prop->val;
				e.manufacturer = p->manufacturer;
			e.sub_ver = p->sub_version;
			e.version = p->version;
			break;
		}
		case HAL_PROP_DEVICE_SERVICE_REC:
		{
			DBG("Device property: HAL_PROP_DEVICE_SERVICE_REC");
			static bt_service_record_t e;
			const struct hal_prop_device_service_rec *p;
			send_props[i].val = &e;
			send_props[i].len = sizeof(e);
				p = (struct hal_prop_device_service_rec *) prop->val;
					memset(&e, 0, sizeof(e));
			memcpy(&e.channel, &p->channel, sizeof(e.channel));
			memcpy(e.uuid.uu, p->uuid, sizeof(e.uuid.uu));
			memcpy(e.name, p->name, p->name_len);
			break;
		}
		default:
			send_props[i].len = prop->len;
			send_props[i].val = prop->val;
			break;
		}

		DBG("prop[%d]: %s, len: %d", i, btproperty2str(&send_props[i]), prop->len);
		len -= sizeof(*prop) + prop->len;
		buf += sizeof(*prop) + prop->len;
		prop = buf;

	}

	if (!len) {
		DBG("-");
		return;
	}

	ERR("invalid device properties (%u bytes left), ", len);
}

static void __bt_hal_handle_adapter_property_changed(void *buf, uint16_t len)
{
	struct hal_ev_adapter_props_changed *ev = (struct hal_ev_adapter_props_changed *)buf;
	bt_property_t props[ev->num_props];
	DBG("+");

	if (!bt_hal_cbacks->adapter_properties_cb)
		return;

	len -= sizeof(*ev);
	__bt_adapter_props_to_hal(props, ev->props, ev->num_props, len);

	if (bt_hal_cbacks->adapter_properties_cb)
		bt_hal_cbacks->adapter_properties_cb(ev->status, ev->num_props, props);
}

static void __bt_hal_handle_adapter_discovery_state_changed(void *buf, uint16_t len)
{
	struct hal_ev_discovery_state_changed *ev = (struct hal_ev_discovery_state_changed *)buf;

	DBG("+");

	if (bt_hal_cbacks->discovery_state_changed_cb)
		bt_hal_cbacks->discovery_state_changed_cb(ev->state);
}

static void __bt_hal_handle_device_found_event(void *buf, uint16_t len)
{
	struct hal_ev_device_found *ev =  (struct hal_ev_device_found *) buf;
	bt_property_t props[ev->num_props];
	DBG("+");

	if (!bt_hal_cbacks->device_found_cb)
		return;

	len -= sizeof(*ev);
	__bt_device_props_to_hal(props, ev->props, ev->num_props, len);

	bt_hal_cbacks->device_found_cb(ev->num_props, props);
}

static void __bt_hal_handle_remote_device_properties_event(void *buf, uint16_t len)
{
	struct hal_ev_remote_device_props *ev = (struct hal_ev_remote_device_props *) buf;
	bt_bdaddr_t bd_addr;
	bt_property_t props[ev->num_props];

	DBG("+");

	if (!bt_hal_cbacks->remote_device_properties_cb)
		return;

	len -= sizeof(*ev);
	__bt_device_props_to_hal(props, ev->props, ev->num_props, len);
	memcpy(bd_addr.address, ev->bdaddr, 6);
	bt_hal_cbacks->remote_device_properties_cb(
			ev->status, &bd_addr, ev->num_props, props);
	DBG("-");
}

static void __bt_hal_handle_bond_state_changed_event(void *buf, uint16_t len)
{

	struct hal_ev_bond_state_changed *ev = (struct hal_ev_bond_state_changed *)buf;
	bt_bdaddr_t bd_addr;
	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	/* BD address*/
	DBG("[0x%x]", bd_addr.address[0]);
	DBG("[0x%x]", bd_addr.address[1]);
	DBG("[0x%x]", bd_addr.address[2]);
	DBG("[0x%x]", bd_addr.address[3]);
	DBG("[0x%x]", bd_addr.address[4]);
	DBG("[0x%x]", bd_addr.address[5]);

	DBG("Bonding  State changed Status [0x%x]", ev->status);
	DBG("Bonding  State [0x%x]", ev->state);

	if (!bt_hal_cbacks->bond_state_changed_cb) {
		ERR("HAL User bond_state_changed_callback is not set!!");
		return;
	}

	bt_hal_cbacks->bond_state_changed_cb(ev->status, &bd_addr, ev->state);
	DBG("-");
}

static void __bt_hal_handle_device_acl_state_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_acl_state_changed *ev = (struct hal_ev_acl_state_changed*)buf;
	bt_bdaddr_t bd_addr;
	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	/* BD address*/
	DBG("[0x%x]", bd_addr.address[0]);
	DBG("[0x%x]", bd_addr.address[1]);
	DBG("[0x%x]", bd_addr.address[2]);
	DBG("[0x%x]", bd_addr.address[3]);
	DBG("[0x%x]", bd_addr.address[4]);
	DBG("[0x%x]", bd_addr.address[5]);

	DBG("ACL Status [0x%x]", ev->status);
	DBG("ACL State  [0x%x]", ev->state);

	if (!bt_hal_cbacks->acl_state_changed_cb) {
		ERR("HAL User acl_state_changed_cb is not set!!");
		return;
	}

	bt_hal_cbacks->acl_state_changed_cb(ev->status, &bd_addr, ev->state);
	DBG("-");
}

static void __bt_hal_handle_device_le_conn_state_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_acl_state_changed *ev = (struct hal_ev_acl_state_changed*)buf;
	bt_bdaddr_t bd_addr;
	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	/* BD address */
	DBG("Address: [%02X:%02X:%02X:%02X:%02X:%02X]", bd_addr.address[0],
			bd_addr.address[1], bd_addr.address[2], bd_addr.address[3],
			bd_addr.address[4], bd_addr.address[5]);
	DBG("LE conn Status [0x%x]", ev->status);
	DBG("LE conn State  [0x%x]", ev->state);

#ifdef TIZEN_BT_HAL
	if (!bt_hal_cbacks->le_conn_state_changed_cb) {
		ERR("HAL User le_conn_state_changed_cb is not set!!");
		return;
	}

	bt_hal_cbacks->le_conn_state_changed_cb(ev->status, &bd_addr, ev->state);
#endif
	DBG("-");
}

static void __bt_hal_handle_authorize_request_event(void *buf, uint16_t len)
{
	struct hal_ev_authorize_request *ev = (struct hal_ev_authorize_request *)buf;
	bt_bdaddr_t bd_addr;

	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);

	/* BD address*/
	DBG("[0x%x]", bd_addr.address[0]);
	DBG("[0x%x]", bd_addr.address[1]);
	DBG("[0x%x]", bd_addr.address[2]);
	DBG("[0x%x]", bd_addr.address[3]);
	DBG("[0x%x]", bd_addr.address[4]);
	DBG("[0x%x]", bd_addr.address[5]);

	DBG("Service Id: [%u]", ev->service_id);

	if (!bt_hal_cbacks->authorize_request_cb) {
		ERR("HAL User authorize_request_cb is not set!!");
		return;
	}

	bt_hal_cbacks->authorize_request_cb(&bd_addr, ev->service_id);
	DBG("-");
}

#ifdef TIZEN_BT_HAL
static void __bt_hal_handle_socket_authorize_request_event(void *buf, uint16_t len)
{
	struct hal_ev_sock_conn_auth *ev = buf;
	bt_bdaddr_t bd_addr;
	bt_uuid_t uuid;

	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	memcpy(uuid.uu, ev->uuid, 16);

	if (!bt_hal_cbacks->socket_authorize_request_cb) {
		ERR("HAL User socket_authorize_request_cb is not set!!");
		return;
	}

	bt_hal_cbacks->socket_authorize_request_cb(&bd_addr, &uuid);
	DBG("-");
}
#endif

static void __bt_hal_handle_ssp_request_event(void *buf, uint16_t len)
{
	struct hal_ev_ssp_request *ev = (struct hal_ev_ssp_request *)buf;
	bt_bdaddr_t bd_addr;
	bt_bdname_t bd_name;
	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	memcpy(bd_name.name, ev->name, 248);

	/* BD address*/
	DBG("[0x%x]", bd_addr.address[0]);
	DBG("[0x%x]", bd_addr.address[1]);
	DBG("[0x%x]", bd_addr.address[2]);
	DBG("[0x%x]", bd_addr.address[3]);
	DBG("[0x%x]", bd_addr.address[4]);
	DBG("[0x%x]", bd_addr.address[5]);

	DBG("COD [%d]", ev->class_of_dev);
	DBG("pairing variant [%d]", ev->pairing_variant);
	DBG("passkey [%d]", ev->passkey);

	if (!bt_hal_cbacks->ssp_request_cb) {
		ERR("HAL User ssp_request_cb is not set!!");
		return;
	}

	bt_hal_cbacks->ssp_request_cb(&bd_addr, &bd_name,
			ev->class_of_dev,
			ev->pairing_variant, ev->passkey);
	DBG("-");
}

static void __bt_hal_handle_pin_request_event(void *buf, uint16_t len)
{
	struct hal_ev_pin_request *ev = (struct hal_ev_pin_request *)buf;
	bt_bdaddr_t bd_addr;
	bt_bdname_t bd_name;
	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	memcpy(bd_name.name, ev->name, 248);

	/* BD address*/
	DBG("[0x%x]", bd_addr.address[0]);
	DBG("[0x%x]", bd_addr.address[1]);
	DBG("[0x%x]", bd_addr.address[2]);
	DBG("[0x%x]", bd_addr.address[3]);
	DBG("[0x%x]", bd_addr.address[4]);
	DBG("[0x%x]", bd_addr.address[5]);

	DBG("COD [%u]", ev->class_of_dev);

	if (!bt_hal_cbacks->pin_request_cb) {
		ERR("HAL User pin_request_cb is not set!!");
		return;
	}

	bt_hal_cbacks->pin_request_cb(&bd_addr, &bd_name,
			ev->class_of_dev);
	DBG("-");
}

static void __bt_hal_handle_device_trust_state_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_device_trust_state_changed *ev = (struct hal_ev_device_trust_state_changed*)buf;
	bt_bdaddr_t bd_addr;
	DBG("+");

	memcpy(bd_addr.address, ev->bdaddr, 6);
	/* BD address*/
	DBG("[0x%x]", bd_addr.address[0]);
	DBG("[0x%x]", bd_addr.address[1]);
	DBG("[0x%x]", bd_addr.address[2]);
	DBG("[0x%x]", bd_addr.address[3]);
	DBG("[0x%x]", bd_addr.address[4]);
	DBG("[0x%x]", bd_addr.address[5]);

	DBG("Device Trusted?  [0x%x]", ev->trust);

	if (!bt_hal_cbacks->device_trust_state_changed_cb) {
		ERR("HAL User device_trust_state_changed_cb is not set!!");
		return;
	}

	bt_hal_cbacks->device_trust_state_changed_cb(&bd_addr, ev->trust);
	DBG("-");
}

static void __bt_hal_handle_device_trusted_profiles_changed_event(void *buf, uint16_t len)
{
	struct hal_ev_device_trusted_profiles_changed *ev = buf;
	bt_bdaddr_t bd_addr;

	DBG("+");
#ifdef TIZEN_BT_HAL
	memcpy(bd_addr.address, ev->bdaddr, 6);
	DBG("[0x%02X:%02X:%02X:%02X:%02X:%02X]",
			bd_addr.address[0], bd_addr.address[1],
			bd_addr.address[2], bd_addr.address[3],
			bd_addr.address[4], bd_addr.address[5]);

	if (!bt_hal_cbacks->device_trusted_profiles_changed_cb) {
		ERR("device_trusted_profiles_changed_cb is NULL");
		return;
	}

	bt_hal_cbacks->device_trusted_profiles_changed_cb(&bd_addr, ev->trust_val);
#endif
	DBG("-");
}

static void __bt_handle_rssi_monitor_state_changed(void *buf, uint16_t len)
{
	struct hal_ev_rssi_monitor_state_changed *ev = buf;
	bt_bdaddr_t bd_addr;

	DBG("+");
#ifdef TIZEN_BT_HAL
	memcpy(bd_addr.address, ev->bdaddr, 6);
	DBG("[0x%02X:%02X:%02X:%02X:%02X:%02X]",
			bd_addr.address[0], bd_addr.address[1],
			bd_addr.address[2], bd_addr.address[3],
			bd_addr.address[4], bd_addr.address[5]);

	if (bt_hal_cbacks->rssi_monitor_state_changed_cb)
		bt_hal_cbacks->rssi_monitor_state_changed_cb(
				&bd_addr, ev->link_type, ev->state);
#endif
	DBG("-");
}

static void __bt_handle_rssi_alert_recieved(void *buf, uint16_t len)
{
	struct hal_ev_rssi_alert_recieved *ev = buf;
	bt_bdaddr_t bd_addr;

	DBG("+");
#ifdef TIZEN_BT_HAL
	memcpy(bd_addr.address, ev->bdaddr, 6);
	DBG("[0x%02X:%02X:%02X:%02X:%02X:%02X]",
			bd_addr.address[0], bd_addr.address[1],
			bd_addr.address[2], bd_addr.address[3],
			bd_addr.address[4], bd_addr.address[5]);

	if (bt_hal_cbacks->rssi_alert_cb)
		bt_hal_cbacks->rssi_alert_cb(
				&bd_addr, ev->link_type, ev->alert_type, ev->rssi);
#endif
	DBG("-");
}

static void __bt_handle_raw_rssi_recieved(void *buf, uint16_t len)
{
	struct hal_ev_raw_rssi_recieved *ev = buf;
	bt_bdaddr_t bd_addr;

	DBG("+");
#ifdef TIZEN_BT_HAL
	memcpy(bd_addr.address, ev->bdaddr, 6);
	DBG("[0x%02X:%02X:%02X:%02X:%02X:%02X]",
			bd_addr.address[0], bd_addr.address[1],
			bd_addr.address[2], bd_addr.address[3],
			bd_addr.address[4], bd_addr.address[5]);

	if (bt_hal_cbacks->raw_rssi_received_cb)
		bt_hal_cbacks->raw_rssi_received_cb(
				&bd_addr, ev->link_type, ev->rssi);
#endif
	DBG("-");
}

static void __bt_hal_handle_stack_messages(int message, void *buf, uint16_t len)
{
	DBG("+");
	switch (message) {
	case HAL_EV_ADAPTER_STATE_CHANGED:
		DBG("Event: HAL_EV_ADAPTER_STATE_CHANGED");
		__bt_hal_handle_adapter_state_changed(buf, len);
		break;
	case HAL_EV_ADAPTER_PROPS_CHANGED:
		DBG("Event: HAL_EV_ADAPTER_PROPS_CHANGED");
		__bt_hal_handle_adapter_property_changed(buf, len);
		break;
	case HAL_EV_DISCOVERY_STATE_CHANGED:
		DBG("Event: HAL_EV_DISCOVERY_STATE_CHANGED");
		__bt_hal_handle_adapter_discovery_state_changed(buf, len);
		break;
	case HAL_EV_DEVICE_FOUND:
		DBG("Event: HAL_EV_DEVICE_FOUND");
		__bt_hal_handle_device_found_event(buf, len);
		break;
	case HAL_EV_REMOTE_DEVICE_PROPS:
		DBG("Event: HAL_EV_REMOTE_DEVICE_PROPS");
		__bt_hal_handle_remote_device_properties_event(buf, len);
		break;
	case HAL_EV_BOND_STATE_CHANGED:
		DBG("Event: HAL_EV_BOND_STATE_CHANGED:");
		__bt_hal_handle_bond_state_changed_event(buf, len);
		break;
	case HAL_EV_ACL_STATE_CHANGED:
		DBG("Event: HAL_ACL_STATE_CONNECTED or Disconnected");
		__bt_hal_handle_device_acl_state_changed_event(buf, len);
		break;
	case HAL_EV_LE_CONN_STATE_CHANGED:
		DBG("Event: HAL_LE_CONN_STATE_CONNECTED or Disconnected");
		__bt_hal_handle_device_le_conn_state_changed_event(buf, len);
		break;
	case HAL_EV_SSP_REQUEST:
		DBG("Event: HAL_EV_SSP_REQUEST");
		__bt_hal_handle_ssp_request_event(buf, len);
		break;
	case HAL_EV_PIN_REQUEST:
		DBG("Event: HAL_EV_PIN_REQUEST");
		__bt_hal_handle_pin_request_event(buf, len);
		break;
	case HAL_EV_AUTHORIZE_REQUEST:
		DBG("Event: HAL_EV_AUTHORIZE_REQUEST");
		__bt_hal_handle_authorize_request_event(buf, len);
		break;
	case HAL_EV_DEVICE_TRUST_CHANGED:
		DBG("Event: HAL_EV_DEVICE_TRUST_CHANGED");
		__bt_hal_handle_device_trust_state_changed_event(buf, len);
		break;
#ifdef TIZEN_BT_HAL
	case HAL_EV_SOCK_AUTHORIZE_REQUEST:
		DBG("Event: HAL_EV_AUTHORIZE_REQUEST");
		__bt_hal_handle_socket_authorize_request_event(buf, len);
		break;
	case HAL_EV_LE_STATE_CHANGED:
		DBG("Event: HAL_EV_LE_STATE_CHANGED");
		__bt_hal_handle_le_state_changed(buf, len);
		break;
#endif
	case HAL_EV_DEVICE_TRUSTED_PROFILES_CHANGED:
		DBG("Event: HAL_EV_DEVICE_TRUST_CHANGED");
		__bt_hal_handle_device_trusted_profiles_changed_event(buf, len);
		break;
	case HAL_EV_RSSI_MONITOR_STATE_CHANGED:
		DBG("Event: HAL_EV_RSSI_MONITOR_STATE_CHANGED");
		__bt_handle_rssi_monitor_state_changed(buf, len);
		break;
	case HAL_EV_RSSI_ALERT_RECIEVED:
		DBG("Event: HAL_EV_RSSI_ALERT_RECIEVED");
		__bt_handle_rssi_alert_recieved(buf, len);
		break;
	case HAL_EV_RAW_RSSI_RECIEVED:
		DBG("Event: HAL_EV_RAW_RSSI_RECIEVED");
		__bt_handle_raw_rssi_recieved(buf, len);
		break;
	default:
		DBG("Event Currently not handled!!");
		break;
	}
	DBG("-");
}
