/*
 * BLUETOOOTH HAL
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */


/*******************************************************************************
 *
 * Description:	GATT Profile Bluetooth Interface
 *
 *******************************************************************************/

#include <hardware/bluetooth.h>
#include <hardware/bt_gatt.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"

#include "bt-hal-adapter-le.h"
#include "bt-hal-gatt-server.h"
#include "bt-hal-gatt-client.h"
#include "bt_gatt_types.h"

/* Advertising report event types */
#define BT_LE_ADV_IND	0x00
#define BT_LE_ADV_DIRECT_IND	0x01
#define BT_LE_ADV_SCAN_IND	0x02
#define BT_LE_ADV_NONCONN_IND	0x03
#define BT_LE_ADV_SCAN_RSP	0x04

#define BT_HAL_ADV_DATA_MAX_SIZE	31
#define BT_HAL_GATTC_READ_VALUE_TYPE_VALUE          0x0000  /* Attribute value itself */
typedef struct {
	char addr[18];
	uint8_t addr_type;
	uint8_t adv_type;
	int rssi;
	int data_len;
	uint8_t data[BT_HAL_ADV_DATA_MAX_SIZE * 2];
	guint timer_id;
} bt_hal_le_adv_info_t;

/****************************************** Global Variables ******************************************/
const btgatt_callbacks_t *bt_gatt_callbacks = NULL;

extern btgatt_client_interface_t btgatt_client_interface;
extern btgatt_server_interface_t btgatt_server_interface;

static GSList *adv_ind_list = NULL;
/****************************************** Forward Declarations *************************************/
static void __bt_handle_gatt_client_registered(void *buf, uint16_t len);
static void __bt_hal_handle_gatt_client_scan_result(void *buf, uint16_t len);
static void __bt_handle_gatt_client_connected(void *buf, uint16_t len);
static void __bt_handle_gatt_client_disconnected(void *buf, uint16_t len);
static void __bt_handle_gatt_client_search_result(void *buf, uint16_t len);
static void __bt_handle_gatt_client_search_complete(void *buf, uint16_t len);
static void __bt_handle_gatt_client_search_char_result(void *buf, uint16_t len);
static void __bt_handle_gatt_client_search_desc_result(void *buf, uint16_t len);
static void __bt_handle_gatt_client_read_charac(void *buf, uint16_t len);
static void __bt_handle_gatt_client_read_desc(void *buf, uint16_t len);
static void __bt_handle_gatt_client_write_char(void *buf, uint16_t len);
static void __bt_handle_gatt_client_write_desc(void *buf, uint16_t len);
static void __bt_handle_gatt_client_watch_notification(void *buf, uint16_t len);
static void __bt_handle_gatt_client_changed_value(void *buf, uint16_t len);
/*****************************************************************************************************/

static bool interface_ready(void)
{
	return bt_gatt_callbacks != NULL;
}

static void __bt_hal_handle_server_instance_initialized(void *buf, uint16_t len)
{
	struct hal_ev_server_instance_registered *ev = buf;

	if (bt_gatt_callbacks->server->register_server_cb)
		bt_gatt_callbacks->server->register_server_cb(ev->status, ev->server_instance, (bt_uuid_t*)&ev->app_uuid);
}

static void __bt_hal_handle_multi_adv_data_set(void *buf, uint16_t len)
{
	struct hal_ev_multi_adv_data_set *ev = buf;

	if (bt_gatt_callbacks->server->multi_adv_data_cb)
		bt_gatt_callbacks->server->multi_adv_data_cb(ev->server_instance, ev->status);
}

static void __bt_hal_handle_multi_adv_data_update(void *buf, uint16_t len)
{
	struct hal_ev_multi_adv_update *ev = buf;

	if (bt_gatt_callbacks->server->multi_adv_update_cb)
		bt_gatt_callbacks->server->multi_adv_update_cb(ev->server_instance, ev->status);
}

static void __bt_hal_handle_multi_adv_enable(void *buf, uint16_t len)
{
	struct hal_ev_multi_adv_enable *ev = buf;

	if (bt_gatt_callbacks->server->multi_adv_enable_cb)
		bt_gatt_callbacks->server->multi_adv_enable_cb(ev->server_instance, ev->status);
}

static void __bt_hal_handle_multi_adv_disable(void *buf, uint16_t len)
{
	struct hal_ev_multi_adv_disable *ev = buf;

	if (bt_gatt_callbacks->server->multi_adv_disable_cb)
		bt_gatt_callbacks->server->multi_adv_disable_cb(ev->server_instance, ev->status);
}

/* Same callback for both Enable and DIsable */
static void __bt_hal_handle_legacy_adv_status(void *buf, uint16_t len)
{
	struct hal_ev_legacy_adv_status *ev = buf;

	if (bt_gatt_callbacks->server->listen_cb)
		bt_gatt_callbacks->server->listen_cb(ev->status, ev->server_instance);
}

static void __bt_hal_handle_service_added(void *buf, uint16_t len)
{
	struct hal_ev_gatt_service_added *ev = buf;
	btgatt_srvc_id_t srvc_id;

	memset(&srvc_id, 0x00, sizeof(btgatt_srvc_id_t));
	srvc_id.is_primary = ev->is_primary;
	srvc_id.id.inst_id = ev->server_instance;
	memcpy(srvc_id.id.uuid.uu, ev->svc_uuid, sizeof(ev->svc_uuid));

	if (bt_gatt_callbacks->server->service_added_cb)
		bt_gatt_callbacks->server->service_added_cb(ev->status, ev->server_instance,
				&srvc_id, ev->service_handle);
}

static void __bt_hal_handle_char_added(void *buf, uint16_t len)
{
	struct hal_ev_gatt_char_added *ev = buf;
	bt_uuid_t uuid;

	memcpy(uuid.uu, ev->char_uuid, sizeof(ev->char_uuid));

	if (bt_gatt_callbacks->server->characteristic_added_cb)
		bt_gatt_callbacks->server->characteristic_added_cb(ev->status, ev->server_instance,
				&uuid, ev->service_handle, ev->char_handle);
}

static void __bt_hal_handle_desc_added(void *buf, uint16_t len)
{
	struct hal_ev_gatt_desc_added *ev = buf;
	bt_uuid_t uuid;

	memcpy(uuid.uu, ev->desc_uuid, sizeof(ev->desc_uuid));

	if (bt_gatt_callbacks->server->descriptor_added_cb)
		bt_gatt_callbacks->server->descriptor_added_cb(ev->status, ev->server_instance,
				&uuid, ev->service_handle, ev->desc_handle);
}

static void __bt_hal_handle_service_started(void *buf, uint16_t len)
{
	struct hal_ev_gatt_service_started *ev = buf;

	if (bt_gatt_callbacks->server->service_started_cb)
		bt_gatt_callbacks->server->service_started_cb(ev->status, ev->server_instance,
				ev->service_handle);
}

static void __bt_hal_handle_service_deleted(void *buf, uint16_t len)
{
	struct hal_ev_gatt_service_deleted *ev = buf;

	if (bt_gatt_callbacks->server->service_deleted_cb)
		bt_gatt_callbacks->server->service_deleted_cb(ev->status, ev->server_instance,
				ev->service_handle);
}

static void __bt_hal_handle_gatt_server_connected(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_connected *ev = buf;
	bt_bdaddr_t bd_addr;

	memcpy(bd_addr.address, ev->bdaddr, 6);

	if (bt_gatt_callbacks->server->connection_cb)
		bt_gatt_callbacks->server->connection_cb(ev->conn_id, ev->server_instance,
				ev->connected, &bd_addr);
}

static void __bt_hal_handle_gatt_server_notification_changed(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_notifcation_change *ev = buf;
	bt_bdaddr_t bd_addr;

	memcpy(bd_addr.address, ev->bdaddr, 6);

	if (bt_gatt_callbacks->server->notif_enabled_cb)
		bt_gatt_callbacks->server->notif_enabled_cb(ev->conn_id, ev->trans_id,
				ev->att_handle, ev->notify, &bd_addr);
}

static void __bt_hal_handle_gatt_server_read_requested(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_read_req *ev = buf;
	bt_bdaddr_t bd_addr;

	memcpy(bd_addr.address, ev->bdaddr, 6);

	if (bt_gatt_callbacks->server->request_read_cb)
		bt_gatt_callbacks->server->request_read_cb(ev->conn_id, ev->trans_id, &bd_addr,
					ev->att_handle, ev->offset, ev->is_long);
}

static void __bt_hal_handle_gatt_server_write_requested(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_write_req *ev = buf;

	if (bt_gatt_callbacks->server->request_write_cb)
		bt_gatt_callbacks->server->request_write_cb(ev->conn_id, ev->trans_id,
				(bt_bdaddr_t *) ev->bdaddr,
				ev->att_handle, ev->offset,
				ev->length, ev->need_rsp,
				ev->is_prep, ev->value);
}

static void __bt_hal_handle_gatt_server_acquire_write_requested(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_acquire_write_res *ev = buf;

	DBG("acquire write requested event recived");

	if (bt_gatt_callbacks->server->request_acquire_write_cb)
		bt_gatt_callbacks->server->request_acquire_write_cb(ev->mtu, ev->conn_id, ev->trans_id, ev->char_handl, (bt_bdaddr_t *)ev->bdaddr);
}

static void __bt_hal_handle_gatt_server_acquire_notify_requested(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_acquire_notify *ev = buf;

	DBG("acquire notify  requested event recived");

	if (bt_gatt_callbacks->server->request_acquire_notify_cb)
		bt_gatt_callbacks->server->request_acquire_notify_cb(ev->mtu, ev->conn_id, ev->trans_id, ev->char_handl);
}

static void __bt_hal_handle_gatt_server_indicate_confirmed(void *buf, uint16_t len)
{
	struct hal_ev_gatt_server_indicate_cfm *ev = buf;

	if (bt_gatt_callbacks->server->indication_confirmation_cb)
		bt_gatt_callbacks->server->indication_confirmation_cb(ev->conn_id, ev->trans_id,
				ev->att_handle, (bt_bdaddr_t *) ev->bdaddr);

}

static void __bt_hal_gatt_events(int message, void *buf, uint16_t len)
{
	DBG("+");
	/* Check if GATT interface is Ready */
	if (!interface_ready())
		return;

	switch (message) {
	case HAL_EV_SERVER_INSTANCE_INITIALIZED: {
		__bt_hal_handle_server_instance_initialized(buf, len);
		break;
	}
	case HAL_EV_MULTI_ADV_ENABLE: {
		__bt_hal_handle_multi_adv_enable(buf, len);
		break;
	}
	case HAL_EV_MULTI_ADV_DISABLE: {
		__bt_hal_handle_multi_adv_disable(buf, len);
		break;
	}
	case HAL_EV_MULTI_ADV_DATA_SET: {
		__bt_hal_handle_multi_adv_data_set(buf, len);
		break;
	}
	case HAL_EV_MULTI_ADV_UPDATE: {
		__bt_hal_handle_multi_adv_data_update(buf, len);
		break;
	}
	case HAL_EV_LEGACY_ADV_ENABLE: {
		__bt_hal_handle_legacy_adv_status(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_REGISTERED: {
		__bt_handle_gatt_client_registered(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_SCAN_RESULT: {
		__bt_hal_handle_gatt_client_scan_result(buf, len);
		break;
	}
	case HAL_EV_GATT_SERVICE_ADDED: {
		__bt_hal_handle_service_added(buf, len);
		break;
	}
	case HAL_EV_GATT_CHAR_ADDED: {
		__bt_hal_handle_char_added(buf, len);
		break;
	}
	case HAL_EV_GATT_DESC_ADDED: {
		__bt_hal_handle_desc_added(buf, len);
		break;
	}
	case HAL_EV_GATT_SERVICE_STARTED: {
		__bt_hal_handle_service_started(buf, len);
		break;
	}
	case HAL_EV_GATT_SERVICE_DELETED: {
		__bt_hal_handle_service_deleted(buf, len);
		break;
	}
	case HAL_EV_GATT_SERVER_CONNECTED: {
		__bt_hal_handle_gatt_server_connected(buf, len);
		break;
	}
	case HAL_EV_GATT_NOTIFICATION_CHANGE: {
		__bt_hal_handle_gatt_server_notification_changed(buf, len);
		break;
	}
	case HAL_EV_GATT_READ_REQUESTED: {
		__bt_hal_handle_gatt_server_read_requested(buf, len);
		break;
	}
	case HAL_EV_GATT_WRITE_REQUESTED: {
		__bt_hal_handle_gatt_server_write_requested(buf, len);
		break;
	}
	case HAL_EV_GATT_INDICATE_CFM: {
		__bt_hal_handle_gatt_server_indicate_confirmed(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_CONNECTED: {
		__bt_handle_gatt_client_connected(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_DISCONNECTED: {
		__bt_handle_gatt_client_disconnected(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_SEARCH_RESULT: {
		__bt_handle_gatt_client_search_result(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_SEARCH_COMPLETE: {
		__bt_handle_gatt_client_search_complete(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_CHARAC_SEARCH_RESULT: {
		__bt_handle_gatt_client_search_char_result(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_DESC_SEARCH_RESULT: {
		__bt_handle_gatt_client_search_desc_result(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_READ_CHARAC: {
		__bt_handle_gatt_client_read_charac(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_READ_DESC: {
		__bt_handle_gatt_client_read_desc(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_WRITE_CHARAC: {
		__bt_handle_gatt_client_write_char(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_WRITE_DESC: {
		__bt_handle_gatt_client_write_desc(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_WATCH_NOTIFICATION: {
		__bt_handle_gatt_client_watch_notification(buf, len);
		break;
	}
	case HAL_EV_GATT_CLIENT_NOTIFY_CHANGED_VALUE: {
		__bt_handle_gatt_client_changed_value(buf, len);
		break;
	}
	case HAL_EV_GATT_SERVER_ACQUIRE_WRITE_RES:{
		__bt_hal_handle_gatt_server_acquire_write_requested(buf, len);
		break;
	}
	case HAL_EV_GATT_SERVER_ACQUIRE_NOTIFY_RES:{
		__bt_hal_handle_gatt_server_acquire_notify_requested(buf, len);
		break;
	}

	default:
		DBG("Event Currently not handled!!");
		break;
	}
	DBG("-");
}

/************************************* GATT CLIENT EVENTS ****************************************/
static void __bt_handle_gatt_client_registered(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_registered *ev = buf;
	bt_uuid_t uuid;

	memcpy(uuid.uu, ev->app_uuid, 16);
	if (bt_gatt_callbacks->client->register_client_cb)
		bt_gatt_callbacks->client->register_client_cb(
				ev->status, ev->client_if, &uuid);
}

static void __bt_hal_send_le_scan_result_event(bt_hal_le_adv_info_t *adv_info)
{
	bt_bdaddr_t bd_addr;

	str2bt_bdaddr_t(adv_info->addr, &bd_addr);
	if (bt_gatt_callbacks->client->scan_result_cb)
		bt_gatt_callbacks->client->scan_result_cb(
				&bd_addr, adv_info->rssi, adv_info->data);
}

static void __bt_handle_gatt_client_connected(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_connected  *ev = buf;
	bt_bdaddr_t bd_addr;

	memcpy(bd_addr.address, ev->bdaddr, 6);
	if (bt_gatt_callbacks->client->open_cb)
		bt_gatt_callbacks->client->open_cb(ev->conn_id,
				ev->status, ev->client_if, &bd_addr);
}

static void __bt_handle_gatt_client_disconnected(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_connected  *ev = buf;
	bt_bdaddr_t bd_addr;

	memcpy(bd_addr.address, ev->bdaddr, 6);
	if (bt_gatt_callbacks->client->close_cb)
		bt_gatt_callbacks->client->close_cb(ev->conn_id,
				ev->status, ev->client_if, &bd_addr);
}

static void __bt_handle_gatt_client_search_result(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_search_result  *ev = buf;
	btgatt_srvc_id_t gatt_srvc_id;

	gatt_srvc_id.is_primary = ev->is_primary;
	gatt_srvc_id.id.inst_id = ev->inst_id;
	memcpy(gatt_srvc_id.id.uuid.uu, ev->uuid, 16);

	if (bt_gatt_callbacks->client->search_result_cb)
		bt_gatt_callbacks->client->search_result_cb(ev->conn_id,
							&gatt_srvc_id);
}

static void __bt_handle_gatt_client_search_complete(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_search_complete  *ev = buf;

	if (bt_gatt_callbacks->client->search_complete_cb)
		bt_gatt_callbacks->client->search_complete_cb(ev->conn_id,
								ev->status);
}

static void __bt_handle_gatt_client_search_char_result(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_char_search_result *ev = buf;
	btgatt_srvc_id_t gatt_srvc_id;
	btgatt_gatt_id_t gatt_char_id;

	gatt_srvc_id.is_primary = ev->is_primary;
	gatt_srvc_id.id.inst_id = ev->inst_id;
	memcpy(gatt_srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	if (BT_STATUS_SUCCESS == ev->status) {
		memcpy(gatt_char_id.uuid.uu, ev->char_uuid, 16);
		gatt_char_id.inst_id = ev->inst_id;
	}

	if (bt_gatt_callbacks->client->get_characteristic_cb)
		bt_gatt_callbacks->client->get_characteristic_cb(ev->conn_id,
			ev->status, &gatt_srvc_id, &gatt_char_id, ev->char_prop);
}

static void __bt_handle_gatt_client_search_desc_result(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_desc_search_result *ev = buf;
	btgatt_srvc_id_t gatt_srvc_id;
	btgatt_gatt_id_t gatt_char_id;
	btgatt_gatt_id_t gatt_desc_id;


	gatt_srvc_id.is_primary = ev->is_primary;
	gatt_srvc_id.id.inst_id = ev->inst_id;
	memcpy(gatt_srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	memcpy(gatt_char_id.uuid.uu, ev->char_uuid, 16);
	gatt_char_id.inst_id = ev->inst_id;

	if (BT_STATUS_SUCCESS == ev->status) {
		memcpy(gatt_desc_id.uuid.uu, ev->desc_uuid, 16);
		gatt_desc_id.inst_id = ev->inst_id;
	}

	if (bt_gatt_callbacks->client->get_descriptor_cb)
		bt_gatt_callbacks->client->get_descriptor_cb(ev->conn_id,
			ev->status, &gatt_srvc_id, &gatt_char_id, &gatt_desc_id);
}

static void __bt_handle_gatt_client_read_charac(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_read_data *ev = buf;
	btgatt_read_params_t char_read_parm;

	char_read_parm.srvc_id.is_primary = ev->is_primary;
	char_read_parm.srvc_id.id.inst_id = ev->inst_id;
	memcpy(char_read_parm.srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	char_read_parm.char_id.inst_id = ev->inst_id;
	memcpy(char_read_parm.char_id.uuid.uu, ev->char_uuid, 16);

	char_read_parm.value.len = ev->len;
	if (ev->len > 0) {
		memcpy(char_read_parm.value.value, ev->value, ev->len);
		char_read_parm.value_type = BT_HAL_GATTC_READ_VALUE_TYPE_VALUE;
	}

	if (bt_gatt_callbacks->client->read_characteristic_cb)
		bt_gatt_callbacks->client->read_characteristic_cb(ev->conn_id,
					ev->status, &char_read_parm);
}

static void __bt_handle_gatt_client_read_desc(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_read_data *ev = buf;
	btgatt_read_params_t desc_read_parm;

	desc_read_parm.srvc_id.is_primary = ev->is_primary;
	desc_read_parm.srvc_id.id.inst_id = ev->inst_id;
	memcpy(desc_read_parm.srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	desc_read_parm.char_id.inst_id = ev->inst_id;
	memcpy(desc_read_parm.char_id.uuid.uu, ev->char_uuid, 16);

	desc_read_parm.descr_id.inst_id = ev->inst_id;
	memcpy(desc_read_parm.descr_id.uuid.uu, ev->desc_uuid, 16);

	desc_read_parm.value.len = ev->len;
	if (ev->len > 0) {
		memcpy(desc_read_parm.value.value, ev->value, ev->len);
		desc_read_parm.value_type = BT_HAL_GATTC_READ_VALUE_TYPE_VALUE;
	}

	if (bt_gatt_callbacks->client->read_descriptor_cb)
		bt_gatt_callbacks->client->read_descriptor_cb(ev->conn_id,
					ev->status, &desc_read_parm);
}

static void __bt_handle_gatt_client_write_char(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_write_result *ev = buf;
	btgatt_write_params_t char_write_parm;

	char_write_parm.srvc_id.is_primary = ev->is_primary;
	char_write_parm.srvc_id.id.inst_id = ev->inst_id;
	memcpy(char_write_parm.srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	char_write_parm.char_id.inst_id = ev->inst_id;
	memcpy(char_write_parm.char_id.uuid.uu, ev->char_uuid, 16);

	if (bt_gatt_callbacks->client->write_characteristic_cb)
		bt_gatt_callbacks->client->write_characteristic_cb(ev->conn_id,
					ev->status, &char_write_parm);
}

static void __bt_handle_gatt_client_write_desc(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_write_result *ev = buf;
	btgatt_write_params_t desc_write_parm;

	desc_write_parm.srvc_id.is_primary = ev->is_primary;
	desc_write_parm.srvc_id.id.inst_id = ev->inst_id;
	memcpy(desc_write_parm.srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	desc_write_parm.char_id.inst_id = ev->inst_id;
	memcpy(desc_write_parm.char_id.uuid.uu, ev->char_uuid, 16);

	desc_write_parm.descr_id.inst_id = ev->inst_id;
	memcpy(desc_write_parm.descr_id.uuid.uu, ev->desc_uuid, 16);

	if (bt_gatt_callbacks->client->write_descriptor_cb)
		bt_gatt_callbacks->client->write_descriptor_cb(ev->conn_id,
					ev->status, &desc_write_parm);
}

static void __bt_handle_gatt_client_watch_notification(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_watch_notification *ev = buf;
	btgatt_srvc_id_t gatt_srvc_id;
	btgatt_gatt_id_t gatt_char_id;

	gatt_srvc_id.is_primary = ev->is_primary;
	gatt_srvc_id.id.inst_id = ev->inst_id;
	memcpy(gatt_srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	memcpy(gatt_char_id.uuid.uu, ev->char_uuid, 16);
	gatt_char_id.inst_id = ev->inst_id;

	if (bt_gatt_callbacks->client->register_for_notification_cb)
		bt_gatt_callbacks->client->register_for_notification_cb(ev->client_if,
				ev->registered, ev->status, &gatt_srvc_id, &gatt_char_id);
}

static void __bt_handle_gatt_client_changed_value(void *buf, uint16_t len)
{
	struct hal_ev_gatt_client_notify_changed_value *ev = buf;
	btgatt_notify_params_t changd_value_parm;

	changd_value_parm.srvc_id.is_primary = ev->is_primary;
	changd_value_parm.srvc_id.id.inst_id = ev->inst_id;
	memcpy(changd_value_parm.srvc_id.id.uuid.uu, ev->svc_uuid, 16);

	changd_value_parm.char_id.inst_id = ev->inst_id;
	memcpy(changd_value_parm.char_id.uuid.uu, ev->char_uuid, 16);
	changd_value_parm.is_notify = ev->is_notify;

	memcpy(changd_value_parm.bda.address, ev->bdaddr, 6);

	changd_value_parm.len = ev->len;
	if (ev->len > 0)
		memcpy(changd_value_parm.value, ev->value, ev->len);

	if (bt_gatt_callbacks->client->notify_cb)
		bt_gatt_callbacks->client->notify_cb(ev->conn_id, &changd_value_parm);
}

static bt_hal_le_adv_info_t *__bt_hal_get_adv_ind_info(char *addr)
{
	GSList *l;

	if (!addr)
		return NULL;

	for (l = adv_ind_list; NULL != l; l = g_slist_next(l)) {
		bt_hal_le_adv_info_t *adv_info = l->data;

		if (!adv_info)
			continue;

		if (!g_strcmp0(adv_info->addr, addr))
			return adv_info;
	}

	return NULL;
}

static int __bt_hal_add_adv_ind_info(bt_hal_le_adv_info_t *adv_info)
{
	if (!adv_info)
		return -1;

	if (__bt_hal_get_adv_ind_info(adv_info->addr) != NULL) {
		DBG("Adv info already present");
		return -1;
	}

	adv_ind_list = g_slist_append(adv_ind_list, adv_info);
	return 0;
}

static gboolean __bt_hal_adv_scan_req_timeout_cb(gpointer user_data)
{
	bt_hal_le_adv_info_t *adv_info = user_data;

	__bt_hal_send_le_scan_result_event(adv_info);
	adv_ind_list = g_slist_remove(adv_ind_list, adv_info);
	g_free(adv_info);
	return FALSE;
}

static void __bt_hal_handle_gatt_client_scan_result(void *buf, uint16_t len)
{
	bt_bdaddr_t bd_addr;
	char address[18];
	struct hal_ev_gatt_client_scan_result *ev = buf;
	bt_hal_le_adv_info_t *adv_info;
	int data_len = 0;

	memcpy(bd_addr.address, ev->bd_addr, 6);
	bt_bdaddr_t2str(&bd_addr, address);
	data_len = (ev->len < BT_HAL_ADV_DATA_MAX_SIZE) ? ev->len : BT_HAL_ADV_DATA_MAX_SIZE;
	if (data_len == 0 && ev->adv_type != BT_LE_ADV_SCAN_RSP) {
		ERR("Unexpected: Data len is 0");
		return;
	}

	if (ev->adv_type == BT_LE_ADV_SCAN_RSP) { /* SCAN_RSP */
		adv_info = __bt_hal_get_adv_ind_info(address);
		if (adv_info) {
			/* Copy scan response data in data field and send event */
			memcpy(&(adv_info->data[adv_info->data_len]), ev->adv_data, data_len);
			__bt_hal_send_le_scan_result_event(adv_info);
			adv_ind_list = g_slist_remove(adv_ind_list, adv_info);
			if (adv_info->timer_id)
				g_source_remove(adv_info->timer_id);
			g_free(adv_info);
		}
		return;
	}

	/* ADV_IND */
	adv_info = g_malloc0(sizeof(bt_hal_le_adv_info_t));
	if (!adv_info) {
		ERR("Not enough memory");
		return;
	}

	g_strlcpy(adv_info->addr, address, 18);
	adv_info->addr_type = ev->addr_type;
	adv_info->adv_type = ev->adv_type;
	adv_info->rssi = ev->rssi;
	adv_info->data_len = data_len;
	memcpy(adv_info->data, ev->adv_data, data_len);

#ifdef TIZEN_BT_HAL
	if (_bt_hal_gatt_client_get_le_scan_type() == BT_GATTC_LE_SCAN_TYPE_PASSIVE) {
		__bt_hal_send_le_scan_result_event(adv_info);
		g_free(adv_info);
		return;
	}
#endif
	if (__bt_hal_add_adv_ind_info(adv_info))
		adv_info->timer_id = g_timeout_add(1000,
				__bt_hal_adv_scan_req_timeout_cb,
				(void *)adv_info);
}

/*******************************************************************************
 **
 ** Function		gatt_init
 **
 ** Description		Initializes the GATT interface
 **
 ** Returns		bt_status_t
 **
 *******************************************************************************/
static bt_status_t gatt_init(const btgatt_callbacks_t* callbacks)
{
	bt_gatt_callbacks = callbacks;
	DBG("Register A2DP Src events callback function");
	_bt_hal_register_gatt_le_dbus_handler_cb(__bt_hal_gatt_events);
	_bt_hal_register_gatt_server_handler_cb(__bt_hal_gatt_events);
	_bt_hal_register_gatt_client_handler_cb(__bt_hal_gatt_events);
	_bt_hal_register_event_handler_cb(HAL_GATT, __bt_hal_gatt_events);

	return BT_STATUS_SUCCESS;
}

/*******************************************************************************
 **
 ** Function	gatt_cleanup
 **
 ** Description	Closes the GATT interface
 **
 ** Returns	void
 **
 *******************************************************************************/
static void gatt_cleanup(void)
{
	_bt_hal_unregister_gatt_le_dbus_handler_cb();
	_bt_hal_unregister_gatt_client_handler_cb();
	_bt_hal_unregister_gatt_server_handler_cb();
	_bt_hal_unregister_event_handler_cb(HAL_GATT);

	if (bt_gatt_callbacks)
		bt_gatt_callbacks = NULL;
}

static const btgatt_interface_t btgatt_interface = {
	sizeof(btgatt_interface),

	gatt_init,
	gatt_cleanup,

	&btgatt_client_interface,
	&btgatt_server_interface,
};

/*******************************************************************************
 **
 ** Function		bt_get_gatt_interface
 **
 ** Description		Get the gatt callback interface
 **
 ** Returns		btgatt_interface_t
 **
 *******************************************************************************/
const btgatt_interface_t *bt_get_gatt_interface()
{
	return &btgatt_interface;
}
