/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"

#include "bt-hal-event-receiver.h"
#include "bt-hal-hdp-dbus-handler.h"

static const bthl_callbacks_t *bt_hal_hdp_cbacks;

static bool interface_ready(void)
{
	return bt_hal_hdp_cbacks != NULL;
}

static void __bt_hal_handle_app_reg_state(void *buf, uint16_t len)
{
	struct hal_ev_hdp_app_reg_state *ev = buf;

	DBG("+");

	if (!ev) {
		ERR("Event data is NULL, return");
		return;
	}

	if (bt_hal_hdp_cbacks->app_reg_state_cb)
		bt_hal_hdp_cbacks->app_reg_state_cb(ev->app_id, ev->state);

	DBG("-");
}

static void __bt_hal_handle_hdp_conn_state(void *buf, uint16_t len)
{
	struct hal_ev_hdp_conn_state *ev = buf;

	DBG("+");

	if (!ev) {
		ERR("Event data is NULL, return");
		return;
	}

	if (bt_hal_hdp_cbacks->channel_state_cb)
		bt_hal_hdp_cbacks->channel_state_cb(ev->app_id, (bt_bdaddr_t *)(ev->bdaddr),
			ev->mdep_index, ev->channel_id, ev->channel_state, ev->data_fd);

	DBG("-");
}

static void __bt_hal_handle_hdp_events(int message, void *buf, uint16_t len)
{
	DBG("+");

	if (!interface_ready())
		return;

	switch (message) {
	case HAL_EV_HDP_APP_REG_STATE:
		DBG("Event: HAL_EV_HDP_APP_REG_STATE");
		__bt_hal_handle_app_reg_state(buf, len);
		break;
	case HAL_EV_HDP_CONN_STATE:
		DBG("Event: HAL_EV_HDP_CONN_STATE");
		__bt_hal_handle_hdp_conn_state(buf, len);
		break;
	default:
		DBG("Event Currently not handled!!");
		break;
	}

	DBG("-");
}

/** Register HL application */
static bt_status_t hdp_register_application(bthl_reg_param_t *p_reg_param, int *app_id)
{
	int ret = BT_STATUS_SUCCESS;
	int role;
	int data_type;
	int channel_type;
	const char *desc;

	if (!p_reg_param)
		return BT_STATUS_PARM_INVALID;

	if (!p_reg_param->application_name)
		return BT_STATUS_PARM_INVALID;

	if (!app_id)
		return BT_STATUS_PARM_INVALID;

	/* Currently only 1 dep config is supported per app id */
	if (0 > p_reg_param->number_of_mdeps || 1 < p_reg_param->number_of_mdeps) {
		ERR("Currently only 1 dep config is supported per app");
		return BT_STATUS_PARM_INVALID;
	}

	role = p_reg_param->mdep_cfg[0].mdep_role;
	data_type = p_reg_param->mdep_cfg[0].data_type;
	channel_type = p_reg_param->mdep_cfg[0].channel_type;
	desc = p_reg_param->mdep_cfg[0].mdep_description;

	ret = _bt_hal_dbus_handler_hdp_register_application(
			role, data_type, channel_type, desc, app_id);
	if (BT_STATUS_SUCCESS != ret) {
		ERR("_bt_hal_dbus_handler_hdp_register_application failed");
		*app_id = -1;
	} else {
		DBG("app_id: %d", *app_id);
	}

	return ret;
}

/** Unregister HL application */
static bt_status_t hdp_unregister_application(int app_id)
{
	int ret;

	DBG("+");

	ret = _bt_hal_dbus_handler_hdp_unregister_application(app_id);
	if (BT_STATUS_SUCCESS != ret)
		ERR("_bt_hal_dbus_handler_hdp_unregister_application failed");
	DBG("-");

	return ret;
}

/** connect channel */
static bt_status_t hdp_connect_channel(int app_id, bt_bdaddr_t *bd_addr, int mdep_cfg_index, int *channel_id)
{
	int ret;

	DBG("+");

	*channel_id = -1;
	ret = _bt_hal_dbus_handler_hdp_connect_channel(app_id, bd_addr, channel_id);
	if (BT_STATUS_SUCCESS != ret)
		ERR("_bt_hal_dbus_handler_hdp_unregister_application failed");

	DBG("-");
	return ret;
}

/** destroy channel */
static bt_status_t hdp_destroy_channel(int channel_id)
{
	int ret;

	DBG("+");

	ret = _bt_hal_dbus_handler_hdp_destroy_channel(channel_id);
	if (BT_STATUS_SUCCESS != ret)
		ERR("_bt_hal_dbus_handler_hdp_unregister_application failed");

	DBG("-");
	return ret;
}

static bt_status_t init(bthl_callbacks_t *callbacks)
{
	DBG("");

	if (interface_ready())
		return BT_STATUS_DONE;

	bt_hal_hdp_cbacks = callbacks;
	DBG("Register HDP events callback function");
	_bt_hal_register_hdp_dbus_handler_cb(__bt_hal_handle_hdp_events);

	return BT_STATUS_SUCCESS;
}

static void cleanup(void)
{
	DBG("");

	if (!interface_ready())
		return;

	DBG("Un-register HDP events callback function");
	_bt_hal_unregister_hdp_dbus_handler_cb();

	bt_hal_hdp_cbacks = NULL;
}

static bthl_interface_t hl_if = {
	.size = sizeof(hl_if),
	.init = init,
	.register_application = hdp_register_application,
	.unregister_application = hdp_unregister_application,
	.connect_channel = hdp_connect_channel,
	.destroy_channel = hdp_destroy_channel,
	.cleanup = cleanup
};

bthl_interface_t *bt_get_hl_interface(void)
{
	return &hl_if;
}
