/*
 * BLUETOOTH HAL
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <dlog.h>

#include "bt-hal.h"
#include "bt-hal-log.h"
#include "bt-hal-msg.h"
#include "bt-hal-utils.h"

#include "bt-hal-hf-dbus-handler.h"
#include "bt-hal-event-receiver.h"

static const bthf_callbacks_t *bt_hal_hf_cbacks;

static bool interface_ready(void)
{
	return bt_hal_hf_cbacks != NULL;
}

static bt_status_t hf_connect(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_dbus_handler_hf_connect(bd_addr);
}

static bt_status_t hf_disconnect(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return _bt_hal_dbus_handler_hf_disconnect(bd_addr);
}

static bt_status_t hf_connect_audio(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_disconnect_audio(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_start_voice_recognition(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_stop_voice_recognition(bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_voume_control(bthf_volume_type_t type, int volume,
		bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_device_status_notification(bthf_network_state_t ntk_state,
		bthf_service_type_t svc_type, int signal, int batt_chg)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_cops_response(const char *cops, bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_cind_response(int svc, int num_active, int num_held,
		bthf_call_state_t call_setup_state,
		int signal, int roam, int batt_chg,
		bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_formatted_at_response(const char *rsp, bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_at_response(bthf_at_response_t response_code,
		int error_code, bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_clcc_response(int index, bthf_call_direction_t dir,
		bthf_call_state_t state, bthf_call_mode_t mode,
		bthf_call_mpty_type_t mpty, const char *number,
		bthf_call_addrtype_t type, bt_bdaddr_t *bd_addr)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static bt_status_t hf_phone_state_change(int num_active, int num_held, bthf_call_state_t call_setup_state,
		const char *number, bthf_call_addrtype_t type)
{
	DBG("");
	return BT_STATUS_UNSUPPORTED;
}

static void __bt_hal_handle_hf_conn_state(void *buf, uint16_t len)
{
	struct hal_ev_handsfree_conn_state *ev = buf;

	if (bt_hal_hf_cbacks->connection_state_cb)
		bt_hal_hf_cbacks->connection_state_cb(ev->state, (bt_bdaddr_t *) ev->bdaddr);
}

static void __bt_hal_handle_hf_audio_conn_state(void *buf, uint16_t len)
{
	struct hal_ev_handsfree_audio_state *ev = buf;

	if (bt_hal_hf_cbacks->audio_state_cb)
		bt_hal_hf_cbacks->audio_state_cb(ev->state, (bt_bdaddr_t *) ev->bdaddr);
}

static void __bt_hal_handle_hf_events(int message, void *buf, uint16_t len)
{
	DBG("+");
	if (!interface_ready())
		return;
	switch (message) {
	case HAL_EV_HANDSFREE_CONN_STATE:
		DBG("Event: HAL_EV_HANDSFREE_CONN_STATE");
		__bt_hal_handle_hf_conn_state(buf, len);
		break;
	case HAL_EV_HANDSFREE_AUDIO_STATE:
		DBG("Event: HAL_EV_HANDSFREE_AUDIO_STATE");
		__bt_hal_handle_hf_audio_conn_state(buf, len);
		break;
	default:
		DBG("Event Currently not handled!!");
		break;
	}
	DBG("-");
}

static bt_status_t init(bthf_callbacks_t* callbacks, int max_hf_clients)
{
	DBG("");

	if (interface_ready())
		return BT_STATUS_DONE;

	if (max_hf_clients > 1)
		return BT_STATUS_UNSUPPORTED;
	bt_hal_hf_cbacks = callbacks;
	DBG("Register HF AG Role callback function");
	_bt_hal_register_hf_dbus_handler_cb(__bt_hal_handle_hf_events);
	_bt_hal_register_event_handler_cb(HAL_HF_AG, __bt_hal_handle_hf_events);
	return BT_STATUS_SUCCESS;
}

static void cleanup(void)
{
	DBG("");

	if (!interface_ready())
		return;

	_bt_hal_unregister_event_handler_cb(HAL_HF_AG);
	bt_hal_hf_cbacks = NULL;
}

static bthf_interface_t bthf_if = {
	.size = sizeof(bthf_if),
	.init = init,
	.connect = hf_connect,
	.disconnect = hf_disconnect,
	.connect_audio = hf_connect_audio,
	.disconnect_audio = hf_disconnect_audio,
	.start_voice_recognition = hf_start_voice_recognition,
	.stop_voice_recognition = hf_stop_voice_recognition,
	.volume_control = hf_voume_control,
	.device_status_notification = hf_device_status_notification,
	.cops_response = hf_cops_response,
	.cind_response = hf_cind_response,
	.formatted_at_response = hf_formatted_at_response,
	.at_response = hf_at_response,
	.clcc_response = hf_clcc_response,
	.phone_state_change = hf_phone_state_change,
	.cleanup = cleanup
};

/*******************************************************************************
**
** Function         bt_get_hf_interface
**
** Description      Get the hf callback interface
**
** Returns          bthf_interface_t
**
*******************************************************************************/
const bthf_interface_t *bt_get_hf_interface()
{
	return &bthf_if;
}
