/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <vconf.h>

#include "bt-hal-rfcomm-dbus-handler.h"
#include "bt-hal-dbus-common-utils.h"

#define BT_HAL_RFCOMM_ID_MAX 245
#define BT_HAL_RFCOMM_MAX_BUFFER_SIZE 1024

typedef struct {
	char remote_addr[BT_HAL_ADDRESS_STRING_SIZE];
	int hal_fd;
	unsigned int hal_watch;
	int stack_fd;
	unsigned int bt_watch;
} rfcomm_conn_info_t;

typedef struct {
	char uuid[BT_HAL_UUID_STRING_LEN];
	char *device_path;
	char *obj_path;
	int object_id;
	int id;
	rfcomm_conn_info_t *conn_info;
} rfcomm_cb_data_t;

typedef struct {
	char uuid[BT_HAL_UUID_STRING_LEN];
	char *svc_name;
	char *obj_path;
	int object_id;
	int id;
	int server_fd;
	unsigned int server_watch;
	GSList *conn_list;
} rfcomm_server_data_t;

static GSList *rfcomm_clients;
static GSList *rfcomm_servers;
static int latest_id = -1;
static gboolean id_used[BT_HAL_RFCOMM_ID_MAX];

int __rfcomm_assign_id(void)
{
	int index;

	DBG("latest_id: %d", latest_id);

	index = latest_id + 1;
	if (index >= BT_HAL_RFCOMM_ID_MAX)
		index = 0;

	DBG("index: %d", index);

	while (id_used[index] == TRUE) {
		if (index == latest_id) {
			/* No available ID */
			ERR("All request ID is used");
			return -1;
		}

		index++;
		if (index >= BT_HAL_RFCOMM_ID_MAX)
			index = 0;
	}

	latest_id = index;
	id_used[index] = TRUE;
	DBG("Assigned Id: %d", latest_id);

	return latest_id;
}

void __rfcomm_delete_id(int id)
{
	if (id >= BT_HAL_RFCOMM_ID_MAX || id < 0) {
		ERR("Invalid id %d", id);
		return;
	}

	id_used[id] = FALSE;
	latest_id = id - 1;
	DBG("id: %d, latest_id: %d", id, latest_id);
}

/*************************** RFCOMM Client Implementation ***************************/
static rfcomm_cb_data_t *__find_rfcomm_info_from_path(const char *path)
{
	GSList *l;

	for (l = rfcomm_clients; l != NULL; l = l->next) {
		rfcomm_cb_data_t *info = l->data;

		if (info != NULL)
			if (g_strcmp0(info->obj_path, path) == 0)
				return info;
	}

	return NULL;
}

static void __bt_free_conn(rfcomm_conn_info_t *conn)
{
	DBG("+");

	if (conn == NULL)
		return;

	if (0 < conn->hal_fd)
		close(conn->hal_fd);

	if (conn->hal_watch > 0) {
		g_source_remove(conn->hal_watch);
		conn->hal_watch = 0;
	}

	if (0 < conn->stack_fd)
		close(conn->stack_fd);

	if (conn->bt_watch > 0) {
		g_source_remove(conn->bt_watch);
		conn->bt_watch = 0;
	}

	g_free(conn);
	DBG("-");
}

static void __bt_free_cb_data(rfcomm_cb_data_t *cb_data)
{
	DBG("+");

	if (cb_data->id >= 0)
		__rfcomm_delete_id(cb_data->id);

	if (cb_data->object_id > 0)
		_bt_hal_unregister_gdbus_object(cb_data->object_id);

	if (cb_data->obj_path) {
		INFO("Unregister profile");
		_bt_hal_unregister_profile(cb_data->obj_path);
	}

	g_free(cb_data->obj_path);

	g_free(cb_data->device_path);
	g_free(cb_data);

	DBG("-");
}

static void __rfcomm_cb_data_remove(rfcomm_cb_data_t *info)
{
	if (!info) {
		ERR("info == NULL");
		return;
	}

	rfcomm_clients = g_slist_remove(rfcomm_clients, info);
	__bt_free_conn(info->conn_info);
	__bt_free_cb_data(info);
}

static int write_all(int fd, unsigned char *buf, int len)
{
	int sent = 0;

	while (len > 0) {
		int written;

		written = write(fd, buf, len);
		if (written < 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			return -1;
		}

		if (!written)
			return 0;

		len -= written; buf += written; sent += written;
	}

	return sent;
}

static gboolean app_event_cb(GIOChannel *io, GIOCondition cond, gpointer data)
{
	gsize len;
	int sent;
	rfcomm_cb_data_t *info = data;
	rfcomm_conn_info_t *conn_info;
	unsigned char buff[BT_HAL_RFCOMM_MAX_BUFFER_SIZE];
	GError *err = NULL;
	int fd;
	char err_msg[256] = {0, };

	DBG("+");
	fd = g_io_channel_unix_get_fd(io);
	conn_info = info->conn_info;

	if (cond & G_IO_HUP) {
		ERR("Socket %d hang up", fd);
		goto fail;
	}

	if (cond & (G_IO_ERR | G_IO_NVAL)) {
		ERR("Socket %d error", fd);
		goto fail;
	}

	if (!conn_info) {
		ERR("conn_info is NULL");
		return TRUE;
	}

	/* Read data from application */
	if (g_io_channel_read_chars(io, (gchar *)buff,
			BT_HAL_RFCOMM_MAX_BUFFER_SIZE,
			&len, &err) == G_IO_STATUS_ERROR) {
		if (err)
			ERR("IO Channel read error: %s", err->message);
		else
			ERR("IO Channel read error client");
		goto fail;
	}

	DBG("len: %zu", len);
	if (0 == len) {
		ERR("Other end of socket is closed");
		goto fail;
	}

	/* Send data to remote device */
	sent = write_all(conn_info->stack_fd, buff, len);
	if (sent < 0) {
		strerror_r(errno, err_msg, sizeof(err_msg));
		ERR("write(): %s", err_msg);
		goto fail;
	}

	DBG("-");
	return TRUE;
fail:
	__rfcomm_cb_data_remove(info);
	return FALSE;
}

static gboolean stack_event_cb(GIOChannel *io, GIOCondition cond, gpointer data)
{
	unsigned int len;
	int sent;
	rfcomm_cb_data_t *info = data;
	rfcomm_conn_info_t *conn_info;
	unsigned char buff[BT_HAL_RFCOMM_MAX_BUFFER_SIZE];
	GError *err = NULL;
	int fd;
	char err_msg[256] = {0, };

	DBG("+");

	fd = g_io_channel_unix_get_fd(io);
	conn_info = info->conn_info;

	if (cond & G_IO_HUP) {
		ERR("Socket %d hang up", fd);
		goto fail;
	}

	if (cond & (G_IO_ERR | G_IO_NVAL)) {
		ERR("Socket %d error", fd);
		goto fail;
	}

	if (!conn_info) {
		ERR("conn_info is NULL");
		return TRUE;
	}

	/* Read data from remote device */
	if (g_io_channel_read_chars(io, (gchar *)buff,
			BT_HAL_RFCOMM_MAX_BUFFER_SIZE,
			(gsize *)&len, &err) == G_IO_STATUS_ERROR) {
		if (err)
			ERR("IO Channel read error: %s", err->message);
		else
			ERR("IO Channel read error client");
		goto fail;
	}

	DBG("len: %d", len);
	if (0 == len) {
		ERR("Other end of socket is closed");
		goto fail;
	}

	/* Send data to application */
	sent = write_all(conn_info->hal_fd, buff, len);
	if (sent < 0) {
		strerror_r(errno, err_msg, sizeof(err_msg));
		ERR("write(): %s", err_msg);
		goto fail;
	}

	DBG("-");
	return TRUE;
fail:
	__rfcomm_cb_data_remove(info);
	return FALSE;
}

static int __new_connection(const char *path, int fd, bt_bdaddr_t *addr)
{
	char address[BT_HAL_ADDRESS_STRING_SIZE];
	rfcomm_cb_data_t *info;
	rfcomm_conn_info_t *conn_info;
	struct hal_ev_sock_connect ev;
	GIOCondition cond;
	int len;
	GIOChannel *io;
	char err_msg[256] = {0, };

	/* TODO: Temperary, later need to fill correct channel form correct place */
	int chan = 0;

	if (NULL == path || NULL == addr) {
		ERR("NULL == path || NULL = addr");
		return -1;
	}

	info = __find_rfcomm_info_from_path(path);
	if (info == NULL)
		return -1;

	conn_info = info->conn_info;
	_bt_hal_convert_addr_type_to_string(address, addr->address);
	if (conn_info == NULL) {
		ERR("conn_info is NULL for dev:[%s]", address);
		return -1;
	}

	if (write(conn_info->hal_fd, &chan, sizeof(chan)) != sizeof(chan)) {
		ERR("Error sending RFCOMM channel");
		goto fail;
	}

	conn_info->stack_fd = fd;
	DBG("Remote address: %s, RFCOMM fd: %d", address, conn_info->stack_fd);

	/* Send rfcomm connected event */
	memset(&ev, 0, sizeof(ev));
	ev.size = sizeof(ev);
	memcpy(ev.bdaddr, addr->address, 6);
	ev.status = BT_STATUS_SUCCESS;
	len = write_all(conn_info->hal_fd, (unsigned char *)&ev, sizeof(ev));
	if (len < 0) {
		strerror_r(errno, err_msg, sizeof(err_msg));
		ERR("%s", err_msg);
		goto fail;
	}

	if (len != sizeof(ev)) {
		ERR("Error sending connect event");
		goto fail;;
	}

	/* Handle events from App */
	cond = G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL;
	io = g_io_channel_unix_new(conn_info->hal_fd);
	conn_info->hal_watch = g_io_add_watch(io, cond, app_event_cb, info);
	g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK, NULL);
	g_io_channel_unref(io);

	/* Handle rfcomm events from bluez */
	cond = G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL;
	io = g_io_channel_unix_new(conn_info->stack_fd);
	conn_info->bt_watch = g_io_add_watch(io, cond, stack_event_cb, info);
	g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK, NULL);
	g_io_channel_unref(io);

	return 0;
fail:
	__rfcomm_cb_data_remove(info);
	return -1;
}

static void __bt_connect_response_cb(GDBusProxy *proxy,
		GAsyncResult *res, gpointer user_data)
{
	GError *error = NULL;
	rfcomm_cb_data_t *cb_data;

	DBG("+");

	cb_data = user_data;
	if (cb_data == NULL) {
		ERR("cb_data == NULL");
		return;
	}

	if (!g_dbus_proxy_call_finish(proxy, res, &error)) {
		ERR("Error : %s \n", error->message);
		__rfcomm_cb_data_remove(cb_data);
		g_error_free(error);
	}

	if (proxy)
		g_object_unref(proxy);

	DBG("-");
}

static void __bt_discover_service_response_cb(GDBusProxy *proxy,
		GAsyncResult *res, gpointer user_data)
{
	rfcomm_cb_data_t *cb_data;
	int ret = 0;
	GError *err = NULL;
	bt_hal_register_profile_info_t info = {0};
	char dev_address[BT_HAL_ADDRESS_STRING_SIZE];
	const char *path;

	DBG("+");

	cb_data = user_data;
	if (!cb_data) {
		ERR("cb_data == NULL");
		return;
	}

	path = g_dbus_proxy_get_object_path(proxy);
	_bt_hal_convert_device_path_to_address(path, dev_address);
	DBG("Device Adress [%s]", dev_address);

	g_dbus_proxy_call_finish(proxy, res, &err);
	if (proxy)
		g_object_unref(proxy);
	if (err != NULL) {
		ERR("Error occured in Proxy call [%s]\n", err->message);
		__rfcomm_cb_data_remove(cb_data);
		goto done;
	} else {
		INFO("Services are Updated checking required uuid is there");
		/* Check here for uuid present */
		ret = _bt_hal_discover_service_uuids(dev_address, cb_data->uuid);
		if (ret == BT_STATUS_SUCCESS) {
			info.uuid = (char *)cb_data->uuid;
			info.obj_path = cb_data->obj_path;
			info.role = "client";

			ret = _bt_hal_register_profile(&info, FALSE);
			if (ret < 0)
				DBG("Error: register profile");
			ret = _bt_hal_connect_profile(dev_address, cb_data->uuid,
					__bt_connect_response_cb, cb_data);
			if (ret != BT_STATUS_SUCCESS) {
				ERR("ConnectProfile failed");
				__rfcomm_cb_data_remove(cb_data);
				goto done;
			}
		} else {
			ERR("remote uuid not found");
			__rfcomm_cb_data_remove(cb_data);
		}
	}
done:
	if (err)
		g_clear_error(&err);
}

static rfcomm_cb_data_t *__get_rfcomm_cb_data(char *remote_uuid)
{
	int id;
	int object_id;
	char *path;
	rfcomm_cb_data_t *cb_data;

	DBG("+");

	id = __rfcomm_assign_id();
	if (id < 0) {
		ERR("__rfcomm_assign_id failed");
		return NULL;
	}

	path = g_strdup_printf("/org/socket/client/%d/%d", getpid(), id);
	object_id = _bt_hal_register_new_gdbus_object(path, __new_connection);
	if (object_id < 0) {
		ERR("_bt_hal_register_new_gdbus_object failed");
		g_free(path);
		__rfcomm_delete_id(id);
		return NULL;
	}

	cb_data = g_malloc0(sizeof(rfcomm_cb_data_t));
	g_strlcpy(cb_data->uuid, remote_uuid, BT_HAL_UUID_STRING_LEN);
	cb_data->obj_path = path;
	cb_data->object_id = object_id;
	cb_data->id = id;

	DBG("-");
	return cb_data;
}

static rfcomm_conn_info_t *__rfcomm_create_conn_info(char *addr, int *sock)
{
	int fds[2] = {-1, -1};
	rfcomm_conn_info_t *conn;
	char err_msg[256] = {0, };

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fds) < 0) {
		strerror_r(errno, err_msg, sizeof(err_msg));
		ERR("socketpair(): %s", err_msg);
		*sock = -1;
		return NULL;
	}

	conn = g_malloc0(sizeof(rfcomm_conn_info_t));
	g_strlcpy(conn->remote_addr, addr, BT_HAL_ADDRESS_STRING_SIZE);
	conn->hal_fd = fds[0];
	conn->stack_fd = -1;
	*sock = fds[1];

	DBG("hal_fd: %d, sock: %d", conn->hal_fd, *sock);

	return conn;
}

int _bt_hal_dbus_handler_rfcomm_connect(unsigned char *addr, unsigned char *uuid, int *sock)
{
	int ret;
	rfcomm_cb_data_t *cb_data;
	rfcomm_conn_info_t *conn;
	char remote_addr[BT_HAL_ADDRESS_STRING_SIZE];
	char remote_uuid[BT_HAL_UUID_STRING_LEN];

	if (!addr) {
		ERR("remote_addr is NULL");
		return BT_STATUS_PARM_INVALID;
	}

	if (!uuid) {
		ERR("remote_uuid is NULL");
		return BT_STATUS_PARM_INVALID;
	}

	if (!sock) {
		ERR("sock is NULL");
		return BT_STATUS_PARM_INVALID;
	}

	_bt_hal_convert_uuid_type_to_string(remote_uuid, uuid);
	cb_data = __get_rfcomm_cb_data(remote_uuid);
	if (!cb_data)
		return BT_STATUS_FAIL;

	_bt_hal_convert_addr_type_to_string(remote_addr, addr);
	DBG("Connecting to %s, uuid %s", remote_addr, remote_uuid);
	conn = __rfcomm_create_conn_info(remote_addr, sock);
	if (!conn) {
		__bt_free_cb_data(cb_data);
		return BT_STATUS_FAIL;
	}

	cb_data->conn_info = conn;
	ret = _bt_hal_discover_services(remote_addr, (char *)remote_uuid,
			__bt_discover_service_response_cb, cb_data);
	if (ret != BT_STATUS_SUCCESS) {
		ERR("Error returned while service discovery");
		__bt_free_conn(conn);
		__bt_free_cb_data(cb_data);
		return BT_STATUS_FAIL;
	}

	INFO("Adding callback information to rfcomm_clients");
	rfcomm_clients = g_slist_append(rfcomm_clients, cb_data);

	return BT_STATUS_SUCCESS;
}

/*************************** RFCOMM Server Implementation ***************************/
static rfcomm_server_data_t *__find_rfcomm_server_info_with_path(const gchar *path)
{
	GSList *l;

	for (l = rfcomm_servers; l != NULL; l = l->next) {
		rfcomm_server_data_t *info = l->data;

		if (g_strcmp0(info->obj_path, path) == 0)
			return info;
	}

	return NULL;
}

static rfcomm_server_data_t *__find_rfcomm_server_info_from_uuid(const char *uuid)
{
	GSList *l;

	for (l = rfcomm_servers; l != NULL; l = l->next) {
		 rfcomm_server_data_t *info = l->data;

		if (g_strcmp0(info->uuid, uuid) == 0)
			return info;
	}

	return NULL;
}

static int __send_sock_fd(int sock_fd, const void *buf, int size, int send_fd)
{
	ssize_t ret;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	char cmsg_buf[CMSG_SPACE(sizeof(int))];
	char err_msg[256] = {0, };

	if (sock_fd == -1 || send_fd == -1)
		return -1;

	memset(&msg, 0, sizeof(msg));
	memset(cmsg_buf, 0, sizeof(cmsg_buf));

	msg.msg_control = cmsg_buf;
	msg.msg_controllen = sizeof(cmsg_buf);

	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_RIGHTS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(send_fd));

	memcpy(CMSG_DATA(cmsg), &send_fd, sizeof(send_fd));

	iov.iov_base = (unsigned char *) buf;
	iov.iov_len = size;

	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	ret = sendmsg(sock_fd, &msg, MSG_NOSIGNAL);
	if (ret < 0) {
		strerror_r(errno, err_msg, sizeof(err_msg));
		ERR("sendmsg(): sock_fd %d send_fd %d: %s",
				sock_fd, send_fd, err_msg);
	}

	return ret;
}

int __new_server_connection(const char *path, int fd, bt_bdaddr_t *addr)
{
	int ret;
	rfcomm_server_data_t *info;
	struct hal_ev_sock_connect ev;

	DBG("%s %d", path, fd);

	if (0 > fd) {
		ERR("Invalid socket fd received");
		return -1;
	}

	info = __find_rfcomm_server_info_with_path(path);
	if (info == NULL) {
		ERR("rfcomm server info not found");
		return -1;
	}

	/* Send rfcomm client connected event */
	memset(&ev, 0, sizeof(ev));
	ev.size = sizeof(ev);
	memcpy(ev.bdaddr, addr->address, 6);
	ev.status = BT_STATUS_SUCCESS;
	ret = __send_sock_fd(info->server_fd, (void *)&ev, sizeof(ev), fd);
	if (ret < 0) {
		ERR("Error sending connect event");
		close(fd);
		return -1;
	}

	/* Remove local reference to client socket fd */
	close(fd);
	return 0;
}

static void __free_rfcomm_server_data(rfcomm_server_data_t *data)
{
	DBG("+");

	if (!data) {
		ERR("server data is NULL");
		return;
	}

	if (data->id >= 0)
		__rfcomm_delete_id(data->id);

	if (data->object_id > 0)
		_bt_hal_unregister_gdbus_object(data->object_id);

	if (data->obj_path) {
		INFO("Unregister profile");
		_bt_hal_unregister_profile(data->obj_path);
	}

	if (0 < data->server_fd)
		close(data->server_fd);

	if (data->server_watch > 0) {
		g_source_remove(data->server_watch);
		data->server_watch = 0;
	}

	g_free(data->obj_path);
	g_free(data->svc_name);
	g_free(data);

	DBG("-");
}

static void __remove_rfcomm_server(rfcomm_server_data_t *data)
{

	DBG("+");

	rfcomm_servers = g_slist_remove(rfcomm_servers, data);
	__free_rfcomm_server_data(data);

	DBG("-");

}

static int __register_rfcomm_server(rfcomm_server_data_t *server_data)
{
	int ret;
	bt_hal_register_profile_info_t profile_info;

	memset(&profile_info, 0x00, sizeof(bt_hal_register_profile_info_t));

	profile_info.authentication = TRUE;
	profile_info.authorization = TRUE;
	profile_info.obj_path = server_data->obj_path;
	profile_info.role = "server";
	profile_info.service = server_data->uuid;
	profile_info.uuid = server_data->uuid;

	INFO("uuid %s, svc: %s, role: %s",
		profile_info.uuid, profile_info.service, profile_info.role);
	ret = _bt_hal_register_profile(&profile_info, TRUE);
	if (ret < 0) {
		ERR("Error: register profile");
		return BT_STATUS_FAIL;
	}

	return BT_STATUS_SUCCESS;
}

static rfcomm_server_data_t *__create_rfcomm_server(char *uuid, const char *svc_name, int *sock)
{
	int id;
	int ret;
	int object_id;
	char *path;
	int fds[2] = {-1, -1};
	rfcomm_server_data_t *data;
	char err_msg[256] = {0, };

	DBG("+");

	data = __find_rfcomm_server_info_from_uuid(uuid);
	if (data) {
		ERR("RFCOMM Server exists with UUID: %s, sv_name: %s", uuid, data->svc_name);
		return NULL;
	}

	id = __rfcomm_assign_id();
	if (id < 0) {
		ERR("__rfcomm_assign_id failed");
		return NULL;
	}

	path = g_strdup_printf("/org/socket/server/%d/%d", getpid(), id);
	object_id = _bt_hal_register_new_gdbus_object(path, __new_server_connection);
	if (object_id < 0) {
		ERR("_bt_hal_register_new_gdbus_object failed");
		g_free(path);
		__rfcomm_delete_id(id);
		return NULL;
	}

	data = g_malloc0(sizeof(rfcomm_server_data_t));
	g_strlcpy(data->uuid, uuid, BT_HAL_UUID_STRING_LEN);
	data->svc_name = g_strdup(svc_name);
	data->obj_path = path;
	data->object_id = object_id;
	data->id = id;

	ret = __register_rfcomm_server(data);
	if (ret != BT_STATUS_SUCCESS) {
		ERR("Error returned while registering service");
		__free_rfcomm_server_data(data);
		return NULL;
	}

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fds) < 0) {
		strerror_r(errno, err_msg, sizeof(err_msg));
		ERR("socketpair(): %s", err_msg);
		__free_rfcomm_server_data(data);
		*sock = -1;
		return NULL;
	}

	data->server_fd = fds[0];
	*sock = fds[1];

	DBG("server_fd: %d, sock: %d", data->server_fd, *sock);
	return data;
}

static gboolean __server_event_cb(GIOChannel *io, GIOCondition cond, gpointer data)
{
	gsize len;
	rfcomm_server_data_t *info = data;
	unsigned char buff[BT_HAL_RFCOMM_MAX_BUFFER_SIZE];
	GError *err = NULL;

	DBG("+");

	if (cond & G_IO_HUP) {
		ERR("Socket %d hang up", info->server_fd);
		goto fail;
	}

	if (cond & (G_IO_ERR | G_IO_NVAL)) {
		ERR("Socket %d error", info->server_fd);
		goto fail;
	}

	/* Read data from application */
	if (g_io_channel_read_chars(io, (gchar *)buff,
			BT_HAL_RFCOMM_MAX_BUFFER_SIZE,
			&len, &err) == G_IO_STATUS_ERROR) {
		if (err)
			ERR("IO Channel read error: %s", err->message);
		else
			ERR("IO Channel read error client");
		goto fail;
	}

	DBG("len: %zu", len);
	if (0 == len) {
		ERR("Other end of socket is closed");
		goto fail;
	}

	DBG("-");
	return TRUE;
fail:
	__remove_rfcomm_server(info);
	return FALSE;
}

int _bt_hal_dbus_handler_rfcomm_listen(const char *svc_name, unsigned char *uuid, int *sock)
{
	int chan = 0;
	rfcomm_server_data_t *server_data;
	char server_uuid[BT_HAL_UUID_STRING_LEN];
	GIOCondition cond;
	GIOChannel *io;

	if (!svc_name) {
		ERR("svc_name is NULL");
		return BT_STATUS_PARM_INVALID;
	}

	if (!uuid) {
		ERR("uuid is NULL");
		return BT_STATUS_PARM_INVALID;
	}

	if (!sock) {
		ERR("sock is NULL");
		return BT_STATUS_PARM_INVALID;
	}

	_bt_hal_convert_uuid_type_to_string(server_uuid, uuid);
	server_data = __create_rfcomm_server(server_uuid, svc_name, sock);
	if (!server_data)
		return BT_STATUS_BUSY;

	/* TODO: Temperary, later need to fill correct channel form correct place */
	if (write(server_data->server_fd, &chan, sizeof(chan)) != sizeof(chan)) {
		ERR("Error sending RFCOMM channel");
		__free_rfcomm_server_data(server_data);
		*sock = -1;
		return BT_STATUS_FAIL;
	}

	/* Rfcomm server: Handle events from Application */
	cond = G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL;
	io = g_io_channel_unix_new(server_data->server_fd);
	server_data->server_watch = g_io_add_watch(io, cond, __server_event_cb, server_data);
	g_io_channel_set_flags(io, G_IO_FLAG_NONBLOCK, NULL);
	g_io_channel_unref(io);

	INFO("Adding server information to rfcomm_servers list");
	rfcomm_servers = g_slist_append(rfcomm_servers, server_data);

	DBG("-");
	return BT_STATUS_SUCCESS;
}

gboolean _is_rfcomm_server_uuid(const char *uuid)
{
	DBG("+");

	if (!uuid)
		return FALSE;

	if (NULL == __find_rfcomm_server_info_from_uuid(uuid))
		return FALSE;

	DBG("-");
	return TRUE;
}
