/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *			   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include <dlog.h>
#include <string.h>
#include <bluetooth.h>
#include "oal-utils.h"

char *bdt_bd2str(const bt_address_t *bdaddr, bdstr_t *bdstr)
{
	const uint8_t *addr = bdaddr->addr;

	if (bdaddr == NULL) {
		*bdstr[0] = 0;
		return *bdstr;
	}

	snprintf(*bdstr, sizeof(bdstr_t), "%02x:%02x:%02x:%02x:%02x:%02x",
			addr[0], addr[1], addr[2],
			addr[3], addr[4], addr[5]);
	return *bdstr;
}

char* convert_bdaddr_2_str(const bt_bdaddr_t *bd_addr, char *buf)
{
	const uint8_t *p;

	if (!bd_addr)
		return strncpy(buf, "NULL", sizeof("NULL"));
	p = bd_addr->address;

	snprintf(buf, 18, "%02x:%02x:%02x:%02x:%02x:%02x",
			p[0], p[1], p[2], p[3], p[4], p[5]);

	return buf;
}

char *bdaddr_2_str(const bt_bdaddr_t *bd_addr)
{
	static char buf[18];
	return convert_bdaddr_2_str(bd_addr, buf);
}

void string_to_uuid(char *str, service_uuid_t *p_uuid)
{
	uint32_t uuid0, uuid4;
	uint16_t uuid1, uuid2, uuid3, uuid5;

	sscanf(str, "%08x-%04hx-%04hx-%04hx-%08x%04hx",
			&uuid0, &uuid1, &uuid2, &uuid3, &uuid4, &uuid5);

	uuid0 = htonl(uuid0);
	uuid1 = htons(uuid1);
	uuid2 = htons(uuid2);
	uuid3 = htons(uuid3);
	uuid4 = htonl(uuid4);
	uuid5 = htons(uuid5);

	memcpy(&(p_uuid->uuid[0]), &uuid0, 4);
	memcpy(&(p_uuid->uuid[4]), &uuid1, 2);
	memcpy(&(p_uuid->uuid[6]), &uuid2, 2);
	memcpy(&(p_uuid->uuid[8]), &uuid3, 2);
	memcpy(&(p_uuid->uuid[10]), &uuid4, 4);
	memcpy(&(p_uuid->uuid[14]), &uuid5, 2);

	return;
}

void oal_print_device_address_t(const bt_address_t *addr)
{
	BT_INFO("%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X\n", addr->addr[0], addr->addr[1], addr->addr[2],
			addr->addr[3], addr->addr[4], addr->addr[5]);
}

void oal_convert_addr_string_to_type(unsigned char *addr,
		const char *address)
{
	int i;
	char *ptr = NULL;

	for (i = 0; i < BT_ADDRESS_BYTES_NUM; i++) {
		addr[i] = strtol(address, &ptr, 16);
		if (ptr != NULL) {
			if (ptr[0] != ':')
				return;
			address = ptr + 1;
		}
	}
}

int oal_is_address_zero(unsigned char *addr1)
{
	int i;
	for (i = 0; i < BT_ADDRESS_BYTES_NUM; i++) {
		if (addr1[i] == 0)
			continue;

		break;
	}
	if (i == BT_ADDRESS_BYTES_NUM)
		return 1;
	else
		return 0;
}

void uuid_to_string(service_uuid_t *p_uuid, char *str)
{
	uint32_t uuid0, uuid4;
	uint16_t uuid1, uuid2, uuid3, uuid5;

	memcpy(&uuid0, &(p_uuid->uuid[0]), 4);
	memcpy(&uuid1, &(p_uuid->uuid[4]), 2);
	memcpy(&uuid2, &(p_uuid->uuid[6]), 2);
	memcpy(&uuid3, &(p_uuid->uuid[8]), 2);
	memcpy(&uuid4, &(p_uuid->uuid[10]), 4);
	memcpy(&uuid5, &(p_uuid->uuid[14]), 2);

	snprintf((char *)str, BT_UUID_STRING_MAX, "%.8x-%.4x-%.4x-%.4x-%.8x%.4x",
			ntohl(uuid0), ntohs(uuid1),
			ntohs(uuid2), ntohs(uuid3),
			ntohl(uuid4), ntohs(uuid5));
}

void uuid_to_stringname(service_uuid_t *p_uuid, char *str)
{
	uint32_t uuid0, uuid4;
	uint16_t uuid1, uuid2, uuid3, uuid5;
	const char *uuid_name;
	const char *uuid_name1;

	memcpy(&uuid0, &(p_uuid->uuid[0]), 4);
	memcpy(&uuid1, &(p_uuid->uuid[4]), 2);
	memcpy(&uuid2, &(p_uuid->uuid[6]), 2);
	memcpy(&uuid3, &(p_uuid->uuid[8]), 2);
	memcpy(&uuid4, &(p_uuid->uuid[10]), 4);
	memcpy(&uuid5, &(p_uuid->uuid[14]), 2);


	uuid_name = dump_uuid_name(ntohl(uuid0));
	uuid_name1 = dump_uuid_name((ntohl(uuid4) >> 16));

	BT_DBG("UUID Name [%s]", uuid_name);
	BT_DBG("UUID Name Shifted [%s]", uuid_name1);

	snprintf((char *)str, 2*BT_UUID_STRING_MAX, "%.8x-%.4x-%.4x-%.4x-%.8x%.4x--%s",
			ntohl(uuid0), ntohs(uuid1),
			ntohs(uuid2), ntohs(uuid3),
			ntohl(uuid4), ntohs(uuid5),
			!(g_strcmp0(dump_uuid_name(ntohl(uuid0)), "--")) ? dump_uuid_name(((uuid4) >> 16)) : uuid_name);
}

int hex2bin(const char *s)
{
	int ret = 0;
	int i;
	for (i = 0; i < 2; i++) {
		char c = *s++;
		int n = 0;
		if ('0' <= c && c <= '9')
			n = c-'0';
		else if ('a' <= c && c <= 'f')
			n = 10 + c-'a';
		else if ('A' <= c && c <= 'F')
			n = 10 + c-'A';
		ret = n + ret*16;
	}
	return ret;
}

void convert_str_2_hex(unsigned char out[], char in[])
{
	int i = 0;
	for (i = 0; i < 62; i++) {
		out[i] = hex2bin(in); \
			 in += 2; \
	}
}

void convert_hex_2_str(unsigned char * hex, int len, char * str_out)
{
	int i = 0;

	for (i = 0; i < len; i++)
		snprintf(str_out + (i * 3), 3*(len - i), "%02x ", hex[i]);

	str_out[3*len] = 0;
}

void print_bt_properties(int num_properties, bt_property_t *properties)
{
	int i;
	for (i = 0; i < num_properties; i++) {
		bt_property_t prop;
		memcpy(&prop, properties + i, sizeof(prop));
		BT_INFO("prop: %s\n", convert_bt_property_2_str(&prop));
	}
}

char* convert_scan_mode_2_str(bt_scan_mode_t scan_mode)
{
	switch (scan_mode) {
	case BT_SCAN_MODE_NONE:
		return "Non Scannable";
	case BT_SCAN_MODE_CONNECTABLE_DISCOVERABLE:
		return "Connectable And Discoverable";
	case BT_SCAN_MODE_CONNECTABLE:
		return "Connectable";
	default:
		return "Unknown Scan Mode";
	}

}

char* convert_device_type_2_str(bt_device_type_t device_type)
{
	switch (device_type) {
	case BT_DEVICE_DEVTYPE_BREDR:
		return "BREDR Device";
	case BT_DEVICE_DEVTYPE_BLE:
		return "BLE Device";
	case BT_DEVICE_DEVTYPE_DUAL:
		return "Dual Device";
	default:
		return "Unknown Device Type";
	}
}

char *convert_bt_property_2_str(const bt_property_t *property)
{
	static char buf[4096];
	char *p;

	p = buf + snprintf(buf, 4096, "type=%s len=%d val=",
			convert_property_type_2_str(property->type),
			property->len);

	switch (property->type) {
	case BT_PROPERTY_BDNAME:
	case BT_PROPERTY_REMOTE_FRIENDLY_NAME:
		snprintf(p, property->len + 1, "%s",
				((bt_bdname_t *) property->val)->name);
		break;
	case BT_PROPERTY_BDADDR:
		sprintf(p, "%s", bdaddr_2_str((bt_bdaddr_t *) property->val));
		break;
	case BT_PROPERTY_CLASS_OF_DEVICE:
		sprintf(p, "%06x", *((unsigned int *) property->val));
		break;
	case BT_PROPERTY_TYPE_OF_DEVICE:
		sprintf(p, "%s", convert_device_type_2_str(
					*((bt_device_type_t *) property->val)));
		break;
	case BT_PROPERTY_REMOTE_RSSI:
		sprintf(p, "%d", *((char *) property->val));
		break;
	case BT_PROPERTY_ADAPTER_DISCOVERY_TIMEOUT:
		sprintf(p, "%d", *((unsigned int *) property->val));
		break;
	case BT_PROPERTY_ADAPTER_SCAN_MODE:
		sprintf(p, "%s",
				convert_scan_mode_2_str(*((bt_scan_mode_t *) property->val)));
		break;
	case BT_PROPERTY_ADAPTER_BONDED_DEVICES:
		break;
	case BT_PROPERTY_UUIDS:
		break;
	case BT_PROPERTY_SERVICE_RECORD:
		break;
		/* Tizen BlueZ specific Device properties */
	case BT_PROPERTY_REMOTE_PAIRED:
		sprintf(p, "%d", *((uint8_t *) property->val));
		break;
	case BT_PROPERTY_REMOTE_CONNECTED:
		sprintf(p, "%d", *((unsigned int *) property->val));
		break;
	case BT_PROPERTY_REMOTE_TRUST:
		sprintf(p, "%d", *((uint8_t *) property->val));
		break;
	case BT_PROPERTY_PAIRABLE:
		sprintf(p, "%d", *((uint8_t *) property->val));
		break;
	case BT_PROPERTY_VERSION:
		snprintf(p, property->len + 1, "%s",
				((char *) property->val));
		break;
	case BT_PROPERTY_LOCAL_LE_FEATURES:
		local_le_feat_2_string(p, property->val);
		break;
	case BT_PROPERTY_PAIRABLE_TIMEOUT:
		sprintf(p, "%d", *((unsigned int *) property->val));
		break;
	case BT_PROPERTY_IPSP_INITIALIZED:
		sprintf(p, "%d", *((uint8_t *) property->val));
		break;
	case BT_PROPERTY_MODALIAS:
		snprintf(p, property->len + 1, "%s",
				((char *) property->val));
		break;
	case BT_PROPERTY_REMOTE_DEVICE_MANUFACTURER_DATA_LEN:
		sprintf(p, "%d", *((unsigned int *) property->val));
		break;
	case BT_PROPERTY_REMOTE_DEVICE_MANUFACTURER_DATA: {
		int indx;
		char *pppp = property->val;
		for (indx = 0; indx < property->len; indx++)
			p += sprintf(p, " %2.2X", pppp[indx]);
		break;
	}
	case BT_PROPERTY_REMOTE_BLE_ADV_DATA: {
		int indx;
		char *pppp = property->val;
		for (indx = 0; indx < property->len; indx++)
			p += sprintf(p, " %2.2X", pppp[indx]);
		break;
	}
	case BT_PROPERTY_REMOTE_IS_ALIAS_SET: {
		sprintf(p, "%d", *((uint8_t *) property->val));
		break;
	}
	/* End of Tizen BlueZ specific device propeties */
	case BT_PROPERTY_REMOTE_VERSION_INFO:
	case BT_PROPERTY_REMOTE_DEVICE_TIMESTAMP:
	default:
		sprintf(p, "%p", property->val);
		break;
	}
	return buf;
}

char *dump_char_prop(uint8_t prop)
{
	switch (prop) {
	CASE_RETURN_STR(BROADCAST);
	CASE_RETURN_STR(READ);
	CASE_RETURN_STR(WRITE_NR);
	CASE_RETURN_STR(WRITE);
	CASE_RETURN_STR(NOTIFY);
	CASE_RETURN_STR(INDICATE);
	CASE_RETURN_STR(AUTH);
	CASE_RETURN_STR(EXT_PROP);
	default:
		return "--";
	}
}

char *dump_char_perm(uint8_t perm)
{
	switch (perm) {
	CASE_RETURN_STR(RD);
	CASE_RETURN_STR(WR);
	CASE_RETURN_STR(ENCRYPT_RD);
	CASE_RETURN_STR(ENCRYPT_WR);
	CASE_RETURN_STR(ENC_AUTH_RD);
	CASE_RETURN_STR(ENC_AUTH_WR);
	default:
		return "--";
	}
}

char *char_prop_to_string(uint8_t prop, char *str)
{
	int i, ret = 0;
	char *ptr = str;

	for (i = 0; i < 8; i++) {
		if (prop & (1 << i)) {
			ret = snprintf(ptr, 11, "%s|", dump_char_prop(prop & (1 << i)));
			ptr += ret;
		}
	}
	--ptr;
	*ptr = '\0';
	return str;
}

char *char_perm_to_string(uint8_t perm, char *str)
{
	int i, ret = 0;
	char *ptr = str;

	for (i = 0; i < 6; i++) {
		if (perm & (1 << i)) {
			ret = snprintf(ptr, 13, "%s|", dump_char_perm(perm & (1 << i)));
			ptr += ret;
		}
	}
	--ptr;
	*ptr = '\0';
	return str;
}

void local_le_feat_2_string(char *str, const bt_local_le_features_t *f)
{
	uint16_t scan_num;

	str += snprintf(str, strlen(str), "{\n");

	str += sprintf(str, "Privacy supported: %s,\n",
			f->local_privacy_enabled ? "TRUE" : "FALSE");

	str += sprintf(str, "Num of advertising instances: %u,\n",
			f->max_adv_instance);

	str += sprintf(str, "PRA offloading support: %s,\n",
			f->rpa_offload_supported ? "TRUE" : "FALSE");

	str += sprintf(str, "Num of offloaded IRKs: %u,\n",
			f->max_irk_list_size);

	str += sprintf(str, "Num of offloaded scan filters: %u,\n",
			f->max_adv_filter_supported);

	scan_num = f->scan_result_storage_size;

	str += sprintf(str, "Num of offloaded scan results: %u,\n", scan_num);

	str += sprintf(str, "Activity & energy report support: %s\n",
			f->activity_energy_info_supported ? "TRUE" : "FALSE");

	str += sprintf(str, "LE 2M PHY support: %s\n",
			f->le_2m_phy_supported ? "TRUE" : "FALSE");

	str += sprintf(str, "LE Coded PHY support: %s\n",
			f->le_coded_phy_supported ? "TRUE" : "FALSE");

	sprintf(str, "}");
}

char* convert_property_type_2_str(bt_property_type_t prop_type)
{
	switch (prop_type) {
	case BT_PROPERTY_BDNAME:
		return "[Bluetooth Name]";
	case BT_PROPERTY_BDADDR:
		return "[Bluetooth Address]";
	case BT_PROPERTY_UUIDS:
		return "[UUIDS]";
	case BT_PROPERTY_CLASS_OF_DEVICE:
		return "[Class of Device]";
	case BT_PROPERTY_TYPE_OF_DEVICE:
		return "[Bluetooth Type of Device]";
	case BT_PROPERTY_SERVICE_RECORD:
		return "[Bluetooth Service record]";
	case BT_PROPERTY_ADAPTER_SCAN_MODE:
		return "[Bluetooth Adapter Scan Mode]";
	case BT_PROPERTY_ADAPTER_BONDED_DEVICES:
		return "[Bluetooth Bonded Devices]";
	case BT_PROPERTY_ADAPTER_DISCOVERY_TIMEOUT:
		return "[Bluetooth Adapter Discovery Timeout]";
	case BT_PROPERTY_REMOTE_FRIENDLY_NAME:
		return "[Bluetooth Friendly Name]";
	case BT_PROPERTY_REMOTE_RSSI:
		return "[Bluetooth Rmote RSSI]";
	case BT_PROPERTY_REMOTE_VERSION_INFO:
		return "[Bluetooth Version Info]";
	case BT_PROPERTY_LOCAL_LE_FEATURES:
		return "[Bluetooth LE Features]";
	case BT_PROPERTY_REMOTE_PAIRED:
		return "[Bluetooth Remote Paired]";
	case BT_PROPERTY_REMOTE_CONNECTED:
		return "[Bluetooth Remote Connected]";
	case BT_PROPERTY_REMOTE_TRUST:
		return "[Bluetooth Remote TRUST]";
	case BT_PROPERTY_PAIRABLE:
		return "[Bluetooth Pairable]";
	case BT_PROPERTY_PAIRABLE_TIMEOUT:
		return "[Bluetooth Pairable Timeout]";
	case BT_PROPERTY_VERSION:
		return "[Bluetooth Version]";
	case BT_PROPERTY_IPSP_INITIALIZED:
		return "[Bluetooth IPSP Initialized]";
	case BT_PROPERTY_MODALIAS:
		return "[Bluetooth ModAlias]";
	case BT_PROPERTY_REMOTE_DEVICE_TIMESTAMP:
		return "[Bluetooth Remote Device Timestamp]";
	case BT_PROPERTY_REMOTE_DEVICE_MANUFACTURER_DATA_LEN:
		return "[Bluetooth Remote Device Manufacturer Data Len]";
	case BT_PROPERTY_REMOTE_DEVICE_MANUFACTURER_DATA:
		return "[Bluetooth Remote Device Manufacturer Data]";
	case BT_PROPERTY_REMOTE_BLE_ADV_DATA:
		return "[Bluetooth Remote Device LE Advertising Data]";
	default:
		return "[Default Property]";
	}
}

const char *dump_uuid_name(int uuid_no)
{
	switch (uuid_no) {
	CASE_RETURN_STR(BASE)
	CASE_RETURN_STR(SDP)
	CASE_RETURN_STR(UDP)
	CASE_RETURN_STR(RFCOMM)
	CASE_RETURN_STR(TCP)
	CASE_RETURN_STR(TCS_BIN)
	CASE_RETURN_STR(TCS_AT)
	CASE_RETURN_STR(ATT)
	CASE_RETURN_STR(OBEX)
	CASE_RETURN_STR(IP)
	CASE_RETURN_STR(FTP)
	CASE_RETURN_STR(HTTP)
	CASE_RETURN_STR(WSP)
	CASE_RETURN_STR(BNEP)
	CASE_RETURN_STR(UPNP)
	CASE_RETURN_STR(HIDP)
	CASE_RETURN_STR(HardcopyControlChannel)
	CASE_RETURN_STR(HardcopyDataChannel)
	CASE_RETURN_STR(HardcopyNotification)
	CASE_RETURN_STR(AVCTP)
	CASE_RETURN_STR(AVDTP)
	CASE_RETURN_STR(CMTP)
	CASE_RETURN_STR(MCAPControlChannel)
	CASE_RETURN_STR(MCAPDataChannel)
	CASE_RETURN_STR(L2CAP)
	CASE_RETURN_STR(ServiceDiscoveryServerServiceClassID)
	CASE_RETURN_STR(BrowseGroupDescriptorServiceClassID)
	CASE_RETURN_STR(SerialPort)
	CASE_RETURN_STR(LANAccessUsingPPP)
	CASE_RETURN_STR(DialupNetworking)
	CASE_RETURN_STR(IrMCSync)
	CASE_RETURN_STR(OBEXObjectPush)
	CASE_RETURN_STR(OBEXFileTransfer)
	CASE_RETURN_STR(IrMCSyncCommand)
	CASE_RETURN_STR(Headset)
	CASE_RETURN_STR(CordlessTelephony)
	CASE_RETURN_STR(AudioSource)
	CASE_RETURN_STR(AudioSink)
	CASE_RETURN_STR(AV_RemoteControlTarget)
	CASE_RETURN_STR(AdvancedAudioDistribution)
	CASE_RETURN_STR(AV_RemoteControl)
	CASE_RETURN_STR(AV_RemoteControlController)
	CASE_RETURN_STR(Intercom)
	CASE_RETURN_STR(Fax)
	CASE_RETURN_STR(Headset_Audio_Gateway)
	CASE_RETURN_STR(WAP)
	CASE_RETURN_STR(WAP_CLIENT)
	CASE_RETURN_STR(PANU)
	CASE_RETURN_STR(NAP)
	CASE_RETURN_STR(GN)
	CASE_RETURN_STR(DirectPrinting)
	CASE_RETURN_STR(ReferencePrinting)
	CASE_RETURN_STR(Basic_Imaging_Profile)
	CASE_RETURN_STR(ImagingResponder)
	CASE_RETURN_STR(ImagingAutomaticArchive)
	CASE_RETURN_STR(ImagingReferencedObjects)
	CASE_RETURN_STR(Handsfree)
	CASE_RETURN_STR(HandsfreeAudioGateway)
	CASE_RETURN_STR(DirectPrintingReferenceObjectsService)
	CASE_RETURN_STR(ReflectedUI)
	CASE_RETURN_STR(BasicPrinting)
	CASE_RETURN_STR(PrintingStatus)
	CASE_RETURN_STR(HumanInterfaceDeviceService)
	CASE_RETURN_STR(HardcopyCableReplacement)
	CASE_RETURN_STR(HCR_Print)
	CASE_RETURN_STR(HCR_Scan)
	CASE_RETURN_STR(Common_ISDN_Access)
	CASE_RETURN_STR(SIM_Access)
	CASE_RETURN_STR(Phonebook_Access_PCE)
	CASE_RETURN_STR(Phonebook_Access_PSE)
	CASE_RETURN_STR(Phonebook_Access)
	CASE_RETURN_STR(Headset_HS)
	CASE_RETURN_STR(Message_Access_Server)
	CASE_RETURN_STR(Message_Notification_Server)
	CASE_RETURN_STR(Message_Access_Profile)
	CASE_RETURN_STR(GNSS)
	CASE_RETURN_STR(GNSS_Server)
	CASE_RETURN_STR(ThreeD_Display)
	CASE_RETURN_STR(ThreeD_Glasses)
	CASE_RETURN_STR(ThreeD_Synchronization)
	CASE_RETURN_STR(MPS_Profile)
	CASE_RETURN_STR(MPS_SC)
	CASE_RETURN_STR(CTN_Access_Service)
	CASE_RETURN_STR(CTN_Notification_Service)
	CASE_RETURN_STR(CTN_Profile)
	CASE_RETURN_STR(PnPInformation)
	CASE_RETURN_STR(GenericNetworking)
	CASE_RETURN_STR(GenericFileTransfer)
	CASE_RETURN_STR(GenericAudio)
	CASE_RETURN_STR(GenericTelephony)
	CASE_RETURN_STR(UPNP_Service)
	CASE_RETURN_STR(UPNP_IP_Service)
	CASE_RETURN_STR(ESDP_UPNP_IP_PAN)
	CASE_RETURN_STR(ESDP_UPNP_IP_LAP)
	CASE_RETURN_STR(ESDP_UPNP_L2CAP)
	CASE_RETURN_STR(VideoSource)
	CASE_RETURN_STR(VideoSink)
	CASE_RETURN_STR(VideoDistribution)
	CASE_RETURN_STR(HDP)
	CASE_RETURN_STR(HDP_Source)
	CASE_RETURN_STR(HDP_Sink)
	CASE_RETURN_STR(Generic_Access)
	CASE_RETURN_STR(Generic_Attribute)
	CASE_RETURN_STR(Immediate_Alert)
	CASE_RETURN_STR(Link_Loss)
	CASE_RETURN_STR(Tx_Power)
	CASE_RETURN_STR(Current_Time)
	CASE_RETURN_STR(Reference_Time_Update)
	CASE_RETURN_STR(Next_Dst_Change)
	CASE_RETURN_STR(Glucose)
	CASE_RETURN_STR(Health_Thermometer)
	CASE_RETURN_STR(Device_Information)
	CASE_RETURN_STR(Heart_Rate)
	CASE_RETURN_STR(Phone_Alert_Status)
	CASE_RETURN_STR(Battery_Service)
	CASE_RETURN_STR(Blood_Pressure)
	CASE_RETURN_STR(Alert_Notification)
	CASE_RETURN_STR(Human_Interface_Device)
	CASE_RETURN_STR(Scan_Parameters)
	CASE_RETURN_STR(Running_Speed_And_Cadence)
	CASE_RETURN_STR(Automation_IO)
	CASE_RETURN_STR(Cycling_Speed_And_Cadence)
	CASE_RETURN_STR(Cycling_Power)
	CASE_RETURN_STR(Location_And_Navigation)
	CASE_RETURN_STR(Environmental_Sensing)
	CASE_RETURN_STR(Body_Composition)
	CASE_RETURN_STR(User_Data)
	CASE_RETURN_STR(Weight_Scale)
	CASE_RETURN_STR(Bond_Management)
	CASE_RETURN_STR(Continuous_Glucose_Monitoring)
	CASE_RETURN_STR(Internet_Protocol_Support)
	CASE_RETURN_STR(Indoor_Positioning)
	CASE_RETURN_STR(Pulse_Oximeter)
	CASE_RETURN_STR(Http_Proxy)
	CASE_RETURN_STR(Transport_Discovery)
	CASE_RETURN_STR(Object_Transfer)
	CASE_RETURN_STR(Gap_Device_Name)
	CASE_RETURN_STR(Gap_Appearance)
	CASE_RETURN_STR(Gap_Peripheral_Privacy_Flag)
	CASE_RETURN_STR(Gap_Reconnection_Address)
	CASE_RETURN_STR(Gap_Peripheral_Preferred_Connection_Parameters)
	CASE_RETURN_STR(Gatt_Service_Changed)
	CASE_RETURN_STR(Alert_Level)
	CASE_RETURN_STR(Tx_Power_Level)
	CASE_RETURN_STR(Date_Time)
	CASE_RETURN_STR(Day_Of_Week)
	CASE_RETURN_STR(Day_Date_Time)
	CASE_RETURN_STR(Exact_Time_256)
	CASE_RETURN_STR(Dst_Offset)
	CASE_RETURN_STR(Time_Zone)
	CASE_RETURN_STR(Local_Time_Information)
	CASE_RETURN_STR(Time_With_Dst)
	CASE_RETURN_STR(Time_Accuracy)
	CASE_RETURN_STR(Time_Source)
	CASE_RETURN_STR(Reference_Time_Information)
	CASE_RETURN_STR(Time_Update_Control_Point)
	CASE_RETURN_STR(Time_Update_State)
	CASE_RETURN_STR(Glucose_Measurement)
	CASE_RETURN_STR(Battery_Level)
	CASE_RETURN_STR(Temperature_Measurement)
	CASE_RETURN_STR(Temperature_Type)
	CASE_RETURN_STR(Intermediate_Temperature)
	CASE_RETURN_STR(Measurement_Interval)
	CASE_RETURN_STR(Boot_Keyboard_Input_Report)
	CASE_RETURN_STR(System_Id)
	CASE_RETURN_STR(Model_Number_String)
	CASE_RETURN_STR(Serial_Number_String)
	CASE_RETURN_STR(Firmware_Revision_String)
	CASE_RETURN_STR(Hardware_Revision_String)
	CASE_RETURN_STR(Software_Revision_String)
	CASE_RETURN_STR(Manufacturer_Name_String)
	CASE_RETURN_STR(Regulatory_Certification_Data_List)
	CASE_RETURN_STR(Current_Time_Charac)
	CASE_RETURN_STR(Magnetic_Declination)
	CASE_RETURN_STR(Scan_Refresh)
	CASE_RETURN_STR(Boot_Keyboard_Output_Report)
	CASE_RETURN_STR(Boot_Mouse_Input_Report)
	CASE_RETURN_STR(Glucose_Measurement_Context)
	CASE_RETURN_STR(Blood_Pressure_Measurement)
	CASE_RETURN_STR(Intermediate_Cuff_Pressure)
	CASE_RETURN_STR(Heart_Rate_Measurement)
	CASE_RETURN_STR(Body_Sensor_Location)
	CASE_RETURN_STR(Heart_Rate_Control_Point)
	CASE_RETURN_STR(Alert_Status)
	CASE_RETURN_STR(Ringer_Control_Point)
	CASE_RETURN_STR(Ringer_Setting)
	CASE_RETURN_STR(Alert_Category_Id_Bit_Mask)
	CASE_RETURN_STR(Alert_Category_Id)
	CASE_RETURN_STR(Alert_Notification_Control_Point)
	CASE_RETURN_STR(Unread_Alert_Status)
	CASE_RETURN_STR(New_Alert)
	CASE_RETURN_STR(Supported_New_Alert_Category)
	CASE_RETURN_STR(Supported_Unread_Alert_Category)
	CASE_RETURN_STR(Blood_Pressure_Feature)
	CASE_RETURN_STR(Hid_Information)
	CASE_RETURN_STR(Report_Map)
	CASE_RETURN_STR(Hid_Control_Point)
	CASE_RETURN_STR(Report)
	CASE_RETURN_STR(Protocol_Mode)
	CASE_RETURN_STR(Scan_Interval_Window)
	CASE_RETURN_STR(Pnp_Id)
	CASE_RETURN_STR(Glucose_Feature)
	CASE_RETURN_STR(Record_Access_Control_Point)
	CASE_RETURN_STR(Rsc_Measurement)
	CASE_RETURN_STR(Rsc_Feature)
	CASE_RETURN_STR(Sc_Control_Point)
	CASE_RETURN_STR(Digital)
	CASE_RETURN_STR(Analog)
	CASE_RETURN_STR(Aggregate)
	CASE_RETURN_STR(Csc_Measurement)
	CASE_RETURN_STR(Csc_Feature)
	CASE_RETURN_STR(Sensor_Location)
	CASE_RETURN_STR(Plx_Spot_Check_Measurement)
	CASE_RETURN_STR(Plx_Continuous_Measurement)
	CASE_RETURN_STR(Plx_Features)
	CASE_RETURN_STR(Cycling_Power_Measurement)
	CASE_RETURN_STR(Cycling_Power_Vector)
	CASE_RETURN_STR(Cycling_Power_Feature)
	CASE_RETURN_STR(Cycling_Power_Control_Point)
	CASE_RETURN_STR(Location_And_Speed)
	CASE_RETURN_STR(Navigation)
	CASE_RETURN_STR(Position_Quality)
	CASE_RETURN_STR(Ln_Feature)
	CASE_RETURN_STR(Ln_Control_Point)
	CASE_RETURN_STR(Elevation)
	CASE_RETURN_STR(Pressure)
	CASE_RETURN_STR(Temperature)
	CASE_RETURN_STR(Humidity)
	CASE_RETURN_STR(True_Wind_Speed)
	CASE_RETURN_STR(True_Wind_Direction)
	CASE_RETURN_STR(Apparent_Wind_Speed)
	CASE_RETURN_STR(Apparent_Wind_Direction)
	CASE_RETURN_STR(Gust_Factor)
	CASE_RETURN_STR(Pollen_Concentration)
	CASE_RETURN_STR(Uv_Index)
	CASE_RETURN_STR(Irradiance)
	CASE_RETURN_STR(Rainfall)
	CASE_RETURN_STR(Wind_Chill)
	CASE_RETURN_STR(Heat_Index)
	CASE_RETURN_STR(Dew_Point)
	CASE_RETURN_STR(Descriptor_Value_Changed)
	CASE_RETURN_STR(Aerobic_Threshold)
	CASE_RETURN_STR(Age)
	CASE_RETURN_STR(Anaerobic_Heart_Rate_Lower_Limit)
	CASE_RETURN_STR(Anaerobic_Heart_Rate_Upper_Limit)
	CASE_RETURN_STR(Anaerobic_Threshold)
	CASE_RETURN_STR(Aerobic_Heart_Rate_Upper_Limit)
	CASE_RETURN_STR(Date_Of_Birth)
	CASE_RETURN_STR(Date_Of_Threshold_Assessment)
	CASE_RETURN_STR(Email_Address)
	CASE_RETURN_STR(Fat_Burn_Heart_Rate_Lower_Limit)
	CASE_RETURN_STR(Fat_Burn_Heart_Rate_Upper_Limit)
	CASE_RETURN_STR(First_Name)
	CASE_RETURN_STR(Five_Zone_Heart_Rate_Limits)
	CASE_RETURN_STR(Gender)
	CASE_RETURN_STR(Heart_Rate_Max)
	CASE_RETURN_STR(Height)
	CASE_RETURN_STR(Hip_Circumference)
	CASE_RETURN_STR(Last_Name)
	CASE_RETURN_STR(Maximum_Recommended_Heart_Rate)
	CASE_RETURN_STR(Resting_Heart_Rate)
	CASE_RETURN_STR(Sport_Type_For_Aerobic_And_Anaerobic_Thresholds)
	CASE_RETURN_STR(Three_Zone_Heart_Rate_Limits)
	CASE_RETURN_STR(Two_Zone_Heart_Rate_Limit)
	CASE_RETURN_STR(Vo2_Max)
	CASE_RETURN_STR(Waist_Circumference)
	CASE_RETURN_STR(Weight)
	CASE_RETURN_STR(Database_Change_Increment)
	CASE_RETURN_STR(User_Index)
	CASE_RETURN_STR(Body_Composition_Feature)
	CASE_RETURN_STR(Body_Composition_Measurement)
	CASE_RETURN_STR(Weight_Measurement)
	CASE_RETURN_STR(Weight_Scale_Feature)
	CASE_RETURN_STR(User_Control_Point)
	CASE_RETURN_STR(Magnetic_Flux_Density_2D)
	CASE_RETURN_STR(Magnetic_Flux_Density_3D)
	CASE_RETURN_STR(Language)
	CASE_RETURN_STR(Barometric_Pressure_Trend)
	CASE_RETURN_STR(Bond_Management_Control_Point)
	CASE_RETURN_STR(Bond_Management_Feature)
	CASE_RETURN_STR(Gap_Central_Address_Resolution_Support)
	CASE_RETURN_STR(Cgm_Measurement)
	CASE_RETURN_STR(Cgm_Feature)
	CASE_RETURN_STR(Cgm_Status)
	CASE_RETURN_STR(Cgm_Session_Start_Time)
	CASE_RETURN_STR(Cgm_Session_Run_Time)
	CASE_RETURN_STR(Cgm_Specific_Ops_Control_Point)
	CASE_RETURN_STR(Indoor_Positioning_Configuration)
	CASE_RETURN_STR(Latitude)
	CASE_RETURN_STR(Longitude)
	CASE_RETURN_STR(Local_North_Coordinate)
	CASE_RETURN_STR(Local_East_Coordinate)
	CASE_RETURN_STR(Floor_Number)
	CASE_RETURN_STR(Altitude)
	CASE_RETURN_STR(Uncertainty)
	CASE_RETURN_STR(Location_Name)
	CASE_RETURN_STR(Uri)
	CASE_RETURN_STR(Http_Headers)
	CASE_RETURN_STR(Http_Status_Code)
	CASE_RETURN_STR(Http_Entity_Body)
	CASE_RETURN_STR(Http_Control_Point)
	CASE_RETURN_STR(Https_Security)
	CASE_RETURN_STR(Tds_Control_Point)
	CASE_RETURN_STR(Ots_Feature)
	CASE_RETURN_STR(Object_Name)
	CASE_RETURN_STR(Object_Type)
	CASE_RETURN_STR(Object_Size)
	CASE_RETURN_STR(Object_First_Created)
	CASE_RETURN_STR(Object_Last_Modified)
	CASE_RETURN_STR(Object_Id)
	CASE_RETURN_STR(Object_Properties)
	CASE_RETURN_STR(Object_Action_Control_Point)
	CASE_RETURN_STR(Object_List_Control_Point)
	CASE_RETURN_STR(Object_List_Filter)
	CASE_RETURN_STR(Object_Changed)
	CASE_RETURN_STR(Fitness_Machine_Control_Point)
	CASE_RETURN_STR(Gatt_Characteristic_Extended_Properties)
	CASE_RETURN_STR(Gatt_Characteristic_User_Description)
	CASE_RETURN_STR(Gatt_Client_Characteristic_Configuration)
	CASE_RETURN_STR(Gatt_Server_Characteristic_Configuration)
	CASE_RETURN_STR(Gatt_Characteristic_Presentation_Format)
	CASE_RETURN_STR(Gatt_Characteristic_Aggregate_Format)
	CASE_RETURN_STR(Valid_Range)
	CASE_RETURN_STR(External_Report_Reference)
	CASE_RETURN_STR(Report_Reference)
	CASE_RETURN_STR(Number_Of_Digitals)
	CASE_RETURN_STR(Value_Trigger_Setting)
	CASE_RETURN_STR(Es_Configuration)
	CASE_RETURN_STR(Es_Measurement)
	CASE_RETURN_STR(Es_Trigger_Setting)
	CASE_RETURN_STR(Time_Trigger_Setting)
	CASE_RETURN_STR(Gatt_Primary_Service_Declaration)
	CASE_RETURN_STR(Gatt_Secondary_Service_Declaration)
	CASE_RETURN_STR(Gatt_Include_Declaration)
	CASE_RETURN_STR(Gatt_Characteristic_Declaration)
	default:
		return "--";
	}
}
