/*
 * Copyright (C) 2012 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef ANDROID_INCLUDE_BLUETOOTH_H
#define ANDROID_INCLUDE_BLUETOOTH_H

#include <stdbool.h>
#include <stdint.h>
#include <sys/cdefs.h>
#include <sys/types.h>

#include <hardware/hardware.h>

__BEGIN_DECLS

/**
 * The Bluetooth Hardware Module ID
 */

#define BT_HARDWARE_MODULE_ID "bluetooth"
#define BT_STACK_MODULE_ID "bluetooth"
#define BT_STACK_TEST_MODULE_ID "bluetooth_test"


/* Bluetooth profile interface IDs */

#define BT_PROFILE_HANDSFREE_ID "handsfree"
#define BT_PROFILE_HANDSFREE_CLIENT_ID "handsfree_client"
#define BT_PROFILE_ADVANCED_AUDIO_ID "a2dp"
#define BT_PROFILE_ADVANCED_AUDIO_SINK_ID "a2dp_sink"
#define BT_PROFILE_HEALTH_ID "health"
#define BT_PROFILE_SOCKETS_ID "socket"
#define BT_PROFILE_HIDHOST_ID "hidhost"
#define BT_PROFILE_PAN_ID "pan"
#define BT_PROFILE_MAP_CLIENT_ID "map_client"

#define BT_PROFILE_GATT_ID "gatt"
#define BT_PROFILE_AV_RC_ID "avrcp"
#define BT_PROFILE_AV_RC_CTRL_ID "avrcp_ctrl"

/** Bluetooth Address */
typedef struct {
	uint8_t address[6];
} __attribute__((packed))bt_bdaddr_t;

/** Bluetooth Device Name */
typedef struct {
	uint8_t name[249];
} __attribute__((packed))bt_bdname_t;

/** Bluetooth Adapter Visibility Modes*/
typedef enum {
	BT_SCAN_MODE_NONE,
	BT_SCAN_MODE_CONNECTABLE,
	BT_SCAN_MODE_CONNECTABLE_DISCOVERABLE
} bt_scan_mode_t;

/** Bluetooth Adapter State */
typedef enum {
	BT_STATE_OFF,
	BT_STATE_ON
}   bt_state_t;

/** Bluetooth Error Status */
/** We need to build on this */

typedef enum {
	BT_STATUS_SUCCESS,
	BT_STATUS_FAIL,
	BT_STATUS_NOT_READY,
	BT_STATUS_NOMEM,
	BT_STATUS_BUSY,
	BT_STATUS_DONE,        /* request already completed */
	BT_STATUS_UNSUPPORTED,
	BT_STATUS_PARM_INVALID,
	BT_STATUS_UNHANDLED,
	BT_STATUS_AUTH_FAILURE,
	BT_STATUS_RMT_DEV_DOWN,
#ifndef TIZEN_BT_HAL
	BT_STATUS_CONN_TOUT   /* disconnection due to supervision timeout */
#else
	BT_STATUS_CONN_TOUT,   /* disconnection due to supervision timeout */
	BT_STATUS_AUTH_REJECTED,
	BT_STATUS_CONN_TERM_LOCAL_HOST,
	BT_STATUS_CONN_TERM_RMT_HOST
#endif
} bt_status_t;

/** Bluetooth PinKey Code */
typedef struct {
	uint8_t pin[16];
} __attribute__((packed))bt_pin_code_t;

typedef struct {
	uint8_t status;
	uint8_t ctrl_state;     /* stack reported state */
	uint64_t tx_time;       /* in ms */
	uint64_t rx_time;       /* in ms */
	uint64_t idle_time;     /* in ms */
	uint64_t energy_used;   /* a product of mA, V and ms */
} __attribute__((packed))bt_activity_energy_info;

/** Bluetooth Adapter Discovery state */
typedef enum {
	BT_DISCOVERY_STOPPED,
	BT_DISCOVERY_STARTED
} bt_discovery_state_t;

/** Bluetooth ACL connection state */
typedef enum {
	BT_ACL_STATE_CONNECTED,
	BT_ACL_STATE_DISCONNECTED
} bt_acl_state_t;

/** Remote Device Trusted state */
typedef enum {
	BT_DEVICE_TRUSTED,
	BT_DEVICE_NOT_TRUSTED
} bt_device_trust_state_t;

/** Bluetooth 128-bit UUID */
typedef struct {
	uint8_t uu[16];
} bt_uuid_t;

/** Bluetooth SDP service record */
typedef struct {
	bt_uuid_t uuid;
	uint16_t channel;
	char name[256]; // what's the maximum length
} bt_service_record_t;


/** Bluetooth Remote Version info */
typedef struct {
	int version;
	int sub_ver;
	int manufacturer;
} bt_remote_version_t;


typedef struct {
	uint16_t version_supported;
	uint8_t local_privacy_enabled;
	uint8_t max_adv_instance;
	uint8_t rpa_offload_supported;
	uint8_t max_irk_list_size;
	uint8_t max_adv_filter_supported;
	uint8_t activity_energy_info_supported;
	uint16_t scan_result_storage_size;
	uint16_t total_trackable_advertisers;
	bool extended_scan_support;
	bool debug_logging_supported;
	bool le_2m_phy_supported;
	bool le_coded_phy_supported;
	bool le_extended_advertising_supported;
	bool le_periodic_advertising_supported;
	uint16_t le_maximum_advertising_data_length;
} bt_local_le_features_t;

/*
typedef struct {
	uint8_t local_privacy_enabled;
	uint8_t max_adv_instance;
	uint8_t rpa_offload_supported;
	uint8_t max_irk_list_size;
	uint8_t max_adv_filter_supported;
	uint8_t scan_result_storage_size_lobyte;
	uint8_t scan_result_storage_size_hibyte;
	uint8_t activity_energy_info_supported;
} bt_local_le_features_t;*/

/* Bluetooth Adapter and Remote Device property types */
typedef enum {
	/* Properties common to both adapter and remote device */
	/**
	 * Description - Bluetooth Device Name
	 * Access mode - Adapter name can be GET/SET. Remote device can be GET
	 * Data type   - bt_bdname_t
	 */
	BT_PROPERTY_BDNAME = 0x1,
	/**
	 * Description - Bluetooth Device Address
	 * Access mode - Only GET.
	 * Data type   - bt_bdaddr_t
	 */
	BT_PROPERTY_BDADDR,
	/**
	 * Description - Bluetooth Service 128-bit UUIDs
	 * Access mode - Only GET.
	 * Data type   - Array of bt_uuid_t (Array size inferred from property length).
	 */
	BT_PROPERTY_UUIDS,
	/**
	 * Description - Bluetooth Class of Device as found in Assigned Numbers
	 * Access mode - Only GET.
	 * Data type   - uint32_t.
	 */
	BT_PROPERTY_CLASS_OF_DEVICE,
	/**
	 * Description - Device Type - BREDR, BLE or DUAL Mode
	 * Access mode - Only GET.
	 * Data type   - bt_device_type_t
	 */
	BT_PROPERTY_TYPE_OF_DEVICE,
	/**
	 * Description - Bluetooth Service Record
	 * Access mode - Only GET.
	 * Data type   - bt_service_record_t
	 */
	BT_PROPERTY_SERVICE_RECORD,

	/* Properties unique to adapter */
	/**
	 * Description - Bluetooth Adapter scan mode
	 * Access mode - GET and SET
	 * Data type   - bt_scan_mode_t.
	 */
	BT_PROPERTY_ADAPTER_SCAN_MODE,
	/**
	 * Description - List of bonded devices
	 * Access mode - Only GET.
	 * Data type   - Array of bt_bdaddr_t of the bonded remote devices
	 *               (Array size inferred from property length).
	 */
	BT_PROPERTY_ADAPTER_BONDED_DEVICES,
	/**
	 * Description - Bluetooth Adapter Discovery timeout (in seconds)
	 * Access mode - GET and SET
	 * Data type   - uint32_t
	 */
	BT_PROPERTY_ADAPTER_DISCOVERY_TIMEOUT,

	/* Properties unique to remote device */
	/**
	 * Description - User defined friendly name of the remote device
	 * Access mode - GET and SET
	 * Data type   - bt_bdname_t.
	 */
	BT_PROPERTY_REMOTE_FRIENDLY_NAME,
	/**
	 * Description - RSSI value of the inquired remote device
	 * Access mode - Only GET.
	 * Data type   - int32_t.
	 */
	BT_PROPERTY_REMOTE_RSSI,
	/**
	 * Description - Remote version info
	 * Access mode - SET/GET.
	 * Data type   - bt_remote_version_t.
	 */

	BT_PROPERTY_REMOTE_VERSION_INFO,

	/* Tizen Specific Adapter and Remote Device properties */
	/**
	 * Description - Remote Device paired Info
	 * Access mode - GET.
	 * Data type   - uint8_t.
	 */
	BT_PROPERTY_REMOTE_PAIRED,

	/**
	 * Description - Remote Device connected Info
	 * Access mode - GET.
	 * Data type   - unsigned int.
	 */
	BT_PROPERTY_REMOTE_CONNECTED,

	/**
	 * Description - Remote Device Trusted info
	 * Access mode - GET.
	 * Data type   - uint8_t.
	 */
	BT_PROPERTY_REMOTE_TRUST,

	/**
	 * Description - Adapter Pairable info
	 * Access mode - GET\SET.
	 * Data type   - uint8_t
	 */
	BT_PROPERTY_PAIRABLE,

	/**
	 * Description - Adapter pairable Timeout value
	 * Access mode - GET.
	 * Data type   - uint32_t
	 */
	BT_PROPERTY_PAIRABLE_TIMEOUT,

	/**
	 * Description - Adapter version
	 * Access mode - GET
	 * Data type   - Array of character string
	 */
	BT_PROPERTY_VERSION,

	/**
	 * Description - Adapter IPSP initialized state info
	 * Access mode - GET
	 * Data type   - uint8_t
	 */
	BT_PROPERTY_IPSP_INITIALIZED,

	/**
	 * Description - Adapter Modalias info
	 * Access mode - GET
	 * Data type   - Array of character string
	 */
	BT_PROPERTY_MODALIAS,

	/**
	 * Description - BLE Device manufacturer data length
	 * Access mode - GET
	 * Data type   - uint32_t
	 */
	BT_PROPERTY_REMOTE_DEVICE_MANUFACTURER_DATA_LEN,

	/**
	 * Description - BLE Device manufacturer data
	 * Access mode - GET
	 * Data type   - Array of character string
	 */
	BT_PROPERTY_REMOTE_DEVICE_MANUFACTURER_DATA,

	/**
	 * Description - Remote BLE advertising data
	 * Access mode - Only received during device found callback.
	 * Data type   - Array of uint8_t of remote BLE adv data.
	 *               (Array size inferred from property length).
	 */
	BT_PROPERTY_REMOTE_BLE_ADV_DATA,

	/**
	 * Description - Local LE features
	 * Access mode - GET.
	 * Data type   - bt_local_le_features_t.
	 */
	BT_PROPERTY_LOCAL_LE_FEATURES,

	/**
	 * Description - Remote device friendly name set?
	 * Access mode - GET.
	 * Data type   - uint8_t
	 */
	BT_PROPERTY_REMOTE_IS_ALIAS_SET,

	BT_PROPERTY_REMOTE_DEVICE_TIMESTAMP = 0xFF,
} bt_property_type_t;

/** Bluetooth Adapter Property data structure */
typedef struct {
	bt_property_type_t type;
	int len;
	void *val;
} bt_property_t;


/** Bluetooth Device Type */
typedef enum {
	BT_DEVICE_DEVTYPE_BREDR = 0x1,
	BT_DEVICE_DEVTYPE_BLE,
	BT_DEVICE_DEVTYPE_DUAL
} bt_device_type_t;
/** Bluetooth Bond state */
typedef enum {
	BT_BOND_STATE_NONE,
	BT_BOND_STATE_BONDING,
	BT_BOND_STATE_BONDED
} bt_bond_state_t;

/** Bluetooth SSP Bonding Variant */
typedef enum {
	BT_SSP_VARIANT_PASSKEY_CONFIRMATION,
	BT_SSP_VARIANT_PASSKEY_ENTRY,
	BT_SSP_VARIANT_CONSENT,
	BT_SSP_VARIANT_PASSKEY_NOTIFICATION
} bt_ssp_variant_t;

/** Bluetooth Profile Service IDs */
typedef enum {
	BT_RES_SERVICE_ID,              /* Reserved */
	BT_SPP_SERVICE_ID,              /* Serial port profile. */
	BT_DUN_SERVICE_ID,              /* Dial-up networking profile. */
	BT_A2DP_SRC_SERVICE_ID,         /* A2DP Source profile. */
	BT_LAP_SERVICE_ID,              /* LAN access profile. */
	BT_HSP_SERVICE_ID,              /* Headset profile. */
	BT_HFP_SERVICE_ID,              /* Hands-free profile. */
	BT_OPP_SERVICE_ID,              /* Object push  */
	BT_FTP_SERVICE_ID,              /* File transfer */
	BT_AVRCP_CT_SERVICE_ID,         /* AVRC Controller Terminal */
	BT_ICP_SERVICE_ID,              /* Intercom Terminal */
	BT_SYNC_SERVICE_ID,             /* Synchronization */
	BT_BPP_SERVICE_ID,              /* Basic printing profile */
	BT_BIP_SERVICE_ID,              /* Basic Imaging profile */
	BT_PANU_SERVICE_ID,             /* PAN User */
	BT_NAP_SERVICE_ID,              /* PAN Network access point */
	BT_GN_SERVICE_ID,               /* PAN Group Ad-hoc networks */
	BT_SAP_SERVICE_ID,              /* SIM Access profile */
	BT_A2DP_SERVICE_ID,             /* A2DP Sink */
	BT_AVRCP_SERVICE_ID,            /* A/V remote control */
	BT_HID_SERVICE_ID,              /* HID */
	BT_VDP_SERVICE_ID,              /* Video distribution */
	BT_PBAP_SERVICE_ID,             /* PhoneBook Access Server*/
	BT_HSP_HS_SERVICE_ID,           /* HFP HS role */
	BT_HFP_HS_SERVICE_ID,           /* HSP HS role */
	BT_MAP_SERVICE_ID,              /* Message Access Profile */
	BT_MN_SERVICE_ID,               /* Message Notification Service */
	BT_HDP_SERVICE_ID,              /* Health Device Profile */
	BT_PCE_SERVICE_ID,              /* PhoneBook Access Client*/
#ifdef TIZEN_BT_HAL
	BT_IOTIVITY_SERVICE_ID,         /* Custom IOTIVITY UUID */
#endif
} bt_service_id_t;

#define BT_MAX_NUM_UUIDS 32

#ifdef TIZEN_BT_HAL
#define BT_OSP_SERVER_OBEX 0x00
/* TODO: Add more types If and when required */
#define BT_OSP_SERVER_MAX 0xFF

/* Tizen BT discovery Types */
typedef enum {
	BT_DISC_ROLE_BREDR = 0x01,
	BT_DISC_ROLE_LE,
	BT_DISC_ROLE_DUAL
} bt_disc_role_type_t;

/** Bluetooth Trusted Profiles */
typedef enum {
	BT_TRUSTED_PROFILE_PBAP = 1,
	BT_TRUSTED_PROFILE_MAP,
	BT_TRUSTED_PROFILE_SAP,
	BT_TRUSTED_PROFILE_HFP_HF,
	BT_TRUSTED_PROFILE_A2DP,
	BT_TRUSTED_PROFILE_ALL = 0xFFFFFFFF,
} bt_trusted_profile_t;

/** Bluetooth Bonding Authentication Notifications */
typedef enum {
	BT_PASSKEY_CONFIRMATION = 1,
	BT_PASSKEY_DISPLAY,
	BT_PASSKEY_ENTRY,
	BT_PINCODE_ENTRY,
} bt_gap_auth_variant_t;
#endif

/** Bluetooth Interface callbacks */

/** Bluetooth Enable/Disable Callback. */
typedef void (*adapter_state_changed_callback)(bt_state_t state);

#ifdef TIZEN_BT_HAL
/** BLE Enable/Disable Callback. */
typedef void (*le_state_changed_callback)(bt_state_t state);
#endif

/** GET/SET Adapter Properties callback */
/* TODO: For the GET/SET property APIs/callbacks, we may need a session
 * identifier to associate the call with the callback. This would be needed
 * whenever more than one simultaneous instance of the same adapter_type
 * is get/set.
 *
 * If this is going to be handled in the Java framework, then we do not need
 * to manage sessions here.
 */
typedef void (*adapter_properties_callback)(bt_status_t status,
		int num_properties,
		bt_property_t *properties);

/** GET/SET Remote Device Properties callback */
/** TODO: For remote device properties, do not see a need to get/set
 * multiple properties - num_properties shall be 1
 */
typedef void (*remote_device_properties_callback)(bt_status_t status,
		bt_bdaddr_t *bd_addr,
		int num_properties,
		bt_property_t *properties);

/** New device discovered callback */
/** If EIR data is not present, then BD_NAME and RSSI shall be NULL and -1
 * respectively */
typedef void (*device_found_callback)(int num_properties,
		bt_property_t *properties);

/** Discovery state changed callback */
typedef void (*discovery_state_changed_callback)(bt_discovery_state_t state);

/** Bluetooth Legacy PinKey Request callback */
typedef void (*pin_request_callback)(bt_bdaddr_t *remote_bd_addr,
		bt_bdname_t *bd_name, uint32_t cod);

/** Bluetooth SSP Request callback - Just Works & Numeric Comparison*/
/** pass_key - Shall be 0 for BT_SSP_PAIRING_VARIANT_CONSENT &
 *  BT_SSP_PAIRING_PASSKEY_ENTRY */
/* TODO: Passkey request callback shall not be needed for devices with display
 * capability. We still need support this in the stack for completeness */
typedef void (*ssp_request_callback)(bt_bdaddr_t *remote_bd_addr,
		bt_bdname_t *bd_name,
		uint32_t cod,
		bt_ssp_variant_t pairing_variant,
		uint32_t pass_key);

/** Bluetooth Bond state changed callback */
/* Invoked in response to create_bond, cancel_bond or remove_bond */
typedef void (*bond_state_changed_callback)(bt_status_t status,
		bt_bdaddr_t *remote_bd_addr,
		bt_bond_state_t state);

/** Bluetooth ACL connection state changed callback */
typedef void (*acl_state_changed_callback)(bt_status_t status, bt_bdaddr_t *remote_bd_addr,
		bt_acl_state_t state);

typedef enum {
	ASSOCIATE_JVM,
	DISASSOCIATE_JVM
} bt_cb_thread_evt;

/** Thread Associate/Disassociate JVM Callback */
/* Callback that is invoked by the callback thread to allow upper layer to attach/detach to/from
 * the JVM */
typedef void (*callback_thread_event)(bt_cb_thread_evt evt);

/** Bluetooth Test Mode Callback */
/* Receive any HCI event from controller. Must be in DUT Mode for this callback to be received */
typedef void (*dut_mode_recv_callback)(uint16_t opcode, uint8_t *buf, uint8_t len);

/* LE Test mode callbacks
 * This callback shall be invoked whenever the le_tx_test, le_rx_test or le_test_end is invoked
 * The num_packets is valid only for le_test_end command */
typedef void (*le_test_mode_callback)(bt_status_t status, uint16_t num_packets);

/** Callback invoked when energy details are obtained */
/* Ctrl_state-Current controller state-Active-1,scan-2,or idle-3 state as defined by HCI spec.
 * If the ctrl_state value is 0, it means the API call failed
 * Time values-In milliseconds as returned by the controller
 * Energy used-Value as returned by the controller
 * Status-Provides the status of the read_energy_info API call */
typedef void (*energy_info_callback)(bt_activity_energy_info *energy_info);

/* Service level Authorization request callback */
typedef void (*authorize_request_callback) (bt_bdaddr_t *remote_bd_addr, bt_service_id_t service_d);

#ifdef TIZEN_BT_HAL
/* Service level Authorization request callback */
typedef void (*sock_authorize_request_callback) (bt_bdaddr_t *remote_bd_addr, bt_uuid_t *uuid);

/** Bluetooth ACL connection state */
typedef enum {
	BT_LE_CONN_STATE_CONNECTED,
	BT_LE_CONN_STATE_DISCONNECTED
} bt_le_conn_state_t;

/** Bluetooth ACL connection state changed callback */
typedef void (*le_conn_state_changed_callback)(bt_status_t status, bt_bdaddr_t *remote_bd_addr,
		bt_le_conn_state_t state);

/** Remote device trusted profiles changed callback */
typedef void (*device_trusted_profiles_changed_callback)(bt_bdaddr_t *bd_addr, uint32_t trust_val);

/** Remote device RSSI monitoring state changed callback */
typedef void (*rssi_monitor_state_changed_callback)(bt_bdaddr_t *bd_addr, int32_t link_type, uint8_t state);

/** Remote device RSSI alert callback */
typedef void (*rssi_alert_callback)(bt_bdaddr_t *bd_addr, int32_t link_type, int32_t alert_type, int32_t rssi);

/** Remote device Raw RSSI received callback */
typedef void (*raw_rssi_received_callback)(bt_bdaddr_t *bd_addr, int32_t link_type, int32_t rssi);
#endif

/** TODO: Add callbacks for Link Up/Down and other generic
 *  notifications/callbacks */

/** Remote Device Trust state changed callback */
typedef void (*device_trust_state_changed_callback)(bt_bdaddr_t *remote_bd_addr,
		bt_device_trust_state_t trust);

/** Bluetooth DM callback structure. */
typedef struct {
	/** set to sizeof(bt_callbacks_t) */
	size_t size;
	adapter_state_changed_callback adapter_state_changed_cb;
	adapter_properties_callback adapter_properties_cb;
	remote_device_properties_callback remote_device_properties_cb;
	device_found_callback device_found_cb;
	discovery_state_changed_callback discovery_state_changed_cb;
	pin_request_callback pin_request_cb;
	ssp_request_callback ssp_request_cb;
	bond_state_changed_callback bond_state_changed_cb;
	acl_state_changed_callback acl_state_changed_cb;
	callback_thread_event thread_evt_cb;
	dut_mode_recv_callback dut_mode_recv_cb;
	le_test_mode_callback le_test_mode_cb;
	energy_info_callback energy_info_cb;
	authorize_request_callback authorize_request_cb;
	device_trust_state_changed_callback device_trust_state_changed_cb;
#ifdef TIZEN_BT_HAL
	sock_authorize_request_callback socket_authorize_request_cb;
	le_state_changed_callback le_state_changed_cb;
	le_conn_state_changed_callback le_conn_state_changed_cb;
	device_trusted_profiles_changed_callback device_trusted_profiles_changed_cb;
	rssi_monitor_state_changed_callback rssi_monitor_state_changed_cb;
	rssi_alert_callback rssi_alert_cb;
	raw_rssi_received_callback raw_rssi_received_cb;
#endif
} bt_callbacks_t;

typedef void (*alarm_cb)(void *data);
typedef bool (*set_wake_alarm_callout)(uint64_t delay_millis, bool should_wake, alarm_cb cb, void *data);
typedef int (*acquire_wake_lock_callout)(const char *lock_name);
typedef int (*release_wake_lock_callout)(const char *lock_name);

/** The set of functions required by bluedroid to set wake alarms and
 * grab wake locks. This struct is passed into the stack through the
 * |set_os_callouts| function on |bt_interface_t|.
 */
typedef struct {
	/* set to sizeof(bt_os_callouts_t) */
	size_t size;

	set_wake_alarm_callout set_wake_alarm;
	acquire_wake_lock_callout acquire_wake_lock;
	release_wake_lock_callout release_wake_lock;
} bt_os_callouts_t;

/** NOTE: By default, no profiles are initialized at the time of init/enable.
 *  Whenever the application invokes the 'init' API of a profile, then one of
 *  the following shall occur:
 *
 *    1.) If Bluetooth is not enabled, then the Bluetooth core shall mark the
 *        profile as enabled. Subsequently, when the application invokes the
 *        Bluetooth 'enable', as part of the enable sequence the profile that were
 *        marked shall be enabled by calling appropriate stack APIs. The
 *        'adapter_properties_cb' shall return the list of UUIDs of the
 *        enabled profiles.
 *
 *    2.) If Bluetooth is enabled, then the Bluetooth core shall invoke the stack
 *        profile API to initialize the profile and trigger a
 *        'adapter_properties_cb' with the current list of UUIDs including the
 *        newly added profile's UUID.
 *
 *   The reverse shall occur whenever the profile 'cleanup' APIs are invoked
 */

/** Represents the standard Bluetooth DM interface. */
typedef struct {
	/** set to sizeof(bt_interface_t) */
	size_t size;
	/**
	 * Opens the interface and provides the callback routines
	 * to the implemenation of this interface.
	 */
	int (*init)(bt_callbacks_t*callbacks);

	/** Enable Bluetooth. */
	int (*enable)(void);

	/** Disable Bluetooth. */
	int (*disable)(void);

#ifdef TIZEN_BT_HAL
	/** Reset Bluetooth. */
	int (*reset)(void);
#endif

#ifdef TIZEN_BT_HAL
	/** Enable LE Bluetooth. */
	int (*le_enable)(void);

	/** Disable LE Bluetooth. */
	int (*le_disable)(void);

	int (*le_init)(void);

	void (*le_deinit)(void);

	int (*is_advertising) (void);
#endif

	/** Closes the interface. */
	void (*cleanup)(void);

#ifdef TIZEN_BT_HAL
	/** Get Bluetooth Adapter Powered status */
	int (*get_adapter_powered_status)(uint8_t *status);
#endif

	/** Get all Bluetooth Adapter properties at init */
	int (*get_adapter_properties)(void);

	/** Get Bluetooth Adapter property of 'type' */
	int (*get_adapter_property)(bt_property_type_t type);

	/** Set Bluetooth Adapter property of 'type' */
	/* Based on the type, val shall be one of
	 * bt_bdaddr_t or bt_bdname_t or bt_scanmode_t etc
	 */
	int (*set_adapter_property)(const bt_property_t *property);

	/** Get all Remote Device properties */
	int (*get_remote_device_properties)(bt_bdaddr_t *remote_addr);

	/** Get Remote Device property of 'type' */
	int (*get_remote_device_property)(bt_bdaddr_t *remote_addr,
			bt_property_type_t type);

	/** Set Remote Device property of 'type' */
	int (*set_remote_device_property)(bt_bdaddr_t *remote_addr,
			const bt_property_t *property);

	/** Get Remote Device's service record  for the given UUID */
	int (*get_remote_service_record)(bt_bdaddr_t *remote_addr,
			bt_uuid_t *uuid);

	/** Start SDP to get remote services */
	int (*get_remote_services)(bt_bdaddr_t *remote_addr);

	/** Start Discovery */
	int (*start_discovery)(void);

#ifdef TIZEN_BT_HAL
	int (*start_custom_discovery)(bt_disc_role_type_t disc_type);
#endif

	/** Cancel Discovery */
	int (*cancel_discovery)(void);

	/** Create Bluetooth Bonding */
	int (*create_bond)(const bt_bdaddr_t *bd_addr, int transport);

	/** Remove Bond */
	int (*remove_bond)(const bt_bdaddr_t *bd_addr);

	/** Cancel Bond */
	int (*cancel_bond)(const bt_bdaddr_t *bd_addr);

	/**
	 * Get the connection status for a given remote device.
	 * return value of 0 means the device is not connected,
	 * non-zero return status indicates an active connection.
	 */
	int (*get_connection_state)(const bt_bdaddr_t *bd_addr);

	/** BT Legacy PinKey Reply */
	/** If accept==FALSE, then pin_len and pin_code shall be 0x0 */
	int (*pin_reply)(const bt_bdaddr_t *bd_addr, uint8_t accept,
			uint8_t pin_len, bt_pin_code_t *pin_code);

	/** BT SSP Reply - Just Works, Numeric Comparison and Passkey
	 * passkey shall be zero for BT_SSP_VARIANT_PASSKEY_COMPARISON &
	 * BT_SSP_VARIANT_CONSENT
	 * For BT_SSP_VARIANT_PASSKEY_ENTRY, if accept==FALSE, then passkey
	 * shall be zero */
	int (*ssp_reply)(const bt_bdaddr_t *bd_addr, bt_ssp_variant_t variant,
			uint8_t accept, uint32_t passkey);

	/** Get Bluetooth profile interface */
	const void* (*get_profile_interface) (const char *profile_id);

	/** Bluetooth Test Mode APIs - Bluetooth must be enabled for these APIs */
	/* Configure DUT Mode - Use this mode to enter/exit DUT mode */
	int (*dut_mode_configure)(uint8_t enable);

	/* Send any test HCI (vendor-specific) command to the controller. Must be in DUT Mode */
	int (*dut_mode_send)(uint16_t opcode, uint8_t *buf, uint8_t len);
	/** BLE Test Mode APIs */
	/* opcode MUST be one of: LE_Receiver_Test, LE_Transmitter_Test, LE_Test_End */
	int (*le_test_mode)(uint16_t opcode, uint8_t *buf, uint8_t len);

	/* enable or disable bluetooth HCI snoop log */
	int (*config_hci_snoop_log)(uint8_t enable);

	/** Sets the OS call-out functions that bluedroid needs for alarms and wake locks.
	 * This should be called immediately after a successful |init|.
	 */
	int (*set_os_callouts)(bt_os_callouts_t *callouts);

	/** Read Energy info details - return value indicates BT_STATUS_SUCCESS or BT_STATUS_NOT_READY
	 * Success indicates that the VSC command was sent to controller
	 */
	int (*read_energy_info)();

	/* Tizen Specific: Send  service level Authorization response */
	int (*authorize_response)(const bt_bdaddr_t *bd_addr, bt_service_id_t service_id,
			uint8_t authorize, uint8_t save_settings);

	/** Set auto authorization for peer device. Should be a paired device */
	int (*set_authorization)(bt_bdaddr_t *bd_addr, uint8_t auth);
#ifdef TIZEN_BT_HAL
	/**
	 * get the service connection status for a given remote device.
	 * return value of 0 means the device is not connected,
	 * non-zero return status indicates an active connection.
	 */
	int (*get_service_connection_state)(const bt_bdaddr_t *bd_addr, bt_service_id_t rem_svc_id);

	/*
	 * Enable/Disable osp server in gap agent.
	 * enable == 0/1 -> Register/Unregister osp server.
	 */
	int (*set_agent_osp_server)(uint32_t type, uint8_t enable);

	/**
	 * Set profile as trusted for remote device
	 */
	int (*set_trusted_profile)(bt_bdaddr_t *bd_addr, bt_trusted_profile_t profile, uint8_t trust);

	/**
	 * Get profile trusted status for remote device
	 */
	int (*get_trusted_profile)(bt_bdaddr_t *bd_addr, bt_trusted_profile_t profile, uint32_t *trusted);

	/**
	 * Get raw RSSI strength of remote device for connected link type
	 * conn_link_type: (0 = BR/EDR link, 1 = LE link, 0xFF = Any (Default))
	 */
	int (*get_connected_link_rssi_strength)(bt_bdaddr_t *bd_addr, uint32_t conn_link_type);

	/**
	 * Enable RSSI monitoring of remote device for connected link type
	 * conn_link_type: (0 = BR/EDR link, 1 = LE link, 0xFF = Any (Default))
	 * Threshold : (low_threshold, in_range_threshold, high_threshold)
	 */
	int (*enable_rssi_monitoring)(bt_bdaddr_t *bd_addr, uint32_t conn_link_type,
			int low_threshold, int in_range_threshold, int high_threshold);
	/*
	 * Enable/Disable GAP authentication notification to application
	 * enable == 0/1 -> Enable/Disable notification sending
	 * type == value from bt_gap_auth_variant_t
	 */
	int (*enable_gap_auth_notifications)(uint32_t type, uint8_t enable);

	int (*set_le_static_random_address)(uint8_t enable);

	/*
	* Sets is_activating variable in HAL to TRUE. It is required for the
	* adapter switched on event to propagate to upper layers from HAL.
	*/
	int (*set_hal_adapter_request_state)(int enable);

	/*
	* Sets is_activating variable in HAL to TRUE. It is required for the
	* le adapter switched on event to propagate to upper layers from HAL.
	*/
        int (*set_hal_le_request_state)(int enable);

#endif
} bt_interface_t;

/** TODO: Need to add APIs for Service Discovery, Service authorization and
 *       connection management. Also need to add APIs for configuring
 *       properties of remote bonded devices such as name, UUID etc. */

typedef struct {
	struct hw_device_t common;
	const bt_interface_t* (*get_bluetooth_interface)();
} bluetooth_device_t;

typedef bluetooth_device_t bluetooth_module_t;
__END_DECLS

#endif /* ANDROID_INCLUDE_BLUETOOTH_H */
