/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef ANDROID_INCLUDE_BT_GATT_SERVER_H
#define ANDROID_INCLUDE_BT_GATT_SERVER_H

#include <stdint.h>

#include "bt_gatt_types.h"

__BEGIN_DECLS

/** GATT value type used in response to remote read requests */
typedef struct {
	uint8_t           value[BTGATT_MAX_ATTR_LEN];
	uint16_t          handle;
	uint16_t          offset;
	uint16_t          len;
	uint8_t           auth_req;
} btgatt_value_t;

/** GATT remote read request response type */
typedef union {
	btgatt_value_t attr_value;
	uint16_t            handle;
} btgatt_response_t;

typedef struct {
	uint8_t  server_if;
	bool     set_scan_rsp;
	bool     include_name;
	bool     include_txpower;
	uint16_t  appearance;
	char*    manufacturer_data;
	uint16_t manufacturer_data_len;
	char*    service_data;
	uint16_t service_data_len;
	char*    service_uuid;
	uint16_t service_uuid_len;
	char*    solicit_uuid;
	uint16_t solicit_uuid_len;
	uint16_t  min_interval;
	uint16_t  max_interval;
	uint8_t  adv_type;
	uint8_t  chnl_map;
	uint8_t  tx_power;
	uint8_t  timeout_s;
} btgatt_adv_param_setup_t;


/** BT-GATT Server callback structure. */

/** Callback invoked in response to register_server */
typedef void (*register_server_callback)(int status, int server_if,
		bt_uuid_t *app_uuid);

/** Callback indicating that a remote device has connected or been disconnected */
typedef void (*connection_callback)(int conn_id, int server_if, int connected,
		bt_bdaddr_t *bda);

/** Callback invoked in response to create_service */
typedef void (*service_added_callback)(int status, int server_if,
		btgatt_srvc_id_t *srvc_id, int srvc_handle);

/** Callback indicating that an included service has been added to a service */
typedef void (*included_service_added_callback)(int status, int server_if,
		int srvc_handle, int incl_srvc_handle);

/** Callback invoked when a characteristic has been added to a service */
typedef void (*characteristic_added_callback)(int status, int server_if,
		bt_uuid_t *uuid, int srvc_handle, int char_handle);

/** Callback invoked when a descriptor has been added to a characteristic */
typedef void (*descriptor_added_callback)(int status, int server_if,
		bt_uuid_t *uuid, int srvc_handle, int descr_handle);

/** Callback invoked in response to start_service */
typedef void (*service_started_callback)(int status, int server_if,
		int srvc_handle);

/** Callback invoked in response to stop_service */
typedef void (*service_stopped_callback)(int status, int server_if,
		int srvc_handle);

/** Callback triggered when a service has been deleted */
typedef void (*service_deleted_callback)(int status, int server_if,
		int srvc_handle);

/**
 * Callback invoked when indication confirmation
 */
typedef void (*indication_confirmation_callback) (int conn_id, int trans_id, int attr_handle, bt_bdaddr_t *bda);

#ifdef TIZEN_BT_HAL
/**
 * Callback invoked when indication confirmation
 */
typedef void (*notification_enabled_callback) (int conn_id, int trans_id, int attr_handle, bool notify, bt_bdaddr_t *bda);
#endif

/**
 * Callback invoked when a remote device has requested to read a characteristic
 * or descriptor. The application must respond by calling send_response
 */
typedef void (*request_read_callback)(int conn_id, int trans_id, bt_bdaddr_t *bda,
		int attr_handle, int offset, bool is_long);

/**
 * Callback invoked when a remote device has requested to write to a
 * characteristic or descriptor.
 */
typedef void (*request_write_callback)(int conn_id, int trans_id, bt_bdaddr_t *bda,
		int attr_handle, int offset, int length,
		bool need_rsp, bool is_prep, uint8_t* value);

/**
 * Callback invoked when a remote device has requested to acqiore write to a
 * characteristic .
 */
typedef void (*request_acquire_write_callback)(int mtu, int conn_id, int trans_id,
									int attr_handle, bt_bdaddr_t *bda);

typedef void (*request_acquire_notify_callback)(int fd, int conn_id, int trans_id,
									int attr_handle);


/** Callback invoked when a previously prepared write is to be executed */
typedef void (*request_exec_write_callback)(int conn_id, int trans_id,
		bt_bdaddr_t *bda, int exec_write);



/**
 * Callback triggered in response to send_response if the remote device
 * sends a confirmation.
 */
typedef void (*response_confirmation_callback)(int status, int handle);

/**
 * Callback indicationg the status of a listen() operation
 */
typedef void (*listen_callback)(int status, int server_if);

/** Callback invoked when multi-adv enable operation has completed */
typedef void (*multi_adv_enable_callback)(int server_if, int status);

/** Callback invoked when multi-adv param update operation has completed */
typedef void (*multi_adv_update_callback)(int server_if, int status);

/** Callback invoked when multi-adv instance data set operation has completed */
typedef void (*multi_adv_data_callback)(int server_if, int status);

/** Callback invoked when multi-adv disable operation has completed */
typedef void (*multi_adv_disable_callback)(int server_if, int status);

/** Callback invoked when the MTU for a given connection changes */
typedef void (*mtu_changed_callback)(int conn_id, int mtu);

typedef struct {
	register_server_callback        register_server_cb;
	connection_callback             connection_cb;
	service_added_callback          service_added_cb;
	included_service_added_callback included_service_added_cb;
	characteristic_added_callback   characteristic_added_cb;
	descriptor_added_callback       descriptor_added_cb;
	service_started_callback        service_started_cb;
	service_stopped_callback        service_stopped_cb;
	service_deleted_callback        service_deleted_cb;
	indication_confirmation_callback indication_confirmation_cb;
	request_read_callback           request_read_cb;
	request_write_callback          request_write_cb;
	request_exec_write_callback     request_exec_write_cb;
	response_confirmation_callback  response_confirmation_cb;
	listen_callback                 listen_cb;
	multi_adv_enable_callback       multi_adv_enable_cb;
	multi_adv_update_callback       multi_adv_update_cb;
	multi_adv_data_callback         multi_adv_data_cb;
	multi_adv_disable_callback      multi_adv_disable_cb;
	mtu_changed_callback            mtu_changed_cb;
#ifdef TIZEN_BT_HAL
	notification_enabled_callback   notif_enabled_cb;
#endif
	request_acquire_write_callback    request_acquire_write_cb;
	request_acquire_notify_callback  request_acquire_notify_cb;
} btgatt_server_callbacks_t;

/** Represents the standard BT-GATT server interface. */
typedef struct {
	/** Registers a GATT server application with the stack */
	bt_status_t (*register_server)(bt_uuid_t *uuid);

	/** Unregister a server application from the stack */
	bt_status_t (*unregister_server)(int server_if);

	/** Create a connection to a remote peripheral */
	bt_status_t (*connect)(int server_if, const bt_bdaddr_t *bd_addr, bool is_direct);

	/** Disconnect an established connection or cancel a pending one */
	bt_status_t (*disconnect)(int server_if, const bt_bdaddr_t *bd_addr,
			int conn_id);

	/** Create a new service */
	bt_status_t (*add_service)(int server_if, btgatt_srvc_id_t *srvc_id, int num_handles);

	/** Assign an included service to it's parent service */
	bt_status_t (*add_included_service)(int server_if, int service_handle, int included_handle);

	/** Add a characteristic to a service */
	bt_status_t (*add_characteristic)(int server_if,
			int service_handle, bt_uuid_t *uuid,
			int properties, int permissions);

	/** Add a descriptor to a given service */
	bt_status_t (*add_descriptor)(int server_if, int service_handle,
			bt_uuid_t *uuid, int permissions);

	/** Starts a local service */
	bt_status_t (*start_service)(int server_if, int service_handle,
			int transport);

	/** Stops a local service */
	bt_status_t (*stop_service)(int server_if, int service_handle);

	/** Delete a local service */
	bt_status_t (*delete_service)(int server_if, int service_handle);

	/** Send value indication to a remote device */
	bt_status_t (*send_indication)(int server_if, int attribute_handle,
			int conn_id, int len, int confirm,
			char* p_value);

	/** Send a response to a read/write operation */
	bt_status_t (*send_response)(int conn_id, int trans_id,
			int status, btgatt_response_t *response);

	/** Update GATT server attribute value */
	bt_status_t (*update_att_value)(int server_if, int attribute_handle,
			int value_length, char* att_value);

	/** Start or stop advertisements to listen for incoming connections */
	bt_status_t (*listen)(int server_if, bool start);

	/** Set the advertising data or scan response data */
	bt_status_t (*set_adv_data)(int server_if, bool set_scan_rsp, bool include_name,
			bool include_txpower, int min_interval, int max_interval, int appearance,
			uint16_t manufacturer_len, char* manufacturer_data,
			uint16_t service_data_len, char* service_data,
			uint16_t service_uuid_len, char* service_uuid);

	/* Setup the parameters as per spec, user manual specified values and enable multi ADV */
	bt_status_t (*multi_adv_enable)(int server_if);

	/* Update the parameters as per spec, user manual specified values and restart multi ADV */
	bt_status_t (*multi_adv_update)(int server_if, int min_interval, int max_interval, int adv_type,
			int chnl_map, int tx_power, int timeout_s);

	/* Setup the data for the specified instance */
	bt_status_t (*multi_adv_set_inst_data)(btgatt_adv_param_setup_t adv_param_setup);

	/* Disable the multi adv instance */
	bt_status_t (*multi_adv_disable)(int server_if);

	/* Get current att mtu size of active connection */
	bt_status_t (*get_att_mtu)(int conn_id, int *mtu_size);

	/** Send a response to a acquire write/notify  operation */
	bt_status_t (*send_response_acquire)(int conn_id, int trans_id,
			int status, int fd, int mtu, void *);
} btgatt_server_interface_t;

__END_DECLS

#endif /* ANDROID_INCLUDE_BT_GATT_SERVER_H */
