/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Contact: Nilesh Trimbake <t.shripati@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef _OAL_AVRCP_CT_H_
#define _OAL_AVRCP_CT_H_

#include <glib.h>
#include <sys/types.h>
#include <oal-manager.h>
#include "oal-event.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
	OAL_RC_PASS_CMD_PLAY = 0x01,
	OAL_RC_PASS_CMD_PAUSE,
	OAL_RC_PASS_CMD_STOP,
	OAL_RC_PASS_CMD_NEXT,
	OAL_RC_PASS_CMD_PREVIOUS,
	OAL_RC_PASS_CMD_PRESS_FAST_FORWARD,
	OAL_RC_PASS_CMD_RELEASE_FAST_FORWARD,
	OAL_RC_PASS_CMD_PRESS_REWIND,
	OAL_RC_PASS_CMD_RELEASE_REWIND,
	OAL_RC_PASS_CMD_VOLUME_UP,
	OAL_RC_PASS_CMD_VOLUME_DOWN
} avrcp_ct_pass_cmd_key_code_t;

/**
 * @brief Current State of remote key(pressed/released)
 *
 * @see  avrcp_ct_fforward
 * @see  avrcp_ct_rewind
 */
typedef enum {
	PRESS_STATE = 0x00,
	RELEASE_STATE = 0x01
} avrcp_ct_pass_state_t;

typedef struct {
	avrcp_ct_pass_cmd_key_code_t key_code;
	avrcp_ct_pass_state_t key_state;
} avrcp_ct_pass_cmd_t;

typedef struct {
	uint8_t accepted;
} avrcp_ct_playersetting_t;

typedef enum {
	OAL_EQUALIZER = 0x01,
	OAL_REPEAT,
	OAL_SHUFFLE,
	OAL_SCAN,
	OAL_PLAY_STATUS,
	OAL_PLAY_POSITION,
} avrcp_ct_player_property_type_t;

typedef struct {
	avrcp_ct_player_property_type_t type;
	unsigned int value;
} avrcp_ct_property_value_t;

typedef struct {
	uint32_t song_len;
	uint32_t song_pos;
} avrcp_ct_play_position_t;

/**
 * @brief Enable AVRCP Controller Feature
 *
 * @remarks  Devices supporting AVRCP Target will be able to connect.
 * @remarks  Used in conjunction with a2dp_sink_enable.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  avrcp_ct_disable
 * @see  a2dp_sink_enable
 */
oal_status_t avrcp_ct_enable(void);

/**
 * @brief Disable AVRCP Controller Feature
 *
 * @remarks  Devices supporting AVRCP Target will not be able to connect.
 * @remarks  Used in conjunction with a2dp_sink_disable.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AVRCP Controller should be enabled with avrcp_ct_enable().
 *
 * @see  avrcp_ct_enable
 * @see  a2dp_sink_disable
 */
oal_status_t avrcp_ct_disable(void);

/**
 * @brief Initiate a connection with a BT Sound device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre BT Audio should be enabled with avrcp_ct_enable().
 *
 * @see  OAL_EVENT_AVRCP_CT_CONNECTED
 */

oal_status_t avrcp_ct_connect(bt_address_t *device_address);


/**
 * @brief Remove a connection with a BT Sound device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre BT Audio should be connected with a BT Sound device.
 *
 * @see  OAL_EVENT_AVRCP_CT_DISCONNECTED
 */

oal_status_t avrcp_ct_disconnect(bt_address_t *device_address);

/**
 * @brief Play audio on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_play(bt_address_t *device_address);

/**
 * @brief Stop audio on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_stop(bt_address_t *device_address);

/**
 * @brief Pause audio on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_pause(bt_address_t *device_address);

/**
 * @brief Fast Forward audio on the connected remote device.
 *
 * @remarks  PRESS_STATE   - Start Fast Forward
 *           RELEASE_STATE - Stop Fast Forward
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_fforward(bt_address_t *device_address, avrcp_ct_pass_state_t press_state);

/**
 * @brief Rewind audio on the connected remote device.
 *
 * @remarks  PRESS_STATE   - Start Rewind
 *           RELEASE_STATE - Stop Rewind
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_rewind(bt_address_t *device_address, avrcp_ct_pass_state_t press_state);

/**
 * @brief Play next track on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_next_track(bt_address_t *device_address);


/**
 * @brief Play prev track on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_prev_track(bt_address_t *device_address);

/**
 * @brief Play volume up on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_volume_up(bt_address_t *device_address);

/**
 * @brief Play volume down on the connected remote device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 */
oal_status_t avrcp_ct_volume_down(bt_address_t *device_address);


/**
 * @brief Set media player properties to target device
 *
 * @remarks Target device will receive media player properties.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialize with avrcp_ct_enable().
 *
 */
oal_status_t avrcp_ct_set_property(bt_address_t *device_address, avrcp_ct_player_property_type_t type, uint32_t value);

/**
 * @brief Set media transport properties to target device
 *
 * @remarks Target device will receive media player properties.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialize with avrcp_ct_enable().
 *
 */
oal_status_t avrcp_transport_set_property(bt_address_t *device_address, int type, unsigned int value);

/**
 * @brief Get media player properties to target device
 *
 * @remarks Target device will send media player properties.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialize with avrcp_ct_enable().
 *
 */
oal_status_t avrcp_ct_get_property(bt_address_t *device_address, avrcp_ct_player_property_type_t type);

/**
 * @brief Set media player properties to target device
 *
 * @remarks Target device will receive media player properties.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialize with avrcp_ct_enable().
 *
 */
oal_status_t avrcp_ct_get_media_attribute(bt_address_t *device_address);

/**
 * @brief Request current status of the media playing on connected Audio source.
 *
 * @details Result containing - Song Length, Song Position and current status(Stopped/Playing/Paused) will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre A2DP Sink should be connected with a device.
 * @pre AVRCP connection should exist with the device.
 *
 * @see  OAL_EVENT_AVRCP_CT_PLAY_STATUS
 * @see  event_avrcp_ct_play_status_t
 */
oal_status_t avrcp_ct_get_play_status(bt_address_t *device_address);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /*_OAL_AVRCP_CT_H_*/
