/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef _OAL_AVRCP_HOST_H_
#define _OAL_AVRCP_HOST_H_

#include <glib.h>
#include <sys/types.h>
#include <oal-manager.h>
#include "oal-event.h"

#ifdef __cplusplus
extern "C" {
#endif

#define OAL_MAX_ATTR_STR_LEN 255

/* Maximum service name length */
#define BT_AVRCP_NAME_LENGTH_MAX    35

/**
 * @brief AVRCP CMD send to remote BT device
 *
 * @see  avrcp_send_cmd
 */
typedef enum {
	AVRCP_DEVICE_OFF,
	AVRCP_VOLUME_UP,
	AVRCP_VOLUME_DOWN,
	AVRCP_MUTE
} avrcp_cmd;

/**
 * @brief Media's different properties type set to remote BT device
 *
 * @see  avrcp_set_property
 */
typedef enum {
	AVRCP_EQUALIZER = 0x01,
	AVRCP_REPEAT,
	AVRCP_SHUFFLE,
	AVRCP_SCAN,
	AVRCP_STATUS,
	AVRCP_POSITION
} oal_media_prop_type;

/**
 * @brief Player repeat status
 *
 * @see  avrcp_set_property
 */
typedef enum {
	OAL_PLAYER_VAL_OFF_REPEAT = 0x01,
	OAL_PLAYER_VAL_SINGLE_REPEAT = 0x02,
	OAL_PLAYER_VAL_ALL_REPEAT = 0x03,
	OAL_PLAYER_VAL_GROUP_REPEAT = 0x04
} oal_player_repeat_status;

/**
 * @brief Player shuffel mode
 *
 * @see  avrcp_set_property
 */
typedef enum {
	OAL_PLAYER_VAL_OFF_SHUFFLE = 0x01,
	OAL_PLAYER_VAL_ALL_SHUFFLE = 0x02,
	OAL_PLAYER_VAL_GROUP_SHUFFLE = 0x03
} oal_player_shuffle_status;

/**
 * @brief Player play state
 *
 * @see  avrcp_set_property
 */
typedef enum {
	OAL_PLAYSTATE_STOPPED = 0x00,    /* Stopped */
	OAL_PLAYSTATE_PLAYING = 0x01,    /* Playing */
	OAL_PLAYSTATE_PAUSED = 0x02,    /* Paused  */
	OAL_PLAYSTATE_FWD_SEEK = 0x03,    /* Fwd Seek*/
	OAL_PLAYSTATE_REV_SEEK = 0x04,    /* Rev Seek*/
	OAL_PLAYSTATE_ERROR = 0xFF,    /* Error   */
} oal_play_status;

/**
 * @brief Media Track's attibutes set to remote BT device
 *
 * @see  avrcp_set_track_info
 */
typedef struct {
	char title[OAL_MAX_ATTR_STR_LEN];
	char artist[OAL_MAX_ATTR_STR_LEN];
	char album[OAL_MAX_ATTR_STR_LEN];
	char genre[OAL_MAX_ATTR_STR_LEN];
	unsigned int total_tracks;
	unsigned int number;
	unsigned int duration;
} oal_media_metadata_attributes_t;

/**
 * @brief AVRCP Volume and mute
 *
 * @see  OAL_EVENT_AVRCP_VOLUME_MUTE_CHANGED
 */
typedef struct {
	gboolean mute_status ;
	unsigned int volume;
} oal_avrcp_volume_mute_t;

/**
 * @brief AVRCP stream Delay
 *
 * @see  OAL_EVENT_AVRCP_DELAY_CHANGED
 */
typedef struct {
	unsigned int delay;
} oal_avrcp_delay_t;

typedef struct {
	int avrcp_feature;
	bt_address_t address;
	char avrcp_serv_name[BT_AVRCP_NAME_LENGTH_MAX+1];
	char avrcp_prov_name[BT_AVRCP_NAME_LENGTH_MAX+1];
} remote_feature;

/**
 * @brief Enable AVRCP Feature
 *
 * @remarks  AVRCP Feature (like play/pause, media information etc)  will be enabled.
 * @remarks  Used in conjunction with a2dp_enable
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  avrcp_disable()
 * @see  a2dp_enable()
 */
oal_status_t avrcp_enable(void);

/**
 * @brief Disable AVRCP Feature
 *
 * @remarks  AVRCP Feature (like play/pause, media information etc) will be disabled.
 * @remarks  Used in conjunction with a2dp_disable.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AVRCP should be enabled with avrcp_enable().
 *
 * @see  avrcp_enable()
 * @see  a2dp_disable()
 */
oal_status_t avrcp_disable(void);

/**
 * @brief Initiate a connection with AVRCP controller device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre BT Audio should be enabled with avrcp_tg_enable().
 *
 * @see  OAL_EVENT_AVRCP_CONNECTED
 */
oal_status_t avrcp_tg_connect(bt_address_t *rem_addr);


/**
 * @brief Remove a connection with AVRCP controller device
 *
 * @details  Result will be notified through an OAL event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre BT Audio should be connected with a BT Sound device.
 *
 * @see  OAL_EVENT_AVRCP_DISCONNECTED
 */
oal_status_t avrcp_tg_disconnect(bt_address_t *rem_addr);

/**
 * @brief Set the AVRCP property to BT audio remote device
 *
 * @details  Used to set the different type of AVRCP property.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AVRCP should be enabled with avrcp_enable().
 *
 * @see  avrcp_enable()
 */
oal_status_t avrcp_set_property(int type, unsigned int value);

/**
 * @brief Set the media track info to BT audio remote device
 *
 * @details  Used to set the different parameters(title, album, genre etc) of media track.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre AVRCP should be enabled with avrcp_enable().
 *
 * @see  avrcp_enable()
 */
oal_status_t avrcp_set_track_info(oal_media_metadata_attributes_t *meta_data);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /*_OAL_AVRCP_HOST_H_*/
