/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef _OAL_DEVICE_MGR_H_
#define _OAL_DEVICE_MGR_H_

#include <glib.h>
#include <sys/types.h>

#include <oal-manager.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Connection type
 *
 * @see  device_create_bond
 */

typedef enum {
	CONN_TYPE_DEFAULT = 0xFF, /* represents that connection type can both BR/EDR and LE */
	CONN_TYPE_BREDR = 0x00,
	CONN_TYPE_LE = 0x01,
} connection_type_e;

/**
 * @brief OSP server type
 *
 * @see device_set_osp_server()
 */
typedef enum {
	OAL_OSP_SERVER_OBEX = 0x00,
	/*TODO: Add more server type if and when required */
} oal_osp_server_type_e;

/**
 * @brief Trusted profile types
 *
 * @see  device_set_trust_profile/device_get_trust_profile
 */
typedef enum {
	OAL_TRUSTED_PROFILE_PBAP = 1,
	OAL_TRUSTED_PROFILE_MAP,
	OAL_TRUSTED_PROFILE_SAP,
	OAL_TRUSTED_PROFILE_HFP_HF,
	OAL_TRUSTED_PROFILE_A2DP,
	OAL_TRUSTED_PROFILE_ALL = 0xFFFFFFFF,
} oal_trusted_profile_e;

/**
 * @brief GAP authentication types
 *
 * @see  device_enable_gap_auth_notifications
 */
typedef enum {
	OAL_PASSKEY_CONFIRMATION = 1,
	OAL_PASSKEY_DISPLAY,
	OAL_PASSKEY_ENTRY,
	OAL_PINCODE_ENTRY,
} oal_gap_auth_type_e;

/**
 * @brief Request remote device attributes
 *
 * @details Attibutes such as name, vidpid, bond state etc are requested. remote_device_t is provided
 *		   with OAL_EVENT_DEVICE_PROPERTIES
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PROPERTIES
 * @see  remote_device_t
 */
oal_status_t device_query_attributes(bt_address_t * addr);

/**
 * @brief Set alias for remote device
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  remote_device_t
 */
oal_status_t device_set_alias(bt_address_t * addr, char * alias);


/**
 * @brief Initiate bonding with remote device
 *
 * @details Based on IO capabilties of 2 devices, different events can be generated
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PIN_REQUEST
 * @see  OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST
 * @see  OAL_EVENT_DEVICE_PASSKEY_CONFIRMATION_REQUEST
 * @see  OAL_EVENT_DEVICE_PASSKEY_DISPLAY
 * @see  OAL_EVENT_DEVICE_SSP_CONSENT_REQUEST
 * @see  OAL_EVENT_DEVICE_BONDING_SUCCESS
 * @see  OAL_EVENT_DEVICE_BONDING_FAILED
 */
oal_status_t device_create_bond(bt_address_t * addr, connection_type_e transport);

/**
 * @brief Cancel already in-progress bonding procedure
 *
 * @details Based on current progress different events can be recieved.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Bonding must be in progress by calling device_create_bond()
 *
 * @see  OAL_EVENT_DEVICE_BONDING_SUCCESS
 * @see  OAL_EVENT_DEVICE_BONDING_FAILED
 * @see  OAL_EVENT_DEVICE_BONDING_REMOVED
 */
oal_status_t device_stop_bond(bt_address_t * addr);

/**
 * @brief Remove the already created Bond with remote device
 *
 * @details Based on current progress different events can be recieved.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Bond should exist
 *
 * @see  OAL_EVENT_DEVICE_BONDING_REMOVED
 */
oal_status_t device_destroy_bond(bt_address_t * addr);

/**
 * @brief Request services supported by remote device
 *
 * @details List of services in form UUIDs will be provided in event data
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_SERVICES
 * @see  event_dev_services_t
 */
oal_status_t device_query_services(bt_address_t * addr);

/**
 * @brief Cancel already in-progress SDP procedure
 *
 * @details Based on current progress different events can be recieved.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre SDP must be in progress by calling device_query_services()
 *
 */
oal_status_t device_stop_query_sevices(bt_address_t * addr);

/**
 * @brief Accept PIN request as part of Bonding procedure
 *
 * @details Positive response to OAL_EVENT_DEVICE_PIN_REQUEST
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PIN_REQUEST
 */
oal_status_t device_accept_pin_request(bt_address_t * addr, const char * pin);

/**
 * @brief Reject PIN request as part of Bonding procedure
 *
 * @details Negative response to OAL_EVENT_DEVICE_PIN_REQUEST
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PIN_REQUEST
 */
oal_status_t device_reject_pin_request(bt_address_t * addr);

/**
 * @brief Accept passkey entry request as part of Bonding procedure
 *
 * @details Positive response to OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST
 */
oal_status_t device_accept_passkey_entry(bt_address_t * addr, uint32_t passkey);

/**
 * @brief Reject passkey entry request as part of Bonding procedure
 *
 * @details Negative response to OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST
 */
oal_status_t device_reject_passkey_entry(bt_address_t * addr);

/**
 * @brief Reply passkey confirmation request as part of Bonding procedure
 *
 * @details Response to OAL_EVENT_DEVICE_PASSKEY_CONFIRMATION_REQUEST
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_PASSKEY_CONFIRMATION_REQUEST
 */
oal_status_t device_reply_passkey_confirmation(bt_address_t * addr, int accept);

/**
 * @brief Reply SSP consent request as part of Bonding procedure
 *
 * @details Response to OAL_EVENT_DEVICE_SSP_CONSENT_REQUEST
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_SSP_CONSENT_REQUEST
 */
oal_status_t device_reply_ssp_consent(bt_address_t * addr, int accept);

/**
 * @brief Accept or reject authorization request for a connection from remote device
 *
 * @details authorize = TRUE: all connections from this device will be auto accepted without any auth-event
 *                 authorize = FALSE: all connections from this device will result in auth request event
 *
 * @param[in] always  If TRUE, future connection requests will be automatically accepted.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL_EVENT_DEVICE_AUTHORIZE_REQUEST or, OAL_EVENT_SOCKET_AUTHORIZE_REQUEST should have
 * been recieved by application.
 *
 * @see OAL_EVENT_DEVICE_AUTHORIZE_REQUEST
 * @see OAL_EVENT_SOCKET_AUTHORIZE_REQUEST
 */
oal_status_t device_reply_auth_request(bt_address_t * addr, oal_service_t service_type, int accept, int always);

/**
 * @brief Set a device as authorized or unauthorized
 *
 * @details authorize = TRUE: all connections from this device will be auto accepted without any auth-event
 *                 authorize = FALSE: all connections from this device will result in auth request event
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 *
 * @see  OAL_EVENT_DEVICE_AUTHORIZE_REQUEST
 */
oal_status_t device_set_authorized(bt_address_t * addr, int authorize);

/**
 * @brief Get ACL connection state for remote device
 *
 * @details This API is to check whether ACL connection is established with the remote device
 *
 * @param[in] addr: Remote device address.
 *
 * @return TRUE if ACL connection is established with the remote device and FALSE otherwise.
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
gboolean device_get_acl_conn_state(bt_address_t * addr);

/**
 * @brief Get PROFILE connection state for remote device
 *
 * @details This API is to check whether requested profile connection is established with the remote device
 *
 * @param[in] addr: Remote device address.
 * @param[in] svc_id: Profile Id to be checked for connection status.
 *
 * @return TRUE if ACL connection is established with the remote device and FALSE otherwise.
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
gboolean device_get_svc_conn_state(bt_address_t * addr, oal_service_t svc_id);

/**
 * @brief Set OSP server status in GAP agent
 *
 * @details This API is to enable/disable OSP server in GAP agent. if OSP server is enabled,
 * connection authorization events are sent to application otherwise handled internally in HAL.
 *
 * @param[in] tpye : OSP Server type
 * @param[in] enable: 0 -> enable server, 1 -> Disable server
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t device_set_osp_server(oal_osp_server_type_e type, int enable);

/**
 * @brief Set profile trust info for remote device
 *
 * @details This API is used to set profile as trusted/blocked for remote device
 *
 * @param[in] addr: Remote device address.
 * @param[in] profile: Trusted profile enum
 * @param[in] trust: TRUE == Trusted, FALSE == Blocked
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t device_set_trust_profile(bt_address_t *addr, oal_trusted_profile_e profile, gboolean trust);

/**
 * @brief Set profile trust info for remote device
 *
 * @details This API is used to set profile as trusted/blocked for remote device
 *
 * @param[in] addr: Remote device address.
 * @param[in] profile: Trusted profile enum
 * @param[out] trusted: 0 == Blocked, non-zero: Trusted values
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t device_get_trust_profile(bt_address_t *addr, oal_trusted_profile_e profile, unsigned int *trusted);

/**
 * @brief Get RSSI strength of remote device
 *
 * @details This API is used to get RSSI strength of remote device for conncted link type
 *
 * @param[in] addr: Remote device address.
 * @param[in] link_type: conncted link type (0 = BR/EDR link, 1 = LE link, 0xFF = Any (Default))
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t device_get_connected_link_rssi_strength(bt_address_t *addr, unsigned int link_type);

/**
 * @brief Enable remote device RSSI monitoring
 *
 * @details This API is used to enable remote device RSSI monitoring
 *
 * @param[in] addr: Remote device address.
 * @param[in] link_type: conncted link type (0 = BR/EDR link, 1 = LE link, 0xFF = Any (Default))
 * @param[in] low_threshold: Lower limit for RSSI value
 * @param[in] in_range_threshold: in range RSSI values
 * @param[in] high_threshold: Upper limit for RSSI value
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t device_enable_rssi_monitoring(bt_address_t *addr, unsigned int link_type,
		int low_threshold, int in_range_threshold, int high_threshold);

/**
 * @brief Enable/Disable GAP authentication requests to be received from stack
 *
 * @details This API is used to enable/disable GAP authentication requests to be received from stack
 *
 * @param[in] type: GAP authentication types oal_gap_auth_type_e
 * @param[in] enable: TRUE/FALSE
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre Adapter must be enabled with adapter_enable() followed by OAL_EVENT_ADAPTER_ENABLED
 */
oal_status_t device_enable_gap_auth_notifications(oal_gap_auth_type_e type, gboolean enable);

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /*_OAL_DEVICE_MGR_H_*/

