/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef OAL_GATT_H_
#define OAL_GATT_H_

#include <oal-manager.h>

#define OAL_GATT_MAX_ATTR_LEN 600
#define GATTC_READ_VALUE_TYPE_VALUE          0x0000  /* Attribute value itself */
#define GATTC_READ_VALUE_TYPE_AGG_FORMAT     0x2905  /* Characteristic Aggregate Format*/

typedef enum {
	OAL_GATT_WRITE_NO_RSP = 1,
	OAL_GATT_WRITE ,
} oal_gatt_write_type_t;

typedef enum {
	OAL_GATT_AUTH_REQ_NONE = 0,
	OAL_GATT_AUTH_REQ_NO_MITM,                      /* unauthenticated encryption */
	OAL_GATT_AUTH_REQ_MITM,                         /* authenticated encryption */
	OAL_GATT_AUTH_REQ_SIGNED_NO_MITM,
	OAL_GATT_AUTH_REQ_SIGNED_MITM,
} oal_gatt_auth_req_t;

/** GATT value type used in response to remote read requests */
typedef struct {
	uint8_t           value[OAL_GATT_MAX_ATTR_LEN];
	uint16_t          handle;
	uint16_t          offset;
	uint16_t          len;
	uint8_t           auth_req;
} oal_gatt_value_t;

/** GATT remote read request response type */
typedef union {
	oal_gatt_value_t attr_value;
	uint16_t            handle;
} oal_gatt_response_t;

typedef struct {
	uint8_t  server_if;
	uint8_t  set_scan_rsp;
	uint8_t  include_name;
	uint8_t  include_txpower;
	uint16_t  appearance;
	char*    manufacturer_data;
	uint16_t manufacturer_data_len;
	char*    service_data;
	uint16_t service_data_len;
	char*    service_uuid;
	uint16_t service_uuid_len;
	char*    solicit_uuid;
	uint16_t solicit_uuid_len;
	uint16_t min_interval;
	uint16_t max_interval;
	uint8_t  adv_type;
	uint8_t  chnl_map;
	uint16_t  tx_power;
	uint8_t  timeout_s;
} oal_ble_multi_adv_param_setup_t;

/**
 * @brief Enable Bluetooth Low Energy GATT Feature
 *
 * @remarks  Application will be able to advertise and scan other Low Energy devices.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  gatt_disable()
 */
oal_status_t gatt_enable(void);

/**
 * @brief Disable Bluetooth Low Energy GATT Feature
 *
 * @remarks  Application will not be able to advertise or scan other Low Energy devices.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAL API should be enabled with gatt_enable().
 *
 * @see  gatt_enable()
 */
oal_status_t gatt_disable(void);

/**
 * @brief Register the GATT Server
 *
 * @remarks  Application will register the GATT Server
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAl API should be enabled with gatt_enable().
 *
 * @see  gatt_enable()
 */
oal_status_t gatts_register(oal_uuid_t* server_uuid);

/**
 * @brief UnRegister the GATT Server
 *
 * @remarks  Application will unregister the GATT Server
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS	Successful
 *
 * @pre OAl API should be enabled with gatti_enable().
 *
 * @see  gatts_register()
 */
oal_status_t gatts_unregister(int instance_id);

/**
 * @brief Start BLE Advertising feature
 *
 * @remarks  Application will start advertising and behave as a peripheral device
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAl API should be enabled with gatt_enable().
 *
 * @see  gatt_enable()
 */
oal_status_t gatts_start_listen(int instance_id, gboolean enable);

/**
 * @brief Stop BLE Advertising feature
 *
 * @remarks  Application will stop advertising and will no longer behave as a peripheral device.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Advertising should have been started with gatts_start_advertising().
 *
 * @see  gatts_start_listen()
 */
oal_status_t gatts_stop_advertising(int instance_id);

/**
 * @brief Enables the Bluetooth Low Energy Multiple advertisement feature
 *
 * @remarks  The application will be able to do multiple advertising
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  gatts_multi_adv_disable()
 */
oal_status_t gatts_multi_adv_enable(int instance_id);

/**
 * @brief Disables the Bluetooth Low Energy Multiple advertisement feature
 *
 * @remarks  The application will not be able to do multiple advertising
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @see  gatts_multi_adv_enable()
 */
oal_status_t gatts_multi_adv_disable(int instance_id);

/**
 * @brief Updates the parameters to configure the Multiple Advertising Features
 *
 * @remarks  The parameters like min and max interval, advertising type,
 * channel map, tx power and timeout will be updated.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API should be enabled with gatts_multi_adv_enable().
 *
 * @see  gatts_multi_adv_enable()
 */
oal_status_t gatts_multi_adv_update(int instance_id, int min_intv, int max_intv,
		int adv_type, int chnl_map,
		int tx_power, int timeout_s);

/**
 * @brief Sets the the Multiple Advertising Data
 *
 * @remarks  The Multiple advertising data will be set.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API should be enabled with gatts_multi_adv_enable().
 *
 * @see  gatts_multi_adv_enable()
 */
oal_status_t gatts_multi_adv_set_inst_data(int instance_id,
		oal_ble_multi_adv_param_setup_t *adv_param_setup);

/**
 * @brief Get current ATT MTU size of the connection
 *
 * @remarks  The application will get current ATT MTU size of the connection
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS>-Successful
 *
 * @pre OAL API must be initialized with oal_bt_init().
 *
 * @param conn_id: active connection id (input)
 * mtu: att mtu size of the connection (output)
 *
 * @see  gatts_multi_adv_enable()
 */
oal_status_t gatts_get_att_mtu(int conn_id, int *mtu);

/**
 * @brief GATT Server Add Service
 *
 * @remarks  GATT Server will add new service in gatt server database.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Advertising should have been started with gatts_start_advertising().
 *
 * @see  gatts_start_advertising()
 */
oal_status_t gatts_add_service(int instance_id, oal_gatt_srvc_id_t *gatt_serv_id, int num_handles);

/**
 * @brief GATT Server Add Included Service
 *
 * @remarks  GATT Server will add new included service in gatt server database.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Primary service should have been added with gatts_add_service().
 *
 * @see  gatts_add_service()
 */
oal_status_t gatts_add_included_services(int instance_id, int srv_handle, int incld_handle);

/**
 * @brief GATT Server Add Charactertistic
 *
 * @remarks  GATT Server will add new Charactertistic in gatt server database.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Primary service should have been added with gatts_add_service().
 *
 * @see  gatts_add_service()
 */
oal_status_t gatts_add_characteristics(int instance_id, int srv_handle, oal_uuid_t* charc_uuid, int propts, int permsn);

/**
 * @brief GATT Server Add Descriptor
 *
 * @remarks  GATT Server will add new Descriptor in gatt server database.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Primary service should have been added with gatts_add_service().
 *
 * @see  gatts_add_service()
 */
oal_status_t gatts_add_descriptor(int instance_id, int srv_handle, oal_uuid_t* desc_uuid, int permsn);

/**
 * @brief GATT Server make ready the new service for advertiment
 *
 * @remarks  GATT Server will enable the new service for advertiment.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Primary service should have been added with gatts_add_service().
 *
 * @see  gatts_add_service()
 */
oal_status_t gatts_start_service(int innstance_id, int svc_handle, int transport);

/**
 * @brief GATT Server Stop the added service from advertiment
 *
 * @remarks  GATT Server will Stop the added service from advertiment .
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Primary service should have been started with gatts_start_service().
 *
 * @see  gatts_start_service()
 */
oal_status_t gatts_stop_service(int instance_id, int svc_hdl);

/**
 * @brief GATT Server delete the already added service
 *
 * @remarks  GATT Server will delete the already added service in gatt server database.
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre Primary service should have been added with gatts_add_service().
 *
 * @see  gatts_add_service()
 */
oal_status_t gatts_delete_service(int instance_id, int svc_handle);

/**
 * @brief GATT Server send the indication for change in attribute value
 *
 * @remarks  GATT Server will send the indication to client for change in attribute value .
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre A Client should be connected to this server.
 *
 * @see  gatts_connect()
 */
oal_status_t gatts_send_indication(int ins_id, int attr_hndl, int conn_id, int len, int confirm, char* value);

/**
 * @brief GATT Server Update value of the attribute associated with a GATT server instance
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 */
oal_status_t gatts_update_att_value(int ins_id, oal_gatt_value_t *value);

/**
 * @brief GATT Server send the response for attribute value
 *
 * @remarks  GATT Server will send the response for attribute value asked from client .
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre A Client should be connected to this server.
 *
 * @see  gatts_connect()
 */
oal_status_t gatts_send_response(int conn_id, int trans_id, int resp_status, oal_gatt_response_t *response);

oal_status_t gatts_disconnect(int Ins_id, bt_address_t *device_address, int conId);

/**
 * @brief Start LE Device discovery
 *
 * @remarks  BLE discovery will be started
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre OAL API should be enabled with gatt_enable().
 *
 * @see  gatt_enable()
 */
oal_status_t gattc_start_le_discovery(int client_id);
/**
 * @brief Stop LE Device discovery
 *
 * @remarks  BLE discovery will be stopped
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 * @pre BLE Discovery should be started using the gattc_start_le_discovery().
 *
 * @see  gattc_start_le_discovery()
 */
oal_status_t gattc_stop_le_discovery(int client_id);

/**
 * @brief Set scan parameters
 *
 * @remarks  scan parameters will be set
 *
 * @return OAL_STATUS_SUCCESS on success, otherwise a non-zero error value.
 * @retval #OAL_STATUS_SUCCESS  Successful
 *
 */
oal_status_t gattc_set_le_scan_param(int scan_type, int itv, int win);

/** Registers a GATT client application with the stack */
oal_status_t gattc_register(oal_uuid_t* client_uuid);

oal_status_t gattc_deregister(int client_id);

 /** Create a connection to a remote LE or dual-mode device */
oal_status_t gattc_connect(int client_id, bt_address_t *device_address, int isDirect);

oal_status_t gattc_search_service(int conn_id, oal_uuid_t *service_uuid);

oal_status_t gattc_get_characteristic(int conn_id, oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id);

oal_status_t gattc_get_descriptor(int conn_id, oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id,
							oal_gatt_id_t *desc_id);

oal_status_t gattc_read_characteristic(int conn_id, oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id,
									oal_gatt_auth_req_t auth_req);

oal_status_t gattc_read_descriptor(int conn_id, oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id,
						oal_gatt_id_t *desc_id, oal_gatt_auth_req_t auth_req);
oal_status_t gattc_write_characteristic(int conn_id, oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id, oal_gatt_write_type_t write_type,
		int len, oal_gatt_auth_req_t auth_req, char* data);

oal_status_t gattc_acquire_notify(int conn_id, oal_gatt_srvc_id_t *srvc_id,
		oal_gatt_id_t *char_id,  int *fd, int *mtu);

oal_status_t gattc_acquire_write(int conn_id, oal_gatt_srvc_id_t *srvc_id,
		oal_gatt_id_t *char_id,  oal_gatt_auth_req_t auth_req , int *fd, int*mtu);

oal_status_t gattc_write_descriptor(int conn_id, oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id, oal_gatt_id_t *desc_id,
		oal_gatt_write_type_t write_type, int len, oal_gatt_auth_req_t auth_req, char* data);

oal_status_t gattc_disconnect(int client_id, bt_address_t *device_address, int conn_id);

oal_status_t gattc_conn_param_update(bt_address_t * address, int min, int max, int latency, int timeout);

oal_status_t gattc_register_for_notification(int client_id, bt_address_t * address,
				oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id);

oal_status_t gattc_deregister_for_notification(int client_id, bt_address_t * address,
				oal_gatt_srvc_id_t *srvc_id, oal_gatt_id_t *char_id);

oal_status_t gatt_send_response_acquire(int conn_id, int trans_id,
		int status, int fd, int mtu, void *);

#endif /* OAL_GATT_H_ */
