/*
 * Open Adaptation Layer (OAL)
 *
 * Copyright (c) 2014-2015 Samsung Electronics Co., Ltd.
 *
 * Contact: Nilesh Trimbake <t.shripati@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dlog.h>
#include <glib.h>

#include <bluetooth.h>
#include <bt_rc.h>

#include "oal-internal.h"
#include "oal-avrcp-ct.h"
#include "oal-utils.h"
#include "oal-common.h"

#define CHECK_OAL_AVRCP_CTRL_ENABLED() \
	do { \
		if (avrcp_ct_interface == NULL) { \
			BT_ERR("OAL, Audio Not Enabled"); \
			return OAL_STATUS_NOT_READY; \
		} \
	} while (0)

#define NO_OF_ATTRIBUTE 1

typedef struct {
	long int song_pos;
	long int playing_time;
	uint8_t title_info[BTRC_MAX_ATTR_STR_LEN];
	uint8_t artist_info[BTRC_MAX_ATTR_STR_LEN];
	uint8_t album_info[BTRC_MAX_ATTR_STR_LEN];
	uint8_t genre_info[BTRC_MAX_ATTR_STR_LEN];
	long int total_track;
	uint32_t track_number;
} avrcp_ct_track_info_t;

typedef struct {
	bdstr_t bdstr;
	int equalizer;
	btrc_player_repeat_val_t repeat;
	btrc_player_shuffle_val_t shuffle;
	int scan;
	btrc_play_status_t status;
	unsigned int volume;
} avrcp_ct_player_settings_t;

typedef struct {
	char address[BT_ADDRESS_STR_LEN];
	avrcp_ct_player_settings_t player_setting;
	avrcp_ct_track_info_t track_info;
} avrcp_ct_player_info_t;

static GSList *player_list = NULL;

const btrc_ctrl_interface_t * _bt_get_stack_interface(void);
static const btrc_ctrl_interface_t *avrcp_ct_interface = NULL;

static void cb_avrcp_ct_connection_state(bool rc_connect, bool bt_connect, bt_bdaddr_t* bd_addr);
static void cb_avrcp_ct_btrc_playerapplicationsetting_changed(bt_bdaddr_t *bd_addr, btrc_player_settings_t *p_vals);
static void cb_avrcp_ct_btrc_play_position_changed(bt_bdaddr_t *bd_addr, uint32_t song_len, uint32_t song_pos);
static void cb_avrcp_ct_btrc_play_status_changed(bt_bdaddr_t *bd_addr, btrc_play_status_t play_status);
static void cb_avrcp_ct_trak_info_chnaged(bt_bdaddr_t *bd_address, uint8_t num_attr, btrc_element_attr_val_t *p_attrs);
static void cb_avrcp_ct_btrc_passthrough_rsp(bt_bdaddr_t *bd_addr, int id, int key_state);
static void cb_avrcp_ct_btrc_setplayerapplicationsetting_rsp(bt_bdaddr_t *bd_addr, uint8_t accepted);

/** AVRCP Controller callback structure. */
static btrc_ctrl_callbacks_t avrcp_ct_cb = {
	.size = sizeof(avrcp_ct_cb),
	.passthrough_rsp_cb = cb_avrcp_ct_btrc_passthrough_rsp,
	.groupnavigation_rsp_cb = NULL,
	.connection_state_cb = cb_avrcp_ct_connection_state,
	.getrcfeatures_cb = NULL,
	.setplayerappsetting_rsp_cb = cb_avrcp_ct_btrc_setplayerapplicationsetting_rsp,
	.playerapplicationsetting_cb = NULL,
	.playerapplicationsetting_changed_cb = cb_avrcp_ct_btrc_playerapplicationsetting_changed,
	.setabsvol_cmd_cb = NULL,
	.registernotification_absvol_cb = NULL,
	.track_changed_cb = cb_avrcp_ct_trak_info_chnaged,
	.play_position_changed_cb = cb_avrcp_ct_btrc_play_position_changed,
	.play_status_changed_cb = cb_avrcp_ct_btrc_play_status_changed,
	.get_folder_items_cb = NULL,
	.change_folder_path_cb = NULL,
	.set_browsed_player_cb = NULL,
	.set_addressed_player_cb = NULL
};

static void __add_device_to_avrcp_list(bt_address_t *address)
{
	avrcp_ct_player_info_t *player_info;
	avrcp_ct_player_settings_t *player_setting;
	bdstr_t bdstr;
	GSList *l;

	bdt_bd2str(address, &bdstr);
	BT_INFO("Address: %s", bdstr);
	for (l = player_list; NULL != l; l = g_slist_next(l)) {
		player_info = (avrcp_ct_player_info_t *)l->data;

		if (!player_info)
			continue;

		if (!strncmp(bdstr, player_info->address, BT_ADDRESS_STR_LEN)) {
			BT_INFO("Already added");
			return;
		}
	}

	player_info = g_new0(avrcp_ct_player_info_t, 1);

	/* Copy address to player_info->address */
	g_strlcpy(player_info->address, bdstr, BT_ADDRESS_STR_LEN);

	/* Init player settings with default values */
	player_setting = &player_info->player_setting;
	player_setting->equalizer = 0;
	player_setting->repeat = BTRC_PLAYER_VAL_OFF_REPEAT;
	player_setting->shuffle = BTRC_PLAYER_VAL_OFF_SHUFFLE;
	player_setting->scan = 0;
	player_setting->status = BTRC_PLAYSTATE_STOPPED;
	player_setting->volume = 0;

	/* Nothing to do for track_info as values are already initialized to 0 */

	/* Add player_info to player_list */
	player_list = g_slist_append(player_list, player_info);
}

static void __remove_device_from_avrcp_list(bt_address_t *address)
{
	avrcp_ct_player_info_t *player_info = NULL;
	bdstr_t bdstr;
	GSList *l;

	bdt_bd2str(address, &bdstr);
	BT_INFO("Address: %s", bdstr);
	for (l = player_list; NULL != l; l = g_slist_next(l)) {
		player_info = (avrcp_ct_player_info_t *)l->data;

		if (!player_info)
			continue;

		if (!strncmp(bdstr, player_info->address, BT_ADDRESS_STR_LEN)) {
			BT_INFO("Found");
			break;
		}

		player_info = NULL;
	}

	if (!player_info)
		return;

	player_list = g_slist_remove(player_list, player_info);
	g_free(player_info);
}

static avrcp_ct_player_info_t *__get_player_info_from_list(bt_address_t *address)
{
	avrcp_ct_player_info_t *player_info;
	bdstr_t bdstr;
	GSList *l;

	bdt_bd2str(address, &bdstr);
	BT_INFO("Address: %s", bdstr);
	for (l = player_list; NULL != l; l = g_slist_next(l)) {
		player_info = (avrcp_ct_player_info_t *)l->data;

		if (!player_info)
			continue;

		if (!strncmp(bdstr, player_info->address, BT_ADDRESS_STR_LEN)) {
			BT_INFO("Found");
			return player_info;
		}
	}

	return NULL;
}

oal_status_t avrcp_ct_enable(void)
{
	const bt_interface_t* blued_inf;
	int ret;

	API_TRACE("AVRCP Controller Enable");

	if ((blued_inf = adapter_get_stack_interface()) == NULL) {
		BT_ERR("Bluetooth module is not loaded");
		return OAL_STATUS_NOT_READY;
	}

	if (avrcp_ct_interface != NULL) {
		BT_WARN("AVRCP Controller Interface is already initialized...");
		return OAL_STATUS_ALREADY_DONE;
	}

	if ((avrcp_ct_interface = (const btrc_ctrl_interface_t *)blued_inf->get_profile_interface(BT_PROFILE_AV_RC_CTRL_ID)) == NULL) {
		BT_ERR("OAL, Failed to get Bluetooth AVRCP Controller Interface");
		return OAL_STATUS_INTERNAL_ERROR;
	}

	BT_DBG("Got profile interface");
	if ((ret = avrcp_ct_interface->init(&avrcp_ct_cb)) != BT_STATUS_SUCCESS) {
		BT_ERR("Failed to initialize Bluetooth AVRCP Controller, status: %s", status2string(ret));
		avrcp_ct_interface = NULL;
		return convert_to_oal_status(ret);
	}
	BT_DBG("OAL, Bluetooth avrcp controller interface initialised");

	return OAL_STATUS_SUCCESS;
}

/* Audio deinit: Resets all the audio information
 * Note: Adapter should be disabled before calling deinit
 * */
oal_status_t avrcp_ct_disable(void)
{
	API_TRACE("AVRCP Controller Disable");

	CHECK_OAL_AVRCP_CTRL_ENABLED();

	avrcp_ct_interface->cleanup();
	avrcp_ct_interface = NULL;

	return OAL_STATUS_SUCCESS;
}

void avrcp_ct_cleanup(void)
{
	BT_DBG();
	avrcp_ct_interface = NULL;
}


oal_status_t avrcp_ct_connect(bt_address_t *device_address)
{
	int result = OAL_STATUS_SUCCESS;
	bt_status_t status;
	bdstr_t bdstr;

	API_TRACE();

	CHECK_OAL_AVRCP_CTRL_ENABLED();
	OAL_CHECK_PARAMETER(device_address, return);

	BT_INFO("BT Audio Address: %s", bdt_bd2str(device_address, &bdstr));

#ifdef TIZEN_BT_HAL
	status = avrcp_ct_interface->connect((bt_bdaddr_t *)device_address);
	if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
		BT_ERR("Connection could not be established, err: %s", status2string(status));;
		result =  convert_to_oal_status(status);
	}
#else
	BT_INFO("Not Supported");
	result = OAL_STATUS_NOT_SUPPORT;
#endif
	return result;
}

oal_status_t avrcp_ct_disconnect(bt_address_t *device_address)
{
	int result = OAL_STATUS_SUCCESS;
	bdstr_t bdstr;
	bt_status_t status;

	API_TRACE();

	CHECK_OAL_AVRCP_CTRL_ENABLED();
	OAL_CHECK_PARAMETER(device_address, return);

	BT_INFO("BT Audio Address: %s", bdt_bd2str(device_address, &bdstr));

#ifdef TIZEN_BT_HAL
	status = avrcp_ct_interface->disconnect((bt_bdaddr_t *)device_address);
	if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
		BT_ERR("OAL, Disconnection failed err: %s", status2string(status));
		result =  convert_to_oal_status(status);
	}
#else
	BT_INFO("Not Supported");
	result = OAL_STATUS_NOT_SUPPORT;
#endif
	return result;
}

static void cb_avrcp_ct_connection_state(bool rc_connect, bool bt_connect, bt_bdaddr_t* bd_addr)
{
	bt_address_t *event_data = NULL;
	bdstr_t bdstr;
	oal_event_t event;

	ret_if(bd_addr == NULL);
	BT_INFO("rc_connect = %d, bt_connect = %d, BT Address = %s",
			rc_connect, bt_connect, bdt_bd2str((bt_address_t*)bd_addr, &bdstr));

	event_data = g_new0(bt_address_t, 1);
	memcpy(event_data->addr, bd_addr->address, BT_ADDRESS_BYTES_NUM);

	if (rc_connect) {
		event = OAL_EVENT_AVRCP_CT_CONNECTED;
		__add_device_to_avrcp_list(event_data);
	} else {
		event = OAL_EVENT_AVRCP_CT_DISCONNECTED;
		__remove_device_from_avrcp_list(event_data);
	}

	send_event_bda_trace(event, event_data, sizeof(bt_address_t), (bt_address_t*)bd_addr);
}

notif_event_avrcp_ct_play_status_t __hal_to_oal_play_status(btrc_play_status_t play_status)
{
	notif_event_avrcp_ct_play_status_t status = OAL_PLAY_STATUS_ERROR;

	switch (play_status) {
	case BTRC_PLAYSTATE_STOPPED:
		status = OAL_PLAY_STATUS_STOPPED;
		break;
	case BTRC_PLAYSTATE_PLAYING:
		status = OAL_PLAY_STATUS_PLAYING;
		break;
	case BTRC_PLAYSTATE_PAUSED:
		status = OAL_PLAY_STATUS_PAUSED;
		break;
	case BTRC_PLAYSTATE_FWD_SEEK:
		status = OAL_PLAY_STATUS_FWD_SEEK;
		break;
	case BTRC_PLAYSTATE_REV_SEEK:
		status = OAL_PLAY_STATUS_REV_SEEK;
		break;
	default:
		BT_INFO("Incorrect Play status");
		break;
	}

	return status;
}

static avrcp_ct_player_settings_t *__get_player_setting(bt_address_t *address)
{
	avrcp_ct_player_info_t *player_info;

	BT_INFO("");
	player_info = __get_player_info_from_list(address);
	if (!player_info)
		return NULL;

	return &player_info->player_setting;
}

static void cb_avrcp_ct_btrc_playerapplicationsetting_changed(
		bt_bdaddr_t *bd_addr, btrc_player_settings_t *p_vals)
{
	avrcp_ct_property_value_t *event_data = NULL;
	oal_event_t event;
	int i;
	bdstr_t bdstr;
	avrcp_ct_player_settings_t *player_setting;

	bdt_bd2str((bt_address_t *)bd_addr, &bdstr);
	player_setting = __get_player_setting((bt_address_t *)bd_addr);
	if (!player_setting) {
		BT_INFO("player_setting is NULL");
		/* Player info not yet added to player_list, first add it to list */
		__add_device_to_avrcp_list((bt_address_t *)bd_addr);
		player_setting = __get_player_setting((bt_address_t *)bd_addr);
	}

	for (i = 0; i < p_vals->num_attr; i++) {
		event_data = g_new0(avrcp_ct_property_value_t, 1);

		switch (p_vals->attr_ids[i]) {
		case BTRC_PLAYER_ATTR_EQUALIZER:
			event = OAL_EVENT_AVRCP_CT_EQUALIZER_STATUS;
			event_data->type = OAL_EQUALIZER;
			event_data->value = p_vals->attr_values[i];
			player_setting->equalizer = p_vals->attr_values[i];
			break;
		case BTRC_PLAYER_ATTR_REPEAT:
			event = OAL_EVENT_AVRCP_CT_REPEAT_STATUS;
			event_data->type = OAL_REPEAT;
			event_data->value = p_vals->attr_values[i];
			player_setting->repeat = p_vals->attr_values[i];
			break;
		case BTRC_PLAYER_ATTR_SHUFFLE:
			event = OAL_EVENT_AVRCP_CT_SHUFFLE_STATUS;
			event_data->type = OAL_SHUFFLE;
			event_data->value = p_vals->attr_values[i];
			player_setting->shuffle = p_vals->attr_values[i];
			break;
		case BTRC_PLAYER_ATTR_SCAN:
			event = OAL_EVENT_AVRCP_CT_SCAN_STATUS;
			event_data->type = OAL_SCAN;
			event_data->value = p_vals->attr_values[i];
			player_setting->scan = p_vals->attr_values[i];
			break;
		default:
			event = OAL_EVENT_END;
			g_free(event_data);
			break;
		}

		if (OAL_EVENT_END != event)
			send_event(event, event_data, sizeof(avrcp_ct_property_value_t));
	}
}

static avrcp_ct_track_info_t *__get_track_info(bt_address_t *address)
{
	avrcp_ct_player_info_t *player_info;

	BT_INFO("");
	player_info = __get_player_info_from_list(address);
	if (!player_info)
		return NULL;

	return &player_info->track_info;
}

static void cb_avrcp_ct_btrc_play_position_changed(
		bt_bdaddr_t *bd_addr, uint32_t song_len, uint32_t song_pos)
{
	event_notif_avrcp_ct_notif_info_t *event_data = NULL;
	avrcp_ct_track_info_t *track_info;

	track_info = __get_track_info((bt_address_t *)bd_addr);
	if (!track_info) {
		BT_INFO("track_info is NULL");
		/* Player info not yet added to player_list, first add it to list */
		__add_device_to_avrcp_list((bt_address_t *)bd_addr);
		track_info = __get_track_info((bt_address_t *)bd_addr);
	}
	track_info->song_pos = song_pos;

	event_data = g_new0(event_notif_avrcp_ct_notif_info_t, 1);
	event_data->song_pos = song_pos;
	send_event(OAL_EVENT_AVRCP_CT_NOTIF_PLAY_POS_CHANGED,
			event_data, sizeof(event_notif_avrcp_ct_notif_info_t));
}

static void cb_avrcp_ct_btrc_play_status_changed(
		bt_bdaddr_t *bd_addr, btrc_play_status_t play_status)
{
	event_notif_avrcp_ct_notif_info_t *event_data = NULL;
	avrcp_ct_player_settings_t *player_setting;

	player_setting = __get_player_setting((bt_address_t *)bd_addr);
	if (!player_setting) {
		BT_INFO("player_setting is NULL");
		/* Player info not yet added to player_list, first add it to list */
		__add_device_to_avrcp_list((bt_address_t *)bd_addr);
		player_setting = __get_player_setting((bt_address_t *)bd_addr);
	}

	player_setting->status = play_status;

	event_data = g_new0(event_notif_avrcp_ct_notif_info_t, 1);
	event_data->play_status = __hal_to_oal_play_status(play_status);
	send_event(OAL_EVENT_AVRCP_CT_NOTIF_PLAY_STATUS_CHANGED,
			event_data, sizeof(event_notif_avrcp_ct_notif_info_t));
}

static void cb_avrcp_ct_btrc_passthrough_rsp(bt_bdaddr_t *bd_addr, int id, int state)
{
	avrcp_ct_pass_cmd_key_code_t key_code;
	avrcp_ct_pass_state_t key_state;
	avrcp_ct_pass_cmd_t *event_data;

	event_data = g_new0(avrcp_ct_pass_cmd_t, 1);

	key_code = id;
	key_state = state;
	if (key_code) {
		event_data->key_code = key_code;
		event_data->key_state = key_state;
		send_event(OAL_EVENT_AVRCP_CT_PASS_CMD_RES,
				event_data, sizeof(avrcp_ct_pass_cmd_t));
	} else {
		BT_ERR("Invalid pass through command key code");
		g_free(event_data);
	}
}

static void cb_avrcp_ct_btrc_setplayerapplicationsetting_rsp(
		bt_bdaddr_t *bd_addr, uint8_t accepted)
{
	avrcp_ct_playersetting_t *event_data;

	event_data = g_new0(avrcp_ct_playersetting_t, 1);

	event_data->accepted = accepted;
	send_event(OAL_EVENT_AVRCP_CT_PLAYER_SETTING_RES,
			event_data, sizeof(avrcp_ct_playersetting_t));
}

static void cb_avrcp_ct_trak_info_chnaged(bt_bdaddr_t *device_address,
		uint8_t num_attr, btrc_element_attr_val_t *p_attrs)
{
	int idx = 0;
	avrcp_ct_track_info_t *track_info;
	event_avrcp_ct_media_info_t *event_data = NULL;

	track_info = __get_track_info((bt_address_t *)device_address);
	if (!track_info) {
		BT_INFO("track_info is NULL");
		/* Player info not yet added to player_list, first add it to list */
		__add_device_to_avrcp_list((bt_address_t *)device_address);
		track_info = __get_track_info((bt_address_t *)device_address);
	}

	event_data = g_new0(event_avrcp_ct_media_info_t, 1);
	for (idx = 0; idx < num_attr; idx++) {
		BT_INFO("idx [%d]", p_attrs[idx].attr_id);

		switch (p_attrs[idx].attr_id) {
		case BTRC_MEDIA_ATTR_ID_TITLE:
			g_strlcpy((char *)track_info->title_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			g_strlcpy((char *)event_data->title_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			break;
		case BTRC_MEDIA_ATTR_ID_ARTIST:
			g_strlcpy((char *)track_info->artist_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			g_strlcpy((char *)event_data->artist_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			break;
		case BTRC_MEDIA_ATTR_ID_ALBUM:
			g_strlcpy((char *)track_info->album_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			g_strlcpy((char *)event_data->album_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			break;
		case BTRC_MEDIA_ATTR_ID_GENRE:
			g_strlcpy((char *)track_info->genre_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			g_strlcpy((char *)event_data->genre_info, (const char *)&p_attrs[idx].text,
					MEDIA_ATTIRBUTE_STRING_LENGTH);
			break;
		case BTRC_MEDIA_ATTR_ID_NUM_TRACKS:
			if (p_attrs[idx].text != NULL) {
				track_info->total_track = g_ascii_strtoll(
						(const gchar *)p_attrs[idx].text, NULL, 10);
				event_data->total_track = track_info->total_track;
			}
			break;
		case BTRC_MEDIA_ATTR_ID_TRACK_NUM:
			if (p_attrs[idx].text != NULL) {
				track_info->track_number = g_ascii_strtoll(
						(const gchar *)p_attrs[idx].text, NULL, 10);
				event_data->track_number = track_info->track_number;
			}
			break;
		case BTRC_MEDIA_ATTR_ID_PLAYING_TIME:
				track_info->playing_time = g_ascii_strtoll(
						(const gchar *)p_attrs[idx].text, NULL, 10);
				event_data->playing_time = (uint64_t) track_info->playing_time;
			break;
		default:
			break;
		}
	}

	send_event(OAL_EVENT_AVRCP_CT_NOTIF_TRACK_CHANGE,
			event_data, sizeof(event_avrcp_ct_media_info_t));
	return;
}

oal_status_t __avrcp_ct_send_pass_through_cmd(bt_address_t *device_address,
		avrcp_ct_pass_cmd_key_code_t key_code, avrcp_ct_pass_state_t key_state)
{
	int result = OAL_STATUS_SUCCESS;
	bt_status_t status;
	bdstr_t bdstr;

	API_TRACE();

	CHECK_OAL_AVRCP_CTRL_ENABLED();

	BT_INFO("BT Audio Address: %s", bdt_bd2str(device_address, &bdstr));

	status = avrcp_ct_interface->send_pass_through_cmd((bt_bdaddr_t *)device_address,
			key_code, key_state);
	if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
		BT_ERR("OAL, send pass through cmd failed err: %s", status2string(status));
		result =  convert_to_oal_status(status);
	}
	return result;
}

oal_status_t avrcp_ct_play(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_PLAY, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_pause(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_PAUSE, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_stop(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_STOP, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_next_track(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_NEXT, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_prev_track(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_PREVIOUS, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_volume_up(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_VOLUME_UP, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_volume_down(bt_address_t *device_address)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();

	result = __avrcp_ct_send_pass_through_cmd(device_address,
			OAL_RC_PASS_CMD_VOLUME_DOWN, PRESS_STATE);

	return result;
}

oal_status_t avrcp_ct_fforward(bt_address_t *device_address, avrcp_ct_pass_state_t press_state)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();
	switch (press_state) {
	case PRESS_STATE:
		result = __avrcp_ct_send_pass_through_cmd(device_address,
				OAL_RC_PASS_CMD_PRESS_FAST_FORWARD, press_state);
		break;
	case RELEASE_STATE:
		result = __avrcp_ct_send_pass_through_cmd(device_address,
				OAL_RC_PASS_CMD_RELEASE_FAST_FORWARD, press_state);
		break;
	default:
		result = OAL_STATUS_INVALID_PARAM;
		break;
	}

	return result;
}

oal_status_t avrcp_ct_rewind(bt_address_t *device_address, avrcp_ct_pass_state_t press_state)
{
	oal_status_t result = OAL_STATUS_SUCCESS;

	API_TRACE();
	switch (press_state) {
	case PRESS_STATE:
		result = __avrcp_ct_send_pass_through_cmd(device_address,
				OAL_RC_PASS_CMD_PRESS_REWIND, press_state);
		break;
	case RELEASE_STATE:
		result = __avrcp_ct_send_pass_through_cmd(device_address,
				OAL_RC_PASS_CMD_RELEASE_REWIND, press_state);
		break;
	default:
		result = OAL_STATUS_INVALID_PARAM;
		break;
	}

	return result;
}

uint8_t __oal_to_hal_property_type(avrcp_ct_player_property_type_t type)
{
	uint8_t property_type = 0;

	switch (type) {
	case OAL_EQUALIZER:
		property_type = BTRC_PLAYER_ATTR_EQUALIZER;
		break;
	case OAL_REPEAT:
		property_type = BTRC_PLAYER_ATTR_REPEAT;
		break;
	case OAL_SHUFFLE:
		property_type = BTRC_PLAYER_ATTR_SHUFFLE;
		break;
	case OAL_SCAN:
		property_type = BTRC_PLAYER_ATTR_SCAN;
		break;
	default:
		BT_ERR(" Invalid Property type[%d]", type);
		break;
	}

	return property_type;
}

oal_status_t avrcp_ct_set_property(bt_address_t *device_address, avrcp_ct_player_property_type_t type, uint32_t value)
{
	int result = OAL_STATUS_SUCCESS;
	uint8_t property_type;
	bt_status_t status;
	bdstr_t bdstr;

	API_TRACE();

	CHECK_OAL_AVRCP_CTRL_ENABLED();
	BT_INFO("BT Audio Address: %s", bdt_bd2str(device_address, &bdstr));

	property_type = __oal_to_hal_property_type(type);
	if (property_type) {
		status = avrcp_ct_interface->set_player_app_setting_cmd((bt_bdaddr_t *)device_address, NO_OF_ATTRIBUTE,
				(uint8_t*)&property_type, (uint8_t*)&value);
		if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
			BT_ERR("OAL, Set property failed err: %s", status2string(status));
			result =  convert_to_oal_status(status);
		}
	} else {
		result = OAL_STATUS_INTERNAL_ERROR;
	}
	return result;
}

oal_status_t avrcp_transport_set_property(bt_address_t *device_address, int type, unsigned int value)
{
	int result = OAL_STATUS_SUCCESS;
	bt_status_t status;
	bdstr_t bdstr;

	API_TRACE();

	CHECK_OAL_AVRCP_CTRL_ENABLED();
	BT_INFO("BT Audio Address: %s", bdt_bd2str(device_address, &bdstr));

	if (type) {
		status = avrcp_ct_interface->set_transport_setting_cmd((bt_bdaddr_t *)device_address, NO_OF_ATTRIBUTE,
				&type, &value);
		if ((status != BT_STATUS_SUCCESS) && (status != BT_STATUS_DONE)) {
			BT_ERR("OAL, Set property failed err: %s", status2string(status));
			result =  convert_to_oal_status(status);
		}
	} else {
		result = OAL_STATUS_INTERNAL_ERROR;
	}
	return result;
}

static gboolean __send_avrcp_property_event(gpointer data)
{
	avrcp_ct_property_value_t *event_data = (avrcp_ct_property_value_t *)data;

	send_event(OAL_EVENT_AVRCP_CT_GET_PROPERTY_RES,
			event_data, sizeof(avrcp_ct_property_value_t));
	return FALSE;
}

oal_status_t avrcp_ct_get_property(bt_address_t *device_address, avrcp_ct_player_property_type_t type)
{
	avrcp_ct_player_info_t *player_info;
	avrcp_ct_player_settings_t *player_setting;
	avrcp_ct_track_info_t *track_info;
	avrcp_ct_property_value_t *event_data;

	API_TRACE();
	CHECK_OAL_AVRCP_CTRL_ENABLED();

	player_info = __get_player_info_from_list(device_address);
	if (!player_info) {
		BT_ERR("Player info not found");
		return OAL_STATUS_INVALID_PARAM;
	}

	player_setting = &player_info->player_setting;
	track_info = &player_info->track_info;

	event_data = g_new0(avrcp_ct_property_value_t, 1);
	event_data->type = type;

	switch (type) {
	case OAL_EQUALIZER:
		event_data->value = player_setting->equalizer;
		break;
	case OAL_REPEAT:
		event_data->value = player_setting->repeat;
		break;
	case OAL_SHUFFLE:
		event_data->value = player_setting->shuffle;
		break;
	case OAL_SCAN:
		event_data->value = player_setting->scan;
		break;
	case OAL_PLAY_POSITION:
		event_data->value = track_info->song_pos;
		break;
	default:
		BT_ERR("Invalid Type [%d]", type);
		g_free(event_data);
		return OAL_STATUS_INVALID_PARAM;
	}

	g_idle_add(__send_avrcp_property_event, (gpointer)event_data);
	return OAL_STATUS_SUCCESS;
}

static gboolean __send_avrcp_play_status_event(gpointer data)
{
	event_avrcp_ct_play_status_t *event_data = (event_avrcp_ct_play_status_t *)data;

	send_event(OAL_EVENT_AVRCP_CT_PLAY_STATUS,
			event_data, sizeof(event_avrcp_ct_play_status_t));

	return FALSE;
}

oal_status_t avrcp_ct_get_play_status(bt_address_t *device_address)
{
	avrcp_ct_player_info_t *player_info;
	avrcp_ct_player_settings_t *player_setting;
	avrcp_ct_track_info_t *track_info;
	event_avrcp_ct_play_status_t *event_data = NULL;

	API_TRACE();
	CHECK_OAL_AVRCP_CTRL_ENABLED();

	player_info = __get_player_info_from_list(device_address);
	if (!player_info) {
		BT_ERR("Player info not found");
		return OAL_STATUS_INVALID_PARAM;
	}

	player_setting = &player_info->player_setting;
	track_info = &player_info->track_info;

	event_data = g_new0(event_avrcp_ct_play_status_t, 1);
	event_data->song_len = track_info->playing_time;
	event_data->song_pos = track_info->song_pos;

	switch (player_setting->status) {
	case BTRC_PLAYSTATE_STOPPED:
		event_data->play_status = OAL_PLAY_STATUS_STOPPED;
		break;
	case BTRC_PLAYSTATE_PLAYING:
		event_data->play_status = OAL_PLAY_STATUS_PLAYING;
		break;
	case BTRC_PLAYSTATE_PAUSED:
		event_data->play_status = OAL_PLAY_STATUS_PAUSED;
		break;
	case BTRC_PLAYSTATE_FWD_SEEK:
		event_data->play_status = OAL_PLAY_STATUS_FWD_SEEK;
		break;
	case BTRC_PLAYSTATE_REV_SEEK:
		event_data->play_status = OAL_PLAY_STATUS_REV_SEEK;
		break;
	default:
		event_data->play_status = OAL_PLAY_STATUS_ERROR;
		break;
	}

	g_idle_add(__send_avrcp_play_status_event, (gpointer)event_data);
	return OAL_STATUS_SUCCESS;
}

static gboolean __send_media_attribute_event(gpointer data)
{
	event_avrcp_ct_media_info_t *event_data = (event_avrcp_ct_media_info_t *)data;

	send_event(OAL_EVENT_AVRCP_CT_MEDIA_INFO,
			event_data, sizeof(event_avrcp_ct_media_info_t));

	return FALSE;
}

oal_status_t avrcp_ct_get_media_attribute(bt_address_t *device_address)
{
	avrcp_ct_track_info_t *track_info;
	event_avrcp_ct_media_info_t *event_data;

	API_TRACE();

	CHECK_OAL_AVRCP_CTRL_ENABLED();

	OAL_CHECK_PARAMETER(device_address, return);

	track_info = __get_track_info(device_address);
	if (!track_info) {
		BT_ERR("Player info not found");
		return OAL_STATUS_INVALID_PARAM;
	}

	event_data = g_new0(event_avrcp_ct_media_info_t, 1);
	g_strlcpy((char *)event_data->title_info, (const char *)track_info->title_info,
				MEDIA_ATTIRBUTE_STRING_LENGTH);
	g_strlcpy((char *)event_data->album_info, (const char *)track_info->album_info,
				MEDIA_ATTIRBUTE_STRING_LENGTH);
	g_strlcpy((char *)event_data->genre_info, (const char *)track_info->genre_info,
				MEDIA_ATTIRBUTE_STRING_LENGTH);
	g_strlcpy((char *)event_data->artist_info, (const char *)track_info->artist_info,
				MEDIA_ATTIRBUTE_STRING_LENGTH);
	event_data->playing_time = (uint64_t)track_info->playing_time;
	event_data->total_track = track_info->total_track;
	event_data->track_number = track_info->track_number;

	g_idle_add(__send_media_attribute_event, (gpointer)event_data);

	return OAL_STATUS_SUCCESS;
}
