/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Nilesh Trimbake <t.shripati@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <vconf.h>

#include "oal-hardware.h"
#include "oal-device-mgr.h"
#include <oal-manager.h>
#include <oal-a2dp-sink.h>

#include <bt-service-a2dp-sink.h>
#include <bt-service-event.h>

int _bt_a2dp_connect_remote_source(bluetooth_device_address_t *device_address/*,
						bt_audio_function_data_t *func_data*/)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int result = BLUETOOTH_ERROR_NONE;
	BT_INFO("+");

	status = a2dp_sink_connect((bt_address_t*)device_address);
	if (status != OAL_STATUS_SUCCESS) {
		BT_ERR("Connection could not be established, err: [%d]", status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}
	return result;
}

int _bt_a2dp_disconnect_remote_source(bluetooth_device_address_t *device_address/*,
						bt_audio_function_data_t *func_data*/)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int result = BLUETOOTH_ERROR_NONE;
	BT_INFO("+");

	status = a2dp_sink_disconnect((bt_address_t*)device_address);
	if (status != OAL_STATUS_SUCCESS) {
		BT_ERR("DisConnection err: [%d]", status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}
	return result;
}

static void __bt_handle_av_source_connected_state(bluetooth_device_address_t *address)
{
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };

	bt_headset_wait_t *wait_device = NULL;
	bluetooth_device_address_t wait_device_address;
	GVariant *param;
	int result = BLUETOOTH_ERROR_NONE;
	ret_if(NULL == address);
	BT_INFO("+");

	_bt_convert_addr_type_to_string(addr, address->addr);
	BT_INFO("Address of connected device [%s]", addr);

	/* Set VCONF Key for A2DP Connected status */
	_bt_set_device_values(TRUE, VCONFKEY_BT_DEVICE_A2DP_HEADSET_CONNECTED);

	/* Send A2DP(SRC Role) connected event to Application */
	param = g_variant_new("(is)", result, addr);
	_bt_send_event(BT_HEADSET_EVENT, BLUETOOTH_EVENT_AV_SOURCE_CONNECTED, param);

	/* Add data from the connected list */
	_bt_add_headset_to_list(BT_AUDIO_A2DP_SOURCE, BT_STATE_CONNECTED, addr);

	/* Delete waiting device data if present */
	wait_device = _bt_get_audio_wait_data();
	if (wait_device != NULL &&
			(g_strcmp0(wait_device->address, addr) == 0))
		_bt_rel_wait_data();
	else {
		if (wait_device != NULL) {
			BT_INFO("A2DP Profile connected for device [%s] but another device [%s]is waiting to be connnected type [%d]",
					addr, wait_device->address, wait_device->type);
			BT_INFO("Trigger connect for the waiting device [%s], audio type [%d]",
					wait_device->address, wait_device->type);
			_bt_convert_addr_string_to_type(wait_device_address.addr,
					wait_device->address);
			_bt_audio_connect(wait_device->type, &wait_device_address);
			/* Now free the wait list */
			//_bt_rel_wait_data();
		}
	}
	BT_INFO("-");
}

static void __bt_handle_av_source_disconnected_state(bluetooth_device_address_t *address)
{
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int result = BLUETOOTH_ERROR_NONE;
	GArray *out_param;
	bt_headset_wait_t *wait_device;
	bluetooth_device_address_t device_address;
	GVariant *param;
	invocation_info_t *req_info = NULL;
	ret_if(NULL == address);
	BT_INFO("+");

	_bt_convert_addr_type_to_string(addr, address->addr);

	/* Set VCONF status for A2DP Disconnection */
	_bt_set_device_values(FALSE, VCONFKEY_BT_DEVICE_A2DP_HEADSET_CONNECTED);


	/* Remove data from the connected list */
	_bt_remove_headset_from_list(BT_AUDIO_A2DP_SOURCE, addr);

	req_info = _bt_get_request_info_data(BT_AV_SOURCE_DISCONNECT, addr);
	BT_INFO("Address  of disconnected device[%s]", addr);
	if (NULL == req_info) {
		BT_INFO("AV DisConnect request not found or possibly already replied");
		req_info = _bt_get_request_info_data(BT_AV_SOURCE_CONNECT, addr);
		if (NULL == req_info) {
			/* This means, AV Disconnect request has successfully passed, and real disconnect is completed */
			BT_INFO("Neither AV Connect or AV Disconnect request is found..means, AV Profile is disconnected");

			/* Send A2DP(SRC Role) disconnected event to Application */
			result = BLUETOOTH_ERROR_NONE;
			param = g_variant_new("(is)", result, addr);
			_bt_send_event(BT_HEADSET_EVENT, BLUETOOTH_EVENT_AV_SOURCE_DISCONNECTED, param);

			/*
			 * TODO: Below logic is not required for BT_AUDIO_CONNECT/BT_AUDIO_DISCONNECT.
			 * But if Tizen supports a device with A2DP Sink and HFP both profiles (BT_AUDIO_HFP_SOURCE)
			 * then below logic will be required for corresponding connect/disconnect function
			 */
#if 0
			req_info = _bt_get_request_info_data(BT_AUDIO_CONNECT, addr);
			if (req_info == NULL) {
				BT_INFO("Audio Connect All request is also Not found..");

				req_info = _bt_get_request_info_data(BT_AUDIO_DISCONNECT, addr);
				if (req_info == NULL)
					BT_INFO("Audio DisConnect All request is also Not found..");
				else
					BT_INFO("Audio DisConnect All request is found..");
			} else {
				BT_INFO("Audio Connect All request is found..");
				result = BLUETOOTH_ERROR_INTERNAL;
				data = _bt_get_pending_audio_conn_info(addr);
				if (data) {
					BT_INFO("Pending connect is found, delete it..");
					_bt_cleanup_audio_pending_conn_info_and_reply_pending_req(data, BLUETOOTH_ERROR_NONE);
					req_info = NULL;
				} else {
					BT_INFO("Pending connect is Not found..");
				}
			}
#endif
			goto check_wait_device;
		} else {
			BT_ERR("AV Connect request has failed.."); /* DBUS return needed */
			result = BLUETOOTH_ERROR_INTERNAL;

			/* Delete waiting device data if present for this device */
			wait_device = _bt_get_audio_wait_data();
			if (wait_device && strncasecmp(wait_device->address, addr, BT_ADDRESS_STRING_SIZE) == 0) {
				_bt_rel_wait_data();
				goto dbus_return;
			}

			goto check_wait_device;
		}
	} else {
		BT_ERR("AV Disconnect request found for [%s], means disconnect request is successful", addr); /* DBUS return needed */

		/* Send A2DP(SRC Role) disconnected event to Application */
		result = BLUETOOTH_ERROR_NONE;
		param = g_variant_new("(is)", result, addr);
		_bt_send_event(BT_HEADSET_EVENT, BLUETOOTH_EVENT_AV_SOURCE_DISCONNECTED, param);

		goto check_wait_device;
	}
check_wait_device:
	wait_device = _bt_get_audio_wait_data();

	if (wait_device == NULL)
		goto dbus_return;

	if (((wait_device->type == BT_AUDIO_ALL) &&
				(wait_device->ag_flag == TRUE)) ||
			(wait_device->type == BT_AUDIO_A2DP_SOURCE) ||
			(wait_device->disconnection_type == BT_AUDIO_A2DP_SOURCE)) {
		if (g_strcmp0(wait_device->address, addr) != 0) {
			BT_INFO("Trigger connect for the waiting device [%s], audio type [%d]",
					wait_device->address, wait_device->type);
			_bt_convert_addr_string_to_type(device_address.addr,
					wait_device->address);
			_bt_audio_connect(wait_device->type, &device_address);
		}
	}
dbus_return:
	if (req_info) {
		out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
		g_array_append_vals(out_param, addr, BT_ADDRESS_STRING_SIZE);
		_bt_service_method_return(req_info->context,
				out_param, result);
		g_array_free(out_param, TRUE);
		g_free(req_info->user_data);
		_bt_free_info_from_invocation_list(req_info);
	}
}

static void __bt_reply_av_source_connection_pending_request(bluetooth_device_address_t *address)
{
	BT_DBG("+");
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	bluetooth_device_address_t device_address;
	GArray *out_param;
	invocation_info_t *req_info;
	bt_pending_audio_conn_t *data = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	memcpy(device_address.addr, address->addr, BLUETOOTH_ADDRESS_LENGTH);
	_bt_convert_addr_type_to_string(addr, address->addr);

	req_info = _bt_get_request_info_data(BT_AV_SOURCE_CONNECT, addr);
	if (NULL == req_info) {
		BT_INFO("AV source Connect request not found or possibly already replied");
		req_info = _bt_get_request_info_data(BT_AUDIO_CONNECT, addr);

		if (req_info == NULL) {
			BT_INFO("AV Connect request & Audio All connect not found or possibly already replied");
			return;
		} else {
			BT_INFO("Audio All connect request found for address [%s]", addr);
			/* TODO Delete the search info if present */
			data = _bt_get_pending_audio_conn_info(addr);
			if (data) {
				BT_INFO("Audio Connect All request present and pending connect is also present, delete info");
				_bt_cleanup_audio_pending_conn_info_and_reply_pending_req(data, BLUETOOTH_ERROR_NONE);
			} else {
				BT_ERR("Abnormal: Audio Connect All request present but pending connect absent");
			}
		}
	} else {
		BT_INFO("AV Connect request found for [%s]", addr);
	}

	out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
	g_array_append_vals(out_param, addr, BT_ADDRESS_STRING_SIZE);
	_bt_service_method_return(req_info->context,
			out_param, result);
	g_array_free(out_param, TRUE);
	g_free(req_info->user_data);
	_bt_free_info_from_invocation_list(req_info);
}

static void __bt_reply_av_source_disconnection_pending_request(bluetooth_device_address_t *address)
{
	BT_DBG("+");
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	bluetooth_device_address_t device_address;
	GArray *out_param;
	invocation_info_t *req_info;
	int result = BLUETOOTH_ERROR_NONE;
	memcpy(device_address.addr, address->addr, BLUETOOTH_ADDRESS_LENGTH);
	_bt_convert_addr_type_to_string(addr, address->addr);

	req_info = _bt_get_request_info_data(BT_AV_SOURCE_DISCONNECT, addr);
	if (NULL == req_info) {
		BT_INFO("AV DisConnect request not found or possibly already replied");
		req_info = _bt_get_request_info_data(BT_AUDIO_DISCONNECT, addr);
		if (req_info == NULL) {
			BT_INFO("AUDIO Disconnect All request also not found or possibly already replied");
			return;
		} else {
			BT_INFO("Audio All DisConnect request found..");
		}
	} else {
		BT_INFO("AV Source DisConnect request found for [%s]", addr);
	}

	/* In any of the above cases, do DBUS return */
	out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
	g_array_append_vals(out_param, &device_address,
			sizeof(bluetooth_device_address_t));
	_bt_service_method_return(req_info->context,
			out_param, result);
	g_array_free(out_param, TRUE);
	g_free(req_info->user_data);
	_bt_free_info_from_invocation_list(req_info);
}

void _bt_a2dp_sink_handle_incoming_authorization(char *address, int service_id)
{
	char connected_address[BT_ADDRESS_STRING_SIZE + 1];
	gboolean connected;
	bluetooth_device_address_t device_address;
	oal_status_t res = OAL_STATUS_SUCCESS;
	BT_INFO("+");

	_bt_convert_addr_string_to_type(device_address.addr, address);

	switch (service_id) {
	case A2DP_SRC_SERVICE_ID:
		connected = _bt_is_headset_type_connected(BT_AUDIO_A2DP_SOURCE, connected_address);
		if (connected) {
			BT_INFO("A2DP (Remote Source) is already connected..check which device..");
			if (g_strcmp0(connected_address, address) != 0) {
				BT_INFO("Different device is connected[%s] reject current [%s] device authorization request",
						connected_address, address);
				res = device_reply_auth_request((bt_address_t*)&device_address, service_id, FALSE, FALSE);
				if (res != OAL_STATUS_SUCCESS)
					BT_ERR("authorize_response: %d", res);
				return;
			}
		} else {
			res = device_reply_auth_request((bt_address_t*)&device_address, service_id, TRUE, FALSE);
			if (res != OAL_STATUS_SUCCESS)
				BT_ERR("authorize_response: %d", res);
		}
		break;
	default:
		BT_ERR("Invalid Audio Service ID..");
	}

	BT_INFO("-");
}

void _bt_a2dp_sink_event_handler(int oal_event, gpointer event_data)
{
	BT_INFO("+");
	bluetooth_device_address_t * bd_addr = (bluetooth_device_address_t*)event_data;

	switch (oal_event) {
	case OAL_EVENT_A2DP_SINK_CONNECTING:
		BT_INFO("A2DP Sink Profile connecting..");
		__bt_reply_av_source_connection_pending_request(bd_addr);
		break;
	case OAL_EVENT_A2DP_SINK_DISCONNECTING:
		BT_INFO("A2DP Sink Profile disconnecting..");
		__bt_reply_av_source_disconnection_pending_request(bd_addr);
		break;
	case OAL_EVENT_A2DP_SINK_CONNECTED:
		BT_INFO("A2DP Sink Profile connected..");
		__bt_handle_av_source_connected_state(bd_addr);
		break;
	case OAL_EVENT_A2DP_SINK_DISCONNECTED:
		BT_INFO("A2DP Sink Profile dissconnected..");
		__bt_handle_av_source_disconnected_state(bd_addr);
		break;
	default:
		BT_INFO("Invalid Event = %d", oal_event);
		break;
	}
	BT_INFO("-");
}
