/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <gio/gio.h>
#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <dbus/dbus.h>

#include <oal-avrcp-tg.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"
#include "bt-service-common.h"
#include "bt-service-avrcp-tg.h"
#include "bt-service-event.h"
#include "bt-service-event-receiver.h"
#include "bt-service-audio-common.h"
#include "bt-service-util.h"

static invocation_info_t* __bt_get_request_info(int service_function, char *address)
{
	GSList *l;
	invocation_info_t *req_info = NULL;

	BT_DBG("+");

	retv_if(NULL == address, FALSE);

	/* Get method invocation context */
	for (l = _bt_get_invocation_list(); l != NULL; l = g_slist_next(l)) {
		req_info = l->data;
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		if (!strncasecmp((char *)req_info->user_data, address, BT_ADDRESS_STRING_SIZE))
			return req_info;
	}

	return NULL;
}

static int __bt_avrcp_set_equalizer(media_player_equalizer_status equalizer)
{
	unsigned int value;
	oal_status_t result;

	BT_DBG("+");

	switch (equalizer) {
	case EQUALIZER_ON:
		value = 0x01;
		break;
	case EQUALIZER_OFF:
		value = 0x02;
		break;
	default:
		BT_ERR("Invalid Equalizer state");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	result = avrcp_set_property(AVRCP_EQUALIZER, value);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("avrcp_set_property failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static int __bt_avrcp_set_repeat_status(media_player_repeat_status repeat_status)
{
	unsigned int value;
	oal_status_t result;

	BT_DBG("+");

	switch (repeat_status) {
	case REPEAT_MODE_OFF:
		value = OAL_PLAYER_VAL_OFF_REPEAT;
		break;
	case REPEAT_SINGLE_TRACK:
		value = OAL_PLAYER_VAL_SINGLE_REPEAT;
		break;
	case REPEAT_ALL_TRACK:
		value = OAL_PLAYER_VAL_ALL_REPEAT;
		break;
	case REPEAT_GROUP:
		value = OAL_PLAYER_VAL_GROUP_REPEAT;
		break;
	default:
		BT_ERR("Invalid repeat status");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	result = avrcp_set_property(AVRCP_REPEAT, value);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("avrcp_set_property failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static int __bt_avrcp_set_shuffel_mode(media_player_shuffle_status shuffel)
{
	unsigned int value;
	oal_status_t result;

	BT_DBG("+");

	switch (shuffel) {
	case SHUFFLE_MODE_OFF:
		value = OAL_PLAYER_VAL_OFF_SHUFFLE;
		break;
	case SHUFFLE_ALL_TRACK:
		value = OAL_PLAYER_VAL_ALL_SHUFFLE;
		break;
	case SHUFFLE_GROUP:
		value = OAL_PLAYER_VAL_GROUP_SHUFFLE;
		break;
	default:
		BT_ERR("Invalid shuffel mode");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	result = avrcp_set_property(AVRCP_SHUFFLE, value);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("avrcp_set_property failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static int __bt_avrcp_set_scan_mode(media_player_scan_status scan_status)
{
	unsigned int value;
	oal_status_t result;

	BT_DBG("+");

	switch (scan_status) {
	case SCAN_MODE_OFF:
		value = 0x01;
		break;
	case SCAN_ALL_TRACK:
		value = 0x02;
		break;
	case SCAN_GROUP:
		value = 0x03;
		break;
	default:
		BT_ERR("Invalid scan mode");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	result = avrcp_set_property(AVRCP_SCAN, value);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("avrcp_set_property failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static int __bt_avrcp_set_player_status(media_player_status status)
{
	unsigned int value;
	oal_status_t result;

	BT_DBG("+");

	switch (status) {
	case STATUS_STOPPED:
		value = OAL_PLAYSTATE_STOPPED;
		break;
	case STATUS_PLAYING:
		value = OAL_PLAYSTATE_PLAYING;
		break;
	case STATUS_PAUSED:
		value = OAL_PLAYSTATE_PAUSED;
		break;
	case STATUS_FORWARD_SEEK:
		value = OAL_PLAYSTATE_FWD_SEEK;
		break;
	case STATUS_REVERSE_SEEK:
		value = OAL_PLAYSTATE_REV_SEEK;
		break;
	case STATUS_ERROR:
		value = OAL_PLAYSTATE_ERROR;
		break;
	default:
		BT_ERR("Invalid ");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	result = avrcp_set_property(AVRCP_STATUS, value);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("avrcp_set_property failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static int __bt_avrcp_set_play_position(unsigned int position)
{
	oal_status_t result;

	BT_DBG("+");

	result = avrcp_set_property(AVRCP_POSITION, position);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("avrcp_set_property failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_avrcp_connect_remote_ctrl(bluetooth_device_address_t *address)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int result = BLUETOOTH_ERROR_NONE;
	bt_address_t bdaddr;

	BT_DBG("+");

	memcpy(bdaddr.addr, address->addr, BT_ADDRESS_BYTES_NUM);
	status = avrcp_tg_connect(&bdaddr);
	if (status != OAL_STATUS_SUCCESS) {
		BT_ERR("Connection could not be established, err: [%d]", status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return result;
}

int _bt_avrcp_disconnect_remote_ctrl(bluetooth_device_address_t *address)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int result = BLUETOOTH_ERROR_NONE;
	bt_address_t bdaddr;

	BT_DBG("+");

	memcpy(bdaddr.addr, address->addr, BT_ADDRESS_BYTES_NUM);
	status = avrcp_tg_disconnect(&bdaddr);
	if (status != OAL_STATUS_SUCCESS) {
		BT_ERR("DisConnection err: [%d]", status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return result;
}

int _bt_avrcp_set_track_info(media_metadata_attributes_t *meta_data)
{
	oal_status_t result;
	oal_media_metadata_attributes_t media_attr;

	BT_DBG("+");

	retv_if(meta_data == NULL, BLUETOOTH_ERROR_INTERNAL);

	g_strlcpy(media_attr.title, meta_data->title, OAL_MAX_ATTR_STR_LEN);
	g_strlcpy(media_attr.artist, meta_data->artist, OAL_MAX_ATTR_STR_LEN);
	g_strlcpy(media_attr.album, meta_data->album, OAL_MAX_ATTR_STR_LEN);
	g_strlcpy(media_attr.genre, meta_data->genre, OAL_MAX_ATTR_STR_LEN);
	media_attr.total_tracks = meta_data->total_tracks;
	media_attr.number = meta_data->number;
	media_attr.duration = meta_data->duration;

	result = avrcp_set_track_info(&media_attr);
	if (result != OAL_STATUS_SUCCESS) {
		BT_ERR("hid_connect error: [%d]", result);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_avrcp_set_properties(media_player_settings_t *properties)
{
	int ret;
	BT_DBG("+");

	ret = __bt_avrcp_set_equalizer(properties->equalizer);
	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("__bt_avrcp_set_equalizer failed");
		return ret;
	}

	ret = __bt_avrcp_set_repeat_status(properties->repeat);
	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("__bt_avrcp_set_repeat_status failed");
		return ret;
	}

	ret = __bt_avrcp_set_shuffel_mode(properties->shuffle);
	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("__bt_avrcp_set_shuffel_mode failed");
		return ret;
	}

	ret = __bt_avrcp_set_scan_mode(properties->scan);
	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("__bt_avrcp_set_scan_mode failed");
		return ret;
	}

	ret = __bt_avrcp_set_play_position(properties->position);
	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("__bt_avrcp_set_play_position failed");
		return ret;
	}

	ret = _bt_avrcp_set_track_info(&(properties->metadata));
	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("_bt_avrcp_set_track_info failed");
		return ret;
	}

	BT_DBG("-");
	return ret;
}

int _bt_avrcp_set_property(int type, unsigned int value)
{
	int ret;

	BT_DBG("+");

	switch (type) {
	case EQUALIZER:
		ret = __bt_avrcp_set_equalizer(value);
		break;
	case REPEAT:
		ret = __bt_avrcp_set_repeat_status(value);
		break;
	case SHUFFLE:
		ret = __bt_avrcp_set_shuffel_mode(value);
		break;
	case SCAN:
		ret = __bt_avrcp_set_scan_mode(value);
		break;
	case STATUS:
		ret = __bt_avrcp_set_player_status(value);
		break;
	case POSITION:
		ret = __bt_avrcp_set_play_position(value);
		break;
	default:
		BT_DBG("Invalid Type");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	if (ret != BLUETOOTH_ERROR_NONE) {
		BT_ERR("set property: %.2X failed", type);
		return ret;
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static void __handle_player_property_equalizer(unsigned char *value)
{
	media_player_equalizer_status equalizer;

	switch (*value) {
	case 0x01:
		equalizer = EQUALIZER_ON;
		break;
	case 0x02:
		equalizer = EQUALIZER_OFF;
		break;
	default:
		BT_ERR("Unknown equalizer setting");
		equalizer = EQUALIZER_INVALID;
	}

	_bt_send_event(BT_AVRCP_EVENT,
			BLUETOOTH_EVENT_AVRCP_SETTING_EQUALIZER_STATUS,
			g_variant_new("(u)", equalizer));
}

static void __handle_player_property_repeat(unsigned char *value)
{
	media_player_repeat_status status;

	switch (*value) {
	case OAL_PLAYER_VAL_OFF_REPEAT:
		status = REPEAT_MODE_OFF;
		break;
	case OAL_PLAYER_VAL_SINGLE_REPEAT:
		status = REPEAT_SINGLE_TRACK;
		break;
	case OAL_PLAYER_VAL_GROUP_REPEAT:
		status = REPEAT_GROUP;
		break;
	case OAL_PLAYER_VAL_ALL_REPEAT:
		status = REPEAT_ALL_TRACK;
		break;
	default:
		BT_ERR("Invalid repeat setting");
		status = REPEAT_INVALID;
	}

	_bt_send_event(BT_AVRCP_EVENT,
			BLUETOOTH_EVENT_AVRCP_SETTING_REPEAT_STATUS,
			g_variant_new("(u)", status));
}

static void __handle_player_property_shuffle(unsigned char *value)
{
	media_player_shuffle_status status;

	switch (*value) {
	case OAL_PLAYER_VAL_OFF_SHUFFLE:
		status = SHUFFLE_MODE_OFF;
		break;
	case OAL_PLAYER_VAL_GROUP_SHUFFLE:
		status = SHUFFLE_GROUP;
		break;
	case OAL_PLAYER_VAL_ALL_SHUFFLE:
		status = SHUFFLE_ALL_TRACK;
		break;
	default:
		BT_ERR("Invalid shuffle setting");
		status = SHUFFLE_INVALID;
	}

	_bt_send_event(BT_AVRCP_EVENT,
			BLUETOOTH_EVENT_AVRCP_SETTING_SHUFFLE_STATUS,
			g_variant_new("(u)", status));
}

static void __handle_player_property_scan(unsigned char *value)
{
	media_player_scan_status status;

	switch (*value) {
	case 0x01:
		status = SCAN_MODE_OFF;
		break;
	case 0x02:
		status = SCAN_ALL_TRACK;
		break;
	case 0x03:
		status = SCAN_GROUP;
		break;
	default:
		BT_ERR("Unknown scan setting");
		status = SCAN_INVALID;
	}

	_bt_send_event(BT_AVRCP_EVENT,
			BLUETOOTH_EVENT_AVRCP_SETTING_SCAN_STATUS,
			g_variant_new("(u)", status));
}

static void __handle_transport_property_delay(unsigned int *value)
{
	BT_DBG("+");
	unsigned int delay;
	delay = *value;

	_bt_send_event(BT_AVRCP_EVENT,
			BLUETOOTH_EVENT_AVRCP_DELAY_CHANGED,
			g_variant_new("(u)", delay));

	BT_DBG("-");
}

static void __handle_avrcp_connection_event(int event, bt_address_t *bdaddr)
{
	int result = BLUETOOTH_ERROR_NONE;
	char address[BT_ADDRESS_STRING_SIZE];
	GVariant *param = NULL;
	invocation_info_t *req_info;

	_bt_convert_addr_type_to_string(address, bdaddr->addr);

	if (event == BLUETOOTH_EVENT_AVRCP_CONNECTED) {
		/* Reply to async request for AVRCP Target connect, if any */
		req_info = __bt_get_request_info(BT_AVRCP_TARGET_CONNECT, address);
		if (!req_info) {
			/* Check if request for AVRCP Target disconnect failed */
			req_info = __bt_get_request_info(BT_AVRCP_TARGET_DISCONNECT, address);
			if (req_info)
				result = BLUETOOTH_ERROR_INTERNAL;
		}
	} else {
		/* Reply to async request for AVRCP Target disconnect, if any */
		req_info = __bt_get_request_info(BT_AVRCP_TARGET_DISCONNECT, address);
		if (!req_info) {
			/* Check if request for AVRCP Target connect failed */
			req_info = __bt_get_request_info(BT_AVRCP_TARGET_CONNECT, address);
			if (req_info)
				result = BLUETOOTH_ERROR_INTERNAL;
		}
	}

	if (NULL != req_info) {
		GArray *out_param;

		/* Create out param */
		out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
		g_array_append_vals(out_param, address, sizeof(address));

		_bt_service_method_return(req_info->context, out_param, result);
		g_array_free(out_param, TRUE);
		g_free(req_info->user_data);
		_bt_free_info_from_invocation_list(req_info);
	}

	if (BLUETOOTH_ERROR_NONE == result) {
		param = g_variant_new("(is)", result, address);
		/* Send event to application */
		_bt_send_event(BT_AVRCP_EVENT, event, param);

		if (event == BLUETOOTH_EVENT_AVRCP_CONNECTED)
			/* Connected: Add AVRCP TARGET to headset list */
			_bt_add_headset_to_list(BT_AVRCP_TARGET, BT_STATE_CONNECTED, address);
		else
			/* Disconnected: Remove AVRCP TARGET from headset list */
			_bt_remove_headset_from_list(BT_AVRCP_TARGET, address);
	}
}

static void __handle_avrcp_target_events(int event_type, gpointer event_data)
{
	switch (event_type) {
	case OAL_EVENT_AVRCP_CONNECTED:
		__handle_avrcp_connection_event(
				BLUETOOTH_EVENT_AVRCP_CONNECTED,
				(bt_address_t *)event_data);
		break;
	case OAL_EVENT_AVRCP_DISCONNECTED:
		__handle_avrcp_connection_event(
				BLUETOOTH_EVENT_AVRCP_DISCONNECTED,
				(bt_address_t *)event_data);
		break;
	case OAL_EVENT_AVRCP_SETTING_EQUALIZER_STATUS:
		__handle_player_property_equalizer((unsigned char *)event_data);
		break;
	case OAL_EVENT_AVRCP_SETTING_REPEAT_STATUS:
		__handle_player_property_repeat((unsigned char *)event_data);
		break;
	case OAL_EVENT_AVRCP_SETTING_SHUFFLE_STATUS:
		__handle_player_property_shuffle((unsigned char *)event_data);
		break;
	case OAL_EVENT_AVRCP_SETTING_SCAN_STATUS:
		__handle_player_property_scan((unsigned char *)event_data);
		break;
	case OAL_EVENT_AVRCP_REMOTE_FEATURES:
		break;
	case OAL_EVENT_AVRCP_VOLUME_MUTE_CHANGED:
		break;
	case OAL_EVENT_AVRCP_DELAY_CHANGED:
		 __handle_transport_property_delay((unsigned int *)event_data);
		break;
	default:
		BT_ERR("Unhandled AVRCP target event: %d", event_type);
	}
}

int _bt_service_avrcp_enable(void)
{
	oal_status_t status = OAL_STATUS_SUCCESS;

	status = avrcp_enable();
	if (OAL_STATUS_SUCCESS != status) {
		BT_ERR("Failed to initialize Bluetooth AVRCP Target Profile, status: %d", status);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Register AVRCP target event handler */
	_bt_service_register_event_handler_callback(BT_AVRCP_MODULE, __handle_avrcp_target_events);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_service_avrcp_disable(void)
{
	oal_status_t status = OAL_STATUS_SUCCESS;

	status = avrcp_disable();
	if (OAL_STATUS_SUCCESS != status) {
		BT_ERR("Failed to initialize Bluetooth A2DP Source Profile, status: %d", status);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Register AVRCP target event handler */
	_bt_service_unregister_event_handler_callback(BT_AVRCP_MODULE);

	return BLUETOOTH_ERROR_NONE;
}
