/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <sys/un.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <vconf.h>

#include <oal-event.h>
#include <oal-hardware.h>
#include <oal-audio-src.h>
#include <oal-a2dp-sink.h>

#include "bt-service-common.h"
#include "bt-request-handler.h"
#include "bt-service-audio-common.h"
#include "bt-service-core-device.h"
#include "bt-service-a2dp-src.h"
#include "bt-service-a2dp-sink.h"
#include "bt-service-avrcp-tg.h"
#include "bt-service-avrcp-ctrl.h"
#include "bt-service-hf.h"

#ifdef TIZEN_SUPPORT_DUAL_HF
#define VCONF_KEY_BT_HOST_BT_MAC_ADDR "db/wms/host_bt_mac"
#endif

/* Static variables and Macros */
static GList *g_connected_list;
static bt_headset_wait_t *g_wait_data;
typedef struct {
	unsigned int type;
	int device_state;
	char device_address[BT_ADDRESS_STRING_SIZE + 1];
} bt_connected_headset_data_t;

/* List for saving device service search info */
static GSList *pending_audio_conn_list = NULL;

/* Headset connection data structures */
gboolean connection_local = FALSE;

static int curr_audio_role = BLUETOOTH_A2DP_SOURCE;
static int pending_audio_role = -1;

static void __bt_remove_device_from_wait_list(char *prev_waiting_device);

static void __bt_audio_event_handler(int oal_event, gpointer event_data);

static int __bt_process_audio_connect_all(bt_pending_audio_conn_t *info);

static int __bt_is_headset_connecting(int type);

static int __bt_is_headset_disconnecting(int type);

static void __bt_audio_cleanup_resources(void);

static gboolean __handle_pending_audio_select_role(gpointer data);

static void __bt_reply_pending_audio_connect_req(char *address, int result);

#if 0
void _bt_headset_set_local_connection(gboolean value)
{
	BT_INFO("setting connection_local to %d", value);
	connection_local = value;
}

gboolean _bt_headset_get_local_connection()
{
	return connection_local;
}
#endif

void _bt_set_audio_wait_data_flag(gboolean flag)
{
	BT_DBG("_bt_set_audio_wait_data_flag \n");
	g_wait_data->ag_flag = flag;
}

bt_headset_wait_t *_bt_get_audio_wait_data(void)
{
	BT_DBG("_bt_get_audio_wait_data \n");
	return g_wait_data;
}

static void __bt_free_wait_data(void)
{
	if (g_wait_data != NULL) {
		g_free(g_wait_data->address);
		g_free(g_wait_data);
		g_wait_data = NULL;
	}
}

void _bt_rel_wait_data(void)
{
	BT_DBG("_bt_rel_wait_data \n");
	__bt_free_wait_data();
}

void _bt_add_headset_to_list(int type, int status, const char *address)
{
	bt_connected_headset_data_t *connected_device;
	bt_connected_headset_data_t *device;
	GList *node;

	BT_DBG("_bt_add_headset_to_list \n");

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		device = (bt_connected_headset_data_t *)node->data;

		if (g_strcmp0(device->device_address, address) == 0) {
			BT_DBG("Address match, update connection type \n");
			if (status == BT_STATE_CONNECTED)
				device->type |= type;
			device->device_state = status;
			return;
		}
		node = g_list_next(node);
	}

	connected_device = g_malloc0(sizeof(bt_connected_headset_data_t));
	/* Fix : NULL_RETURNS */
	if (connected_device == NULL) {
		BT_ERR("No memory allocated");
		return;
	}

	connected_device->device_state = status;
	//if ((status == BT_STATE_CONNECTED) || (status == BT_STATE_CONNECTING))
	if (status == BT_STATE_CONNECTED)
		connected_device->type |= type;
	g_strlcpy(connected_device->device_address, address,
			sizeof(connected_device->device_address));
	g_connected_list = g_list_append(g_connected_list, connected_device);
	BT_INFO("Added device[%s] in connected list, device state [%d] Device Type [%d]",
				address, connected_device->device_state, connected_device->type);
}

gboolean _bt_is_headset_type_connected(int type, char *address)
{
	GList *node;

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		bt_connected_headset_data_t *connected_device = node->data;

		if (connected_device->type & type) {
			if (address != NULL)
				g_strlcpy(address, connected_device->device_address,
						BT_ADDRESS_STRING_SIZE);
			return TRUE;
		}

		node = g_list_next(node);
	}
	return FALSE;
}


gboolean _bt_is_headset_address_type_connected(int type, const char *address)
{
	GList *node;

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		bt_connected_headset_data_t *connected_device = node->data;

		if (connected_device && (connected_device->type & type)) {
			if (memcmp(connected_device->device_address, address, 19) == 0)
				return TRUE;
		}

		node = g_list_next(node);
	}
	return FALSE;
}

static void __bt_set_headset_disconnection_type(const char *address)
{
	bt_connected_headset_data_t *connected_device;
	GList *node;

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		connected_device = node->data;
		if (g_strcmp0(connected_device->device_address, address) == 0) {
			g_wait_data->disconnection_type = connected_device->type;
			return;
		}
		node = g_list_next(node);
	}
}

#ifdef TIZEN_SUPPORT_DUAL_HF
gboolean __bt_is_companion_device(const char *addr)
{
	if (TIZEN_PROFILE_WEARABLE) {
		char *host_device_address = NULL;
		host_device_address = vconf_get_str(VCONF_KEY_BT_HOST_BT_MAC_ADDR);

		if (!host_device_address) {
			BT_INFO("Failed to get a companion device address");
			return FALSE;
		}

		if (g_strcmp0(host_device_address, addr) == 0) {
			BT_INFO("addr[%s] is companion device", addr);
			return TRUE;
		}

		return FALSE;
	} else {
	/* TODO : Need to add companion device check condition for Phone models */
	return FALSE;
	}
}
#endif

static int __bt_is_headset_disconnecting(int type)
{
	bt_connected_headset_data_t *connected_device = NULL;

	/* Check if any other headset is connected */
	GList *node = NULL;

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		connected_device = node->data;
		if (connected_device->device_state == BT_STATE_DISCONNECTING)
			return BLUETOOTH_ERROR_CONNECTION_BUSY;
		node = g_list_next(node);
	}

	return BLUETOOTH_ERROR_NONE;
}

static int __bt_is_headset_connecting(int type)
{
	bt_connected_headset_data_t *connected_device = NULL;

	/* Check if any other headset is connected */
	GList *node = NULL;

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		connected_device = node->data;
		if (connected_device->device_state == BT_STATE_CONNECTING) {
			BT_ERR("@@Device [%s] is already under connecting state", connected_device->device_address);
			return BLUETOOTH_ERROR_CONNECTION_BUSY;
		}
		node = g_list_next(node);
	}

	return BLUETOOTH_ERROR_NONE;
}

static int __bt_is_headset_connected(int current_conn_type, const char *address)
{
	gboolean connected = FALSE;
	int disconn_type;
	char connected_address[BT_ADDRESS_STRING_SIZE + 1];
	bluetooth_device_address_t device_address;
	bt_connected_headset_data_t *connected_device = NULL;
#ifdef TIZEN_SUPPORT_DUAL_HF
	gboolean is_companion_device = FALSE;
#endif

	/* Check if any other headset is connected */
	GList *node = NULL;;

	BT_INFO("Checking  if any Headset connected or not: current device [%s] current dev type [%d]", address, current_conn_type);
	node = g_list_first(g_connected_list);
	while (node != NULL) {
		connected_device = node->data;
		BT_INFO(" A Device is already connected found in list [%s] conn type [%d]",
				connected_device->device_address, connected_device->type);
		if ((connected_device->type & current_conn_type)) {
			g_strlcpy(connected_address, connected_device->device_address,
					BT_ADDRESS_STRING_SIZE + 1);
#ifdef TIZEN_SUPPORT_DUAL_HF
			is_companion_device = __bt_is_companion_device(connected_address);
			BT_INFO(" is_companion_device[%d]", is_companion_device);

			if (!is_companion_device) {
				connected = TRUE;
				break;
			}
#else
			connected = TRUE;
			break;
#endif
		}
		node = g_list_next(node);
	}

	if (!connected) {
		BT_INFO("There is no connected device with connection type [%d]", current_conn_type);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	BT_DBG("connected headset %s", connected_address);

	if (g_strcmp0(connected_address, address) == 0)
		return BLUETOOTH_ERROR_ALREADY_CONNECT;
#ifdef TIZEN_SUPPORT_DUAL_HF
	else if (TRUE == __bt_is_companion_device(address))
		return BLUETOOTH_ERROR_NOT_CONNECTED;
#endif

	/* Convert BD address from string type */
	_bt_convert_addr_string_to_type(device_address.addr, connected_address);
	int value = BLUETOOTH_ERROR_NONE;
	BT_DBG("Already connected headset addr  [%s] connected headset type [0x%x] current dev conn type [0x%x]",
				connected_address, connected_device->type, current_conn_type);
	disconn_type = connected_device->type & current_conn_type;
	BT_DBG("Attempt disconnection of Type [0x%x] of already connected device" , disconn_type);
	value = _bt_audio_disconnect(disconn_type, &device_address);

	/* If already one device is waiting, remove current waiting device and add new */
	if (value == BLUETOOTH_ERROR_NONE) {
		if (g_wait_data != NULL) {
			if (g_strcmp0(g_wait_data->address, address) != 0) {
				BT_INFO("Already one device was waiting for connection [%s], remove it", g_wait_data->address);
				__bt_remove_device_from_wait_list(g_wait_data->address);
				__bt_free_wait_data();
			}
		}

		if (g_wait_data == NULL) {
			BT_INFO("Add current device [%s] into waiting list", address);
			g_wait_data = g_malloc0(sizeof(bt_headset_wait_t));
			g_wait_data->address = g_strdup(address);
			g_wait_data->type = current_conn_type;
			g_wait_data->ag_flag = FALSE;

			/* Set disconnection type */
			__bt_set_headset_disconnection_type(connected_address);
		}
	}

	return value;
}

static void __bt_remove_device_from_wait_list(char *prev_waiting_device)
{
	/* Before deleting the request update the UI */
	bluetooth_device_address_t device_address;
	int result = BLUETOOTH_ERROR_INTERNAL;
	GArray *out_param;
	invocation_info_t *req_info;
	BT_INFO("We are about to DBUS return previous waiting device [%s]", prev_waiting_device);

	_bt_convert_addr_string_to_type(device_address.addr, prev_waiting_device);

	req_info = _bt_get_request_info_data(BT_AV_CONNECT, prev_waiting_device);
	if (req_info == NULL) {
		BT_ERR("req_info == NULL");
		return;
	} else {
		out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
		g_array_append_vals(out_param, &device_address,
				sizeof(bluetooth_device_address_t));
		_bt_service_method_return(req_info->context,
				out_param, result);
		g_array_free(out_param, TRUE);
		g_free(req_info->user_data);
		_bt_free_info_from_invocation_list(req_info);
	}
}

static int __bt_process_audio_profile_connect(bt_audio_type_t type, bluetooth_device_address_t *address)
{
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int result = BLUETOOTH_ERROR_NONE;
	BT_INFO("+");

	_bt_convert_addr_type_to_string(addr,
			(unsigned char *)address->addr);
	/* Attempt profile level connection based on type */
	switch (type) {
	case BT_AUDIO_HSP:
		result = _bt_connect_remote_hfp(address);
		break;
	case BT_AUDIO_A2DP:
		result = _bt_a2dp_connect_remote_sink(address);
		break;
	case BT_AVRCP_TARGET:
		result = _bt_avrcp_connect_remote_ctrl(address);
		break;
	case BT_AVRCP:
		result = _bt_avrcp_connect_remote_target(address);
		break;
	case BT_AUDIO_A2DP_SOURCE:
		return _bt_a2dp_connect_remote_source(address);
		break;
	default:
		BT_ERR("Unknown role");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	if (result == BLUETOOTH_ERROR_NONE) {
		//_bt_headset_set_local_connection(TRUE);
		/* Add data to the connected list */
		_bt_add_headset_to_list(type, BT_STATE_CONNECTING, addr);
	} else {
		BT_ERR("Profile [%d] connect failed!!", type);
	}
	BT_INFO("-");
	return result;
}

static int __bt_process_audio_profile_disconnect(bt_audio_type_t type, bluetooth_device_address_t *address)
{
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int result = BLUETOOTH_ERROR_NONE;
	GList *node;
	BT_INFO("+");

	_bt_convert_addr_type_to_string(addr,
			(unsigned char *)address->addr);
	/* Attempt profile level connection based on type */
	switch (type) {
	case BT_AUDIO_HSP:
		result = _bt_disconnect_remote_hfp(address);
		break;
	case BT_AUDIO_A2DP:
		result = _bt_a2dp_disconnect_remote_sink(address);
		break;
	case BT_AVRCP_TARGET:
		result = _bt_avrcp_disconnect_remote_ctrl(address);
		break;
	case BT_AVRCP:
		result = _bt_avrcp_disconnect_remote_target(address);
		break;
	case BT_AUDIO_A2DP_SOURCE:
		return _bt_a2dp_disconnect_remote_source(address);
		break;
	default:
		BT_ERR("Unknown role");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	if (result == BLUETOOTH_ERROR_NONE) {
		/*
		 *      This logic is added for dual HF mode issue.
		 */
		node = g_list_first(g_connected_list);
		while (node != NULL) {
			bt_connected_headset_data_t *connected_device = node->data;

			if (g_strcmp0(connected_device->device_address, addr) == 0) {
				BT_DBG("Connection type update");
				type = connected_device->type;
				break;
			}
			node = g_list_next(node);
		}

		/* Update device status in connected list */
		_bt_add_headset_to_list(type, BT_STATE_DISCONNECTING, addr);
	} else {
		BT_ERR("Profile [%d] connect failed!!", type);
	}
	BT_INFO("-");
	return result;
}

int _bt_audio_initialize(bt_service_module_t module)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int ret  = BLUETOOTH_ERROR_NONE;

	BT_INFO("_bt_audio_initialize: Module [%d]", module);

	switch (module) {
	case BT_A2DP_SOURCE_MODULE: {
		status = audio_enable(NULL, NULL);
		if (OAL_STATUS_SUCCESS != status) {
			BT_ERR("Failed to initialize Bluetooth A2DP Source Profile, status: %d", status);
			return BLUETOOTH_ERROR_INTERNAL;
		}
		/* Register Audio module event handler */
		_bt_service_register_event_handler_callback(module, _bt_a2dp_source_event_handler);
		break;
	}
	case BT_A2DP_SINK_MODULE: {
		status = a2dp_sink_enable(NULL, NULL);
		if (status != OAL_STATUS_SUCCESS) {
			BT_ERR("Failed to initialize Bluetooth A2DP Sink Profile, status: %d", status);
			return BLUETOOTH_ERROR_INTERNAL;
		}
		/* Register Audio sink module event handler */
		_bt_service_register_event_handler_callback(module, _bt_a2dp_sink_event_handler);
		break;
	}
	case BT_HFP_MODULE: {
		return BLUETOOTH_ERROR_NOT_SUPPORT;
	}
	case BT_AG_MODULE: {
		status = hfp_enable(1);
		if (OAL_STATUS_SUCCESS != status) {
			BT_ERR("Failed to initialize Bluetooth HFP Audio Gateway Profile, status: %d", status);
			return BLUETOOTH_ERROR_INTERNAL;
		}
		/* Register Audio module event handler */
		_bt_service_register_event_handler_callback(module, _bt_hf_event_handler);
		break;
	}
	case BT_AUDIO_ALL_MODULE: {
		/* Register Audio module event handler */
		_bt_service_register_event_handler_callback(module, __bt_audio_event_handler);
		break;
	}
	case BT_AVRCP_CTRL_MODULE: {
		status = avrcp_ct_enable();
		if (status != OAL_STATUS_SUCCESS) {
			BT_ERR("Failed to initialize Bluetooth AVRCP Controller Profile, status: %d", status);
			return BLUETOOTH_ERROR_INTERNAL;
		}
		/* Register Avrcp Controller event handler */
		_bt_service_register_event_handler_callback(module, _bt_avrcp_ctrl_event_handler);
		break;
	}
	case BT_AVRCP_MODULE: {
		/* Initialize AVRCP Target */
		ret = _bt_service_avrcp_enable();
		break;
	}
	default:
		BT_ERR("Not Supported: Module [%d]", module);
		return BLUETOOTH_ERROR_NOT_SUPPORT;
	}

	BT_INFO("Bluetooth audio interface initialised");
	return ret;
}

static void __bt_audio_event_handler(int oal_event, gpointer event_data)
{
	BT_INFO("+");

	switch (oal_event) {
	case OAL_EVENT_ADAPTER_DISABLED: {
		BT_INFO("Adapter Disabled..cleanup resources");
		__bt_audio_cleanup_resources();
		break;
	}
	case OAL_EVENT_ADAPTER_ENABLED: {
		/*TODO Currently not handled */
		BT_INFO("Adapter Enabled..");
		break;
	}
	default:
		BT_ERR("Unknown event..[%d]", oal_event);
	}
	BT_INFO("-");
}

static void __bt_audio_cleanup_resources(void)
{
	GList *node;
	GSList *l;
	bt_pending_audio_conn_t *info = NULL;
	BT_INFO("+");

	/* Remove connected device list */
	node = g_list_first(g_connected_list);
	while (node != NULL) {
		bt_connected_headset_data_t *connected_device = node->data;
		BT_INFO("Data found [%s] in connected device list...remove it..", connected_device->device_address);
		g_connected_list = g_list_remove(g_connected_list, connected_device);
		node = g_list_next(node);
	}

	/* Remove pending connection list */
	for (l = pending_audio_conn_list; l != NULL; l = g_slist_next(l)) {
		info = (bt_pending_audio_conn_t*)l->data;

		if (info) {
			BT_INFO("Info found for addr [%s], remove it...", info->address);
			pending_audio_conn_list = g_slist_remove(pending_audio_conn_list, info);
			g_free(info->address);
			g_free(info);
			info = NULL;
		}
	}

	BT_INFO("-");
}

static int __bt_process_audio_connect_all(bt_pending_audio_conn_t *info)
{
	bluetooth_device_address_t device_address;
	int result = BLUETOOTH_ERROR_NONE;
	int type = BT_AUDIO_ALL;
	BT_INFO("+");

	_bt_convert_addr_string_to_type(device_address.addr, info->address);

	/* Attempt profile level connection */
	if (info->is_hfp_supported) {
		type = BT_AUDIO_HSP;
		result = _bt_connect_remote_hfp(&device_address);

		if (info->is_a2dp_sink_supported)
			BT_INFO("A2DP is supported by [%s]", info->address);
		else
			BT_INFO("A2DP is not supported");
	} else if (info->is_a2dp_sink_supported) {
		type = BT_AUDIO_A2DP;
		result = _bt_a2dp_connect_remote_sink(&device_address);
	} else if (info->is_a2dp_src_supported) {
		type = BT_AUDIO_A2DP_SOURCE;
		result = _bt_a2dp_connect_remote_source(&device_address);
	}
	if (result == BLUETOOTH_ERROR_NONE) {
		//_bt_headset_set_local_connection(TRUE);
		/* Add data to the connected list */
		_bt_add_headset_to_list(type, BT_STATE_CONNECTING, info->address);
	} else {
		BT_ERR("Profile connect failed!!");
	}
	BT_INFO("-");
	return result;
}

bt_pending_audio_conn_t* _bt_get_pending_audio_conn_info(char *address)
{
	BT_INFO("+");
	bt_pending_audio_conn_t *info = NULL;
	GSList *l;

	for (l = pending_audio_conn_list; l != NULL; l = g_slist_next(l)) {
		info = (bt_pending_audio_conn_t*)l->data;

		if (g_strcmp0(info->address, address) == 0 && info->search_status == SERVICE_SEARCH_DONE) {
			BT_INFO("Pending data found for addr [%s]", info->address);
			return info;
		}
	}
	BT_INFO("Pending data not found!!");
	return NULL;
}

static void __bt_reply_pending_audio_connect_req(char *address, int result)
{
	GArray *out_param;
	invocation_info_t *req_info;

	BT_DBG("+");
	ret_if(NULL == address);

	/* Reply to async request for Audio connect, if any */
	req_info = _bt_get_request_info_data(BT_AUDIO_CONNECT, address);
	ret_if(NULL == req_info);

	out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
	g_array_append_vals(out_param, address, BT_ADDRESS_STRING_SIZE);
	_bt_service_method_return(req_info->context, out_param, result);

	g_array_free(out_param, TRUE);
	g_free(req_info->user_data);
	_bt_free_info_from_invocation_list(req_info);

	BT_DBG("-");
}

void _bt_cleanup_audio_pending_conn_info_and_reply_pending_req(bt_pending_audio_conn_t *info, int result)
{
	ret_if(NULL == info);

	/* Reply to async request for Audio connect, if any */
	__bt_reply_pending_audio_connect_req(info->address, result);

	/* Remove pending audio conn info */
	pending_audio_conn_list = g_slist_remove(pending_audio_conn_list, info);
	g_free(info->address);
	g_free(info);
}

gboolean _bt_audio_check_pending_connection(char *address)
{
	bt_pending_audio_conn_t *data = NULL;
	bluetooth_device_address_t device_address;
	BT_INFO("+");

	data = _bt_get_pending_audio_conn_info(address);

	if (data) {
		BT_INFO("A2DP Connection is pending..initiate it...");
		if (data->is_a2dp_sink_supported && data->conn_req_type == BT_AUDIO_ALL) {
			_bt_convert_addr_string_to_type(device_address.addr, address);
			if (_bt_audio_connect(BT_AUDIO_A2DP, &device_address) != BLUETOOTH_ERROR_NONE) {
				BT_INFO("A2DP connect triggered for [%s]  but it failed!!", data->address);
				_bt_cleanup_audio_pending_conn_info_and_reply_pending_req(data, BLUETOOTH_ERROR_INTERNAL);
			} else {
				BT_INFO("A2DP connect triggered successfully for [%s]", data->address);
			}
		} else if (data->is_a2dp_src_supported && data->conn_req_type == BT_AUDIO_HFP_SOURCE) {
			_bt_convert_addr_string_to_type(device_address.addr, address);
			if (_bt_audio_connect(BT_AUDIO_A2DP_SOURCE, &device_address) != BLUETOOTH_ERROR_NONE) {
				BT_INFO("A2DP sink connect triggered for [%s]  but it failed!!", data->address);
				_bt_cleanup_audio_pending_conn_info_and_reply_pending_req(data, BLUETOOTH_ERROR_INTERNAL);
			} else {
				BT_INFO("A2DP sink connect triggered successfully for [%s]", data->address);
			}
		}
		return TRUE;
	} else {
		BT_INFO("A2DP Connection is not pending..");
	}
	BT_INFO("-");
	return FALSE;
}

void _bt_audio_check_pending_disconnection(char *address, int type)
{
	int ret = BLUETOOTH_ERROR_NONE;
	bluetooth_device_address_t device_address;
	BT_INFO("+");

	if (_bt_is_service_connected(address, type)) {
		BT_INFO("Service [%d] is connected with device [%s], disconnect it...", type, address);
		_bt_convert_addr_string_to_type(device_address.addr, address);
		ret = __bt_process_audio_profile_disconnect(type, &device_address);

		if (ret != BLUETOOTH_ERROR_NONE)
			BT_ERR("Disconnecting service [%d] with device [%s] failed!!", type, address);
	} else {
		BT_INFO("Service [%d] is Not connected with device [%s],..", type, address);
	}
	BT_INFO("-");
}

int __bt_handle_audio_all_connect(bt_pending_audio_conn_t *info)
{
	bt_remote_dev_info_t *rem_info;
	int ret;
	int indx;

	BT_DBG("+");

	retv_if(NULL == info, BLUETOOTH_ERROR_INTERNAL);
	retv_if(NULL == info->address, BLUETOOTH_ERROR_INTERNAL);

	rem_info = _bt_service_get_remote_dev_info(info->address);
	if (!rem_info) {
		BT_ERR("_bt_service_get_remote_dev_info returned NULL");
		ret = BLUETOOTH_ERROR_NOT_PAIRED;
		goto fail;
	}

	for (indx = 0; indx < rem_info->uuid_count; indx++) {
		BT_INFO("UUID [%s]", rem_info->uuids[indx]);
		if (0 == g_strcmp0(A2DP_SINK_UUID, rem_info->uuids[indx])) {
			BT_INFO("Device supports A2DP Sink Profile");
			info->is_a2dp_sink_supported = TRUE;
		} else if (0 == g_strcmp0(A2DP_SOURCE_UUID, rem_info->uuids[indx])) {
			BT_INFO("Device supports A2DP Source Profile");
			info->is_a2dp_src_supported = TRUE;
		} else if ((g_strcmp0(HFP_HS_UUID, rem_info->uuids[indx]) == 0)) {
			BT_INFO("Device supports HFP Profile");
			info->is_hfp_supported = TRUE;
		}
	}

	if (info->is_a2dp_sink_supported == FALSE &&
			info->is_hfp_supported == FALSE &&
			info->is_a2dp_src_supported == FALSE) {
		BT_ERR("No audio profile is supported");
		ret = BLUETOOTH_ERROR_NOT_SUPPORT;
		goto fail;
	}

	BT_INFO("AUDIO_CONNECT_ALL request for [%s]", info->address);
	info->search_status = SERVICE_SEARCH_DONE;

	/* Give preference to HFP over A2DP for outgoing connection sequence for AUDIO_ALL_CONNECT */
	if (info->is_hfp_supported)
		info->type = BT_AUDIO_HSP;
	else if (info->is_a2dp_sink_supported)
		info->type = BT_AUDIO_A2DP;
	else if (info->is_a2dp_src_supported)
		info->type = BT_AUDIO_A2DP_SOURCE;

	ret = __bt_is_headset_connected(info->type, info->address);
	if (ret == BLUETOOTH_ERROR_ALREADY_CONNECT) {
		BT_ERR("Can't start Audio All connect for [%s], reason: [%s]",
				info->address, "Already connected...");
		goto fail;
	}

	if (ret == BLUETOOTH_ERROR_IN_PROGRESS) {
		BT_ERR("Can't start Audio All connect for [%s], reason: [%s]",
				info->address, "Already in progress...");
		goto fail;
	}

	if (ret == BLUETOOTH_ERROR_NOT_CONNECTED) {
		BT_ERR("Great, can start Audio All connect for [%s]", info->address);
		ret = __bt_is_headset_connecting(info->type);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Can't start Audio All connect for [%s], reason: [%s]",
					info->address, "some other device conn in progress");
			ret = BLUETOOTH_ERROR_IN_PROGRESS;
			goto fail;
		}

		ret = __bt_process_audio_connect_all(info);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("__bt_process_audio_connect_all failed");
			goto fail;
		}

		/* If multiple profiles are supported, queue pending connection info */
		if ((info->is_hfp_supported && info->is_a2dp_sink_supported) ||
				(info->is_hfp_supported && info->is_a2dp_src_supported)) {
			BT_INFO("[%s] Supports HFP and (A2DP_Src or, A2DP_Snk)", info->address);
			pending_audio_conn_list = g_slist_append(pending_audio_conn_list, (gpointer)info);
		} else {
			BT_INFO("[%s] Supports one profile only", info->address);
			/*
			 * It means, we dont need pending connect info as only A2DP (Src/Snk) or
			 * HFP will be connected for present device, so lets free pending info.
			 */
			g_free(info->address);
			g_free(info);
		}
	} else if (ret == BLUETOOTH_ERROR_NONE) {
		BT_INFO("Waiting for disconnection...");
		/*
		 * It means, we dont need pending connect info as only A2DP (Src/Snk) or
		 * HFP will be connected for present device, so lets free pending info.
		 */
		g_free(info->address);
		g_free(info);
	}

	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;

fail:
	BT_ERR("Audio connect all failed");
	g_free(info->address);
	g_free(info);
	return ret;
}

static gboolean __audio_get_bonded_info(gpointer data)
{
	bt_pending_audio_conn_t *info = data;
	char address[BT_ADDRESS_STRING_SIZE];

	BT_DBG("+");

	retv_if(NULL == info, FALSE);
	retv_if(NULL == info->address, FALSE);

	if ((info->bonded_info_retries > 0) && !_bt_is_bonded_devices_retrived()) {
		info->bonded_info_retries--;
		return TRUE;
	}

	g_strlcpy(address, info->address, BT_ADDRESS_STRING_SIZE);
	if (info->bonded_info_retries <= 0) {
		BT_ERR_C("Even after all retries, bonded devices stil not retrieved");
		g_free(info->address);
		g_free(info);
		goto failed;
	}

	if (BLUETOOTH_ERROR_NONE != __bt_handle_audio_all_connect(info)) {
		BT_ERR("__bt_handle_audio_all_connect failed");
		goto failed;
	}

	BT_DBG("-");
	return FALSE;
failed:
	__bt_reply_pending_audio_connect_req(address, BLUETOOTH_ERROR_INTERNAL);
	return FALSE;
}

int _bt_audio_connect(int type,	bluetooth_device_address_t *device_address)
{
	int ret = BLUETOOTH_ERROR_NONE;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	bt_pending_audio_conn_t *info = NULL;

	BT_INFO("+");

	BT_CHECK_PARAMETER(device_address, return);

	_bt_convert_addr_type_to_string(address, device_address->addr);
	BT_INFO("Adderss: [%s], request type: [%d]", address, type);

	/*
	 * If type is BT_AUDIO_ALL or BT_AUDIO_HFP_SOURCE, enqueue search info in
	 * order to find out supported services by the remote device.
	 */
	if (type == BT_AUDIO_ALL || type ==  BT_AUDIO_HFP_SOURCE) {
		if (NULL != _bt_get_pending_audio_conn_info(address)) {
			BT_ERR("Connection for same device already in progress");
			return BLUETOOTH_ERROR_IN_PROGRESS;
		}

		info = g_malloc0(sizeof(bt_pending_audio_conn_t));
		info->search_status = SERVICE_SEARCH_STARTED;
		info->address = g_strdup(address);
		info->type = type;
		info->conn_req_type = type;

		/* Request bonded device info */
		if (!_bt_is_bonded_devices_retrived()) {
			info->bonded_info_retries = 5;
			/* Wait till bonded devices retrival is completed */
			g_timeout_add_seconds(1, __audio_get_bonded_info, info);
			return BLUETOOTH_ERROR_NONE;
		}

		return __bt_handle_audio_all_connect(info);
	}

	/* If type is A2DP Sink or HSP/HFP, check further */
	ret =  __bt_is_headset_connected(type, address);
	if (ret == BLUETOOTH_ERROR_ALREADY_CONNECT) {
		BT_ERR("Failed with BLUETOOTH_ERROR_ALREADY_CONNECT");
		return BLUETOOTH_ERROR_ALREADY_CONNECT;
	}

	if (ret == BLUETOOTH_ERROR_IN_PROGRESS) {
		BT_ERR("Failed with BLUETOOTH_ERROR_IN_PROGRESS");
		return BLUETOOTH_ERROR_IN_PROGRESS;
	}

	if (ret == BLUETOOTH_ERROR_NONE) {
		BT_INFO("Waiting for disconnection...");
	} else if (ret == BLUETOOTH_ERROR_NOT_CONNECTED) {
		ret = __bt_process_audio_profile_connect(type, device_address);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Failed with %d", ret);
			return ret;
		}
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_audio_disconnect(int type, bluetooth_device_address_t *device_address)
{
	int ret = BLUETOOTH_ERROR_NONE;
	int value = BLUETOOTH_ERROR_NONE;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	_bt_convert_addr_type_to_string(address, device_address->addr);
	BT_INFO("Audio Diconnect Request: type[%d] address [%s]", type, address);

	if (type == BT_AUDIO_ALL) {
		if (_bt_is_service_connected(address, BT_AUDIO_A2DP)) {
			type = BT_AUDIO_A2DP;
		} else if (_bt_is_service_connected(address, BT_AUDIO_HSP)) {
			type = BT_AUDIO_HSP;
		} else {
			BT_ERR("No audio service connected");
			return BLUETOOTH_ERROR_NOT_CONNECTED;
		}
	} else if (type == BT_AUDIO_HFP_SOURCE) {
		if (_bt_is_service_connected(address, BT_AUDIO_A2DP_SOURCE)) {
			type = BT_AUDIO_A2DP_SOURCE;
		} else if (_bt_is_service_connected(address, BT_AUDIO_HSP)) {
			type = BT_AUDIO_HSP;
		} else {
			BT_ERR("No audio service connected");
			return BLUETOOTH_ERROR_NOT_CONNECTED;
		}
	}

	value = __bt_is_headset_disconnecting(type);
	if (value != BLUETOOTH_ERROR_NONE) {
		BT_INFO("Disconnect in progress");
		return BLUETOOTH_ERROR_IN_PROGRESS;
	}

	ret = __bt_process_audio_profile_disconnect(type, device_address);
	return ret;
}

gboolean _bt_is_service_connected(char* address, int type)
{
	GList *node;
	node = g_list_first(g_connected_list);
	while (node != NULL) {
		bt_connected_headset_data_t *conn_device = node->data;

		if ((g_strcmp0(conn_device->device_address, address) == 0) &&
				(conn_device->type & type)) {
			BT_INFO("Service connected");
			return TRUE;
		}

		node = g_list_next(node);
	}
	BT_INFO("Service not connected");
	return FALSE;
}

void _bt_remove_headset_from_list(int type, const char *address)
{
	GList *node;

	BT_DBG("_bt_remove_headset_from_list \n");

	node = g_list_first(g_connected_list);
	while (node != NULL) {
		bt_connected_headset_data_t *connected_device = node->data;

		if (g_strcmp0(connected_device->device_address, address) != 0) {
			node = g_list_next(node);
			continue;
		}

		BT_DBG("Address match \n");

		BT_DBG("Connection type = %x\n", connected_device->type);

		switch (type) {
		case BT_AUDIO_A2DP:
			if (connected_device->type & BT_AUDIO_A2DP)
				connected_device->type &= ~(BT_AUDIO_A2DP);
			break;
		case BT_AUDIO_HSP:
			if (connected_device->type & BT_AUDIO_HSP)
				connected_device->type &= ~(BT_AUDIO_HSP);
			break;
		case BT_AUDIO_ALL:
			if (connected_device->type & BT_AUDIO_ALL)
				connected_device->type &= ~(BT_AUDIO_ALL);
			break;
		case BT_AVRCP_TARGET:
			if (connected_device->type & BT_AVRCP_TARGET)
				connected_device->type &= ~(BT_AVRCP_TARGET);
			break;
		case BT_AVRCP:
			if (connected_device->type & BT_AVRCP)
				connected_device->type &= ~(BT_AVRCP);
			break;
		case BT_AUDIO_A2DP_SOURCE:
			if (connected_device->type & BT_AUDIO_A2DP_SOURCE)
				connected_device->type &= ~(BT_AUDIO_A2DP_SOURCE);
			break;
		default:
			break;
		}

		BT_DBG("Connection type = %x\n", connected_device->type);

		if (connected_device->type == 0x00) {
			BT_INFO("Device will be completely removed from connected list as the only profile connected or connecting got disconnected");
			g_connected_list = g_list_remove(g_connected_list, connected_device);
			g_free(connected_device);
		} else {
			connected_device->device_state = BT_STATE_CONNECTED;
		}

		node = g_list_next(node);
	}

	/*
	 * If Audio role selection request is pending (pending_audio_role > 0) then process pending
	 * audio role select request on BT_AUDIO_A2DP/BT_AUDIO_A2DP_SOURCE device dicaonnection.
	 */
	if (pending_audio_role >= 0 &&
		(type == BT_AUDIO_A2DP || type == BT_AUDIO_A2DP_SOURCE))
		g_idle_add(__handle_pending_audio_select_role, NULL);
}

static void __handle_pending_requests(int result, int service_function, void *user_data, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info;

	BT_DBG("+");

	/* Get method invocation context */
	for (l = _bt_get_invocation_list(); l != NULL; ) {
		req_info = l->data;
		l = g_slist_next(l);

		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		/* Create out param */
		out_param = g_array_new(FALSE, FALSE, sizeof(gchar));

		switch (service_function) {
		case BT_AUDIO_SELECT_ROLE:
			_bt_service_method_return(req_info->context, out_param, result);
			g_array_free(out_param, TRUE);
			_bt_free_info_from_invocation_list(req_info);
			return;
		default:
			BT_ERR("Unknown service function[%d]", service_function);
			return;
		}
	}

	BT_DBG("-");
}

static gboolean __handle_pending_audio_select_role(gpointer data)
{
	int ret;
	oal_status_t status;

	BT_DBG("+");
	if (BLUETOOTH_A2DP_SOURCE == pending_audio_role) {
		/* Disable A2DP Sink role */
		status = a2dp_sink_disable();
		if (status != OAL_STATUS_SUCCESS) {
			BT_ERR("Failed to Disable Bluetooth A2DP Sink Profile, status: %d", status);
			ret = BLUETOOTH_ERROR_INTERNAL;
			goto done;
		}

		/* Disable AVRCP CT role */
		status = avrcp_ct_disable();
		if (status != OAL_STATUS_SUCCESS) {
			BT_ERR("Failed to Disable Bluetooth AVRCP Controller Profile, status: %d", status);
			ret = BLUETOOTH_ERROR_INTERNAL;
			goto done;
		}

		/* Enable A2DP Source role */
		ret = _bt_audio_initialize(BT_A2DP_SOURCE_MODULE);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Failed to enable Bluetooth A2DP Source Profile");
			goto done;
		}

		/* Enable AVRCP TG role */
		ret = _bt_audio_initialize(BT_AVRCP_MODULE);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Failed to enable Bluetooth AVRCP Target Profile");
			goto done;
		}
	} else {
		/* Disable A2DP Source role */
		status = audio_disable();
		if (OAL_STATUS_SUCCESS != status) {
			BT_ERR("Failed to disable Bluetooth A2DP Source Profile, status: %d", status);
			ret = BLUETOOTH_ERROR_INTERNAL;
			goto done;
		}

		/* Disable AVRCP TG role */
		ret = _bt_service_avrcp_disable();
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Failed to Disable Bluetooth AVRCP Target Profile");
			goto done;
		}

		/* Enable A2DP Sink role */
		ret = _bt_audio_initialize(BT_A2DP_SINK_MODULE);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Failed to enable Bluetooth A2DP Sink Profile");
			goto done;
		}

		/* Enable AVRCP CT role */
		ret = _bt_audio_initialize(BT_AVRCP_CTRL_MODULE);
		if (ret != BLUETOOTH_ERROR_NONE) {
			BT_ERR("Failed to enable Bluetooth AVRCP Controller Profile");
			goto done;
		}
	}

done:
	if (ret == BLUETOOTH_ERROR_NONE)
		_bt_audio_set_current_role(pending_audio_role);

	__handle_pending_requests(ret, BT_AUDIO_SELECT_ROLE, NULL, 0);
	pending_audio_role = -1;
	BT_DBG("-");
	return FALSE;
}

int _bt_audio_select_role(bluetooth_audio_role_t role)
{
	int result = BLUETOOTH_ERROR_NONE;
	char address[BT_ADDRESS_STRING_SIZE];
	bluetooth_device_address_t dev_addr;

	retv_if(BLUETOOTH_A2DP_SINK < role, BLUETOOTH_ERROR_INVALID_PARAM);
	retv_if(pending_audio_role >= 0, BLUETOOTH_ERROR_IN_PROGRESS);

	BT_DBG("+");

	if (curr_audio_role == role) {
		BT_ERR("Desired audio role already enabled, return");
		return BLUETOOTH_ERROR_ALREADY_INITIALIZED;
	}

	pending_audio_role = role;

	if (BLUETOOTH_A2DP_SINK == role) {
		if (_bt_is_headset_type_connected(BT_AUDIO_A2DP, address)) {
			_bt_convert_addr_string_to_type(dev_addr.addr, address);
			result = _bt_audio_disconnect(BT_AUDIO_A2DP, &dev_addr);
			if (result != BLUETOOTH_ERROR_NONE) {
				BT_ERR("_bt_audio_disconnect(BT_AUDIO_A2DP, %s) Failed", address);
				result = BLUETOOTH_ERROR_INTERNAL;
			}
		} else {
			BT_INFO("No BT_AUDIO_A2DP device is connected, proceed with role switch");
			g_idle_add(__handle_pending_audio_select_role, NULL);
			result = BLUETOOTH_ERROR_NONE;
		}
	} else {
		if (_bt_is_headset_type_connected(BT_AUDIO_A2DP_SOURCE, address)) {
			_bt_convert_addr_string_to_type(dev_addr.addr, address);
			result = _bt_audio_disconnect(BT_AUDIO_A2DP_SOURCE, &dev_addr);
			if (result != BLUETOOTH_ERROR_NONE) {
				BT_ERR("_bt_audio_disconnect(BT_AUDIO_A2DP_SOURCE, %s) Failed", address);
				result = BLUETOOTH_ERROR_INTERNAL;
			}
		} else {
			BT_INFO("No BT_AUDIO_A2DP_SOURCE device is connected, proceed with role switch");
			g_idle_add(__handle_pending_audio_select_role, NULL);
			result = BLUETOOTH_ERROR_NONE;
		}
	}

	BT_DBG("-");
	return result;
}

void _bt_audio_set_current_role(bluetooth_audio_role_t role)
{
	BT_INFO("role: %s", (role == BLUETOOTH_A2DP_SINK) ? "AUDIO_SINK" : "AUDIO_SOURCE");
	curr_audio_role = role;
}
