/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Anupam Roy <anupam.r@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <vconf.h>

#include "oal-hardware.h"
#include "oal-manager.h"
#include "oal-audio-src.h"
#include "oal-hfp.h"
#include "oal-device-mgr.h"

#include <bt-service-a2dp-src.h>
#include <bt-service-hf.h>
#include <bt-service-event.h>

static void __bt_hf_handle_audio_disconnection_state(bt_address_t *address)
{
	BT_DBG("+");
	GVariant *param;
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int result = BLUETOOTH_ERROR_NONE;
	bt_headset_wait_t *wait_list;
	bluetooth_device_address_t device_address;
	ret_if(NULL == address);

	_bt_convert_addr_type_to_string(addr, address->addr);

	/* Send HF(AG Role) disconnected event to Application */
	param = g_variant_new("(is)", result, addr);
	_bt_send_event(BT_HEADSET_EVENT, BLUETOOTH_EVENT_AG_DISCONNECTED, param);

	/*Handle Platform logics */
	/* Remove data from the connected list */
	_bt_remove_headset_from_list(BT_AUDIO_HSP, addr);

	wait_list = _bt_get_audio_wait_data();
	if (wait_list == NULL)
		return;
	_bt_set_audio_wait_data_flag(TRUE);
	_bt_convert_addr_string_to_type(device_address.addr,
			wait_list->address);
	BT_INFO("Trigger connect for the waiting device [%s], audio type [%d]",
			wait_list->address, wait_list->type);
	_bt_audio_connect(wait_list->type, &device_address);
	_bt_rel_wait_data();

	/*Handle Product logics */
	BT_DBG("-");
}

static void __bt_reply_hf_disconnection_pending_request(bt_address_t *address)
{
	BT_DBG("+");
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	bluetooth_device_address_t device_address;
	GArray *out_param;
	invocation_info_t *req_info;
	bt_headset_wait_t *wait_list = NULL;
	bt_pending_audio_conn_t* data = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	ret_if(NULL == address);

	memcpy(device_address.addr, address->addr, BLUETOOTH_ADDRESS_LENGTH);
	_bt_convert_addr_type_to_string(addr, address->addr);

	/* Remove data from the connected list */
	_bt_remove_headset_from_list(BT_AUDIO_HSP, addr);

	/* Find Async request information*/
	req_info = _bt_get_request_info_data(BT_AG_DISCONNECT, addr);
	if (NULL == req_info) {
		BT_INFO("AG DisConnect request not found or possibly already replied..");

		req_info = _bt_get_request_info_data(BT_AG_CONNECT, addr);
		if (req_info == NULL) {
			BT_INFO("AG Connect request also not found..");

			req_info = _bt_get_request_info_data(BT_AUDIO_DISCONNECT, addr);
			if (req_info == NULL) {
				BT_INFO("AUDIO Disconnect request also not found..");

				req_info = _bt_get_request_info_data(BT_AUDIO_CONNECT, addr);
				if (req_info == NULL) {
					BT_INFO("AUDIO Connect request also not found..");
					return;
				} else {
					BT_INFO("AG Audio All Connect request found..");
					result = BLUETOOTH_ERROR_INTERNAL;
				}
			} else {
				BT_INFO("Audio All DisConnect request found..");
			}
		} else {
			BT_INFO("AG Connect request found..");
			result = BLUETOOTH_ERROR_INTERNAL;
		}
	} else {
		BT_INFO("AG DisConnect request found..");
	}

	/* Cleanup device service search info if Aysnc request was Audio Connect All */
	if (req_info->service_function == BT_AUDIO_CONNECT) {
		/* Check if pending connect was present */
		data = _bt_get_pending_audio_conn_info(addr);
		if (data) {
			BT_INFO("HF Connection failed during Audio All connect, & pending connect present");
			_bt_cleanup_audio_pending_conn_info_and_reply_pending_req(data, BLUETOOTH_ERROR_INTERNAL);
			goto try_waiting_device;
		} else {
			BT_INFO("Abnormal:HF Connection failed during Audio All connect, but pending connect not present");
		}
	}

	/* In any of the above cases, reply DBUS context */
	out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
	g_array_append_vals(out_param, addr, BT_ADDRESS_STRING_SIZE);
	_bt_service_method_return(req_info->context,
			out_param, result);
	g_array_free(out_param, TRUE);
	g_free(req_info->user_data);
	_bt_free_info_from_invocation_list(req_info);

try_waiting_device:
	wait_list = _bt_get_audio_wait_data();
	if (wait_list == NULL)
		return;
	_bt_convert_addr_string_to_type(device_address.addr,
			wait_list->address);
	BT_INFO("Trigger connect for the waiting device [%s], audio type [%d]",
			wait_list->address, wait_list->type);
	_bt_audio_connect(wait_list->type, &device_address);
	_bt_rel_wait_data();
}

static void __bt_hf_handle_audio_connection_state(bt_address_t *bd_addr)
{
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	bt_headset_wait_t *wait_list = NULL;
	GArray *out_param;
	unsigned int result = BLUETOOTH_ERROR_NONE;
	invocation_info_t *req_info;
	GVariant *param;
	BT_INFO("+");

	_bt_convert_addr_type_to_string(address, bd_addr->addr);
	BT_INFO("HF(AG Role) Connected for address [%s]", address);

	/* Send HF(AG Role) connected event to Application. Even though it is AG Audio
	   Connected, still send AG Connected event as per logic implemented in Tizen  */
	param = g_variant_new("(is)", result, address);
	_bt_send_event(BT_HEADSET_EVENT, BLUETOOTH_EVENT_AG_CONNECTED, param);

	/* Add data to the connected list */
	_bt_add_headset_to_list(BT_AUDIO_HSP,
			BT_STATE_CONNECTED, address);

	wait_list = _bt_get_audio_wait_data();
	if (wait_list != NULL &&
			(g_strcmp0(wait_list->address, address) == 0)) {
		BT_INFO("Same Device [%s] is under wait, connection type [%d], going to delete it",
				wait_list->address, wait_list->type);
		_bt_rel_wait_data();
	} else {
		if (wait_list != NULL)
			BT_INFO("Some Device [%s] is under wait, connection type [%d], ", wait_list->address, wait_list->type);
	}

	/* DBUS Context return */
	req_info = _bt_get_request_info_data(BT_AUDIO_CONNECT, address);

	if (NULL != req_info) {
		BT_INFO("Audio All Connect request found");
		goto check_pending_a2dp_connect;
	} else {
		BT_INFO("Audio All Connect Request not found..");
		req_info = _bt_get_request_info_data(BT_AG_CONNECT, address);

		if (req_info) {
			BT_INFO("AG Connect Request found..");
			goto dbus_return;
		} else {
			BT_ERR("AG Connect Request also not found..Abnormal Case!!!");
			return;
		}
	}

check_pending_a2dp_connect:
	/* Check pending A2DP connection */
	BT_INFO("Check A2DP pending connect");
	if (_bt_audio_check_pending_connection(address) == TRUE) {
		BT_INFO("Pending A2DP connect is present..dont return DBUS for AUDIO_CONNECT_ALL");
		return;
	}
	BT_INFO("Pending A2DP connect is Not present..return DBUS for AUDIO_CONNECT_ALL");
dbus_return:
	out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
	g_array_append_vals(out_param, address, BT_ADDRESS_STRING_SIZE);
	_bt_service_method_return(req_info->context,
			out_param, BLUETOOTH_ERROR_NONE);
	g_array_free(out_param, TRUE);
	g_free(req_info->user_data);
	_bt_free_info_from_invocation_list(req_info);
}

static void __bt_reply_hf_connection_pending_request(bt_address_t *address)
{
	BT_DBG("+");
}

/* This event handler process events for HF (AG role) */
void _bt_hf_event_handler(int oal_event, gpointer event_data)
{
	BT_INFO("+");
	bt_address_t * bt_addr = event_data;

	switch (oal_event) {
	case OAL_EVENT_HFP_DISCONNECTED:
		/* Reply to async request for HF connect or disconnect request, if any */
		BT_INFO("HF Profile disconnected, reply pending DBUS request and check waiting device");
		__bt_reply_hf_disconnection_pending_request(bt_addr);
		break;
	case OAL_EVENT_HFP_AUDIO_DISCONNECTED:
		/* Reply to async request for HF connect or disconnect request, if any */
		BT_INFO("HF SCO disconnected, reply pending request if any and send event");
		__bt_hf_handle_audio_disconnection_state(bt_addr);
		break;
	case OAL_EVENT_HFP_CONNECTED:
		BT_INFO("HFP Profile connected, Event & DBUS context will be handled at finalizing SCO connect..");
		__bt_reply_hf_connection_pending_request(bt_addr);
		break;
	case OAL_EVENT_HFP_AUDIO_CONNECTED:
		BT_INFO("HFP SCO conencted, handle the event..");
		__bt_hf_handle_audio_connection_state(bt_addr);
		break;
	case OAL_EVENT_HFP_CONNECTING:
		BT_INFO("HFP Profile connection successful, wait for Audio connect..");
		break;
	case OAL_EVENT_HFP_DISCONNECTING:
		BT_INFO("HFP Connecting or Disconnecting..No need to send event to app");
		break;
	case OAL_EVENT_HFP_AUDIO_CONNECTING:
	case OAL_EVENT_HFP_AUDIO_DISCONNECTING:
		BT_INFO("HFP SCO conencting or disconnecting, Not handled..");
		break;
	default:
		break;
	}
}

int _bt_connect_remote_hfp(bluetooth_device_address_t *device_address)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int result = BLUETOOTH_ERROR_NONE;
	BT_INFO("+");

	status = hfp_connect((bt_address_t *)device_address);
	if (status != OAL_STATUS_SUCCESS) {
		BT_ERR("Connection could not be established, err: [%d]", status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}
	return result;
}

int _bt_disconnect_remote_hfp(bluetooth_device_address_t *device_address)
{
	oal_status_t status = OAL_STATUS_SUCCESS;
	int result = BLUETOOTH_ERROR_NONE;
	BT_INFO("+");

	status = hfp_disconnect((bt_address_t *)device_address);
	if (status != OAL_STATUS_SUCCESS) {
		BT_ERR("DisConnection err: [%d]", status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}
	return result;
}
