/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#ifdef TIZEN_FEATURE_BT_DPM

#include <glib.h>
#include <gio/gio.h>
#include <dlog.h>
#include <string.h>
#include <bundle_internal.h>
#include <vconf.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-service-common.h"
#include "bt-service-core-adapter.h"
#include "bt-service-dpm.h"

static dpm_policy_t policy_table[DPM_POLICY_END] = {
	[DPM_POLICY_ALLOW_BLUETOOTH] = { {DPM_BT_ERROR} },
	[DPM_POLICY_BLUETOOTH_DEVICE_RESTRICTION] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_UUID_RESTRICTION] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST] = { { } },
	[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST] = { { } },
	[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST] = { { } },
	[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST] = { { } },
	[DPM_POLICY_ALLOW_BLUETOOTH_OUTGOING_CALL] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_PAIRING_STATE] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_DESKTOP_CONNECTIVITY_STATE] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_DISCOVERABLE_STATE] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_LIMITED_DISCOVERABLE_STATE] = { {DPM_STATUS_ERROR} },
	[DPM_POLICY_BLUETOOTH_DATA_TRANSFER_STATE] = { {DPM_STATUS_ERROR} },
};


/**
 * @brief DPM profile state
 * @see
 */
static dpm_profile_state_t dpm_profile_state[DPM_PROFILE_NONE]  = {
	[DPM_POLICY_BLUETOOTH_A2DP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_AVRCP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_BPP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_DUN_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_FTP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_HFP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_HSP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_PBAP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_SAP_PROFILE_STATE] = {DPM_STATUS_ERROR},
	[DPM_POLICY_BLUETOOTH_SPP_PROFILE_STATE] = {DPM_STATUS_ERROR},
};

int _bt_dpm_set_allow_bluetooth_mode(dpm_bt_allow_t value)
{
	BT_INFO("_bt_dpm_set_allow_bluetooth_mode");

#if 0
	if (value == DPM_BT_ALLOWED && value == DPM_BT_HANDSFREE_ONLY) {
		/* Update Bluetooth DPM Status to notify other modules */
		if (vconf_set_int(VCONFKEY_BT_DPM_STATUS, value) != 0)
			BT_ERR("Set vconf failed\n");
			return DPM_RESULT_FAIL;
	} else {
		/* Update Bluetooth DPM Status to notify other modules */
		if (vconf_set_int(VCONFKEY_BT_DPM_STATUS, VCONFKEY_BT_DPM_STATUS_RESTRICTED) != 0)
			BT_ERR("Set vconf failed\n");
			return DPM_RESULT_FAIL;
	}
#endif
	policy_table[DPM_POLICY_ALLOW_BLUETOOTH].value	= value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_allow_bluetooth_mode(int *value)
{
	BT_INFO("_bt_dpm_get_allow_bluetooth_mode");

	*value = policy_table[DPM_POLICY_ALLOW_BLUETOOTH].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_activate_bluetooth_device_restriction(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;
	BT_INFO("_bt_dpm_activate_bluetooth_device_restriction");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	policy_table[DPM_POLICY_BLUETOOTH_DEVICE_RESTRICTION].value  = value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_is_bluetooth_device_restriction_active(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_is_bluetooth_device_restriction_active");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_DEVICE_RESTRICTION].value;

	return BLUETOOTH_ERROR_NONE;
}

dpm_result_t _bt_dpm_activate_bluetoooth_uuid_restriction(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_activate_bluetooth_device_restriction");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	policy_table[DPM_POLICY_BLUETOOTH_UUID_RESTRICTION].value  = value;

	return BLUETOOTH_ERROR_NONE;
}

dpm_status_t _bt_dpm_is_bluetooth_uuid_restriction_active(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_is_bluetooth_uuid_restriction_active");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_UUID_RESTRICTION].value;

	return BLUETOOTH_ERROR_NONE;
}

dpm_result_t _bt_dpm_add_bluetooth_devices_to_blacklist(bluetooth_device_address_t *bd_addr)
{
	char device_address[BT_ADDRESS_STRING_SIZE] = { 0 };
	char *dev_addr = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_add_bluetooth_devices_to_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	BT_CHECK_PARAMETER(bd_addr, return);

	_bt_convert_addr_type_to_string(device_address,
			(unsigned char *)bd_addr->addr);

	dev_addr = g_strdup(device_address);
	if (!dev_addr)
		return BLUETOOTH_ERROR_MEMORY_ALLOCATION;

	policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list  = g_slist_append(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list, dev_addr);

	/* Disconnect if connected blacklist device is existing */
	// TODO: need to implement disconnect logic

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_devices_from_blacklist(GArray **out_param1)
{
	int ret = BLUETOOTH_ERROR_INTERNAL;
	int allow_bt = DPM_BT_ERROR;
	bt_dpm_device_list_t device_list = { 0, };
	GSList *list = policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list;
	int i = 0;

	BT_INFO("_bt_dpm_get_bluetooth_devices_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	if (list) {
		ret = BLUETOOTH_ERROR_NONE;
		for (; list; list = list->next, i++) {
			memset(device_list.addresses[i].addr, 0, BLUETOOTH_ADDRESS_LENGTH);
			_bt_convert_addr_string_to_type(device_list.addresses[i].addr, list->data);
		}
		device_list.count = g_slist_length(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list);
		g_array_append_vals(*out_param1, &device_list, sizeof(bt_dpm_device_list_t));
	} else {
		ret = BLUETOOTH_ERROR_NONE;
		device_list.count = 0;
		g_array_append_vals(*out_param1, &device_list, sizeof(bt_dpm_device_list_t));
	}
	return ret;
}

int _bt_dpm_add_bluetooth_devices_to_whitelist(bluetooth_device_address_t *bd_addr)
{
	char device_address[BT_ADDRESS_STRING_SIZE] = { 0 };
	char *dev_addr = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_add_bluetooth_devices_to_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	BT_CHECK_PARAMETER(bd_addr, return);

	_bt_convert_addr_type_to_string(device_address,
			(unsigned char *)bd_addr->addr);

	dev_addr = g_strdup(device_address);
	if (!dev_addr)
		return BLUETOOTH_ERROR_MEMORY_ALLOCATION;
	policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list  = g_slist_append(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list, dev_addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_devices_from_whitelist(GArray **out_param1)
{
	dpm_result_t ret = BLUETOOTH_ERROR_INTERNAL;
	int allow_bt = DPM_BT_ERROR;
	bt_dpm_device_list_t device_list = { 0, };
	GSList *list = policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list;
	int i = 0;

	BT_INFO("_bt_dpm_get_bluetooth_devices_from_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	if (list) {
		ret = BLUETOOTH_ERROR_NONE;
		for (; list; list = list->next, i++) {
			memset(device_list.addresses[i].addr, 0, BLUETOOTH_ADDRESS_LENGTH);
			_bt_convert_addr_string_to_type(device_list.addresses[i].addr, list->data);

		}
		device_list.count = g_slist_length(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list);
		g_array_append_vals(*out_param1, &device_list, sizeof(bt_dpm_device_list_t));
	} else {
		ret = BLUETOOTH_ERROR_NONE;
		device_list.count = 0;
		g_array_append_vals(*out_param1, &device_list, sizeof(bt_dpm_device_list_t));
	}
	return ret;
}

int _bt_dpm_add_bluetooth_uuids_to_blacklist(const char *uuid)
{
	char *l_uuid;
	int allow_bt = DPM_BT_ERROR;
	GArray *addr_list = NULL;
	int i = 0;

	BT_INFO("_bt_dpm_add_bluetooth_uuids_to_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	l_uuid = g_strdup(uuid);
	if (!l_uuid)
		return BLUETOOTH_ERROR_MEMORY_ALLOCATION;

	policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list  = g_slist_append(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list, l_uuid);

	/* Disconnect if connected blacklist uuid is existing */
	addr_list = g_array_new(FALSE, FALSE, sizeof(gchar));
	_bt_get_profile_connected_devices((char *)uuid, &addr_list);
	for (i = 0; i < (addr_list->len / sizeof(bluetooth_device_address_t)); i++) {
		char address[BT_ADDRESS_STRING_SIZE];

		bluetooth_device_address_t *addr = &g_array_index(addr_list, bluetooth_device_address_t, i);
		_bt_convert_addr_type_to_string(address, addr->addr);
		BT_INFO("device[%s] is in blacklist uuid, will be disconnected", address);
		// TODO: need to implement disconnect logic
	}
	g_array_free(addr_list, TRUE);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_uuids_from_blacklist(GArray **out_param1)
{
	int ret = BLUETOOTH_ERROR_INTERNAL;
	int allow_bt = DPM_BT_ERROR;
	bt_dpm_uuids_list_t uuids_list = {0, { {0}, } };
	GSList *list = policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list;
	int i = 0;

	BT_INFO("_bt_dpm_get_bluetooth_uuids_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	if (list) {
		ret = BLUETOOTH_ERROR_NONE;
		uuids_list.count = g_slist_length(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list);
		for (; list; list = list->next, i++) {
			memset(uuids_list.uuids[i], 0, BLUETOOTH_UUID_STRING_MAX);
			g_strlcpy(uuids_list.uuids[i], list->data,
				BLUETOOTH_UUID_STRING_MAX);
		}
		g_array_append_vals(*out_param1, &uuids_list, sizeof(bt_dpm_uuids_list_t));
	} else {
		ret = BLUETOOTH_ERROR_NONE;
		uuids_list.count = 0;
		g_array_append_vals(*out_param1, &uuids_list, sizeof(bt_dpm_uuids_list_t));
	}

	return ret;
}

int _bt_dpm_add_bluetooth_uuids_to_whitelist(const char *uuid)
{
	char *l_uuid;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_add_bluetooth_uuids_to_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	l_uuid = g_strdup(uuid);
	if (!l_uuid)
		return BLUETOOTH_ERROR_MEMORY_ALLOCATION;

	policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list  = g_slist_append(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list, l_uuid);
	return BLUETOOTH_ERROR_NONE;
}


int _bt_dpm_get_bluetooth_uuids_from_whitelist(GArray **out_param1)
{
	int ret = BLUETOOTH_ERROR_INTERNAL;
	int allow_bt = DPM_BT_ERROR;
	bt_dpm_uuids_list_t uuids_list = {0, { {0}, } };
	GSList *list = policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list;
	int i = 0;

	BT_INFO("_bt_dpm_get_bluetooth_uuids_from_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	if (list) {
		ret = BLUETOOTH_ERROR_NONE;
		uuids_list.count = g_slist_length(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list);
		for (; list; list = list->next, i++) {
			memset(uuids_list.uuids[i], 0, BLUETOOTH_UUID_STRING_MAX);
			g_strlcpy(uuids_list.uuids[i], list->data,
				BLUETOOTH_UUID_STRING_MAX);
		}
		g_array_append_vals(*out_param1, &uuids_list, sizeof(bt_dpm_uuids_list_t));
	} else {
		ret = BLUETOOTH_ERROR_NONE;
		uuids_list.count = 0;
		g_array_append_vals(*out_param1, &uuids_list, sizeof(bt_dpm_uuids_list_t));
	}

	return ret;

}

int _bt_dpm_set_allow_bluetooth_outgoing_call(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_activate_bluetooth_device_restriction");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	policy_table[DPM_POLICY_ALLOW_BLUETOOTH_OUTGOING_CALL].value  = value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_allow_bluetooth_outgoing_call(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_allow_bluetooth_outgoing_call");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_ALLOW_BLUETOOTH_OUTGOING_CALL].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_clear_bluetooth_devices_from_blacklist(void)
{
	GSList *l = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_clear_bluetooth_devices_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	for (l = policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list; l; l = g_slist_next(l)) {
		char *address = l->data;
		if (address) {
			policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list, address);
			g_free(address);
		}
	}
	g_slist_free(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list);
	policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list = NULL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_clear_bluetooth_devices_from_whitelist(void)
{
	GSList *l = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_clear_bluetooth_devices_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	for (l = policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list; l; l = g_slist_next(l)) {
		char *address = l->data;
		if (address) {
			policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list, address);
			g_free(address);
		}
	}
	g_slist_free(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list);
	policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list = NULL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_clear_bluetooth_uuids_from_blacklist(void)
{
	GSList *l = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_clear_bluetooth_devices_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	for (l = policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list; l; l = g_slist_next(l)) {
		char *l_uuid = l->data;
		if (l_uuid) {
			policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list, l_uuid);
			g_free(l_uuid);
		}
	}
	g_slist_free(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list);
	policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list = NULL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_clear_bluetooth_uuids_from_whitelist(void)
{
	GSList *l = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_clear_bluetooth_uuids_from_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	for (l = policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list; l; l = g_slist_next(l)) {
		char *l_uuid = l->data;
		if (l_uuid) {
			policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list, l_uuid);
			g_free(l_uuid);
		}
	}
	g_slist_free(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list);
	policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list = NULL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_set_bluetooth_pairing_state(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_set_bluetooth_pairing_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	policy_table[DPM_POLICY_BLUETOOTH_PAIRING_STATE].value = value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_pairing_state(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_bluetooth_pairing_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_PAIRING_STATE].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_set_bluetooth_profile_state(dpm_profile_t profile, dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_set_bluetooth_profile_state");

	if (profile >= DPM_PROFILE_NONE)
		return BLUETOOTH_ERROR_INTERNAL;

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	dpm_profile_state[profile].value = value;

	/* In case of restriction, disconnect if connected profile is existing */
	if (value == DPM_RESTRICTED) {
		char *uuid = NULL;
		GArray *addr_list = NULL;
		int i = 0;

		switch (profile) {
		case DPM_POLICY_BLUETOOTH_A2DP_PROFILE_STATE:
			uuid = BT_A2DP_UUID;
			break;
		case DPM_POLICY_BLUETOOTH_AVRCP_PROFILE_STATE:
			uuid = BT_AVRCP_TARGET_UUID;
			break;
		case DPM_POLICY_BLUETOOTH_BPP_PROFILE_STATE:
			uuid = "00001118-0000-1000-8000-00805f9b34fb"; // TODO: need to add definition and below also.
			break;
		case DPM_POLICY_BLUETOOTH_DUN_PROFILE_STATE:
			uuid = "00001103-0000-1000-8000-00805f9b34fb";
			break;
		case DPM_POLICY_BLUETOOTH_FTP_PROFILE_STATE:
			uuid = BT_FTP_UUID;
			break;
		case DPM_POLICY_BLUETOOTH_HFP_PROFILE_STATE:
			uuid = "0000111e-0000-1000-8000-00805f9b34fb";
			break;
		case DPM_POLICY_BLUETOOTH_HSP_PROFILE_STATE:
			uuid = "00001108-0000-1000-8000-00805f9b34fb";
			break;
		case DPM_POLICY_BLUETOOTH_PBAP_PROFILE_STATE:
			uuid = BT_OBEX_PSE_UUID;
			break;
		case DPM_POLICY_BLUETOOTH_SAP_PROFILE_STATE:
			uuid = "0000112d-0000-1000-8000-00805f9b34fb";
			break;
		case DPM_POLICY_BLUETOOTH_SPP_PROFILE_STATE:
			uuid = BT_SPP_UUID;
			break;
		default:
			BT_ERR("Unknown profile %d", profile);
			return BLUETOOTH_ERROR_INTERNAL;
		}

		addr_list = g_array_new(FALSE, FALSE, sizeof(gchar));
		_bt_get_profile_connected_devices(uuid, &addr_list);
		for (i = 0; i < (addr_list->len / sizeof(bluetooth_device_address_t)); i++) {
			char address[BT_ADDRESS_STRING_SIZE];

			bluetooth_device_address_t *addr = &g_array_index(addr_list, bluetooth_device_address_t, i);
			_bt_convert_addr_type_to_string(address, addr->addr);
			BT_INFO("device[%s] is in blacklist uuid, will be disconnected", address);
			// TODO: need to implement disconnect logic
		}
		g_array_free(addr_list, TRUE);
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_profile_state(dpm_profile_t profile, int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_bluetooth_profile_state");

	if (profile >= DPM_PROFILE_NONE)
		return BLUETOOTH_ERROR_INTERNAL;

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = dpm_profile_state[profile].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_set_bluetooth_desktop_connectivity_state(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_set_bluetooth_desktop_connectivity_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	policy_table[DPM_POLICY_BLUETOOTH_DESKTOP_CONNECTIVITY_STATE].value = value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_desktop_connectivity_state(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_bluetooth_desktop_connectivity_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_DESKTOP_CONNECTIVITY_STATE].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_set_bluetooth_discoverable_state(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_set_bluetooth_discoverable_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	if (value == DPM_RESTRICTED) {
		/* Since Discoverable mode is restricted, stop the ongoing discoverable mode */
		_bt_set_discoverable_mode(BLUETOOTH_DISCOVERABLE_MODE_CONNECTABLE, 0);
	}

	policy_table[DPM_POLICY_BLUETOOTH_DISCOVERABLE_STATE].value = value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_discoverable_state(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_bluetooth_discoverable_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_DISCOVERABLE_STATE].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_set_bluetooth_limited_discoverable_state(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_set_bluetooth_limited_discoverable_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	if (value == DPM_RESTRICTED) {
		/* Since Discoverable mode is restricted, stop the ongoing discoverable mode */
		_bt_set_discoverable_mode(BLUETOOTH_DISCOVERABLE_MODE_CONNECTABLE, 0);
	}

	policy_table[DPM_POLICY_BLUETOOTH_LIMITED_DISCOVERABLE_STATE].value = value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_bluetooth_limited_discoverable_state(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_bluetooth_limited_discoverable_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_LIMITED_DISCOVERABLE_STATE].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_set_bluetooth_data_transfer_state(dpm_status_t value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_set_bluetooth_data_transfer_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	policy_table[DPM_POLICY_BLUETOOTH_DATA_TRANSFER_STATE].value = value;
	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_get_allow_bluetooth_data_transfer_state(int *value)
{
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_get_allow_bluetooth_data_transfer_state");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	*value = policy_table[DPM_POLICY_BLUETOOTH_DATA_TRANSFER_STATE].value;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_remove_bluetooth_devices_from_whitelist(bluetooth_device_address_t *device_address)
{
	GSList *l = NULL;
	char bd_addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_remove_bluetooth_devices_from_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	_bt_convert_addr_type_to_string(bd_addr,
			(unsigned char *)device_address->addr);

	for (l = policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list; l; l = g_slist_next(l)) {
		char *l_device = l->data;
		if (l_device && g_strcmp0(l_device, bd_addr)) {
			policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list, l_device);
			g_free(l_device);
		}
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_remove_bluetooth_devices_from_blacklist(bluetooth_device_address_t *device_address)
{
	GSList *l = NULL;
	char bd_addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_remove_bluetooth_devices_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	_bt_convert_addr_type_to_string(bd_addr,
			(unsigned char *)device_address->addr);

	for (l = policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list; l; l = g_slist_next(l)) {
		char *l_device = l->data;
		if (l_device && g_strcmp0(l_device, bd_addr)) {
			policy_table[DPM_POLICY_BLUETOOTH_DEVICES_BLACKLIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_DEVICES_WHITELIST].list, l_device);
			g_free(l_device);
		}
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_remove_bluetooth_uuids_from_whitelist(const char *uuids)
{
	GSList *l = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_remove_bluetooth_uuids_from_whitelist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	for (l = policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list; l; l = g_slist_next(l)) {
		char *l_uuid = l->data;
		if (l_uuid && g_strcmp0(l_uuid, uuids)) {
			policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_WHITELIST].list, l_uuid);
			g_free(l_uuid);
		}
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_remove_bluetooth_uuids_from_blacklist(const char *uuids)
{
	GSList *l = NULL;
	int allow_bt = DPM_BT_ERROR;

	BT_INFO("_bt_dpm_remove_bluetooth_uuids_from_blacklist");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	for (l = policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list; l; l = g_slist_next(l)) {
		char *l_uuid = l->data;
		if (l_uuid && g_strcmp0(l_uuid, uuids)) {
			policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list = g_slist_remove(policy_table[DPM_POLICY_BLUETOOTH_UUIDS_BLACKLIST].list, l_uuid);
			g_free(l_uuid);
		}
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_dpm_clear_bluetooth_uuids_from_list(void)
{
	int allow_bt = DPM_BT_ERROR;
	int err = BLUETOOTH_ERROR_INTERNAL;

	BT_INFO("_bt_dpm_clear_bluetooth_uuids_from_list");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	err = _bt_dpm_clear_bluetooth_uuids_from_blacklist();
	if (!err)
		err = _bt_dpm_clear_bluetooth_uuids_from_blacklist();

	return err;
}

int _bt_dpm_clear_bluetooth_devices_from_list(void)
{
	int allow_bt = DPM_BT_ERROR;
	int err = BLUETOOTH_ERROR_INTERNAL;

	BT_INFO("_bt_dpm_clear_bluetooth_devices_from_list");

	_bt_dpm_get_allow_bluetooth_mode(&allow_bt);
	if (allow_bt == DPM_BT_RESTRICTED)
		return BLUETOOTH_ERROR_PERMISSION_DEINED;

	err = _bt_dpm_clear_bluetooth_devices_from_blacklist();
	if (!err)
		err = _bt_dpm_clear_bluetooth_devices_from_blacklist();

	return err;
}
#endif /* #ifdef TIZEN_FEATURE_BT_DPM */
