/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2015 - 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Atul Kumar Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <string.h>
#include <dlog.h>
#include <vconf.h>
#include <vconf-internal-bt-keys.h>

#include <oal-event.h>

#include "bt-service-common.h"
#include "bt-service-event-receiver.h"

typedef struct {
	int type;
	void *data;
} bt_service_oal_event_data_t;

_bt_service_event_handler_callback adapter_cb;
_bt_service_event_handler_callback device_cb;
_bt_service_event_handler_callback hid_cb;
_bt_service_event_handler_callback socket_cb;
_bt_service_event_handler_callback audio_event_cb;
_bt_service_event_handler_callback a2dp_source_cb;
_bt_service_event_handler_callback a2dp_sink_cb;
_bt_service_event_handler_callback ag_cb;
_bt_service_event_handler_callback hdp_cb;
_bt_service_event_handler_callback avrcp_ctrl_cb;
_bt_service_event_handler_callback avrcp_cb;
_bt_service_event_handler_callback adapter_le_cb;
_bt_service_event_handler_callback adapter_gatt_cb;

void _bt_service_register_event_handler_callback(
		bt_service_module_t module, _bt_service_event_handler_callback cb)
{
	switch (module) {
	case BT_ADAPTER_MODULE:
		BT_INFO("Register BT_ADAPTER_MODULE Callback");
		adapter_cb = cb;
		break;
	case BT_DEVICE_MODULE:
		BT_INFO("Register BT_DEVICE_MODULE Callback");
		device_cb = cb;
		break;
	case BT_HID_MODULE:
		BT_INFO("Register BT_HID_MODULE Callback");
		hid_cb = cb;
		break;
	case BT_SOCKET_MODULE:
		BT_INFO("Register BT_SOCKET_MODULE Callback");
		socket_cb = cb;
		break;
	case BT_AUDIO_ALL_MODULE:
		BT_INFO("Register BT_AUDIO_ALL_MODULE Callback");
		audio_event_cb = cb;
		break;
	case BT_A2DP_SOURCE_MODULE:
		BT_INFO("Register BT_A2DP_SOURCE_MODULE Callback");
		a2dp_source_cb = cb;
		break;
	case BT_A2DP_SINK_MODULE:
		BT_INFO("Register BT_A2DP_SINK_MODULE Callback");
		a2dp_sink_cb = cb;
		break;
	case BT_AG_MODULE:
		BT_INFO("Register BT_AG_MODULE callback");
		ag_cb = cb;
		break;
	case BT_HEALTH_MODULE:
		BT_INFO("Register BT_HEALTH_MODULE callback");
		hdp_cb = cb;
		break;
	case BT_AVRCP_MODULE:
		BT_INFO("Register BT_AVRCP_MODULE callback");
		avrcp_cb = cb;
		break;
	case BT_AVRCP_CTRL_MODULE:
		BT_INFO("Register BT_AVRCP_CTRL_MODULE callback");
		avrcp_ctrl_cb = cb;
		break;
	case BT_ADAPTER_LE_MODULE:
		BT_INFO("Register BT_ADAPTER_LE_MODULE callback");
		adapter_le_cb = cb;
		break;
	case BT_GATT_MODULE:
		BT_INFO("Register BT_GATT_MODULE callback");
		adapter_gatt_cb = cb;
		break;
	default:
		BT_INFO("Unknown module");
	}
}

void _bt_service_unregister_event_handler_callback(bt_service_module_t module)
{
	switch (module) {
	case BT_ADAPTER_MODULE:
		BT_INFO("Un-Register BT_ADAPTER_MODULE Callback");
		adapter_cb = NULL;
		break;
	case BT_DEVICE_MODULE:
		BT_INFO("Un-Register BT_DEVICE_MODULE Callback");
		device_cb = NULL;
		break;
	case BT_HID_MODULE:
		BT_INFO("Un-Register BT_HID_MODULE Callback");
		hid_cb = NULL;
		break;
	case BT_SOCKET_MODULE:
		BT_INFO("Un-Register BT_SOCKET_MODULE Callback");
		socket_cb = NULL;
		break;
	case BT_A2DP_SINK_MODULE:
		BT_INFO("Un-Register BT_A2DP_SINK_MODULE Callback");
		a2dp_sink_cb = NULL;
		break;
	case BT_HEALTH_MODULE:
		BT_INFO("Un-Register BT_HEALTH_MODULE callback");
		hdp_cb = NULL;
		break;
	case BT_AVRCP_MODULE:
		BT_INFO("Unregister BT_AVRCP_MODULE callback");
		avrcp_cb = NULL;
		break;
	case BT_AVRCP_CTRL_MODULE:
		BT_INFO("Un-Register BT_AVRCP_CTRL_MODULE callback");
		avrcp_ctrl_cb = NULL;
		break;
	case BT_ADAPTER_LE_MODULE:
		BT_INFO("Un-Register BT_ADAPTER_MODULE callback");
		adapter_le_cb = NULL;
		break;
	case BT_GATT_MODULE:
		BT_INFO("Un-Register BT_GATT_MODULE callback");
		adapter_gatt_cb = NULL;
		break;
	default:
		BT_INFO("Unknown module");
	}
}

static gboolean __bt_handle_oal_events(gpointer data)
{
	bt_service_oal_event_data_t *oal_event = data;
	int event_type = oal_event->type;
	gpointer event_data = oal_event->data;

	switch (event_type) {
	case OAL_EVENT_ADAPTER_ENABLED:
	case OAL_EVENT_ADAPTER_DISABLED:
		if (audio_event_cb)
			audio_event_cb(event_type, event_data);
		if (adapter_cb)
			adapter_cb(event_type, event_data);
		break;
	case OAL_EVENT_OAL_INITIALISED_SUCCESS:
	case OAL_EVENT_OAL_INITIALISED_FAILED:
	case OAL_EVENT_ADAPTER_PROPERTY_ADDRESS:
	case OAL_EVENT_ADAPTER_PROPERTY_NAME:
	case OAL_EVENT_ADAPTER_PROPERTY_VERSION:
	case OAL_EVENT_ADAPTER_PROPERTY_SERVICES:
	case OAL_EVENT_ADAPTER_BONDED_DEVICE_LIST:
	case OAL_EVENT_ADAPTER_MODE_NON_CONNECTABLE:
	case OAL_EVENT_ADAPTER_MODE_CONNECTABLE:
	case OAL_EVENT_ADAPTER_MODE_DISCOVERABLE:
	case OAL_EVENT_ADAPTER_MODE_DISCOVERABLE_TIMEOUT:
	case OAL_EVENT_ADAPTER_INQUIRY_STARTED:
	case OAL_EVENT_ADAPTER_INQUIRY_FINISHED:
		if (adapter_cb)
			adapter_cb(event_type, event_data);
		break;
	case OAL_EVENT_ADAPTER_INQUIRY_RESULT_BREDR_ONLY:
	case OAL_EVENT_ADAPTER_INQUIRY_RESULT_BLE:
	case OAL_EVENT_DEVICE_BONDING_SUCCESS:
	case OAL_EVENT_DEVICE_BONDING_REMOVED:
	case OAL_EVENT_DEVICE_BONDING_FAILED:
	case OAL_EVENT_DEVICE_ACL_CONNECTED:
	case OAL_EVENT_DEVICE_ACL_DISCONNECTED:
	case OAL_EVENT_DEVICE_LE_CONNECTED:
	case OAL_EVENT_DEVICE_LE_DISCONNECTED:
	case OAL_EVENT_DEVICE_PIN_REQUEST:
	case OAL_EVENT_DEVICE_PASSKEY_ENTRY_REQUEST:
	case OAL_EVENT_DEVICE_PASSKEY_CONFIRMATION_REQUEST:
	case OAL_EVENT_DEVICE_PASSKEY_DISPLAY:
	case OAL_EVENT_DEVICE_SSP_CONSENT_REQUEST:
	case OAL_EVENT_DEVICE_SERVICES:
	case OAL_EVENT_DEVICE_AUTHORIZE_REQUEST:
	case OAL_EVENT_DEVICE_TRUSTED:
	case OAL_EVENT_DEVICE_UNTRUSTED:
	case OAL_EVENT_DEVICE_NAME:
	case OAL_EVENT_DEVICE_TRUSTED_PROFILES_CHANGED:
	case OAL_EVENT_RSSI_MONITORING_ENABLED:
	case OAL_EVENT_RSSI_MONITORING_DISABLED:
	case OAL_EVENT_RAW_RSSI_RECEIVED:
	case OAL_EVENT_RSSI_ALERT_RECEIVED:
		if (device_cb)
			device_cb(event_type, event_data);
		if (adapter_gatt_cb)
			adapter_gatt_cb(event_type, event_data);
		break;
	case OAL_EVENT_DEVICE_PROPERTIES:
		if (audio_event_cb) {
			BT_INFO("Device properties received, Audio callback registered, call it...");
			audio_event_cb(event_type, event_data);
		}
#if 0
		if (hid_cb)
			hid_cb(event_type, event_data);
#endif
		if (device_cb)
			device_cb(event_type, event_data);
		break;
	case OAL_EVENT_HID_CONNECTED:
	case OAL_EVENT_HID_DISCONNECTED:
		if (hid_cb)
			hid_cb(event_type, event_data);
		break;
	case OAL_EVENT_SOCKET_OUTGOING_CONNECTED:
	case OAL_EVENT_SOCKET_DISCONNECTED:
	case OAL_EVENT_SOCKET_AUTHORIZE_REQUEST:
		if (socket_cb)
			socket_cb(event_type, event_data);
		break;
	case OAL_EVENT_AUDIO_CONNECTING:
	case OAL_EVENT_AUDIO_DISCONNECTING:
	case OAL_EVENT_AUDIO_CONNECTED:
	case OAL_EVENT_AUDIO_DISCONNECTED:
	case OAL_EVENT_AUDIO_STOPPED:
	case OAL_EVENT_AUDIO_STARTED:
		if (a2dp_source_cb)
			a2dp_source_cb(event_type, event_data);
		break;
	case OAL_EVENT_A2DP_SINK_DISCONNECTED:
	case OAL_EVENT_A2DP_SINK_CONNECTED:
	case OAL_EVENT_A2DP_SINK_DISCONNECTING:
	case OAL_EVENT_A2DP_SINK_CONNECTING:
		if (a2dp_sink_cb)
			a2dp_sink_cb(event_type, event_data);
		break;
	case OAL_EVENT_HFP_CONNECTING:
	case OAL_EVENT_HFP_DISCONNECTING:
	case OAL_EVENT_HFP_CONNECTED:
	case OAL_EVENT_HFP_DISCONNECTED:
	case OAL_EVENT_HFP_AUDIO_CONNECTED:
	case OAL_EVENT_HFP_AUDIO_DISCONNECTED:
	case OAL_EVENT_HFP_AUDIO_CONNECTING:
	case OAL_EVENT_HFP_AUDIO_DISCONNECTING:
		if (ag_cb)
			ag_cb(event_type, event_data);
		break;
	case OAL_EVENT_HDP_APP_REGISTERED:
	case OAL_EVENT_HDP_APP_UNREGISTERED:
	case OAL_EVENT_HDP_CHANNEL_CONNECTED:
	case OAL_EVENT_HDP_CHANNEL_DESTROYED:
		if (hdp_cb)
			hdp_cb(event_type, event_data);
		break;
	case OAL_EVENT_AVRCP_CONNECTED:
	case OAL_EVENT_AVRCP_DISCONNECTED:
	case OAL_EVENT_AVRCP_REMOTE_FEATURES:
	case OAL_EVENT_AVRCP_SETTING_EQUALIZER_STATUS:
	case OAL_EVENT_AVRCP_SETTING_REPEAT_STATUS:
	case OAL_EVENT_AVRCP_SETTING_SHUFFLE_STATUS:
	case OAL_EVENT_AVRCP_SETTING_SCAN_STATUS:
	case OAL_EVENT_AVRCP_VOLUME_MUTE_CHANGED:
	case OAL_EVENT_AVRCP_DELAY_CHANGED:
		if (avrcp_cb)
			avrcp_cb(event_type, event_data);
		break;
	case OAL_EVENT_AVRCP_CT_CONNECTED:
	case OAL_EVENT_AVRCP_CT_DISCONNECTED:
	case OAL_EVENT_AVRCP_CT_EQUALIZER_STATUS:
	case OAL_EVENT_AVRCP_CT_REPEAT_STATUS:
	case OAL_EVENT_AVRCP_CT_SHUFFLE_STATUS:
	case OAL_EVENT_AVRCP_CT_SCAN_STATUS:
	case OAL_EVENT_AVRCP_CT_NOTIF_PLAY_STATUS_CHANGED:
	case OAL_EVENT_AVRCP_CT_NOTIF_PLAY_POS_CHANGED:
	case OAL_EVENT_AVRCP_CT_NOTIF_TRACK_CHANGE:
	case OAL_EVENT_AVRCP_CT_MEDIA_INFO:
	case OAL_EVENT_AVRCP_CT_PLAY_STATUS:
	case OAL_EVENT_AVRCP_CT_PASS_CMD_RES:
	case OAL_EVENT_AVRCP_CT_PLAYER_SETTING_RES:
	case OAL_EVENT_AVRCP_CT_GET_PROPERTY_RES:
		if (avrcp_ctrl_cb)
			avrcp_ctrl_cb(event_type, event_data);
		break;
	case OAL_EVENT_BLE_ENABLED:
	case OAL_EVENT_BLE_DISABLED:
	case OAL_EVENT_BLE_SERVER_INSTANCE_INITIALISED:
	case OAL_EVENT_BLE_ADVERTISING_STARTED:
	case OAL_EVENT_BLE_ADVERTISING_STOPPED:
	case OAL_EVENT_BLE_MULTI_ADVERTISING_ENABLE:
	case OAL_EVENT_BLE_MULTI_ADVERTISING_DISABLE:
	case OAL_EVENT_BLE_MULTI_ADVERTISING_SET_INST_DATA:
	case OAL_EVENT_BLE_MULTI_ADVERTISING_UPDATE:
	/* GATT Server Events */
	case OAL_EVENT_GATTS_SERVICE_ADDED:                     /* gatts Service Added Done */\
	case OAL_EVENT_GATTS_CHARACTERISTIC_ADDED:              /* gatts Characteristic Added Done */\
	case OAL_EVENT_GATTS_DESCRIPTOR_ADDED:                  /* gatts Descriptor added done */\
	case OAL_EVENT_GATTS_SERVICE_STARTED:                   /* gatts Service Started  */\
	case OAL_EVENT_GATTS_SERVICE_STOPED:                    /* gatts Service Stoped */\
	case OAL_EVENT_GATTS_SERVICE_DELETED:           /* gatts Service Deleted  */\
	case OAL_EVENT_GATTS_CONNECTION_COMPLETED:              /* gatts Connection completed */\
	case OAL_EVENT_GATTS_DISCONNECTION_COMPLETED:           /* gatts Disconnection completed */\
	case OAL_EVENT_GATTS_REQUEST_READ:                      /* gatts Request Read from client */\
	case OAL_EVENT_GATTS_REQUEST_WRITE:                     /* gatts Request Write from client */\
	case OAL_EVENT_GATTS_IND_CONFIRM:			/* gatts Indicate confirmation from remote client */\
	case OAL_EVENT_GATTS_REQUEST_ACQUIRE_WRITE:
	case OAL_EVENT_GATTS_REQUEST_ACQUIRE_NOTIFY:
	/* Tizen Platform Specific */
	case OAL_EVENT_GATTS_NOTIFICATION:			/* gatts Notification changed event */\
	case OAL_EVENT_GATTS_MTU_CHANGED:			/* gatts MTU changed event */\
		if (adapter_gatt_cb)
			adapter_gatt_cb(event_type, event_data);
		if (adapter_le_cb)
			adapter_le_cb(event_type, event_data);
		break;
#ifdef TIZEN_GATT_CLIENT
	case OAL_EVENT_GATTC_DISCONNECTION_COMPLETED:
	case OAL_EVENT_GATTC_SERVICE_SEARCH_RESULT:
	case OAL_EVENT_GATTC_SERVICE_SEARCH_DONE:
	case OAL_EVENT_GATTC_INCL_SRVC_SERACH_RESULT:
	case OAL_EVENT_GATTC_CHARAC_SERACH_RESULT:
	case OAL_EVENT_GATTC_DESC_SERACH_RESULT:
	case OAL_EVENT_GATTC_READ_CHARAC:
	case OAL_EVENT_GATTC_WRITE_CHARAC:
	case OAL_EVENT_GATTC_READ_DESCR:
	case OAL_EVENT_GATTC_WRITE_DESCR:
	case OAL_EVENT_GATTC_SERVICE_CHANGED_IND:
	case OAL_EVENT_GATTC_MTU_EXCHANGE_COMPLETED:
	case OAL_EVENT_GATTC_REMOTE_RSSI:
	case OAL_EVENT_GATTC_NOTIFICATION_REGISTERED:
	case OAL_EVENT_GATTC_NOTIFICATION_DEREGISTERED:
	case OAL_EVENT_GATTC_NOTIFY_DATA:
	case OAL_EVENT_GATTC_CONNECTION_COMPLETED:
	case OAL_EVENT_GATTC_REGISTRATION:
		if (adapter_gatt_cb)
			adapter_gatt_cb(event_type, event_data);
		if (adapter_le_cb)
			adapter_le_cb(event_type, event_data);
		break;
#endif
	case OAL_EVENT_BLE_DISCOVERY_STARTED:
	case OAL_EVENT_BLE_DISCOVERY_STOPPED:
	case OAL_EVENT_BLE_REMOTE_DEVICE_FOUND:
	case OAL_EVENT_BLE_LOCAL_FEATURES:
		if (adapter_le_cb)
			adapter_le_cb(event_type, event_data);
		break;
	default:
		BT_ERR("Unhandled Event: %d", event_type);
		break;
	}

	if (oal_event->data)
		g_free(oal_event->data);
	g_free(oal_event);
	return FALSE;
}

void _bt_service_oal_event_receiver(int event_type, gpointer event_data, gsize len)
{
	bt_service_oal_event_data_t *oal_event = g_new0(bt_service_oal_event_data_t, 1);

	BT_INFO("event_type: [%d], data size: [%zu]", event_type, len);

	oal_event->type = event_type;
	if (event_data)
		oal_event->data = g_memdup(event_data, len);

	g_idle_add(__bt_handle_oal_events, oal_event);
}
