/*
 * Copyright (c) 2015 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Contact: Atul Rai <a.rai@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <gio/gio.h>
#include <glib.h>
#include <dlog.h>
#include <string.h>

/*bt-service headers */
#include "bt-internal-types.h"
#include "bt-service-common.h"
#include "bt-service-util.h"
#include "bt-service-event-receiver.h"
#include "bt-request-handler.h"
#include "bt-service-event.h"
#include "bt-service-core-adapter.h"
#include "bt-service-core-device.h"
#include "bt-service-device-internal.h"

/* OAL headers */
#include <oal-event.h>
#include <oal-manager.h>

static GSList *paired_list;
static GSList *bonded_addr_list;
static unsigned int bonded_list_index = 0;
static gboolean bonded_list_retrieved = FALSE;

static void __bt_service_get_bonded_dev_list(GArray **dev_list)
{
	GSList *l;

	BT_DBG("+");

	ret_if(NULL == dev_list);

	for (l = paired_list; NULL != l; l = g_slist_next(l)) {
		bt_remote_dev_info_t *rem_info = l->data;
		bluetooth_device_info_t dev_info;

		if (!rem_info)
			continue;

		_bt_copy_remote_device(rem_info, &dev_info);
		g_array_append_vals(*dev_list, &dev_info, sizeof(bluetooth_device_info_t));
	}

	BT_DBG("+");
	return;
}

bt_remote_dev_info_t *_bt_service_get_remote_dev_info(const char *address)
{
	GSList *l;

	BT_DBG("+");
	retv_if(NULL == address, NULL);

	for (l = paired_list; NULL != l; l = g_slist_next(l)) {
		bt_remote_dev_info_t *rem_info = l->data;

		if (!rem_info)
			continue;

		if (!strncasecmp(rem_info->address, address, BT_ADDRESS_STRING_SIZE)) {
			BT_DBG("Device found");
			return rem_info;
		}
	}

	BT_DBG("-");
	return NULL;
}

static void __bt_handle_pending_request(int result, int service_function)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info = NULL;

	BT_DBG("+");

	/* Get method invocation context */
	for (l = _bt_get_invocation_list(); l != NULL; ) {
		req_info = l->data;
		l = g_slist_next(l);
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		switch (service_function) {
		case BT_GET_BONDED_DEVICE: {
			char *address;
			bt_remote_dev_info_t *rem_info;
			bluetooth_device_info_t dev_info;

			address = (char *)req_info->user_data;
			if (!address) {
				BT_ERR_C("Unexpected: address is NULL");
				continue;
			}

			BT_DBG("BT_GET_BONDED_DEVICE for [%s] found", address);
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			rem_info = _bt_service_get_remote_dev_info(address);
			if (rem_info) {
				result = BLUETOOTH_ERROR_NONE;
				_bt_copy_remote_device(rem_info, &dev_info);
				g_array_append_vals(out_param, &dev_info,
						sizeof(bluetooth_device_info_t));
			} else {
				BT_DBG("Device not found in paired list");
				result = BLUETOOTH_ERROR_NOT_PAIRED;
			}

			_bt_service_method_return(req_info->context, out_param, result);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_GET_BONDED_DEVICES: {
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			__bt_service_get_bonded_dev_list(&out_param);
			_bt_service_method_return(req_info->context, out_param, result);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_GET_IS_ALIAS_SET: {
			char *address;
			bt_remote_dev_info_t *rem_info;

			address = (char *)req_info->user_data;
			if (!address) {
				BT_ERR_C("Unexpected: address is NULL");
				continue;
			}

			BT_DBG("BT_GET_IS_ALIAS_SET for [%s] found", address);
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			rem_info = _bt_service_get_remote_dev_info(address);
			if (rem_info) {
				result = BLUETOOTH_ERROR_NONE;
				g_array_append_vals(out_param, &(rem_info->is_alias_set),
						sizeof(gboolean));
			} else {
				BT_DBG("Device not found in paired list");
				result = BLUETOOTH_ERROR_NOT_PAIRED;
			}

			_bt_service_method_return(req_info->context, out_param, result);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		case BT_GET_CONNECTED_LINK_TYPE: {
			char *address;
			bt_remote_dev_info_t *rem_info;

			address = (char *)req_info->user_data;
			if (!address) {
				BT_ERR_C("Unexpected: address is NULL");
				continue;
			}

			BT_DBG("BT_GET_CONNECTED_LINK_TYPE for [%s] found", address);
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
			rem_info = _bt_service_get_remote_dev_info(address);
			if (rem_info) {
				result = BLUETOOTH_ERROR_NONE;
				g_array_append_vals(out_param, &(rem_info->connected),
						sizeof(bluetooth_connected_link_t));
			} else {
				BT_DBG("Device not found in paired list");
				result = BLUETOOTH_ERROR_NOT_PAIRED;
			}

			_bt_service_method_return(req_info->context, out_param, result);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
		default:
			BT_ERR_C("Unknown service function %d", service_function);
		}
	}
}

static void __bt_service_fetch_dev_info_for_bonded_addr_list(void)
{
	unsigned int count = g_slist_length(bonded_addr_list);

	BT_DBG("+");

	BT_INFO("bonded_list_index: %d", bonded_list_index);
	for (; bonded_list_index < count; bonded_list_index++) {
		GSList *l = g_slist_nth(bonded_addr_list, bonded_list_index);
		bluetooth_device_address_t *dev_addr = l->data;

		if (!dev_addr)
			continue;

		if (BLUETOOTH_ERROR_NONE != _bt_device_get_bonded_device_info(dev_addr)) {
			char address[BT_ADDRESS_STRING_SIZE];
			_bt_convert_addr_type_to_string(address, dev_addr->addr);
			BT_ERR("_bt_device_get_bonded_device_info Failed for [%s]", address);
			continue;
		}
		return;
	}

	/* Retrieval of bonded device list completed */
	bonded_list_index = 0;
	bonded_list_retrieved = TRUE;
	g_slist_free_full(bonded_addr_list, g_free);
	bonded_addr_list = NULL;

	/* Send reply if request for BT_GET_BONDED_DEVICES is pending */
	__bt_handle_pending_request(
			BLUETOOTH_ERROR_NONE, BT_GET_BONDED_DEVICES);

	/* Send reply if request for BT_GET_BONDED_DEVICE is pending */
	__bt_handle_pending_request(
			BLUETOOTH_ERROR_NONE, BT_GET_BONDED_DEVICE);

	__bt_handle_pending_request(
			BLUETOOTH_ERROR_NONE, BT_GET_IS_ALIAS_SET);
	__bt_handle_pending_request(
			BLUETOOTH_ERROR_NONE, BT_GET_CONNECTED_LINK_TYPE);
	BT_DBG("-");
}

static void __bt_update_connected_status(bt_remote_dev_info_t *rem_info, gboolean connected, unsigned char type)
{
	BT_DBG("+");

	if (connected) {
		if (0 == type) {
			switch (rem_info->connected) {
			case BLUETOOTH_CONNECTED_LINK_NONE:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_BREDR;
				break;
			case BLUETOOTH_CONNECTED_LINK_LE:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_BREDR_LE;
				break;
			default:
				BT_ERR_C("rem_info->connected is already set to desired value: %d",
						rem_info->connected);
			}
		} else {
			switch (rem_info->connected) {
			case BLUETOOTH_CONNECTED_LINK_NONE:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_LE;
				break;
			case BLUETOOTH_CONNECTED_LINK_BREDR:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_BREDR_LE;
				break;
			default:
				BT_ERR_C("rem_info->connected is already set to desired value: %d",
						rem_info->connected);
			}
		}
	} else {
		if (0 == type) {
			switch (rem_info->connected) {
			case BLUETOOTH_CONNECTED_LINK_BREDR:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_NONE;
				break;
			case BLUETOOTH_CONNECTED_LINK_BREDR_LE:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_LE;
				break;
			default:
				BT_ERR_C("rem_info->connected is already set to desired value: %d",
						rem_info->connected);
			}
		} else {
			switch (rem_info->connected) {
			case BLUETOOTH_CONNECTED_LINK_LE:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_NONE;
				break;
			case BLUETOOTH_CONNECTED_LINK_BREDR_LE:
				rem_info->connected = BLUETOOTH_CONNECTED_LINK_BREDR;
				break;
			default:
				BT_ERR_C("rem_info->connected is already set to desired value: %d",
						rem_info->connected);
			}
		}
	}

	BT_DBG("-");
}

void _bt_update_remote_dev_property(
		const char *address, bt_device_prop_type_e type, void *value)
{
	bt_remote_dev_info_t *rem_info;

	BT_DBG("+");

	BT_DBG("Property type: %d", type);
	rem_info = _bt_service_get_remote_dev_info(address);
	ret_if(NULL == rem_info);

	switch (type) {
	case DEV_PROP_CONNECTED: {
		bt_device_conn_info_t *conn_info = value;

		__bt_update_connected_status(
			rem_info, conn_info->connected, conn_info->type);
		break;
	}
	case DEV_PROP_NAME: {
		char *name = value;

		if (name && strlen(name) > 0) {
			g_free(rem_info->name);
			rem_info->name = g_strdup(name);
		}
		break;
	}
	case DEV_PROP_ALIAS: {
		char *alias = value;

		g_free(rem_info->alias);
		rem_info->alias = NULL;
		rem_info->is_alias_set = FALSE;

		if (alias && strlen(alias) > 0) {
			rem_info->alias = g_strdup(alias);
			rem_info->is_alias_set = TRUE;
		}
		break;
	}
	case DEV_PROP_TRUST: {
		gboolean *trusted = value;
		rem_info->trust = *trusted;
		break;
	}
	case DEV_PROP_SERVICES: {
		bt_remote_dev_info_t *src = value;

		if (src->uuid_count > 0 && src->uuids) {
			int i;

			/* Release previously allocated UUIDs */
			if (rem_info->uuid_count > 0 && rem_info->uuids)
				for (i = 0; i < rem_info->uuid_count; i++)
					g_free(rem_info->uuids[i]);
			g_free(rem_info->uuids);

			/* Copy new UUID list */
			rem_info->uuids = g_malloc0(src->uuid_count * sizeof(char *));
			for (i = 0; i < src->uuid_count; i++)
				rem_info->uuids[i] = g_strdup(src->uuids[i]);

			rem_info->uuid_count = src->uuid_count;
			BT_DBG("UUID count: %d", rem_info->uuid_count);
		}
		break;
	}
	default:
		BT_DBG("Unknown property");
		return;
	}

	BT_DBG("-");
}

void _bt_service_add_device_to_bonded_list(bt_remote_dev_info_t *rem_info)
{
	bt_remote_dev_info_t *dev_info;

	BT_DBG("+");
	ret_if(NULL == rem_info);

	dev_info = _bt_service_get_remote_dev_info(rem_info->address);
	if (!dev_info) {
		BT_DBG("Device info not present, allocate memory");
		dev_info = g_malloc0(sizeof(bt_remote_dev_info_t));
		paired_list = g_slist_append(paired_list, dev_info);
	}

	_bt_copy_remote_dev_info(dev_info, rem_info);

	if (!bonded_list_retrieved) {
		char addr_str[BT_ADDRESS_STRING_SIZE];
		bluetooth_device_address_t *dev_addr;
		GSList *l;

		l = g_slist_nth(bonded_addr_list, bonded_list_index);
		if (!l) {
			BT_ERR_C("l is NULL unexpectedly, need to check");
			return;
		}

		dev_addr = l->data;
		if (!dev_addr) {
			BT_ERR_C("dev_addr is NULL unexpectedly, need to check");
			return;
		}

		_bt_convert_addr_type_to_string(addr_str, dev_addr->addr);
		if (!strncasecmp(addr_str, rem_info->address, BT_ADDRESS_STRING_SIZE)) {
			bonded_list_index++;
			__bt_service_fetch_dev_info_for_bonded_addr_list();
		}
	}

	BT_DBG("-");
}

void _bt_service_remove_device_from_bonded_list(const char *address)
{
	bt_remote_dev_info_t *rem_info;

	BT_DBG("+");
	ret_if(NULL == address);

	rem_info = _bt_service_get_remote_dev_info(address);
	if (!rem_info) {
		BT_DBG("Device not found in list");
		return;
	}

	paired_list = g_slist_remove(paired_list, rem_info);
	_bt_free_remote_dev(rem_info);
	BT_DBG("-");
}

void _bt_device_handle_paired_address_list(bluetooth_device_address_t *addr_list, int count)
{
	int indx;

	BT_DBG("+");

	if (bonded_list_retrieved) {
		BT_ERR_C("Unexpected bonded address list received");
		return;
	}

	if (0 == count) {
		BT_DBG("No paired device found");
		bonded_list_index = 0;
		bonded_list_retrieved = TRUE;
		return;
	}

	ret_if(NULL == addr_list);

	for (indx = 0; indx < count; indx++) {
		bluetooth_device_address_t *dev_addr;

		dev_addr = g_memdup(&(addr_list[indx]), sizeof(bluetooth_device_address_t));
		bonded_addr_list = g_slist_append(bonded_addr_list, dev_addr);
	}

	__bt_service_fetch_dev_info_for_bonded_addr_list();
	BT_DBG("-");
}

static gboolean __check_bonded_devices(gpointer data)
{
	BT_DBG("+");

	retv_if(bonded_list_retrieved, FALSE);

	/* Bonded devices are not yet retrived, retry */
	if (BLUETOOTH_ERROR_NONE != _bt_adapter_get_bonded_devices())
		BT_ERR_C("_bt_adapter_get_bonded_devices failed");

	BT_DBG("-");
	return TRUE;
}

void _bt_device_handle_adapter_state(gboolean enabled)
{
	BT_DBG("+");

	if (!enabled) {
		BT_DBG("Adapter disabled");
		g_slist_free_full(paired_list, g_free);
		paired_list = NULL;
		bonded_list_retrieved = FALSE;
		return;
	}

	if (BLUETOOTH_ERROR_NONE != _bt_adapter_get_bonded_devices())
		BT_ERR_C("_bt_adapter_get_bonded_devices failed");

	g_timeout_add_seconds(5, __check_bonded_devices, NULL);
	BT_DBG("-");
}

static gboolean __reply_pending_request(gpointer data)
{
	int *request = data;

	BT_INFO("Request: %d", *request);
	__bt_handle_pending_request(BLUETOOTH_ERROR_NONE, *request);
	g_free(request);
	return FALSE;
}

int _bt_get_bonded_device_info(bluetooth_device_address_t *dev_addr)
{
	char address[BT_ADDRESS_STRING_SIZE];
	bt_remote_dev_info_t *rem_info;
	int request = BT_GET_BONDED_DEVICE;

	BT_DBG("+");

	retv_if(!dev_addr, BLUETOOTH_ERROR_INVALID_PARAM);

	if (!bonded_list_retrieved) {
		/*
		 * Blonded list retrival is in progress, return BLUETOOTH_ERROR_NONE here.
		 * reply will be sent once bonded list is retrieved completely.
		 */
		return BLUETOOTH_ERROR_NONE;
	}

	_bt_convert_addr_type_to_string(address, dev_addr->addr);

	rem_info = _bt_service_get_remote_dev_info(address);
	if (!rem_info) {
		BT_ERR("Bonded device not found in the list, return error");
		return BLUETOOTH_ERROR_NOT_PAIRED;
	}

	g_idle_add(__reply_pending_request, g_memdup(&request, sizeof(int)));
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_bonded_devices(void)
{
	int request = BT_GET_BONDED_DEVICES;

	BT_DBG("+");
	if (!bonded_list_retrieved) {
		/*
		 * Blonded list retrival is in progress, return BLUETOOTH_ERROR_NONE here.
		 * reply will be sent once bonded list is retrieved completely.
		 */
		return BLUETOOTH_ERROR_NONE;
	}

	g_idle_add(__reply_pending_request, g_memdup(&request, sizeof(int)));
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_is_alias_set(bluetooth_device_address_t *device_address)
{
	char address[BT_ADDRESS_STRING_SIZE];
	bt_remote_dev_info_t *rem_info;
	int request = BT_GET_IS_ALIAS_SET;

	BT_DBG("+");
	if (!bonded_list_retrieved) {
		/*
		 * Blonded list retrival is in progress, return BLUETOOTH_ERROR_NONE here.
		 * reply will be sent once bonded list is retrieved completely.
		 */
		return BLUETOOTH_ERROR_NONE;
	}

	_bt_convert_addr_type_to_string(address, device_address->addr);
	rem_info = _bt_service_get_remote_dev_info(address);
	if (!rem_info) {
		BT_ERR("Bonded device not found in the list, return error");
		return BLUETOOTH_ERROR_NOT_PAIRED;
	}

	g_idle_add(__reply_pending_request, g_memdup(&request, sizeof(int)));
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_connected_link(bluetooth_device_address_t *device_address)
{
	char address[BT_ADDRESS_STRING_SIZE];
	bt_remote_dev_info_t *rem_info;
	int request = BT_GET_CONNECTED_LINK_TYPE;

	BT_DBG("+");
	if (!bonded_list_retrieved) {
		/*
		 * Blonded list retrival is in progress, return BLUETOOTH_ERROR_NONE here.
		 * reply will be sent once bonded list is retrieved completely.
		 */
		return BLUETOOTH_ERROR_NONE;
	}

	_bt_convert_addr_type_to_string(address, device_address->addr);
	rem_info = _bt_service_get_remote_dev_info(address);
	if (!rem_info) {
		BT_ERR("Bonded device not found in the list, return error");
		return BLUETOOTH_ERROR_NOT_PAIRED;
	}

	g_idle_add(__reply_pending_request, g_memdup(&request, sizeof(int)));
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

gboolean _bt_is_bonded_devices_retrived(void)
{
	return bonded_list_retrieved;
}
