/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2016-2017 Samsung Electronics Co., Ltd.
 *
 * Contact: Anupam Roy (anupam.r@samsung.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <glib.h>
#include <dlog.h>
#include "bt-service-common.h"
#include "bt-service-core-adapter.h"
#include "bt-service-event-receiver.h"
#include "bt-request-handler.h"
#include "bluetooth-api.h"

#include "bluetooth-api.h"
#include "bt-internal-types.h"
#include "bt-service-util.h"
#include "bt-service-common.h"
#include "bt-service-event.h"

#include "bt-internal-types.h"
#include "bt-service-gatt.h"
#include "bt-service-gatt.h"
#include "bt-service-core-device.h"
#include "bt-service-core-adapter-le.h"
#ifdef TIZEN_GATT_CLIENT
#include "bluetooth-gatt-client-api.h"
#endif

#include <oal-hardware.h>
#include <oal-manager.h>
#include <oal-event.h>
#include <oal-adapter-mgr.h>
#include <oal-device-mgr.h>
#include <oal-gatt.h>

#define BT_GATT_TRANSPORT_LE 0
#define BT_GATT_TRANSPORT_BR_EDR 1
#define BT_GATT_TRANSPORT_LE_BR_EDR 2
#define BT_UUID_STRING_MAX 64
#define BT_SENDER_MAX_LENGTH 50
#define MAX_APPS_SUPPORTED 11 /* Slot 0 is not used */

#ifdef TIZEN_GATT_CLIENT
#define NUM_UUID 20
#else
#define NUM_UUID 10
#endif


#define UUID_MAX_LEN 50


#define BDADDR_ANY   (&(bluetooth_device_address_t) {{0, 0, 0, 0, 0, 0} })

#ifdef TIZEN_GATT_CLIENT
char uuid_list[NUM_UUID][BT_UUID_STRING_MAX] = {"0000b00b-0000-0000-f065-080080fa49b5", /* Used by BLEAPP */
	"0000b00b-1111-1111-0123-456789ab0cd2", /* Used by BLEAPP */
	"0000b00b-2222-1111-0123-456789ab0cd2",
	"0000b00b-3333-1111-0123-456789ab0cd2",
	"0000b00b-4444-1111-0123-456789ab0cd2",
	"0000b00b-5555-1111-0123-456789ab0cd2",
	"0000b00b-6666-1111-0123-456789ab0cd2",
	"0000b00b-7777-1111-0123-456789ab0cd2",
	"0000b00b-8888-1111-0123-456789ab0cd2",
	"0000b00b-9999-1111-0123-456789ab0cd2",
	"0000b00b-aaaa-1111-0123-456789ab0cd2",
	"0000b00b-bbbb-1111-0123-456789ab0cd2",
	"0000b00b-cccc-1111-0123-456789ab0cd2",
	"0000b00b-dddd-1111-0123-456789ab0cd2",
	"0000b00b-eeee-1111-0123-456789ab0cd2",
	"0000b00b-ffff-1111-0123-456789ab0cd2",
	"0000b00c-0000-1111-0123-456789ab0cd2",
	"0000b00c-1111-1111-0123-456789ab0cd2",
	"0000b00c-2222-1111-0123-456789ab0cd2",
	"0000b00c-3333-1111-0123-456789ab0cd2"};
#else
char uuid_list[NUM_UUID][BT_UUID_STRING_MAX] = {"0000b00b-0000-0000-f065-080080fa49b5", /* Used by BLEAPP */
	"0000b00b-1111-1111-0123-456789ab0cd2", /* Used by BLEAPP */
	"0000b00b-2222-1111-0123-456789ab0cd2",
	"0000b00b-3333-1111-0123-456789ab0cd2",
	"0000b00b-4444-1111-0123-456789ab0cd2",
	"0000b00b-5555-1111-0123-456789ab0cd2",
	"0000b00b-6666-1111-0123-456789ab0cd2",
	"0000b00b-7777-1111-0123-456789ab0cd2",
	"0000b00b-8888-1111-0123-456789ab0cd2",
	"0000b00b-9999-1111-0123-456789ab0cd2"};
#endif

#ifdef TIZEN_GATT_CLIENT
/* Reserved GATT client Instance UUID. This is used only internally by bt-service */
#define DEFAULT_GATT_CLIENT_UUID "0000a00a-1111-1111-0123-456789abcdef"

static int gatt_default_client = -1;

typedef struct {
	gboolean is_registered;
	bluetooth_device_address_t addr;
	unsigned char svc_uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int svc_inst;
	unsigned char char_uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int char_inst;
} bt_gatt_notif_reg_info_t;

struct gatt_out_conn_info_t {
	int client_id;                                   /* This value unique identifies a GATT Client instance */
	char *addr;                                      /* Remote GATT Server address */
};

/* Linked List of outgoing gatt connection list
   Note: This is valid only for local GATT client  */
static GSList *outgoing_gatt_conn_list = NULL;

/* GATT Server Info(Local Client) List Structure */
struct gatt_server_info_t {
	int connection_id;                               /* This value will uniquely identify a GATT client-server connection */
	int client_id;                                   /* This value unique identifies a GATT Client instance */
	char *addr;                                      /* Remote GATT Server address */
};

/* Linked List of connected Remote GATT Servers */
static GSList *gatt_server_info_list = NULL;

typedef struct {
	gboolean is_changed;
	unsigned char uuid[BLUETOOTH_UUID_HEX_MAX_LEN]; /* If any service added */
} bt_gatt_svc_changed_info_t;


typedef struct {
	int conn_id;
	//      bluetooth_device_address_t address; /* Remote BLE Device Address */
	GSList *services;                   /* List of all services of above remote device */
	int count;                          /* Number of services browsed from remote device */
	bt_gatt_svc_changed_info_t info;
} bt_gatt_service_info_list_t;

typedef struct {
	unsigned char uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int inst_id;
	int is_primary;
	GSList *chars;
	GSList *included_svcs;
	gboolean is_removed;            /* 0 => Service is added, 1=> removed */
} bt_gatt_service_info_t;

typedef struct {
	unsigned char uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int inst_id;
	int props;
	int val_len;
	unsigned char val[BLUETOOTH_GATT_ATT_DATA_LENGTH_MAX];
	GSList *descs;
} bt_gatt_char_info_t;

typedef struct {
	unsigned char uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int inst_id;
	int val_len;
	unsigned char val[BLUETOOTH_GATT_ATT_DATA_LENGTH_MAX];
} bt_gatt_descriptor_info_t;

typedef struct {
	unsigned char uuid[BLUETOOTH_UUID_HEX_MAX_LEN];
	int inst_id;
} bt_gatt_included_service_info_t;

GSList *list_gatt_info;


#endif

/* App Information structure */
typedef struct {
	int adv_handle;
	char sender[BT_SENDER_MAX_LENGTH];
	char uuid[BT_UUID_STRING_MAX];
	int instance_id;
	int adv_instance;
	bluetooth_advertising_data_t adv_data; /* Will store adv data for specific slot */
	int adv_data_len;
	bluetooth_scan_resp_data_t scan_rsp;   /* Will store scan rsp data for specific slot */
	int scan_rsp_len;
	gboolean is_initialized;
	GSList *service_handles;
#ifdef TIZEN_GATT_CLIENT
	int client_id;                      /* GATT Client instance ID */
	bluetooth_device_address_t address; /* Remote BLE Device Address */
	gboolean is_watcher_enabled;
#endif
} bt_service_app_info_t;

/* GATT Server Request Info Structure */
struct gatt_server_req_info {
	int connection_id;                               /* This value will uniquely identify a GATT client-server connection */
	int request_id;                                  /* This is an unique transaction ID assigned against each request by stack */
	int attribute_handle;                            /* GATT server attribute handle */
	int offset;                                      /* GATT server attribute offset on which request is invoked by GATT client */
	bluetooth_gatt_att_request_type_e request_type;  /* Read or Write request */
	char *addr;                                      /* Remote GATT client address */
};

/* GATT Client Info List Structure */
struct gatt_client_info_t {
	int connection_id;                               /* This value will uniquely identify a GATT client-server connection */
	int instance_id;                                 /* This value unique identifies a GATT server instance */
	char *addr;                                      /* Remote GATT client address */
};

/* GATT Indicate confirm result  */
struct gatt_indicate_cfm_result_info_t {
	int result;                                      /* Result of event */
	char *addr;                                      /* Remote GATT client address */
	int att_hdl;                                     /* Characteristic Attribute handle */
	int completed;                                   /* 1 if last event, otheriwse 0 */
};

/* Request Search Utility method */
static struct gatt_server_req_info *__bt_gatt_server_find_request_info(int request_id,
		bluetooth_gatt_att_request_type_e req_type);

static int __bt_gatt_send_indication_to_all_connected_clients(bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_server_indication_params_t *param);

static void __bt_remove_all_service_handles(int instance_id);

#ifdef TIZEN_GATT_CLIENT
static void __bt_handle_client_instance_registered(event_gattc_register_t *data);
static void __bt_handle_client_connected(event_gattc_conn_t *event_data);
static void __bt_handle_client_disconnected(event_gattc_conn_t *event_data);
static void __bt_handle_client_service_search_completed(event_gattc_conn_status_t *event_data);
static void __bt_handle_client_service_search_result(event_gattc_service_result_t *event_data);
static void __bt_handle_client_characteristic_search_result(
		event_gattc_characteristic_result_t *event_data);
static void __bt_handle_client_descriptor_search_result(event_gattc_descriptor_result_t *event_data);
static void __bt_handle_client_characteristic_read_data(event_gattc_read_data *event_data);
static void __bt_handle_client_descriptor_read_data(event_gattc_read_data *event_data);
static void __bt_handle_client_characteristic_write_data(event_gattc_write_data *event_data);
static void __bt_handle_client_descriptor_write_data(event_gattc_write_data *event_data);
static int __bt_do_unregister_gatt_instance(int instance_id);
static void __bt_hanlde_le_device_disconnection(event_dev_conn_status_t *event_data);
static void __bt_handle_client_notification_registered(event_gattc_regdereg_notify_t *event_data,
		gboolean is_registered);
static void __bt_handle_client_notification_data(event_gattc_notify_data *event_data);

#endif

static int __bt_do_unregister_server_instance(int server_instance);

/* Linked List of GATT requests from Remote GATT Clients */
static GSList *gatt_server_requests = NULL;

/* Linked List of connected Remote GATT clients */
static GSList *gatt_client_info_list = NULL;

/* Number of clients to be notified to */
static int num_indicate_clients;

/* List of applications */
static bt_service_app_info_t numapps[MAX_APPS_SUPPORTED];

static void __bt_gatt_handle_pending_request_info(int result,
		int service_function, void *data, unsigned int size);

static void __bt_handle_server_instance_registered(event_gatts_register_t *data);

static void __bt_gatt_event_handler(int event_type, gpointer event_data);

void _bt_check_adv_app_termination(const char *name)
{
	bt_service_app_info_t *app = NULL;
	int k;
	int apps[MAX_APPS_SUPPORTED];

	ret_if(NULL == name);

	memset(&apps, -1, sizeof(apps));

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		app = &numapps[k];

		/* Search for a app which has same sender and stop adv is running */
		if (!strncasecmp(app->sender, name, strlen(name)) && app->is_initialized == TRUE) {
			BT_DBG("Match found, name: %s", name);

			/* Unregister all service handles with stack */
			__bt_remove_all_service_handles(app->instance_id);

			/* If Advertising is enabled, stop it */
			if (app->adv_handle > 0) {
				BT_INFO("Stop advertising on instance ID [%d]", app->instance_id);
				/* Disable adv if running */
				BT_INFO("Disable Advertising Adv Handle [%d] sender [%s]",
						app->adv_handle, name);
				_bt_set_advertising(app->sender, app->adv_handle, FALSE, FALSE);
			}

			/* Save instances of all apps that need to be unregistered */
			apps[app->instance_id] = 1;
		}
	}

	/* Unregister all apps one by one */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (apps[k] == 1) {
			BT_INFO("Unregister app[%d]", k);
			/* Unregister server instance */
			__bt_do_unregister_server_instance(k);
		}
	}
}

static int __bt_gatt_send_indication_to_all_connected_clients(bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_server_indication_params_t *param)
{
	GSList *l;
	int ret = OAL_STATUS_SUCCESS;
	int result = OAL_STATUS_INTERNAL_ERROR;

	BT_INFO("Current total number of connected clients [%d]", g_slist_length(gatt_client_info_list));
	for (l = gatt_client_info_list; l != NULL; l = l->next) {
		struct gatt_client_info_t *info = l->data;

		if (info) {
			BT_INFO("GATT Remote client address [%s] connection Id [%d]", info->addr, info->connection_id);

			ret = gatts_send_indication(param->instance_id, param->atrribute_handle,
					info->connection_id, data->length,
					param->need_confirmation, (char *)(&data->data[0]));

			BT_INFO("Send Indication to GATT client [%s] result: [%d]", info->addr, ret);
			if (ret == OAL_STATUS_SUCCESS) {
				BT_INFO("Send Indication sent successfully to GATT client [%s]", info->addr);
				result = ret;
				num_indicate_clients++;
			}
		}
	}
	BT_INFO("Indication sending done for total number of clients [%d]", num_indicate_clients);
	return result;
}

static struct gatt_server_req_info *__bt_gatt_server_find_request_info(int request_id,
		bluetooth_gatt_att_request_type_e req_type)
{
	GSList *l;

	for (l = gatt_server_requests; l != NULL; l = l->next) {
		struct gatt_server_req_info *req_info = l->data;

		if (req_info && req_info->request_id == request_id && req_info->request_type == req_type) {
			BT_INFO("GATT Server request info found Req ID [%d] handle [%d] conn ID [%d]",
					req_info->request_id, req_info->attribute_handle, req_info->connection_id);
			return req_info;
		}
	}
	BT_ERR("Gatt Request not found");
	return NULL;
}

void _bt_get_adv_handle_from_instance(int server_inst, int *adv_handle)
{
	BT_DBG("+");
	int k;
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (numapps[k].is_initialized == 1 && numapps[k].instance_id == server_inst) {
			*adv_handle = numapps[k].adv_handle;
			break;
		}
	}
}

#ifdef TIZEN_GATT_CLIENT
char * _bt_gatt_get_default_gatt_client_uuid(void)
{
	return g_strdup(DEFAULT_GATT_CLIENT_UUID);
}


static void __bt_register_default_gatt_client()
{
	char *uuid_str;
	oal_uuid_t uuid;

	uuid_str = _bt_gatt_get_default_gatt_client_uuid();
	_bt_string_to_uuid(uuid_str, (service_uuid_t*)&uuid);

	BT_INFO("Register Default GATT client uuid [%s]", uuid_str);

	if (OAL_STATUS_SUCCESS != gattc_register(&uuid)) /* for only Smart Control */
		BT_ERR("gattc register failed");

	g_free(uuid_str);
}
#endif

int _bt_gatt_init(void)
{
	BT_DBG("+");

	if (OAL_STATUS_SUCCESS != gatt_enable()) {
		BT_ERR("gatt Init failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Register gatt event handler */
	_bt_service_register_event_handler_callback(BT_GATT_MODULE, __bt_gatt_event_handler);
#ifdef TIZEN_GATT_CLIENT
	/*In bluedroid product, defacult gatt client is handled differently*/
	__bt_register_default_gatt_client();
#endif
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

static void __bt_service_reset_gatt_data()
{
	int k;

	BT_INFO("Rest numapp");

	/* Reset data */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
			numapps[k].is_initialized = 0;
			numapps[k].instance_id = -1;
			numapps[k].adv_handle = 0;
			numapps[k].adv_instance = -1;
			memset(numapps[k].sender, 0x00, sizeof(numapps[k].sender));
			memset(numapps[k].uuid, 0x00, sizeof(numapps[k].uuid));
			memset(numapps[k].adv_data.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			memset(numapps[k].scan_rsp.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			numapps[k].adv_data_len = 0;
			numapps[k].scan_rsp_len = 0;
	}
}

void _bt_gatt_deinit(void)
{
	BT_INFO("GATT deinit");

	if (OAL_STATUS_SUCCESS != gatt_disable())
		BT_ERR("gatt deinit failed");

	/* Un-register gatt event handler */
	_bt_service_unregister_event_handler_callback(BT_GATT_MODULE);

	__bt_service_reset_gatt_data();
}

void _bt_update_adv_handle(const char *sender, int adv_handle)
{
	int k;
	BT_DBG("Sender [%s] Adv handle [%d]", sender, adv_handle);
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];
		/* Search for a app which has same sender and adv handle as 0 */
		if (!g_strcmp0(info->sender, sender) && info->adv_handle == 0)
			info->adv_handle = adv_handle;
	}
}

int _bt_is_sender_gatt_server_with_no_adv(const char *sender, int adv_handle)
{
	int k;
	BT_DBG("Sender [%s] Adv handle [%d]", sender, adv_handle);
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];
		/* Search for a app which has same sender and adv handle as 0
		   It is possible that same sender but different adv handle */
		if (!g_strcmp0(info->sender, sender) && info->adv_handle == 0) {
			//info->adv_handle = adv_handle;
			return info->instance_id;
		}
	}
	return -1;
}

int _bt_get_allocated_server_instance(const char *sender, int adv_handle, gboolean use_reserved_slot)
{
	int k;
	BT_DBG("Sender [%s] Adv handle [%d]", sender, adv_handle);
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		/* Exact matching of Adv handle + sender combination */
		if (!g_strcmp0(info->sender, sender) && info->adv_handle == adv_handle)
			return info->instance_id;
#if 0
		if (!g_strcmp0(info->sender, sender) && info->adv_handle == -1)
			return info->instance_id;
#endif
	}
	return -1;
}

char * _bt_get_sender_and_handle(int server_instance, int *adv_handle)
{
	int k;
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == server_instance && info->adv_handle != -1) {
			*adv_handle = info->adv_handle;
			BT_DBG("Server instance [%d] Adv handle [%d] Sender [%s]", server_instance, *adv_handle, info->sender);
			return g_strdup(info->sender);
		}
	}
	return NULL;
}

void _bt_set_new_adv_data(bluetooth_advertising_data_t *adv, int len, int instance)
{
	int k;
	BT_DBG("+");
	bt_service_app_info_t *info = NULL;
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == instance) {
			memcpy(info->adv_data.data, &adv->data, len);
			break;
		}
	}
	BT_DBG("-");
}

void _bt_set_new_scan_rsp_data(bluetooth_scan_resp_data_t *scan, int len, int instance)
{
	int k;
	BT_DBG("+");
	bt_service_app_info_t *info = NULL;
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == instance) {
			memcpy(info->scan_rsp.data, &scan->data, len);
			break;
		}
	}
	BT_DBG("-");
}

void _bt_get_previous_adv_data(bluetooth_advertising_data_t *adv, int *len, int instance)
{
	int k;
	BT_DBG("+");
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == instance) {
			memcpy(&adv->data, info->adv_data.data, info->adv_data_len);
			*len = info->adv_data_len;
			break;
		}
	}
}

void _bt_get_previous_scan_rsp_data(bluetooth_scan_resp_data_t *scan, int *len, int instance)
{
	int k;
	BT_DBG("+");
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == instance) {
			memcpy(&scan->data, info->scan_rsp.data, info->scan_rsp_len);
			*len = info->scan_rsp_len;
			break;
		}
	}
}

#ifdef TIZEN_GATT_CLIENT
static int __bt_do_unregister_gatt_instance(int instance_id)
{
	int ret = OAL_STATUS_SUCCESS;
	int k;

	BT_INFO("DeAllocate server or client instance ID [%d]", instance_id);

	/* Reset data: instance_id parameter could be either for GATT Server or for GATT client  */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (numapps[k].instance_id == instance_id) {
			BT_INFO("This is a GATT server app, unregister: Slot [%d] vacant", k);
			numapps[k].is_initialized = FALSE;
			numapps[k].instance_id = -1;
			numapps[k].adv_handle = 0;
			numapps[k].adv_instance = -1;
			memset(numapps[k].sender, 0x00, sizeof(numapps[k].sender));
			memset(numapps[k].uuid, 0x00, sizeof(numapps[k].uuid));
			memset(numapps[k].adv_data.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			memset(numapps[k].scan_rsp.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			numapps[k].adv_data_len = 0;
			numapps[k].scan_rsp_len = 0;

			/* Its a GATT Server Instance */
			ret = gatts_unregister(instance_id);
			if (ret != OAL_STATUS_SUCCESS) {
				BT_ERR("DeAllocate server instance with stack Fail ret: %d", ret);
				return BLUETOOTH_ERROR_INTERNAL;
			}
			break;
		} else if (numapps[k].client_id == instance_id) {
			BT_INFO("This is a GATT client app, unregister: Slot [%d] vacant", k);
			numapps[k].client_id = -1;
			numapps[k].is_initialized = FALSE;
			memset(numapps[k].sender, 0x00, sizeof(numapps[k].sender));
			memset(numapps[k].uuid, 0x00, sizeof(numapps[k].uuid));
			memset(&numapps[k].address.addr, 0x00, sizeof(bluetooth_device_address_t));

			/* Its a GATT Client Instance */
			ret = gattc_deregister(instance_id);
			if (ret != OAL_STATUS_SUCCESS) {
				BT_ERR("DeAllocate GATT Client instance with stack Fail ret: %d", ret);
				return BLUETOOTH_ERROR_INTERNAL;
			}
			break;
		}
	}
	return BLUETOOTH_ERROR_NONE;
}
#endif

static int __bt_do_unregister_server_instance(int server_instance)
{
	int ret = OAL_STATUS_SUCCESS;
	int k;

	/* Unregister the server instance */
	ret = gatts_unregister(server_instance);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("DeAllocate server instance with stack Fail ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	BT_INFO("DeAllocated server instance with stack successful..");

	/* Reset data */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (numapps[k].instance_id == server_instance) {
			numapps[k].is_initialized = 0;
			numapps[k].instance_id = -1;
			numapps[k].adv_handle = 0;
			numapps[k].adv_instance = -1;
			memset(numapps[k].sender, 0x00, sizeof(numapps[k].sender));
			memset(numapps[k].uuid, 0x00, sizeof(numapps[k].uuid));
			memset(numapps[k].adv_data.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			memset(numapps[k].scan_rsp.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			numapps[k].adv_data_len = 0;
			numapps[k].scan_rsp_len = 0;
			break;
		}
	}
	BT_DBG("Going8 to reset numapp block num [%d]", k);
	return BLUETOOTH_ERROR_NONE;
}

static void __bt_remove_all_service_handles(int instance_id)
{
	GSList *l;
	int *handle = NULL;
	int ret = OAL_STATUS_SUCCESS;
	int count;

	count = g_slist_length(numapps[instance_id].service_handles);
	BT_INFO("Before handle removal: current count [%d]", count);

	for (l = numapps[instance_id].service_handles; l != NULL;) {
		handle = l->data;
		l = g_slist_next(l);

		if (handle) {
			BT_INFO("Server Handle to be Removed [%d] Instance ID [%d]", *handle, instance_id);
			if (_bt_gatt_server_stop_service(numapps[instance_id].sender, *handle, instance_id) != BLUETOOTH_ERROR_NONE)
				continue;

			ret = gatts_delete_service(instance_id, *handle);
			if (ret != OAL_STATUS_SUCCESS) {
				BT_ERR("ret: %d", ret);
				continue;
			} else {
				numapps[instance_id].service_handles = g_slist_remove(numapps[instance_id].service_handles, handle);
				g_free(handle);
				handle = NULL;
				count = g_slist_length(numapps[instance_id].service_handles);
				BT_INFO("After deleting current count [%d]", count);
			}
		}
	}
}

int _bt_unregister_server_instance(const char *sender, int adv_handle)
{
	BT_INFO("Unregister Allocated server instance request Sender [%s] Adv handle [%d]", sender, adv_handle);
	int result = BLUETOOTH_ERROR_NONE;
	int apps[MAX_APPS_SUPPORTED];
	int server_instance;
	int k;

	memset(&apps, -1, sizeof(apps));

	if (adv_handle == 0) {
		BT_DBG("Its a direct GATT Server app request to unregister");
		/* Unregister server instance for each app with same sender (case: GATT Server with multiple adv handle) */
		bt_service_app_info_t *info = NULL;

		for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
			info = &numapps[k];

			/* Exact matching of sender */
			if (!g_strcmp0(info->sender, sender)) {
				BT_INFO("Unregister GATT server instance [%d]", info->instance_id);
				/* Unregister all service handles with stack */
				__bt_remove_all_service_handles(info->instance_id);

				/* Disable adv if running */
				BT_INFO("Disable Advertising Adv Handle [%d] sender [%s]",
						info->adv_handle, sender);
				_bt_set_advertising(sender, info->adv_handle, FALSE, FALSE);

				/* Save all instances which need to be unregistered */
				apps[info->instance_id] = 1;
			}
		}
	} else {
		BT_DBG("Its an Internal unregister request by adv application");
		server_instance = _bt_get_allocated_server_instance(sender, adv_handle, FALSE);
		BT_DBG("Its an Internal unregister request by adv application: Adv disabled srv instance [%d]", server_instance);
		if (server_instance == -1) {
			BT_ERR("No allocated server instance to be removed");
			return BLUETOOTH_ERROR_INVALID_PARAM;
		}
		if (!numapps[server_instance].service_handles) {
			BT_INFO("There are no Service handles with this app, safe to unregister");
			/* Unregister server instance only if this sender does not have any gatt services in it */
			result = __bt_do_unregister_server_instance(server_instance);
		} else {
			numapps[server_instance].adv_handle = 0;
			memset(numapps[server_instance].adv_data.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			numapps[server_instance].adv_data_len = 0;
			memset(numapps[server_instance].scan_rsp.data, 0x00, BLUETOOTH_ADVERTISING_DATA_LENGTH_MAX);
			numapps[server_instance].scan_rsp_len = 0;
		}
		return result;
	}

	/* Unregister all apps one by one */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (apps[k] == 1) {
			BT_INFO("Unregister app[%d]", k);
			/* Unregister server instance */
			__bt_do_unregister_server_instance(k);
		}
	}

	return result;
}

int _bt_register_server_instance(const char *sender, int adv_handle)
{
	int ret = OAL_STATUS_SUCCESS;
	char *uuid_string = NULL;
	int slot = -1;
	int k;
	oal_uuid_t uuid;

	BT_INFO("###Check on which instance Server instance can be initialized....");
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (numapps[k].is_initialized == 1) {
			BT_INFO("Instance ID [%d] is already in use..Check next slot", numapps[k].instance_id);
		} else {
			slot = k;
			BT_INFO("Time to register GATT Server..UUID to be used is [%s] slot [%d]", uuid_list[slot-1], slot);
			break;
		}
	}

	if (slot == -1) {
		BT_ERR("No Slot if free for GATT Server registration..");
		return BLUETOOTH_ERROR_REGISTRATION_FAILED;
	}

	uuid_string = g_malloc0(BT_UUID_STRING_MAX);
	_bt_string_to_uuid(uuid_list[slot-1], (service_uuid_t*)&uuid);
	g_strlcpy(uuid_string, uuid_list[slot-1], BT_UUID_STRING_MAX);
	BT_INFO("Copied UUID string [%s]", uuid_string);

	/* Register GATT Server */
	ret = gatts_register(&uuid);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(uuid_string);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	BT_INFO("GATT Server registration call successfully accepted by OAL..wait for Instance Initialized event from OAL..");
	/* Return & wait for GATT Server Instance Initialization event */
	memset(numapps[slot].sender, 0x00, sizeof(numapps[slot].sender));
	memset(numapps[slot].uuid, 0x00, sizeof(numapps[slot].uuid));

	g_strlcpy(numapps[slot].sender, sender, sizeof(numapps[slot].sender));
	g_strlcpy(numapps[slot].uuid, uuid_string, sizeof(numapps[slot].uuid));

	numapps[slot].is_initialized = 0; /* Set initialization from app registered callback */
	numapps[slot].adv_handle = adv_handle;

	g_free(uuid_string);
	return BLUETOOTH_ERROR_NONE;
}

/* Event handlers */
static void __bt_gatt_handle_pending_request_info(int result,
		int service_function, void *data, unsigned int size)
{
	GSList *l;
	GArray *out_param;
	invocation_info_t *req_info = NULL;
	ret_if(data == NULL);
	BT_DBG("+");

	for (l = _bt_get_invocation_list(); l != NULL; ) {
		req_info = l->data;
		l = g_slist_next(l);
		if (req_info == NULL || req_info->service_function != service_function)
			continue;

		switch (service_function) {
#ifndef GATT_DIRECT
		case BT_GATT_SERVER_REGISTER: {
			bt_service_app_info_t *param = (bt_service_app_info_t*)data;
			BT_INFO("Sender [%s] Request Sender [%s]", param->sender, req_info->sender);
			if (!g_strcmp0(req_info->sender, param->sender)) {
				BT_DBG("GATT Server app found [%s]", req_info->sender);

				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &param->instance_id, sizeof(int));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_GATT_SERVER_START_SERVICE:
		case BT_GATT_SERVER_DELETE_SERVICE: {
			bt_service_app_info_t *param = (bt_service_app_info_t*)data;
			BT_INFO("Sender [%s] Request Sender [%s]", param->sender, req_info->sender);

			int *saved_instance_id = (int*)req_info->user_data;
			if (!g_strcmp0(req_info->sender, param->sender) && param->instance_id == *saved_instance_id) {
				BT_DBG("GATT Server app found [%s] Instance ID [%d] Reply DBUS",
						req_info->sender, *saved_instance_id);

				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &saved_instance_id, sizeof(int));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_GATT_SERVER_ADD_SERVICE:
		case BT_GATT_SERVER_ADD_DESCRIPTOR:
		case BT_GATT_SERVER_ADD_CHARACTERISTIC: {
			int *handle = (int*)data;
			BT_INFO("Characteristic added: Handle [%d]", *handle);
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));

			g_array_append_vals(out_param, handle, sizeof(int));
			_bt_service_method_return(req_info->context, out_param, result);

			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);
			g_array_free(out_param, TRUE);
			break;
		}
#endif
		case BT_CONNECT_LE:
		case BT_DISCONNECT_LE: {
#ifdef TIZEN_GATT_CLIENT
		       char *addr = (char*)req_info->user_data;
		       bluetooth_device_address_t address;

			if (!g_strcmp0(addr, (char*)data)) {
				BT_INFO("GATT Client connect-disconnect call pending for app [%s] addr [%s]",
						req_info->sender, addr);
				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				_bt_convert_addr_string_to_type(address.addr, addr);

				g_array_append_vals(out_param, (bluetooth_device_address_t*)&address,
						sizeof(bluetooth_device_address_t));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
#else
			out_param = g_array_new(FALSE, FALSE, sizeof(gchar));

			g_array_append_vals(out_param, (bluetooth_device_address_t*)data,
					sizeof(bluetooth_device_address_t));
			_bt_service_method_return(req_info->context, out_param, result);

			g_free(req_info->user_data);
			_bt_free_info_from_invocation_list(req_info);

			g_array_free(out_param, TRUE);
#endif
			break;
		}
#ifdef TIZEN_GATT_CLIENT
		case BT_GATT_CLIENT_REGISTER: {
			bt_service_app_info_t *param = (bt_service_app_info_t*)data;
			BT_INFO("Sender [%s] Request Sender [%s]", param->sender, req_info->sender);
			if (!g_strcmp0(req_info->sender, param->sender)) {
				BT_DBG("GATT Client app found [%s] created client ID [%d]",
						req_info->sender, param->client_id);

				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &param->client_id, sizeof(int));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_GATT_GET_PRIMARY_SERVICES: {
			bt_services_browse_info_t *param = (bt_services_browse_info_t*)data;
			char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);
			_bt_convert_addr_type_to_string(address,
					(unsigned char *)(&param->device_addr.addr));

			BT_INFO("Request Sender [%s] address [%s]", req_info->sender, address);

			/* Match address to determine same request */
			if (!g_strcmp0((char*)req_info->user_data, address)) {
				BT_DBG("GATT Client app found [%s] Remote address [%s]",
						req_info->sender, address);

				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, param,
						sizeof(bt_services_browse_info_t));

				//g_array_append_vals(out_param, &param->client_id, sizeof(int));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			g_free(address);
			break;
		}
		case BT_GATT_GET_SERVICE_PROPERTIES: {
			bt_char_browse_info_t param;
			memcpy((void*)&param, data, sizeof(bt_char_browse_info_t));
			//bt_char_browse_info_t *param = (bt_char_browse_info_t*)data;
			BT_INFO("Request Sender [%s]", req_info->sender);
			bluetooth_gatt_client_svc_prop_info_t *prop = (bluetooth_gatt_client_svc_prop_info_t*)req_info->user_data;

			/* Match both address and service properties to determine same request */
			if (!memcmp(param.device_addr.addr,
					prop->device_address.addr,
					sizeof(bluetooth_device_address_t)) &&
					!memcmp(param.svc_uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) &&
					param.svc_inst_id == prop->svc.instance_id) {
				BT_INFO("Returning Service properties");

				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, &param, sizeof(bt_char_browse_info_t));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_GATT_GET_CHARACTERISTIC_PROPERTIES: {
			bt_descriptor_browse_info_t *param = (bt_descriptor_browse_info_t*)data;

			BT_INFO("Request Sender [%s]", req_info->sender);
			bluetooth_gatt_client_char_prop_info_t *prop = (bluetooth_gatt_client_char_prop_info_t*)req_info->user_data;

			/* Match both address, service properties &char properties to determine same request */
			if (!memcmp(param->device_addr.addr,
					prop->device_address.addr,
					sizeof(bluetooth_device_address_t)) &&
					!memcmp(param->svc_uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) &&
					param->svc_inst_id == prop->svc.instance_id &&
					!memcmp(param->char_uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) &&
					param->char_inst_id == prop->characteristic.instance_id) {
				BT_INFO("Returning Characteristic properties");
				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
						g_array_append_vals(out_param, param, sizeof(bt_descriptor_browse_info_t));
				_bt_service_method_return(req_info->context, out_param, result);

				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_GATT_WATCH_CHARACTERISTIC: {
			bt_gatt_notif_reg_info_t *param = (bt_gatt_notif_reg_info_t*)data;
			bluetooth_gatt_client_char_prop_info_t *prop = (bluetooth_gatt_client_char_prop_info_t*)req_info->user_data;
			BT_INFO("Request Sender [%s]", req_info->sender);
			/* Match both address, service properties &char properties to determine same request */
			if (!memcmp(param->addr.addr,
					prop->device_address.addr,
					sizeof(bluetooth_device_address_t)) &&
					!memcmp(param->svc_uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) &&
					param->svc_inst == prop->svc.instance_id &&
					!memcmp(param->char_uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) &&
					param->char_inst == prop->characteristic.instance_id) {
				BT_INFO("Characteristic Watch Successful: Is registered [%d]",
						param->is_registered);
				out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				g_array_append_vals(out_param, param, sizeof(bt_gatt_notif_reg_info_t));
				_bt_service_method_return(req_info->context, out_param, result);
				g_free(req_info->user_data);
				_bt_free_info_from_invocation_list(req_info);
				g_array_free(out_param, TRUE);
			}
			break;
		}
		case BT_GATT_READ_CHARACTERISTIC:
		case BT_GATT_WRITE_CHARACTERISTIC_VALUE_BY_TYPE: {
			 bluetooth_gatt_client_char_prop_info_t *param = (bluetooth_gatt_client_char_prop_info_t*)data;
			 BT_INFO("Request Sender [%s]", req_info->sender);

			 char *addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
			 bluetooth_gatt_client_char_prop_info_t *prop = (bluetooth_gatt_client_char_prop_info_t*)req_info->user_data;
			 _bt_convert_addr_type_to_string(addr, prop->device_address.addr);

			 if (!memcmp(param, prop, sizeof(bluetooth_gatt_client_char_prop_info_t))) {
				 BT_INFO("Gatt Char read or write request matched for address [%s]", addr);
				 out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				 g_array_append_vals(out_param, param, sizeof(bluetooth_gatt_client_char_prop_info_t));
				 _bt_service_method_return(req_info->context, out_param, result);

				 g_free(req_info->user_data);
				 _bt_free_info_from_invocation_list(req_info);
				 g_array_free(out_param, TRUE);
			}
			g_free(addr);
			 break;
		}
		case BT_GATT_READ_DESCRIPTOR_VALUE:
		case BT_GATT_WRITE_DESCRIPTOR_VALUE: {
			 bluetooth_gatt_client_desc_prop_info_t *param = (bluetooth_gatt_client_desc_prop_info_t*)data;
			 BT_INFO("Request Sender [%s]", req_info->sender);

			 char *addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
			 bluetooth_gatt_client_desc_prop_info_t *prop = (bluetooth_gatt_client_desc_prop_info_t*)req_info->user_data;
			 _bt_convert_addr_type_to_string(addr, prop->device_address.addr);

			 if (!memcmp(param, prop, sizeof(bluetooth_gatt_client_desc_prop_info_t))) {
				 BT_INFO("Descriptor read or write request matched for address [%s]", addr);
				 out_param = g_array_new(FALSE, FALSE, sizeof(gchar));
				 g_array_append_vals(out_param, param, sizeof(bluetooth_gatt_client_desc_prop_info_t));
				 _bt_service_method_return(req_info->context, out_param, result);

				 g_free(req_info->user_data);
				 _bt_free_info_from_invocation_list(req_info);
				 g_array_free(out_param, TRUE);
			}
			g_free(addr);
			break;
		}
#endif
		default:
			break;
		}
	}
	BT_DBG("-");
}

static void __bt_handle_server_instance_registered(event_gatts_register_t *data)
{
	bt_service_app_info_t *info = NULL;
	int k;
	char *uuid_string = g_malloc0(BT_UUID_STRING_MAX);

	_bt_uuid_to_string(&(data->server_uuid), uuid_string);
	BT_INFO("Instance ID is Intialized [%d] UUID initialized [%s]", data->server_inst, uuid_string);

	/* Check if the just registered Instance ID belongs to BLEAPP or GATT Server */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (g_strcmp0(info->uuid, uuid_string) == 0) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			info->is_initialized = TRUE;
			info->instance_id = data->server_inst;
			info->adv_instance = data->server_inst;
			__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_NONE, BT_GATT_SERVER_REGISTER,
					(void*)info, sizeof(bt_service_app_info_t));
			break;
		}
	}
	g_free(uuid_string);
}

static void __bt_handle_gatt_server_service_added(event_gatts_srvc_prm_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	int svc_handle = 0;
	int *handle;
	int count;
	bt_service_app_info_t *info = NULL;
	int k;

	char *uuid_str = g_malloc0(BT_UUID_STRING_MAX);
	_bt_uuid_to_string(&(event->gatt_srvc_id.id.uuid), uuid_str);
	BT_INFO("GATT Added Service  UUID: [%s] Server Ins [%d] Service hdl [%d] Is Primary [%d]\n",
			uuid_str, event->gatt_srvc_stat.server_inst, svc_handle, event->gatt_srvc_id.is_prmry);

	if (event->gatt_srvc_stat.status != OAL_STATUS_SUCCESS) {
		result = BLUETOOTH_ERROR_INTERNAL;
		svc_handle = 0;  /* Service handle set to 0 indicates.
				    0 is reserved by BT SIG, cant be used by app*/
	} else
		svc_handle = event->gatt_srvc_stat.servic_hndl;

	BT_INFO("GATT Added Service  Status [%d] Server Ins [%d] Service hdl [%d] Is Primary [%d]\n",
			event->gatt_srvc_stat.status, event->gatt_srvc_stat.server_inst, svc_handle, event->gatt_srvc_id.is_prmry);

	/* Check if the just registered Instance ID belongs to BLEAPP or GATT Server */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->gatt_srvc_stat.server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_GATT_SERVER_ADD_SERVICE,
					(int*)&svc_handle, sizeof(int));

			/* Add Service Handle */
			if (svc_handle > 0) {
				handle = g_malloc0(sizeof(int));
				*handle = svc_handle;
				numapps[event->gatt_srvc_stat.server_inst].service_handles = g_slist_append(numapps[event->gatt_srvc_stat.server_inst].service_handles, handle);
				count = g_slist_length(numapps[event->gatt_srvc_stat.server_inst].service_handles);
				BT_INFO("Added Service handle [%d] to list..current count [%d]", svc_handle, count);
			}
			break;
		}
	}

	g_free(uuid_str);
}

static void __bt_handle_gatt_server_characteristic_added(event_gatts_srvc_charctr_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	int char_handle;
	bt_service_app_info_t *info = NULL;
	int k;
	char *uuid_str = g_malloc0(BT_UUID_STRING_MAX);

	BT_INFO("GATT Server Char added status [%d]", event->gatt_srvc_stat.status);
	BT_INFO("GATT Server_inst: [%d]", event->gatt_srvc_stat.server_inst);
	BT_INFO("GATT Service Handle: [%d]", event->gatt_srvc_stat.servic_hndl);
	BT_INFO("GATT Add characteristic  Status: [%d]", event->gatt_srvc_stat.status);
	BT_INFO("GATT Service characteristic Handle: [%d]", event->charctr_hndl);

	if (event->gatt_srvc_stat.status != OAL_STATUS_SUCCESS) {
		result = BLUETOOTH_ERROR_INTERNAL;
		char_handle = 0;  /* characteristic handle set to 0 indicates.
				     0 is reserved by BT SIG, cant be used by app*/
	} else
		char_handle = event->charctr_hndl;


	/* Check if the just registered Instance ID belongs to BLEAPP or GATT Server */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->gatt_srvc_stat.server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_GATT_SERVER_ADD_CHARACTERISTIC,
					(int*)&char_handle, sizeof(int));
			break;
		}
	}


	_bt_uuid_to_string(&(event->charctr_uuid), uuid_str);
	BT_INFO("GATT Added Characteristic:  UUID: [%s]", uuid_str);

	g_free(uuid_str);
}

static void __bt_handle_gatt_server_descriptor_added(event_gatts_srvc_descr_t* event)
{
	int result = BLUETOOTH_ERROR_NONE;
	int desc_handle;
	bt_service_app_info_t *info = NULL;
	int k;
	char *uuid_str = g_malloc0(BT_UUID_STRING_MAX);

	BT_INFO("GATT Server Descriptor added status [%d]", event->gatt_srvc_stat.status);
	BT_INFO("GATT Server_inst: [%d]", event->gatt_srvc_stat.server_inst);
	BT_INFO("GATT Service Handle: [%d]", event->gatt_srvc_stat.servic_hndl);
	BT_INFO("GATT Add Descriptor  Status: [%d]", event->gatt_srvc_stat.status);
	BT_INFO("GATT Service Descriptor Handle: [%d]", event->descrptr_hndl);

	if (event->gatt_srvc_stat.status != OAL_STATUS_SUCCESS) {
		result = BLUETOOTH_ERROR_INTERNAL;
		desc_handle = 0;  /* Service handle set to 0 indicates.
				     0 is reserved by BT SIG, cant be used by app*/
	} else
		desc_handle = event->descrptr_hndl;

	/* Check if the just registered Instance ID belongs to BLEAPP or GATT Server */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->gatt_srvc_stat.server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_GATT_SERVER_ADD_DESCRIPTOR,
					(int*)&desc_handle, sizeof(int));
			break;
		}
	}

	_bt_uuid_to_string(&(event->descrptr_uuid), uuid_str);
	BT_INFO("GATT Added Descriptor:  UUID: [%s]", uuid_str);

	g_free(uuid_str);
}

static void __bt_handle_gatt_server_service_started(event_gatts_srvc_t *event)
{
	bt_service_app_info_t *info = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	int k;
	BT_INFO("GATT Server Service Started..");


	BT_INFO("GATT Server_inst: [%d]", event->server_inst);
	BT_INFO("GATT Service Handle: [%d]", event->servic_hndl);
	BT_INFO("GATT Service Started Status: [%d]", event->status);

	if (event->status != OAL_STATUS_SUCCESS) {
		BT_ERR("GATT Server Service Start Failed Err: [%d]", event->status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	/* Check if the just registered Instance ID belongs to requester */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_GATT_SERVER_START_SERVICE,
					(void*)info, sizeof(bt_service_app_info_t));
			break;
		}
	}
}

static void __bt_handle_gatt_server_service_stopped(event_gatts_srvc_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	bt_service_app_info_t *info = NULL;
	int k;
	BT_INFO("GATT Server Service Stopped..");

	BT_INFO("GATT Server_inst: [%d]", event->server_inst);
	BT_INFO("GATT Service Handle: [%d]", event->servic_hndl);
	BT_INFO("GATT Service Stopped Status: [%d]", event->status);

	if (event->status != OAL_STATUS_SUCCESS) {
		BT_ERR("GATT Server Service Stop Failed Err: [%d]", event->status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	/* Check if the just registered Instance ID belongs to requester */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_GATT_SERVER_STOP_SERVICE,
					(void*)info, sizeof(bt_service_app_info_t));
			break;
		}
	}
}

static void __bt_handle_gatt_server_service_deleted(event_gatts_srvc_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	bt_service_app_info_t *info = NULL;
	int k;
	BT_INFO("GATT Server Service Deleted..");

	BT_INFO("GATT Server_inst: [%d]", event->server_inst);
	BT_INFO("GATT Service Handle: [%d]", event->servic_hndl);
	BT_INFO("GATT Service Deleted Status: [%d]", event->status);

	if (event->status != OAL_STATUS_SUCCESS) {
		BT_ERR("GATT Server Service Delete Failed Err: [%d]", event->status);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	/* Check if the just registered Instance ID belongs to requester */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_GATT_SERVER_DELETE_SERVICE,
					(void*)info, sizeof(bt_service_app_info_t));
			break;
		}
	}
}

static struct gatt_client_info_t *__bt_find_remote_gatt_client_info(char *address)
{
	GSList *l;
	struct gatt_client_info_t *info = NULL;
	for (l = gatt_client_info_list; l != NULL; l = g_slist_next(l)) {
		info = (struct gatt_client_info_t*)l->data;
		if (info == NULL)
			continue;

		if (!g_strcmp0(info->addr, address)) {
			BT_INFO("Remote GATT client found addr[%s]", info->addr);
			return info;
		}
	}
	return NULL;
}

static struct gatt_client_info_t *__bt_find_remote_gatt_client_info_from_conn_id(int conn_id)
{
	GSList *l;
	struct gatt_client_info_t *info = NULL;

	for (l = gatt_client_info_list; l != NULL; l = g_slist_next(l)) {
		info = (struct gatt_client_info_t*)l->data;
		if (info == NULL)
			continue;

		if (info->connection_id == conn_id) {
			BT_INFO("Remote GATT client found addr[%s]", info->addr);
			return info;
		}
	}
	return NULL;
}

#ifdef TIZEN_GATT_CLIENT
static struct gatt_server_info_t *__bt_find_remote_gatt_server_info(char *address)
{
	GSList *l;
	struct gatt_server_info_t *info = NULL;
	for (l = gatt_server_info_list; l != NULL; l = g_slist_next(l)) {
		info = (struct gatt_server_info_t*)l->data;
		if (info == NULL)
			continue;

		if (!g_strcmp0(info->addr, address)) {
			BT_INFO("Remote GATT Server found addr[%s]", info->addr);
			return info;
		}
	}
	return NULL;
}

static struct gatt_out_conn_info_t* __bt_find_gatt_outgoing_conn_info(char *address)
{
	GSList *l;
	struct gatt_out_conn_info_t *info = NULL;
	for (l = outgoing_gatt_conn_list; l != NULL; l = g_slist_next(l)) {
		info = (struct gatt_out_conn_info_t*)l->data;
		if (info == NULL)
			continue;

		if (!g_strcmp0(info->addr, address)) {
			BT_INFO("Outgoing connection info found addr[%s]", info->addr);
			return info;
		}
	}
	return NULL;
}

static struct gatt_server_info_t *__bt_find_remote_gatt_server_info_from_client_if(int client_if)
{
	GSList *l;
	struct gatt_server_info_t *info = NULL;

	for (l = gatt_server_info_list; l != NULL; l = g_slist_next(l)) {
		info = (struct gatt_server_info_t*)l->data;
		if (info == NULL)
			continue;

		if (info->client_id == client_if) {
			BT_INFO("Remote GATT server found addr[%s]", info->addr);
			return info;
		}
	}
	return NULL;
}
#endif




#ifdef TIZEN_GATT_CLIENT
static void __bt_handle_gatt_server_connection_state(event_gatts_conn_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_client_info_t *conn_info = NULL;
	bluetooth_device_address_t dev_addr;
	GVariant *param = NULL;

	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);

	memcpy(dev_addr.addr, event->address.addr, 6);

	/* REPLY dbus Context to application which called BT_CONNECT_LE. There is status
	   in this event */
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	if (event->status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;

	__bt_gatt_handle_pending_request_info(result, BT_CONNECT_LE,
			address, BT_ADDRESS_STRING_SIZE);

	BT_INFO("GATT Server Connedted: Remote Client addr[%s] Server Instance [%d] Connection ID [%d]",
			address, event->server_inst, event->conn_id);


	/* Check if device is already in connected list */
	conn_info = __bt_find_remote_gatt_client_info(address);

	if (!conn_info) {
		BT_INFO("Conn Info absent: But no need to Send Local GATT Server Connected event to apps");

		param = g_variant_new("(is)", result, address);

		/* Send event to application */
		_bt_send_event(BT_DEVICE_EVENT,
				BLUETOOTH_EVENT_GATT_SERVER_CONNECTED, /* Local device is GATT server */
				param);

		/* Save Connection info */
		conn_info = g_new0(struct gatt_client_info_t, 1);
		conn_info->addr = g_strdup(address);
		BT_INFO("Added GATT client addr[%s]", conn_info->addr);
		conn_info->connection_id = event->conn_id;
		conn_info->instance_id = event->server_inst;
		gatt_client_info_list = g_slist_append(gatt_client_info_list, conn_info);
		BT_INFO("Total num of connected Remote GATT Clients [%d]", g_slist_length(gatt_client_info_list));
	}

	g_free(address);
}

/* GATT Server Dis connected */
static void __bt_handle_gatt_server_disconnection_state(event_gatts_conn_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_client_info_t *conn_info = NULL;
	bluetooth_device_address_t dev_addr;
	GVariant *param = NULL;
	char address[BT_ADDRESS_STRING_SIZE];

	memcpy(dev_addr.addr, event->address.addr, 6);

	/* REPLY dbus Context to application which called BT_DISCONNECT_LE. There is status
	   in this event */
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	if (event->status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;

	if (NULL ==  _bt_get_request_info_data(BT_DISCONNECT_LE, address)) {
		if (NULL !=  _bt_get_request_info_data(BT_CONNECT_LE, address)) {
			result = BLUETOOTH_ERROR_INTERNAL;

			__bt_gatt_handle_pending_request_info(result, BT_CONNECT_LE,
					address, BT_ADDRESS_STRING_SIZE);

			BT_ERR("Failed to connect Local GATT Server Remote Client addr[%s]", address);
			return;
		}
	} else {
		__bt_gatt_handle_pending_request_info(result, BT_DISCONNECT_LE,
				address, BT_ADDRESS_STRING_SIZE);
	}

	BT_INFO("Local GATT Server DisConnected: Remote Client addr[%s] Server Instance [%d] Connection ID [%d]",
			address, event->server_inst, event->conn_id);
	/* Remove Connection info */
	conn_info = __bt_find_remote_gatt_client_info(address);
	if (conn_info) {
		BT_INFO("No need to Send Local GATT Server Disconnected event to apps, just remove remote client info");

		param = g_variant_new("(is)", result, address);
		/* Send event to application */
		_bt_send_event(BT_DEVICE_EVENT,
				BLUETOOTH_EVENT_GATT_SERVER_DISCONNECTED, /* Local device is GATT server */
				param);

		/* Remove info from List */
		gatt_client_info_list = g_slist_remove(gatt_client_info_list, conn_info);
		BT_INFO("Total num of connected GATT clients [%d]", g_slist_length(gatt_client_info_list));
		g_free(conn_info->addr);
		g_free(conn_info);
	}
}
#else

static void __bt_handle_gatt_server_connection_state(event_gatts_conn_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_client_info_t *conn_info = NULL;
	bt_service_app_info_t *info = NULL;
	bluetooth_device_address_t dev_addr;
	GVariant *param = NULL;
	int k;

	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);

	memcpy(dev_addr.addr, event->address.addr, 6);


	/* Check if the just registered Instance ID belongs to BLEAPP or GATT Server */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_CONNECT_LE,
					(bluetooth_device_address_t*)&dev_addr, sizeof(bluetooth_device_address_t));

			break;
		}
	}

	/* Send event to BT-API */
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	BT_INFO("GATT Server Connedted: Remote Client addr[%s] Server Instance [%d] Connection ID [%d]",
			address, event->server_inst, event->conn_id);


	/* Check if device is already in connected list */
	conn_info = __bt_find_remote_gatt_client_info(address);

	if (!conn_info) {
		param = g_variant_new("(is)", result, address);
		/* Send event to application */
		_bt_send_event(BT_GATT_SERVER_EVENT,
				BLUETOOTH_EVENT_GATT_CONNECTED,
				param);

		/* Save Connection info */
		conn_info = g_new0(struct gatt_client_info_t, 1);
		conn_info->addr = g_strdup(address);
		BT_INFO("Added GATT client addr[%s]", conn_info->addr);
		conn_info->connection_id = event->conn_id;
		conn_info->instance_id = event->server_inst;
		gatt_client_info_list = g_slist_append(gatt_client_info_list, conn_info);
		BT_INFO("Total num of connected GATT clients [%d]", g_slist_length(gatt_client_info_list));
	}

	g_free(address);
}

/* GATT Server Dis connected */
static void __bt_handle_gatt_server_disconnection_state(event_gatts_conn_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_client_info_t *conn_info = NULL;
	bluetooth_device_address_t dev_addr;
	bt_service_app_info_t *info = NULL;
	GVariant *param = NULL;
	int k;

	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);

	memcpy(dev_addr.addr, event->address.addr, 6);

	/* Check if the just registered Instance ID belongs to BLEAPP or GATT Server */
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (info->instance_id == event->server_inst) {
			BT_INFO("Found GATT Server.. UUID [%s], sender [%s]", info->uuid, info->sender);
			__bt_gatt_handle_pending_request_info(result, BT_DISCONNECT_LE,
					(bluetooth_device_address_t*)&dev_addr, sizeof(bluetooth_device_address_t));

			break;
		}
	}

	/* Send event to BT-API */
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	BT_INFO("GATT Server DisConnected: Remote Client addr[%s] Server Instance [%d] Connection ID [%d]",
			address, event->server_inst, event->conn_id);

	param = g_variant_new("(is)", result, address);
	/* Send event to application */
	_bt_send_event(BT_GATT_SERVER_EVENT,
			BLUETOOTH_EVENT_GATT_DISCONNECTED,
			param);

	/* Remove Connection info */
	conn_info = __bt_find_remote_gatt_client_info(address);
	if (conn_info) {
		BT_INFO("Remove GATT client info from List..");
		/* Remove info from List */
		gatt_client_info_list = g_slist_remove(gatt_client_info_list, conn_info);
		BT_INFO("Total num of connected GATT clients [%d]", g_slist_length(gatt_client_info_list));
		g_free(conn_info->addr);
		g_free(conn_info);
	}

	g_free(address);
}
#endif


static void __bt_handle_gatt_server_acquire_write_requested(event_gatts_srvc_acquire_attr_t *event)
{
	GVariant *param = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_req_info *req_info = NULL;
	bluetooth_device_address_t dev_addr;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	BT_INFO("GATT Server ACQUIRE  Write Req Connection ID: [%d]", event->attr_trans.conn_id);
	BT_INFO("GATT Server  ACQUIRE Write Req Transaction ID:[%d]", event->attr_trans.trans_id);
	BT_INFO("GATT Server ACQUIRE Write Req Attribute Handle: [%d]", event->attr_trans.attr_handle);

	//address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	memcpy(dev_addr.addr, event->address.addr, 6);

	req_info = g_new0(struct gatt_server_req_info, 1);
	req_info->request_id = event->attr_trans.trans_id;
	req_info->attribute_handle = event->attr_trans.attr_handle;
	req_info->connection_id = event->attr_trans.conn_id;
	req_info->request_type = BLUETOOTH_GATT_REQUEST_TYPE_ACQUIRE_WRITE;
	gatt_server_requests = g_slist_append(gatt_server_requests, req_info);

	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	param = g_variant_new("(iiiiiis)", result,
			event->attr_trans.conn_id,
			event->attr_trans.trans_id,
			event->attr_trans.attr_handle,
			event->mtu, event->attr_trans.offset, address);
	BT_INFO("GATT Server ACQUIRE Write Req Attribute remote address : [%s]", address);

	_bt_send_event(BT_GATT_SERVER_EVENT,
			BLUETOOTH_EVENT_GATT_SERVER_ACQUIRE_WRITE,
			param);

}

static void __bt_handle_gatt_server_acquire_notify_requested(event_gatts_srvc_acquire_attr_t *event)
{
	GVariant *param = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_req_info *req_info = NULL;
	bluetooth_device_address_t dev_addr;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };

	BT_INFO("GATT Server ACQUIRE  Notify Req Connection ID: [%d]", event->attr_trans.conn_id);
	BT_INFO("GATT Server ACQUIRE Notify  Req Transaction ID:[%d]", event->attr_trans.trans_id);
	BT_INFO("GATT Server ACQUIRE Notify Req Attribute Handle: [%d]", event->attr_trans.attr_handle);
	BT_INFO("GATT Server ACQUIRE notify Req address : [%s]", address);

	memcpy(dev_addr.addr, event->address.addr, 6);
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);
	BT_INFO("GATT Server ACQUIRE notify Req remote address : [%s]", address);

	req_info = g_new0(struct gatt_server_req_info, 1);
	req_info->request_id = event->attr_trans.trans_id;
	req_info->attribute_handle = event->attr_trans.attr_handle;
	req_info->connection_id = event->attr_trans.conn_id;
	req_info->request_type = BLUETOOTH_GATT_REQUEST_TYPE_ACQUIRE_NOTIFY;
	gatt_server_requests = g_slist_append(gatt_server_requests, req_info);

	param = g_variant_new("(iiiiiis)", result,
			event->attr_trans.conn_id,
			event->attr_trans.trans_id,
			event->attr_trans.attr_handle,
			event->mtu, event->attr_trans.offset,
			address);

	BT_INFO("GATT Server ACQUIRE Notify  Req Attribute : ");

	 _bt_send_event(BT_GATT_SERVER_EVENT,
				 BLUETOOTH_EVENT_GATT_SERVER_ACQUIRE_NOTIFY,
				 param);
}

static void __bt_handle_gatt_server_write_requested(event_gatts_srvc_write_attr_t *event)
{
	char *address;
	bluetooth_device_address_t dev_addr;
	GVariant *param = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	gboolean need_resp;
	gboolean is_prepare_write;
	char *write_val;
	GVariant *data = NULL;
	int i;
	struct gatt_server_req_info *req_info = NULL;
	BT_INFO("GATT Server Write Requested");

	memcpy(dev_addr.addr, event->address.addr, 6);

	BT_INFO("GATT Server Write Req Connection ID: [%d]", event->attr_trans.conn_id);
	BT_INFO("GATT Server Write Req Transaction ID:[%d]", event->attr_trans.trans_id);
	BT_INFO("GATT Server Write Req Attribute Handle: [%d]", event->attr_trans.attr_handle);
	BT_INFO("GATT Server Write Req Attribute Offset: [%d]", event->attr_trans.offset);
	BT_INFO("GATT Server Write Req value length [%d]", event->length);
	BT_INFO("GATT Server Write Req needs response: [%d]", event->need_rsp);
	BT_INFO("GATT Server Write Req Is Prep: [%d]", event->is_prep);

	if (event->length <= 0) {
		BT_INFO("GATT Server write requested, but length of attr value is 0");
		return;
	}

	need_resp = event->need_rsp;
	is_prepare_write = event->is_prep;

	write_val = g_memdup(&event->value[0], event->length);

	address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	BT_INFO("GATT Server Write Request from remote client [%s]", address);
	/* DEBUG */
	if (event->length > 0) {
		for (i = 0; i < event->length; i++)
			BT_DBG("Data[%d] = [0x%x]", i, event->value[i]);
	}
	/* Save Write Request Info */
	req_info = g_new0(struct gatt_server_req_info, 1);
	req_info->request_id = event->attr_trans.trans_id;
	req_info->attribute_handle = event->attr_trans.attr_handle;
	req_info->connection_id = event->attr_trans.conn_id;
	req_info->addr = address;
	req_info->offset = event->attr_trans.offset;
	req_info->request_type = BLUETOOTH_GATT_REQUEST_TYPE_WRITE;
	gatt_server_requests = g_slist_append(gatt_server_requests, req_info);

	data = g_variant_new_from_data(
			G_VARIANT_TYPE_BYTESTRING,
			write_val,
			event->length,
			TRUE, NULL, NULL);

	param = g_variant_new("(iiiiiibbsn@ay)", result,
			event->attr_trans.conn_id,
			event->attr_trans.trans_id,
			event->attr_trans.attr_handle,
			event->attr_trans.offset,
			event->length,
			need_resp,
			is_prepare_write,
			address,
			event->length,
			data);

	 _bt_send_event(BT_GATT_SERVER_EVENT,
			 BLUETOOTH_EVENT_GATT_SERVER_VALUE_CHANGED,
			param);

	g_free(write_val);
}

static void __bt_handle_gatt_server_read_requested(event_gatts_srvc_read_attr_t *event)
{
	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	bluetooth_device_address_t dev_addr;
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_req_info *req_info = NULL;
	GVariant *param = NULL;
	gboolean is_long;
	BT_INFO("GATT Server Read Requested");

	memcpy(dev_addr.addr, event->address.addr, 6);

	BT_INFO("GATT Server-Client Connection ID: [%d]", event->attr_trans.conn_id);
	BT_INFO("GATT Server-Client Transaction ID: [%d]", event->attr_trans.trans_id);
	BT_INFO("GATT Server Attribute Handle: [%d]", event->attr_trans.attr_handle);
	BT_INFO("GATT Server Attribute Offset for read: [%d]", event->attr_trans.offset);
	BT_INFO("GATT Server Attribute is long: [%d]", event->is_long);

	is_long = event->is_long;

	/* Save Read Request Info */
	req_info = g_new0(struct gatt_server_req_info, 1);
	req_info->request_id = event->attr_trans.trans_id;
	req_info->attribute_handle = event->attr_trans.attr_handle;
	req_info->connection_id = event->attr_trans.conn_id;
	req_info->addr = address;
	req_info->offset = event->attr_trans.offset;
	req_info->request_type = BLUETOOTH_GATT_REQUEST_TYPE_READ;
	gatt_server_requests = g_slist_append(gatt_server_requests, req_info);

	/* Send event to BT-API */
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	BT_INFO("GATT Server Read Request from remote client [%s]", address);

	param = g_variant_new("(iiiiibs)", result,
			event->attr_trans.conn_id,
			event->attr_trans.trans_id,
			event->attr_trans.attr_handle,
			event->attr_trans.offset,
			is_long,
			address);

	_bt_send_event(BT_GATT_SERVER_EVENT,
			BLUETOOTH_EVENT_GATT_SERVER_READ_REQUESTED,
			param);
}

static void __bt_handle_gatt_server_indicate_confirmed(event_gatts_ind_cnfrm_t *event)
{
	bluetooth_device_address_t dev_addr;
	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	int cur_connected_clients;
	static int recvd = 0;
	gboolean completed = 0;
	GVariant *param = NULL;

	/* OAL event does provide error, so MW assumes event will never contain wrong data,
	   incase of any issues, check with OAL */
	int result = BLUETOOTH_ERROR_NONE;

	memcpy(dev_addr.addr, event->address.addr, 6);
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	BT_INFO("Indication sent to GATT client [%s] conn_ ID [%d] transaction ID [%d] Att handle [%d]",
			address, event->conn_id, event->trans_id, event->attr_handle);


	cur_connected_clients = g_slist_length(gatt_client_info_list);
	BT_INFO("Number of connected clients during sending Indication [%d] & current connected count [%d]",
			num_indicate_clients, cur_connected_clients);

	recvd++;
	if (recvd == num_indicate_clients) {
		BT_INFO("Gatt indication confirm event for last GATT client.. [%s]", address);
		completed = 1; /* Last event */
		recvd = 0; /* Reset */
		num_indicate_clients = 0;
	}

	param = g_variant_new("(isib)",
			result,
			address,
			event->attr_handle,
			completed);

	/* Send event to BT-API */
	_bt_send_event(BT_GATT_SERVER_EVENT,
			BLUETOOTH_EVENT_GATT_SERVER_NOTIFICATION_COMPLETED,
			param);

	BT_INFO("Received Indication confirm for client number [%d]", recvd);
	g_free(address);
}

/* Tizen Platform Specific */
static void __bt_handle_gatt_server_notification_changed(event_gatts_notif_t *event)
{
	bluetooth_device_address_t dev_addr;
	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	GVariant *param = NULL;
	gboolean notify;

	/* OAL event does provide error, so MW assumes event will never contain wrong data,
	   incase of any issues, check with OAL */
	int result = BLUETOOTH_ERROR_NONE;

	memcpy(dev_addr.addr, event->address.addr, 6);
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)dev_addr.addr);

	BT_INFO("notification_changed [%s] conn_ ID [%d] transaction ID [%d] Att handle [%d] Notify[%d]",
			address, event->conn_id, event->trans_id, event->attr_handle, event->notify);

	/* Set Notifcation status */
	notify = event->notify;

	param = g_variant_new("(isib)",
			result,
			address,
			event->attr_handle,
			notify);

	/* Send event to BT-API */
	_bt_send_event(BT_GATT_SERVER_EVENT,
			BLUETOOTH_EVENT_GATT_SERVER_NOTIFICATION_STATE_CHANGED,
			param);

	g_free(address);
}

static void __bt_handle_gatt_mtu_changed_event(event_gatts_mtu_changed_t *event)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_client_info_t *conn_info = NULL;
	GVariant *param = NULL;
	guint8 status = 0;
	BT_INFO("GATT Server MTU changed event");

	conn_info = __bt_find_remote_gatt_client_info_from_conn_id(event->conn_id);
	if (conn_info == NULL) {
		BT_ERR("Cant find connection Information");
		return;
	}
	BT_INFO("Got connection Info GATT client [%s] MTU Size [%d]",
			conn_info->addr, event->mtu_size);

	param = g_variant_new("(isqy)",
			result,
			conn_info->addr,
			event->mtu_size,
			status);

	/* Send event to BT-API */
	_bt_send_event(BT_GATT_SERVER_EVENT,
			BLUETOOTH_EVENT_GATT_ATT_MTU_CHANGED,
			param);
}

static void __bt_gatt_event_handler(int event_type, gpointer event_data)
{
	BT_INFO("OAL event = 0x%x, \n", event_type);
	switch (event_type) {
		case OAL_EVENT_BLE_SERVER_INSTANCE_INITIALISED: {
		BT_INFO("OAL Event: Server Instance Registered");
		/* GATT Server Registered event is handled in MAIN thread context */
		__bt_handle_server_instance_registered((event_gatts_register_t *)event_data);
		break;
								}
	case OAL_EVENT_GATTS_SERVICE_ADDED: {
		BT_INFO("OAL Event: GATT Service added");
		__bt_handle_gatt_server_service_added((event_gatts_srvc_prm_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_CHARACTERISTIC_ADDED: {
		BT_INFO("OAL Event: GATT characteristic added");
		__bt_handle_gatt_server_characteristic_added((event_gatts_srvc_charctr_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_DESCRIPTOR_ADDED: {
		BT_INFO("OAL Event: GATT descriptor added");
		__bt_handle_gatt_server_descriptor_added((event_gatts_srvc_descr_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_SERVICE_STARTED: {
		BT_INFO("OAL Event: GATT Service started");
		__bt_handle_gatt_server_service_started((event_gatts_srvc_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_SERVICE_STOPED: {
		BT_INFO("OAL Event: GATT Service stopped");
		__bt_handle_gatt_server_service_stopped((event_gatts_srvc_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_SERVICE_DELETED: {
		BT_INFO("OAL Event: GATT Service deleted");
		__bt_handle_gatt_server_service_deleted((event_gatts_srvc_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTS_CONNECTION_COMPLETED: {
		BT_INFO("OAL Event: GATT Server Connected");
		__bt_handle_gatt_server_connection_state((event_gatts_conn_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_DISCONNECTION_COMPLETED: {
		BT_INFO("OAL Event: GATT Server Disconnected");
		__bt_handle_gatt_server_disconnection_state((event_gatts_conn_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_REQUEST_READ: {
		BT_INFO("OAL Event: GATT Server Read Request");
		__bt_handle_gatt_server_read_requested((event_gatts_srvc_read_attr_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_REQUEST_WRITE: {
		BT_INFO("OAL Event: GATT Server Write Request");
		__bt_handle_gatt_server_write_requested((event_gatts_srvc_write_attr_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_REQUEST_ACQUIRE_WRITE: {
		BT_INFO("OAL Event: GATT Server Acquire  Write Request");
		__bt_handle_gatt_server_acquire_write_requested((event_gatts_srvc_acquire_attr_t*)event_data);
		break;
	}
	case OAL_EVENT_GATTS_REQUEST_ACQUIRE_NOTIFY: {
		BT_INFO("OAL Event: GATT ServerAcquire Notify  Request");
		__bt_handle_gatt_server_acquire_notify_requested((event_gatts_srvc_acquire_attr_t*)event_data);
		break;
	}
	case OAL_EVENT_GATTS_IND_CONFIRM: {
		BT_INFO("OAL Event: GATT Server Indication confirmed");
		__bt_handle_gatt_server_indicate_confirmed((event_gatts_ind_cnfrm_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_NOTIFICATION: { /* Tizen Platform Specific */
		BT_INFO("OAL Event: GATT Server DisConnected");
		__bt_handle_gatt_server_notification_changed((event_gatts_notif_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTS_MTU_CHANGED: {
		BT_INFO("OAL Event: GATT Server MTU changed event callback");
		__bt_handle_gatt_mtu_changed_event((event_gatts_mtu_changed_t *)event_data);
		break;
	}
#ifdef TIZEN_GATT_CLIENT
	case OAL_EVENT_GATTC_REGISTRATION: {
		BT_INFO("OAL Event: GATT Client instance Registered");
		__bt_handle_client_instance_registered((event_gattc_register_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_CONNECTION_COMPLETED: {
		BT_INFO("OAL Event: GATT Client Connected");
		__bt_handle_client_connected((event_gattc_conn_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_DISCONNECTION_COMPLETED: {
		BT_INFO("OAL Event: GATT Client DisConnected");
		__bt_handle_client_disconnected((event_gattc_conn_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_SERVICE_SEARCH_RESULT: {
		BT_INFO("OAL Event: GATT Client Service Search Result");
		__bt_handle_client_service_search_result((event_gattc_service_result_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_SERVICE_SEARCH_DONE: {
		BT_INFO("OAL Event: GATT Client Service Completed");
		__bt_handle_client_service_search_completed((event_gattc_conn_status_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_CHARAC_SERACH_RESULT: {
		BT_INFO("OAL Event: GATT Client Characteristic Search Result");
		__bt_handle_client_characteristic_search_result((event_gattc_characteristic_result_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_DESC_SERACH_RESULT: {
		BT_INFO("OAL Event: GATT Client Descriptor Search Result");
		__bt_handle_client_descriptor_search_result((event_gattc_descriptor_result_t *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_READ_CHARAC: {
		BT_INFO("OAL Event: GATT Client Characteristic Read Data");
		__bt_handle_client_characteristic_read_data((event_gattc_read_data *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_READ_DESCR: {
		BT_INFO("OAL Event: GATT Client Descriptor Read Data");
		__bt_handle_client_descriptor_read_data((event_gattc_read_data *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_WRITE_CHARAC: {
		BT_INFO("OAL Event: GATT Client Characteristic Write Data");
		__bt_handle_client_characteristic_write_data((event_gattc_write_data *) event_data);
		break;
	}
	case OAL_EVENT_GATTC_WRITE_DESCR: {
		BT_INFO("OAL Event: GATT Client Descriptor Write Data");
		__bt_handle_client_descriptor_write_data((event_gattc_write_data *) event_data);
		break;
	}
	case OAL_EVENT_DEVICE_LE_DISCONNECTED: {
		BT_INFO("OAL Event: LE device disconnected");
		__bt_hanlde_le_device_disconnection((event_dev_conn_status_t *)event_data);
		break;
	}
	case OAL_EVENT_GATTC_NOTIFICATION_REGISTERED: {
		BT_INFO("OAL Event: GATT Client Notification Registered");
		__bt_handle_client_notification_registered((event_gattc_regdereg_notify_t *) event_data, TRUE);
		break;
	}
	case OAL_EVENT_GATTC_NOTIFICATION_DEREGISTERED: {
		BT_INFO("OAL Event: GATT Client Notification Registered");
		__bt_handle_client_notification_registered((event_gattc_regdereg_notify_t *) event_data, FALSE);
		break;
	}
	case OAL_EVENT_GATTC_NOTIFY_DATA: {
		BT_INFO("OAL Event: GATT Client Notification Data");
		__bt_handle_client_notification_data((event_gattc_notify_data *) event_data);

	}
#endif
	default:
		break;
	}
	BT_DBG("-");
}

int _bt_gatt_server_add_service(char *sender, int service_type,
		int num_handles, char *svc_uuid, int instance_id)
{
	BT_CHECK_PARAMETER(svc_uuid, return);
	BT_CHECK_PARAMETER(sender, return);
	int ret = OAL_STATUS_SUCCESS;

	oal_gatt_srvc_id_t svc_data;

	svc_data.is_prmry = service_type;
	svc_data.id.inst_id = instance_id;

	BT_INFO("Service UUID [%s] Num handles [%d] Instance ID [%d]", svc_uuid, num_handles, instance_id);
	_bt_string_to_uuid(svc_uuid, (service_uuid_t*)&svc_data.id.uuid);

	ret = gatts_add_service(instance_id, &svc_data, num_handles);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}


int _bt_gatt_server_add_included_service(char *sender, int instance_id,
		int service_handle, int included_svc_handle)
{
	BT_CHECK_PARAMETER(sender, return);
	int ret = OAL_STATUS_SUCCESS;

	ret = gatts_add_included_services(instance_id, service_handle, included_svc_handle);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_add_characteristic(char *sender, char *char_uuid,
		bluetooth_gatt_server_attribute_params_t *param)
{
	BT_CHECK_PARAMETER(char_uuid, return);
	BT_CHECK_PARAMETER(sender, return);
	BT_CHECK_PARAMETER(param, return);
	int ret = OAL_STATUS_SUCCESS;

	oal_uuid_t uuid = {{0} };

	BT_INFO("Char UUID [%s] Instance ID [%d]", char_uuid, param->instance_id);
	_bt_string_to_uuid(char_uuid, (service_uuid_t*)&uuid);

	BT_INFO("Char permission From API [0x%x]", param->permissions);

	ret = gatts_add_characteristics(param->instance_id, param->service_handle, &uuid,
			param->properties, (int)param->permissions);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_add_descriptor(char *sender, char *desc_uuid,
		bt_gatt_permission_t *param, int service_handle, int instance_id)
{
	BT_CHECK_PARAMETER(desc_uuid, return);
	BT_CHECK_PARAMETER(sender, return);
	BT_CHECK_PARAMETER(param, return);
	int ret = OAL_STATUS_SUCCESS;

	oal_uuid_t uuid = {{0} };

	BT_INFO("Descriptor UUID [%s] Instance ID [%d] Service handle [%d]",
			desc_uuid, service_handle, instance_id);

	_bt_string_to_uuid(desc_uuid, (service_uuid_t*)&uuid);

	BT_INFO("Descriptor permission From API [0x%x]", *param);
	ret = gatts_add_descriptor(instance_id, service_handle, &uuid, (int)*param);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_start_service(char *sender, int service_handle, int instance_id)
{
	BT_CHECK_PARAMETER(sender, return);
	int ret = OAL_STATUS_SUCCESS;

	ret = gatts_start_service(instance_id, service_handle, BT_GATT_TRANSPORT_LE);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_stop_service(char *sender, int service_handle, int instance_id)
{
	BT_CHECK_PARAMETER(sender, return);
	int ret = OAL_STATUS_SUCCESS;

	ret = gatts_stop_service(instance_id, service_handle);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_delete_service(char *sender, int service_handle, int instance_id)
{
	BT_CHECK_PARAMETER(sender, return);
	int ret = OAL_STATUS_SUCCESS;
	GSList *l;
	int *handle = NULL;

	ret = gatts_delete_service(instance_id, service_handle);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Remove the Service Handle */
	for (l = numapps[instance_id].service_handles; l != NULL;) {
		handle = l->data;
		l = g_slist_next(l);
		if (handle && *handle == service_handle) {
			BT_INFO("Remove Service handle [%d]", *handle);
			numapps[instance_id].service_handles = g_slist_remove(numapps[instance_id].service_handles, handle);
			g_free(handle);
			handle = NULL;
		}
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_send_response(char *sender, bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_server_response_params_t *param)
{
	BT_CHECK_PARAMETER(sender, return);
	BT_CHECK_PARAMETER(data, return);
	BT_CHECK_PARAMETER(param, return);
	struct gatt_server_req_info *req_info = NULL;
	int ret = OAL_STATUS_SUCCESS;

	oal_gatt_response_t response;

	BT_INFO("GATT Server Response: Req Type [%d] req_id [%d] status [%d] auth_req [%d] offset[%d] data len[%d]",
			param->req_type, param->request_id,
			param->response_status, param->auth_req,
			data->offset, data->length);

	/* Search for matching Request in List */
	req_info = __bt_gatt_server_find_request_info(param->request_id, param->req_type);
	if (!req_info) {
		BT_ERR("GATT Server Req Info not found for current response..return Error");
		return BLUETOOTH_ERROR_NOT_IN_OPERATION;
	}

	memset(&response, 0x00, sizeof(oal_gatt_response_t));

	response.handle = req_info->attribute_handle;
	response.attr_value.auth_req = param->auth_req;
	response.attr_value.handle = req_info->attribute_handle;
	response.attr_value.offset = data->offset;
	response.attr_value.len = data->length;
	memcpy(&response.attr_value.value, &data->data, data->length);


	ret = gatts_send_response(req_info->connection_id, param->request_id,
			param->response_status, &response);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("GATT Server Response successfully sent");
	/* Remove GATT server request from list */
	gatt_server_requests = g_slist_remove(gatt_server_requests, req_info);
	g_free(req_info->addr);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_acquire_send_response(char *sender, bluetooth_gatt_server_acquire_response_params_t *param ,	void *fd_list)
{
	BT_CHECK_PARAMETER(sender, return);
	BT_CHECK_PARAMETER(param, return);
	struct gatt_server_req_info *req_info = NULL;
	int ret = OAL_STATUS_SUCCESS;


	BT_INFO("GATT acquire Server Response: Req Type [%d] req_id [%d] fd  [%d] mtu[%d]",
			param->req_type, param->request_id,
			 param->fd,
			param->mtu);

	/* Search for matching Request in List */
	req_info = __bt_gatt_server_find_request_info(param->request_id, param->req_type);
	if (!req_info) {
		BT_ERR("GATT acquire Server Req Info not found for current response..return Error");
		return BLUETOOTH_ERROR_NOT_IN_OPERATION;
	}

	ret = gatt_send_response_acquire(req_info->connection_id, param->request_id, 0, param->fd, param->mtu, fd_list);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("GATT acquire Server Response successfully sent");
	/* Remove GATT server request from list */
	gatt_server_requests = g_slist_remove(gatt_server_requests, req_info);
	g_free(req_info->addr);

	return BLUETOOTH_ERROR_NONE;
}



int _bt_gatt_server_send_indication(char *sender, bluetooth_device_address_t *dev_addr,
		bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_server_indication_params_t *param)
{
	BT_CHECK_PARAMETER(sender, return);
	BT_CHECK_PARAMETER(data, return);
	BT_CHECK_PARAMETER(param, return);
	char *address;
	gboolean all_send = FALSE;
	int ret = OAL_STATUS_SUCCESS;
	struct gatt_client_info_t *conn;

	address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(address, dev_addr->addr);

	if (memcmp(dev_addr->addr, BDADDR_ANY, 6) == 0) {
		BT_INFO("GATT Server: Send Indication to all connected GATT clients..");
		all_send = TRUE;
	} else {
		BT_INFO("GATT Server: Send Indication to connected GATT client addr [%s]", address);
	}

	/* Attempt to send Notification/Indication to all Connected GATT clients */
	if (all_send) {
		ret = __bt_gatt_send_indication_to_all_connected_clients(data, param);
		if (ret != OAL_STATUS_SUCCESS) {
			BT_ERR("ret: %d", ret);
			g_free(address);
			return BLUETOOTH_ERROR_INTERNAL;
		}

	} else {
		conn = __bt_find_remote_gatt_client_info(address);
		if (conn) {
			ret = gatts_send_indication(param->instance_id, param->atrribute_handle,
					conn->connection_id, data->length,
					param->need_confirmation, (char *)(&data->data[0]));

			if (ret != OAL_STATUS_SUCCESS) {
				BT_ERR("ret: %d", ret);
				BT_INFO("Indication failed to send to Remote GATT Client [%s]", address);
				g_free(address);
				return BLUETOOTH_ERROR_INTERNAL;
			}
			BT_INFO("Indication sent to Remote GATT Client [%s] wait for Notification completed event from OAL", address);
			g_free(address);
			num_indicate_clients = 1;
			return BLUETOOTH_ERROR_NONE;
		}
		BT_ERR("Remote GATT client [%s] is not connected..Cant send Indication!!", address);
		g_free(address);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}
	g_free(address);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_server_update_attribute_value(char *sender, int instance_id,
		bluetooth_gatt_server_update_value_t *param)
{
	BT_CHECK_PARAMETER(sender, return);
	BT_CHECK_PARAMETER(param, return);
	int ret = OAL_STATUS_SUCCESS;

	oal_gatt_value_t value;
	BT_INFO("GATT Server Update value: Instance ID [%d] attr handle [%d] Value len [%d]",
			instance_id, param->attribute_handle, param->length);


	memset(&value, 0x00, sizeof(oal_gatt_value_t));

	value.handle = param->attribute_handle;
	value.len = param->length;
	memcpy(&value.value, &param->data.data, param->length);

	ret = gatts_update_att_value(instance_id, &value);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("GATT Server Update Attribute Value successfully");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_get_att_mtu(bluetooth_device_address_t *address,
		unsigned int *mtu)
{
	BT_CHECK_PARAMETER(address, return);
	BT_CHECK_PARAMETER(mtu, return);
	struct gatt_client_info_t *conn_info = NULL;
	char addr[BT_ADDRESS_STRING_SIZE] = { 0 };
	int ret = OAL_STATUS_SUCCESS;
	int stack_mtu;

	_bt_convert_addr_type_to_string(addr, address->addr);

	BT_INFO("Get current MTU size for the remote client:DevAddress:[%s]", addr);

	conn_info = __bt_find_remote_gatt_client_info(addr);
	if (conn_info) {
		BT_INFO("GATT Client [%s] is connected, conn Id [%d] Instance ID [%d]",
				conn_info->addr, conn_info->connection_id, conn_info->instance_id);
	} else {
		BT_ERR("GATT Client [%s] is not yet connected..", addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	ret = gatts_get_att_mtu(conn_info->connection_id, &stack_mtu);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("ATT MTU received from OAL [%d]", stack_mtu);
	*mtu = (unsigned int)stack_mtu;

	if (*mtu == 0) {
		BT_ERR("MTU value is zero, GATT Client [%s] is not yet connected..", addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	return BLUETOOTH_ERROR_NONE;
}

#ifdef TIZEN_GATT_CLIENT
/* GATT Client utility static functions */
static bt_gatt_service_info_list_t * __bt_get_service_info_list(int conn_id)
{
	GSList *l;
	bt_gatt_service_info_list_t *info = NULL;

	for (l = list_gatt_info; l != NULL; l = g_slist_next(l)) {
		info = (bt_gatt_service_info_list_t*)l->data;
		if (info == NULL)
			continue;

		if (info->conn_id == conn_id)
			return info;

	}
	return NULL;
}

static bt_gatt_service_info_t* __bt_find_matching_service(
		bt_gatt_service_info_list_t *svc_list, oal_gatt_srvc_id_t *svc)
{
	GSList *l;
	bt_gatt_service_info_t *info = NULL;

	for (l = svc_list->services; l != NULL; l = g_slist_next(l)) {
		info = (bt_gatt_service_info_t*)l->data;
		if (info == NULL)
			continue;

		/* Match UUID and instance ID */
		if (!memcmp(&svc->id.uuid.uuid, &info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN)
				&& (svc->id.inst_id == info->inst_id)) {
			return info;
		}
	}
	return NULL;
}

static bt_gatt_char_info_t* __bt_find_matching_charc(
		bt_gatt_service_info_t *svc_info, oal_gatt_id_t *charc)
{
	GSList *l;
	bt_gatt_char_info_t *info = NULL;

	for (l = svc_info->chars; l != NULL; l = g_slist_next(l)) {
		info = (bt_gatt_char_info_t*)l->data;
		if (info == NULL)
			continue;

		/* Match UUID and instance ID */
		if (!memcmp(&charc->uuid.uuid, &info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN)
				&& (charc->inst_id == info->inst_id)) {
			return info;
		}
	}
	return NULL;
}

static bt_gatt_descriptor_info_t* __bt_find_matching_desc(
		bt_gatt_char_info_t *char_info, oal_gatt_id_t *desc)
{
	GSList *l;
	bt_gatt_descriptor_info_t *info = NULL;

	for (l = char_info->descs; l != NULL; l = g_slist_next(l)) {
		info = (bt_gatt_descriptor_info_t*)l->data;
		if (info == NULL)
			continue;

		/* Match UUID and instance ID */
		if (!memcmp(&desc->uuid, &info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN)
				&& (desc->inst_id == info->inst_id)) {
			return info;
		}
	}
	return NULL;
}


static struct gatt_server_info_t *__bt_find_remote_gatt_server_info_from_conn_id(int conn_id)
{
	GSList *l;
	struct gatt_server_info_t *info = NULL;

	for (l = gatt_server_info_list; l != NULL; l = g_slist_next(l)) {
		info = (struct gatt_server_info_t*)l->data;
		if (info == NULL)
			continue;

		if (info->connection_id == conn_id) {
			BT_INFO("Remote GATT server found addr[%s]", info->addr);
			return info;
		}
	}
	return NULL;
}

static bt_gatt_service_info_t* __bt_find_removed_service(bt_gatt_service_info_list_t *svc_list)
{
	GSList *l;
	bt_gatt_service_info_t *info = NULL;

	for (l = svc_list->services; l != NULL; l = g_slist_next(l)) {
		info = (bt_gatt_service_info_t*)l->data;
		if (info == NULL)
			continue;

		/* Service is marked a removed */
		if (info->is_removed == 1)
			return info;
	}
	return NULL;
}

static void __bt_remove_service_info_from_list(bt_gatt_service_info_t *svc_info)
{
#if 0
	GSList *l;
	GSList *l1;
	GSList *l2;
	bt_gatt_char_info_t *charc = NULL;
	bt_gatt_included_service_info_t *incl = NULL;
	bt_gatt_descriptor_info_t *desc = NULL;

	/* Remove all Characteristic and Descriptors within characteristic */
	for (l = svc_info->chars; l != NULL;) {
		charc = (bt_gatt_char_info_t*)l->data;
		l = g_slist_next(l); /* Incase if l is removed, saving next to l */

		if (charc == NULL)
			continue;

		/* Inside Characteristic */
		for (l1 = charc->descs; l1 != NULL;) {

			desc = (bt_gatt_descriptor_info_t*)l1->data;
			l1 = g_slist_next(l1);

			if (desc == NULL)
				continue;

			/* Remove Descriptor */
			charc->descs = g_slist_remove(charc->descs, desc);
			g_free(desc);
		}
		/* Remove Characteristic */
		svc_info->chars = g_slist_remove(svc_info->chars, charc);
		g_free(charc);
	}

	/* Remove all Included Services */
	for (l2 = svc_info->included_svcs; l2 != NULL;) {
		incl = (bt_gatt_included_service_info_t*)l2->data;
		l2 = g_slist_next(l2); /* Incase if l is removed, saving next to l */

		if (incl == NULL)
			continue;

		/* Remove included service */
		svc_info->included_svcs = g_slist_remove(svc_info->included_svcs, incl);
		g_free(incl);
	}
#endif
}


static void __bt_build_service_browse_info(int conn_id,
		bt_services_browse_info_t* info)
{
	GSList *l;
	bt_gatt_service_info_list_t *svc_info_list;
	bt_gatt_service_info_t *svc_info;

	service_uuid_t uuid;
	struct gatt_server_info_t *conn_info = NULL;
	int count = 0;
	char uuid_string[BLUETOOTH_UUID_STRING_MAX];
	BT_INFO("+");

	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(conn_id);

	if (!conn_info)
		return;

	_bt_convert_addr_string_to_type(info->device_addr.addr, conn_info->addr);

	svc_info_list = __bt_get_service_info_list(conn_id);

	if (!svc_info_list)
		return;

	info->count = g_slist_length(svc_info_list->services);
	BT_INFO("Total services present in the svc info list for this conn id [%d] is [%d]",
			conn_id, info->count);

	for (l = svc_info_list->services; l != NULL; l = g_slist_next(l)) {
		svc_info = (bt_gatt_service_info_t*)l->data;
		if (svc_info == NULL)
			continue;

		memcpy(&uuid.uuid, &svc_info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
		_bt_uuid_to_string(&uuid, uuid_string);

		BT_INFO("Service UUID formed [%s] strlen [%zd]", uuid_string, strlen(uuid_string));

		/* Fill UUID of service */
		g_strlcpy(info->uuids[count], uuid_string,
				BLUETOOTH_UUID_STRING_MAX);

		BT_INFO("Service UUID formed  TO be sent [%s] strlen [%zd]",
				info->uuids[count], strlen(info->uuids[count]));
		/* Fill instance ID of service */
		info->inst_id[count] = svc_info->inst_id;

		/* Fill primary service or not info */
		info->primary[count] = svc_info->is_primary;

		/* Increment count of services browsed */
		count++;
	}

	BT_INFO("Total services browsed [%d]", count);
}

static void __bt_build_char_browse_info(int conn_id,
		bt_gatt_service_info_t *svc_info,
		bt_char_browse_info_t* info)
{
	GSList *l;
	bt_gatt_char_info_t *char_info;
	service_uuid_t uuid;

	struct gatt_server_info_t *conn_info = NULL;
	int count = 0;
	char uuid_string[BLUETOOTH_UUID_STRING_MAX];

	BT_INFO("+");

	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(conn_id);
	if (!conn_info)
		return;

	/* Fill default data, this will be required even in case of failure */
	_bt_convert_addr_string_to_type(info->device_addr.addr, conn_info->addr);
	memcpy(&info->svc_uuid, svc_info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	info->svc_inst_id = svc_info->inst_id;

	if (!svc_info->chars) {
		BT_ERR("No Chars browsed for address [%s]", conn_info->addr);
		return;
	}

	info->count = g_slist_length(svc_info->chars);
	BT_INFO("Total count of Characteristics [%d]", info->count);

	for (l = svc_info->chars; l != NULL; l = g_slist_next(l)) {
		char_info = (bt_gatt_char_info_t*)l->data;
		if (char_info == NULL)
			continue;

		memcpy(&uuid.uuid, &char_info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
		_bt_uuid_to_string(&uuid, uuid_string);

		/* Fill UUID of characteristic */
		g_strlcpy(info->uuids[count], uuid_string,
				BLUETOOTH_UUID_STRING_MAX);

		/* Fill instance ID of characteristic */
		info->inst_id[count] = char_info->inst_id;

		/* Fill property of characteristic */
		info->props[count] = char_info->props;

		/* Increment count of services browsed */
		count++;
	}
	BT_INFO("Total characteristics browsed [%d]", count);
}

static void __bt_build_descriptor_browse_info(int conn_id,
		bt_gatt_service_info_t *svc_info,
		bt_gatt_char_info_t *char_info,
		bt_descriptor_browse_info_t* info)
{
	GSList *l;
	bt_gatt_descriptor_info_t *desc_info;

	service_uuid_t uuid;
	struct gatt_server_info_t *conn_info = NULL;
	int count = 0;
	char uuid_string[BLUETOOTH_UUID_STRING_MAX];

	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(conn_id);

	/* Fill default data, this will be required even in case of failure */
	_bt_convert_addr_string_to_type(info->device_addr.addr, conn_info->addr);
	memcpy(&info->svc_uuid, svc_info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	info->svc_inst_id = svc_info->inst_id;
	memcpy(&info->char_uuid, char_info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	info->char_inst_id = char_info->inst_id;

	/* Fill property of the parent characteristic of this descriptor */
	info->char_props_map = char_info->props;

	info->count = g_slist_length(char_info->descs);
	BT_INFO("Total count of Descriptors [%d]", info->count);

	if (!char_info->descs) {
		BT_ERR("No Descriptors browsed for address [%s]", conn_info->addr);
		return;
	}

	for (l = char_info->descs; l != NULL; l = g_slist_next(l)) {
		desc_info = (bt_gatt_descriptor_info_t*)l->data;
		if (desc_info == NULL)
			continue;

		memcpy(&uuid.uuid, &desc_info->uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
		_bt_uuid_to_string(&uuid, uuid_string);

		/* Fill UUID of Descriptor */
		g_strlcpy(info->uuids[count], uuid_string,
				BLUETOOTH_UUID_STRING_MAX);

		/* Fill instance ID of Descriptor */
		info->inst_id[count] = desc_info->inst_id;


		/* Increment count of Descriptor browsed */
		count++;
	}

	BT_INFO("Total descriptors browsed [%d]", count);
}


static void __bt_cleanup_remote_services(struct gatt_server_info_t *conn_info)
{
	GSList *l;
	GSList *ll;
	GSList *lll;
	GSList *llll;

	bt_gatt_service_info_list_t * svc_info_list = NULL;
	bt_gatt_service_info_t *svc = NULL;
	bt_gatt_char_info_t *chr = NULL;
	bt_gatt_descriptor_info_t *desc = NULL;
	bt_gatt_included_service_info_t *incl_svc = NULL;

	BT_INFO("Start Cleanup of all services");

	svc_info_list = __bt_get_service_info_list(conn_info->connection_id);
	if (!svc_info_list) {
		BT_INFO("Could not find Svc Info list for the connection ID [%d]",
				conn_info->connection_id);
		return;
	}

	BT_INFO("Num Services [%d]", g_slist_length(svc_info_list->services));
	for (l = svc_info_list->services; l;) {
		svc = (bt_gatt_service_info_t*)l->data;
		l = g_slist_next(l);
		if (svc == NULL)
			continue;

		BT_INFO("Service info Is Prim[%d] Inst ID [%d]",
				svc->is_primary, svc->inst_id);
		BT_INFO("Num chars [%d]", g_slist_length(svc->chars));
		/* Delete all chars and its descriptors */
		for (ll = svc->chars; ll;) {
			chr = (bt_gatt_char_info_t*)ll->data;
			ll = g_slist_next(ll);
			if (chr == NULL)
				continue;

			BT_INFO("Num descs [%d]", g_slist_length(chr->descs));
			for (lll = chr->descs; lll;) {
				desc = (bt_gatt_descriptor_info_t *)lll->data;
				lll = g_slist_next(lll);
				if (desc == NULL)
					continue;
				chr->descs = g_slist_remove(chr->descs, desc);
				g_free(desc);
			}
			svc->chars = g_slist_remove(svc->chars, chr);
			g_free(chr);
		}

		BT_INFO("Num incl svcs [%d]", g_slist_length(svc->included_svcs));
		/* Delete all included services */
		for (llll = svc->included_svcs; llll;) {
			incl_svc = (bt_gatt_included_service_info_t*)llll->data;
			llll = g_slist_next(llll);
			if (incl_svc == NULL)
				continue;

			svc->included_svcs = g_slist_remove(svc->included_svcs, incl_svc);
			g_free(incl_svc);
		}
		svc_info_list->services = g_slist_remove(svc_info_list->services, svc);
		g_free(svc);
	}

	list_gatt_info = g_slist_remove(list_gatt_info, svc_info_list);
	g_free(svc_info_list);
	BT_INFO("Cleanup of all services done");
}

int _bt_register_gatt_client_instance(const char *sender,
		bluetooth_device_address_t *address)
{
	int ret = OAL_STATUS_SUCCESS;
	char *uuid_string = NULL;
	int slot = -1;
	int k;
	oal_uuid_t uuid;

	/* App should ensure that it should not send */
	BT_INFO("###Check on which instance GATT Client instance can be initialized....");
	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		if (numapps[k].is_initialized == 1) {
			BT_INFO("Instance ID [%d] is already in use..Check next slot",
					numapps[k].instance_id);
		} else {
			slot = k;
			BT_INFO("Time to register GATT client instancer..UUID to be used is [%s] slot [%d]",
					uuid_list[slot-1], slot);
			break;
		}
	}

	if (slot == -1) {
		BT_ERR("No Slot if free for GATT Client registration..");
		return BLUETOOTH_ERROR_REGISTRATION_FAILED;
	}

	uuid_string = g_malloc0(BT_UUID_STRING_MAX);
	_bt_string_to_uuid(uuid_list[slot-1], (service_uuid_t*)&uuid);
	g_strlcpy(uuid_string, uuid_list[slot-1], BT_UUID_STRING_MAX);
	BT_INFO("Copied UUID string [%s]", uuid_string);

	/* Register GATT Client */
	ret = gattc_register(&uuid);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(uuid_string);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	BT_INFO("GATT Client registration call successfully accepted by OAL..wait for Instance Initialized event from OAL..");

	/* Return & wait for GATT Client Instance Initialization event */
	memset(numapps[slot].sender, 0x00, sizeof(numapps[slot].sender));
	memset(numapps[slot].uuid, 0x00, sizeof(numapps[slot].uuid));

	g_strlcpy(numapps[slot].sender, sender, sizeof(numapps[slot].sender));
	g_strlcpy(numapps[slot].uuid, uuid_string, sizeof(numapps[slot].uuid));

	/* Address is saved here. When event comes, sender + address are matched for replying pending
	   request. It is impossible for same sender to have requests with two same addresses */
	memcpy(&numapps[slot].address.addr, address->addr, sizeof(bluetooth_device_address_t));

	BT_INFO("Slot [%d] occupied", slot);
	numapps[slot].is_initialized = TRUE; /* Set initialization to true here itself */

	g_free(uuid_string);
	return BLUETOOTH_ERROR_NONE;

}



/* GATT client events */
static void __bt_handle_client_instance_registered(event_gattc_register_t *data)
{
	bt_service_app_info_t *info = NULL;
	int k;
	char *uuid_string = g_malloc0(BT_UUID_STRING_MAX);

	_bt_uuid_to_string(&(data->client_uuid), uuid_string);
	BT_INFO("CLient ID is Initialized [%d] UUID initialized [%s]", data->client_if, uuid_string);

	/* Platform GATT client framwork does not use Default GATT client instance
	   This GATT client instance is never deregistred in the lifetime of bt-service */
	BT_INFO("Default UUID [%s] current registered uuid [%s]",
			DEFAULT_GATT_CLIENT_UUID, uuid_string);
	if (g_strcmp0(uuid_string, DEFAULT_GATT_CLIENT_UUID) == 0) {
		BT_INFO("Default client Instance Registered [%s] Client instance [%d]",
				uuid_string, data->client_if);
		gatt_default_client = data->client_if;
		g_free(uuid_string);
		return;
	}

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (g_strcmp0(info->uuid, uuid_string) == 0) {
			BT_INFO("Found GATT client.. UUID [%s], sender [%s]", info->uuid, info->sender);
			BT_INFO("Slot [%d] occupied", k);
			info->is_initialized = TRUE;
			info->client_id = data->client_if;
			__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_NONE, BT_GATT_CLIENT_REGISTER,
					(void*)info, sizeof(bt_service_app_info_t));
			break;
		}
	}
	g_free(uuid_string);
}

static void __bt_handle_client_connected(event_gattc_conn_t *event_data)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_info_t *conn_info = NULL;
	struct gatt_out_conn_info_t *out_conn_info = NULL;

	GVariant *param = NULL;

	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)event_data->address.addr);

	if (event_data->status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;

	/* DBUS Return fo BT_CONNECT_LE for all the apps */
	__bt_gatt_handle_pending_request_info(result, BT_CONNECT_LE, address,
			BT_ADDRESS_STRING_SIZE);

	BT_INFO("Local GATT Client Connected: Remote addr[%s] Client Interface [%d] Connection ID [%d] status[%d]",
			address, event_data->client_if, event_data->conn_id, event_data->status);

	if (result == BLUETOOTH_ERROR_NONE) {
		/* Check if device is already in connected list */
		conn_info = __bt_find_remote_gatt_server_info(address);

		if (!conn_info) {
			/* Send event to BT-API */
			param = g_variant_new("(is)", result, address);
			_bt_send_event(BT_DEVICE_EVENT,
					BLUETOOTH_EVENT_GATT_CLIENT_CONNECTED, /* Local device is GATT client */
					param);

			/* Save Connection info */
			conn_info = g_new0(struct gatt_server_info_t, 1);
			conn_info->addr = g_strdup(address);
			conn_info->client_id = event_data->client_if;
			BT_INFO("Added GATT server addr[%s]", conn_info->addr);
			conn_info->connection_id = event_data->conn_id;
			gatt_server_info_list = g_slist_append(gatt_server_info_list, conn_info);
			BT_INFO("Total num of connected Remote GATT server devices [%d]",
					g_slist_length(gatt_server_info_list));

#if 0
			BT_INFO("Do a Internal refresh");
			if (OAL_STATUS_SUCCESS != gattc_refresh(conn_info->client_id, &event_data->address))
				BT_ERR("GATT database refresh failed!!");
			else
				BT_INFO("GATT database refresh Success!!");
#endif
		} else
			BT_ERR("Local GATT Client connected event for addr[%s], but device is in connected list already", address);
	} else
		BT_ERR("GATT Client Connection failed!!");

	/* If outgoing connection Info is present, then remove it */
	out_conn_info = __bt_find_gatt_outgoing_conn_info(address);
	if (out_conn_info) {
		BT_ERR("Outgoing Client connect request was sent");
		outgoing_gatt_conn_list = g_slist_remove(outgoing_gatt_conn_list, out_conn_info);
		g_free(out_conn_info->addr);
		g_free(out_conn_info);
	}
	g_free(address);
}

static void __bt_handle_client_disconnected(event_gattc_conn_t *event_data)
{
	int result = BLUETOOTH_ERROR_NONE;

	struct gatt_server_info_t *conn_info = NULL;
	struct gatt_out_conn_info_t *out_conn_info = NULL;

	GVariant *param = NULL;

	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(address,
			(unsigned char *)event_data->address.addr);

	if (event_data->status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;

	if (NULL ==  _bt_get_request_info_data(BT_DISCONNECT_LE, address)) {
		if (NULL !=  _bt_get_request_info_data(BT_CONNECT_LE, address)) {
			result = BLUETOOTH_ERROR_INTERNAL;
			__bt_gatt_handle_pending_request_info(result, BT_CONNECT_LE,
					address, BT_ADDRESS_STRING_SIZE);
			BT_ERR("Failed to connect Local GATT Remote addr[%s]", address);
			g_free(address);
			return;
		}
	} else {
		/* DBUS Return for BT_DISCONNECT_LE for all the apps */
		__bt_gatt_handle_pending_request_info(result, BT_DISCONNECT_LE, address,
				BT_ADDRESS_STRING_SIZE);
	}
	BT_INFO("Local GATT Client DisConnected: Remote addr[%s] Client Interface [%d] Connection ID [%d] status [%d]",
			address, event_data->client_if, event_data->conn_id, event_data->status);

	/* Remove Connection info */
	conn_info = __bt_find_remote_gatt_server_info(address);

	if (conn_info) {
		param = g_variant_new("(is)", result, address);
		/* Send event to application */
		_bt_send_event(BT_DEVICE_EVENT,
				BLUETOOTH_EVENT_GATT_CLIENT_DISCONNECTED,
				param);

		BT_INFO("Remove GATT server info from List..");
		/* Remove all services from info list_gatt_info */
		__bt_cleanup_remote_services(conn_info);

		/* Remove info from List */
		gatt_server_info_list = g_slist_remove(gatt_server_info_list, conn_info);
		BT_INFO("Total num of connected GATT servers [%d]", g_slist_length(gatt_server_info_list));
		g_free(conn_info->addr);
		g_free(conn_info);
	} else
		BT_INFO("Can not find conn info, already removed!");

	/* If outgoing connection Info is present, then remove it */
	out_conn_info = __bt_find_gatt_outgoing_conn_info(address);
	if (out_conn_info) {
		BT_ERR("Client Disconnected event, but outgoing connect request was sent");
		outgoing_gatt_conn_list = g_slist_remove(outgoing_gatt_conn_list, out_conn_info);
		g_free(out_conn_info->addr);
		g_free(out_conn_info);
	}
	g_free(address);
}


static void __bt_handle_client_service_search_result(
		event_gattc_service_result_t *event_data)
{
	BT_INFO("+");
	BT_INFO("received the gatt service search result");
	/* Pre: status is never fail from OAL */

	/* Find service list from address */
	bt_gatt_service_info_list_t *svc_info_list;
	bt_gatt_service_info_t *svc_info;
	BT_INFO("Search Result: status [%d] conn_id [%d]",
			event_data->conn_status.status,
			event_data->conn_status.conn_id);

	svc_info_list = __bt_get_service_info_list(event_data->conn_status.conn_id);
	if (!svc_info_list) {
		BT_INFO("Service info list not present for this connection ID, means first time browse");
		/* Means for this conn_id, no services are ever browsed, first time,
		   create service info list for this conn_id */
		svc_info_list = g_malloc0(sizeof(bt_gatt_service_info_list_t));
		svc_info_list->conn_id = event_data->conn_status.conn_id;
		list_gatt_info = g_slist_append(list_gatt_info, svc_info_list);
	} else {
		BT_INFO("Service info list Already present for this connection ID, means not first time browse for this conn ID ");
	}

	/* send list and current service's uuid and instance id to find it  */
	svc_info = __bt_find_matching_service(svc_info_list, &event_data->srvc_id);


	/* If not found, check if service changed, if yes, means this is a new service added
	   in remote GATT device, update uuid info in svc info list structure, to be used when
	   search is completed */
	if (!svc_info) {
		BT_INFO("Service Not found in svc info list for this connection ID");
		if (svc_info_list->info.is_changed) {
			BT_INFO("Service Changed indication already found for this connection ID");
			memcpy(svc_info_list->info.uuid, event_data->srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
		}
		/* Create and add new service in service list */
		svc_info = g_malloc0(sizeof(bt_gatt_service_info_t));
		memcpy(svc_info->uuid, event_data->srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
		svc_info->inst_id = event_data->srvc_id.id.inst_id;
		svc_info->is_primary = event_data->srvc_id.is_prmry;
		svc_info_list->services = g_slist_append(svc_info_list->services, svc_info);
		BT_INFO("Service created and added in Svc info listf or this connection ID");
	} else {
		BT_INFO("Service Already found to be present inside the svc info list for this connection ID");
		/* If returned matching service info, then just update service_rmeoved value inside it to 0 */
		svc_info->is_removed = 0;
	}
}

static void __bt_handle_client_service_search_completed(
		event_gattc_conn_status_t *event_data)
{
	BT_INFO("+");
	struct gatt_server_info_t *conn_info = NULL;
	bt_gatt_service_info_list_t *svc_info_list;
	bt_gatt_service_info_t *svc_info;
	bt_services_browse_info_t browse_info;
	unsigned char uuid_empty[BLUETOOTH_UUID_HEX_MAX_LEN];

	memset(&uuid_empty, 0x00, BLUETOOTH_UUID_HEX_MAX_LEN);
	memset(&browse_info, 0x00, sizeof(bt_services_browse_info_t));
	BT_INFO("Primary Services browsing completed status[%d] conn ID [%d]",
			event_data->status, event_data->conn_id);

	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(event_data->conn_id);

	svc_info_list = __bt_get_service_info_list(event_data->conn_id);
	if (!svc_info_list) {
		BT_ERR("No services browsed ever for addr [%s]", conn_info->addr);

		/* Just build response and return ERROR */
		__bt_build_service_browse_info(event_data->conn_id, &browse_info);

		__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_INTERNAL,
				BT_GATT_GET_PRIMARY_SERVICES, &browse_info,
				sizeof(bt_services_browse_info_t));
		return;
	}

	/* If fail, then send event with error  */
	if (event_data->status != OAL_STATUS_SUCCESS) {
		/* Just build response and return ERROR */
		__bt_build_service_browse_info(event_data->conn_id, &browse_info);

		__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_INTERNAL,
				BT_GATT_GET_PRIMARY_SERVICES, &browse_info,
				sizeof(bt_services_browse_info_t));
		return;
	}

	/* If success, then find service info list from address */

	/* If svc_changed == 1 and uuid valid, means a new service is added*/
	if (svc_info_list->info.is_changed && !memcmp(uuid_empty, svc_info_list->info.uuid, BLUETOOTH_UUID_HEX_MAX_LEN)) {
		/* TODO: Send event -Service added with instance ID and UUID of newly added service */
		BT_INFO("new service added");

		BT_INFO("TODO new service added");
	}

	/* If svc_changed == 1 and uuid invalid, then a service is removed */
	if (svc_info_list->info.is_changed && memcmp(uuid_empty, svc_info_list->info.uuid, BLUETOOTH_UUID_HEX_MAX_LEN)) {
		/* Scan through the service info list to find service with is_removed = 1*/
		svc_info = __bt_find_removed_service(svc_info_list);

		/* TODO Send event - Service removed with instance ID and UUID of just rmeoved service */

		/* Remove that service info from service info list */
		svc_info_list->services = g_slist_remove(svc_info_list->services, svc_info);

		/* Delete that service completely from svc_info list*/
		__bt_remove_service_info_from_list(svc_info);
		g_free(svc_info);
	}

	/* Reset svc_changed = 0, and reset UUID = all 0's */
	svc_info_list->info.is_changed = 0;
	memset(&svc_info_list->info.uuid, 0x00, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Build Reply and send to service browse primary services request of pending apps */
	__bt_build_service_browse_info(event_data->conn_id, &browse_info);

	__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_NONE,
			BT_GATT_GET_PRIMARY_SERVICES, &browse_info,
			sizeof(bt_services_browse_info_t));
}


static void __bt_handle_client_characteristic_search_result(
		event_gattc_characteristic_result_t *event_data)
{
	bt_gatt_service_info_list_t *svc_info_list;
	bt_gatt_service_info_t *svc_info;
	bt_gatt_char_info_t *char_info;
	bt_char_browse_info_t browse_info;

	BT_INFO("Characteristic search result status [%d]",
			event_data->conn_status.status);

	memset(&browse_info, 0x00, sizeof(bt_char_browse_info_t));

	/* If success */
	if (event_data->conn_status.status == OAL_STATUS_SUCCESS) {
		/* Find service info list from address */
		svc_info_list = __bt_get_service_info_list(event_data->conn_status.conn_id);
		if (svc_info_list == NULL) {
			BT_ERR("svc_info_list is NULL");
			return;
		}


		/* Find matching service info from svc info list */
		svc_info = __bt_find_matching_service(svc_info_list, &event_data->srvc_id);
		if (svc_info == NULL) {
			BT_ERR("svc_info is NULL");
			return;
		}

		/* Find Matching char from service info in event */
		char_info = __bt_find_matching_charc(svc_info, &event_data->char_id);
		if (char_info == NULL)
			BT_ERR("char_info is NULL");

		/* If not found, then add new characteristic and return */
		if (!char_info) {
			BT_INFO(" add new characteristic");
			char_info = g_malloc0(sizeof(bt_gatt_char_info_t));
			memcpy(char_info->uuid, event_data->char_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

			char_info->inst_id = event_data->char_id.inst_id;
			char_info->props = event_data->char_prop;
			svc_info->chars = g_slist_append(svc_info->chars, char_info);
		} else {
			/* If found, then return */
			BT_INFO("update char property as Characteristic browsed is already present");
			char_info->props |= event_data->char_prop;
		}
	} else {
		/* If Not success: Means Charc browse is completed  */
		/* Find char list from service in event */
		/* Find service list from address */
		svc_info_list = __bt_get_service_info_list(event_data->conn_status.conn_id);

		/* Find service info from service in event */
		svc_info = __bt_find_matching_service(svc_info_list, &event_data->srvc_id);

		/* Build char list from service in event */
		__bt_build_char_browse_info(event_data->conn_status.conn_id,
				svc_info, &browse_info);

		/* Create response and return by sending event*/
		/* Build Reply and send to service browse All Included services request of pending apps */
		__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_NONE,
				BT_GATT_GET_SERVICE_PROPERTIES,
				&browse_info,
				sizeof(bt_char_browse_info_t));
	}
}

static void __bt_handle_client_descriptor_search_result(
		event_gattc_descriptor_result_t *event_data)
{
	bt_gatt_service_info_list_t *svc_info_list;
	bt_gatt_service_info_t *svc_info;
	bt_gatt_char_info_t *char_info;
	bt_gatt_descriptor_info_t *desc_info;
	bt_descriptor_browse_info_t browse_info;

	BT_INFO("descriptor search result status [%d]", event_data->conn_status.status);

	memset(&browse_info, 0x00, sizeof(bt_descriptor_browse_info_t));

	/* If success */
	if (event_data->conn_status.status == OAL_STATUS_SUCCESS) {
		/* Find service list from address */
		svc_info_list = __bt_get_service_info_list(event_data->conn_status.conn_id);
		svc_info = __bt_find_matching_service(svc_info_list, &event_data->srvc_id);
		char_info = __bt_find_matching_charc(svc_info, &event_data->char_id);
		desc_info = __bt_find_matching_desc(char_info, &event_data->descr_id);

		/* If not found, add new descriptor and return */
		if (!desc_info) {
			desc_info = g_malloc0(sizeof(bt_gatt_descriptor_info_t));
			memcpy(desc_info->uuid, event_data->descr_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

			desc_info->inst_id = event_data->descr_id.inst_id;
			char_info->descs = g_slist_append(char_info->descs, desc_info);

		} else {
			/* If found, then return */
			BT_INFO("Characteristic browsed is already presesnt");
		}
	} else {
		/* If Not success */
		/* Find service list from address */
		/* Find included service list from service in event */
		/* Create response and return by sending event*/
		svc_info_list = __bt_get_service_info_list(event_data->conn_status.conn_id);

		/* Find service info from service in event */
		svc_info = __bt_find_matching_service(svc_info_list, &event_data->srvc_id);

		/* Find char info from char in event */
		char_info = __bt_find_matching_charc(svc_info, &event_data->char_id);

		/* Build descriptor list from char in event */
		__bt_build_descriptor_browse_info(event_data->conn_status.conn_id,
				svc_info, char_info, &browse_info);

		/* DBUS returni */
		__bt_gatt_handle_pending_request_info(BLUETOOTH_ERROR_NONE,
				BT_GATT_GET_CHARACTERISTIC_PROPERTIES,
				&browse_info,
				sizeof(bt_descriptor_browse_info_t));
	}
}

static void __bt_handle_client_characteristic_read_data(
		event_gattc_read_data *event_data)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_info_t *conn_info = NULL;
	bluetooth_gatt_client_char_prop_info_t read_info;

	/* Read Information data structures */
	GVariant *param = NULL;
	GVariant *data = NULL;
	GVariant *data_svc_uuid = NULL;
	GVariant *data_char_uuid = NULL;
	char *read_val = NULL;
	char *svc_uuid = NULL;
	char *char_uuid = NULL;
	int i;
	int uuid_len = 16;

	BT_INFO("+");

	//memset(&read_info, 0x00, sizeof(bt_gatt_handle_property_t));
	memset(&read_info, 0x00, sizeof(bluetooth_gatt_client_char_prop_info_t));

	/* Extract Address from conn_id of event data */
	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(
			event_data->uuid_status.conn_status.conn_id);

	BT_INFO("Characteristic Read result from addr [%s] status [%d]",
			conn_info->addr, event_data->uuid_status.conn_status.status);

	/* Fill char in buffer */
	memcpy(&read_info.characteristic.uuid,
			event_data->uuid_status.char_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	read_info.characteristic.instance_id = event_data->uuid_status.char_id.inst_id;

	/* Fill Service in buffer */
	memcpy(&read_info.svc.uuid,
			event_data->uuid_status.srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	read_info.svc.instance_id = event_data->uuid_status.srvc_id.id.inst_id;

	/* Fill remote device address */
	_bt_convert_addr_string_to_type(read_info.device_address.addr, conn_info->addr);

	/* Fill data and reply to all apps waiting for Read result on the same characteristic
Note: Even in case of failure, address, handles and result code should be returned */
	if (event_data->uuid_status.conn_status.status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;
	else {
		BT_ERR("read data len is [%d]", event_data->data_len);
		if (event_data->data_len > 0) {
			/* DEBUG */
			for (i = 0; i < event_data->data_len; i++)
				BT_INFO("Data[%d] = [0x%x]", i, event_data->data[i]);

			/* Read data */
			read_val = g_memdup(&event_data->data[0], event_data->data_len);

			data = g_variant_new_from_data(
					G_VARIANT_TYPE_BYTESTRING,
					read_val,
					event_data->data_len,
					TRUE, NULL, NULL);
		} else {
			BT_ERR("Characteristic Read success, but no data!!!");

			data = g_variant_new_from_data(
					G_VARIANT_TYPE_BYTESTRING,
					NULL,
					0,
					FALSE, NULL, NULL);
		}

		/* SVC uuid */
		svc_uuid = g_memdup(&event_data->uuid_status.srvc_id.id.uuid.uuid[0], uuid_len);

		data_svc_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				svc_uuid,
				16,
				TRUE, NULL, NULL);

		/* Char uuid */
		char_uuid = g_memdup(&event_data->uuid_status.char_id.uuid.uuid[0], uuid_len);

		data_char_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				char_uuid,
				16,
				TRUE, NULL, NULL);

		param = g_variant_new("(isn@ayin@ayin@ay)", result,
				conn_info->addr,
				16,
				data_svc_uuid,
				event_data->uuid_status.srvc_id.id.inst_id,
				16,
				data_char_uuid,
				event_data->uuid_status.char_id.inst_id,
				event_data->data_len,
				data);

		/* Send Event */
		_bt_send_event(BT_GATT_CLIENT_EVENT,
				BLUETOOTH_EVENT_GATT_READ_CHAR,
				param);
	}
	/* Send DBUS return */
	__bt_gatt_handle_pending_request_info(result,
			BT_GATT_READ_CHARACTERISTIC,
			&read_info,
			sizeof(bluetooth_gatt_client_char_prop_info_t));

	if (read_val)
		g_free(read_val);
	if (svc_uuid)
		g_free(svc_uuid);
	if (char_uuid)
		g_free(char_uuid);
}

/* Modified */
static void __bt_handle_client_descriptor_read_data(
		event_gattc_read_data *event_data)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_info_t *conn_info = NULL;
	bluetooth_gatt_client_desc_prop_info_t read_info;

	/* Read Information data structures */
	GVariant *param = NULL;
	GVariant *data = NULL;
	GVariant *data_svc_uuid = NULL;
	GVariant *data_char_uuid = NULL;
	GVariant *data_desc_uuid = NULL;
	char *read_val = NULL;
	char *svc_uuid = NULL;
	char *char_uuid = NULL;
	char *desc_uuid = NULL;
	int i;
	int uuid_len = 16;
	BT_INFO("+");

	memset(&read_info, 0x00, sizeof(bluetooth_gatt_client_desc_prop_info_t));

	/* Extract Address from conn_id of event data */
	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(
			event_data->uuid_status.conn_status.conn_id);

	BT_INFO("Descriptor Read result from addr [%s] status [%d]",
			conn_info->addr, event_data->uuid_status.conn_status.status);

	/* Fill descriptor informations in buffer */
	memcpy(&read_info.descriptor.uuid,
			event_data->uuid_status.descr_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	read_info.descriptor.instance_id = event_data->uuid_status.descr_id.inst_id;

	/* Fill Characteristic informations in buffer */
	memcpy(&read_info.characteristic.uuid,
			event_data->uuid_status.char_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	read_info.characteristic.instance_id = event_data->uuid_status.char_id.inst_id;

	/* Fill Service informations in buffer */
	memcpy(&read_info.svc.uuid,
			event_data->uuid_status.srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	read_info.svc.instance_id = event_data->uuid_status.srvc_id.id.inst_id;

	/* Fill remote device address */
	_bt_convert_addr_string_to_type(read_info.device_address.addr, conn_info->addr);

	/* Fill data and reply to all apps waiting for Read result on the same characteristic */
	if (event_data->uuid_status.conn_status.status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;
	else {
		BT_INFO("desc data len:%d", event_data->data_len);
		if (event_data->data_len > 0) {
			/* DEBUG */
			for (i = 0; i < event_data->data_len; i++)
				BT_DBG("Data[%d] = [0x%x]", i, event_data->data[i]);

			/* Read data */
			read_val = g_memdup(&event_data->data[0], event_data->data_len);

			data = g_variant_new_from_data(
					G_VARIANT_TYPE_BYTESTRING,
					read_val,
					event_data->data_len,
					TRUE, NULL, NULL);
		} else {
			BT_INFO("Descriptor Read success, but no data!!!");

			data = g_variant_new_from_data(
					G_VARIANT_TYPE_BYTESTRING,
					NULL,
					0,
					FALSE, NULL, NULL);
		}
		/* SVC uuid */
		svc_uuid = g_memdup(&event_data->uuid_status.srvc_id.id.uuid.uuid[0], uuid_len);

		data_svc_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				svc_uuid,
				16,
				TRUE, NULL, NULL);

		/* Char uuid */
		char_uuid = g_memdup(&event_data->uuid_status.char_id.uuid.uuid[0], uuid_len);

		data_char_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				char_uuid,
				16,
				TRUE, NULL, NULL);

		/* Desc uuid */
		desc_uuid = g_memdup(&event_data->uuid_status.descr_id.uuid.uuid[0], uuid_len);

		data_desc_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				desc_uuid,
				16,
				TRUE, NULL, NULL);

		param = g_variant_new("(isn@ayin@ayin@ayin@ay)", result,
				conn_info->addr,
				uuid_len,
				data_svc_uuid,
				event_data->uuid_status.srvc_id.id.inst_id,
				16,
				data_char_uuid,
				event_data->uuid_status.char_id.inst_id,
				16,
				data_desc_uuid,
				event_data->uuid_status.descr_id.inst_id,
				event_data->data_len,
				data);

		/* Send Event */
		_bt_send_event(BT_GATT_CLIENT_EVENT,
					BLUETOOTH_EVENT_GATT_READ_DESC,
					param);
	}
	BT_INFO("Send DBUS rpely for GATT Read Descriptor");
	/* Send DBUS return */
	__bt_gatt_handle_pending_request_info(result,
			BT_GATT_READ_DESCRIPTOR_VALUE,
			&read_info,
			sizeof(bluetooth_gatt_client_desc_prop_info_t));
	if (read_val)
		g_free(read_val);
	if (svc_uuid)
		g_free(svc_uuid);
	if (char_uuid)
		g_free(char_uuid);
	if (desc_uuid)
		g_free(desc_uuid);
}

static void __bt_handle_client_characteristic_write_data(
		event_gattc_write_data *event_data)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_info_t *conn_info = NULL;
	bluetooth_gatt_client_char_prop_info_t write_info;

	/* Read Information data structures */
	GVariant *param = NULL;
	GVariant *data_svc_uuid = NULL;
	GVariant *data_char_uuid = NULL;
	char *svc_uuid = NULL;
	char *char_uuid = NULL;
	int uuid_len = 16;
	BT_INFO("+");

	memset(&write_info, 0x00, sizeof(bluetooth_gatt_client_char_prop_info_t));

	/* Extract Address from conn_id of event data */
	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(
			event_data->conn_status.conn_id);

	BT_INFO("Characteristic Write callback from addr [%s] status [%d]",
			conn_info->addr, event_data->conn_status.status);

	/* Fill char in buffer */
	memcpy(&write_info.characteristic.uuid,
			event_data->char_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	write_info.characteristic.instance_id = event_data->char_id.inst_id;

	/* Fill Service in buffer */
	memcpy(&write_info.svc.uuid,
			event_data->srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	write_info.svc.instance_id = event_data->srvc_id.id.inst_id;

	/* Fill remote device address */
	_bt_convert_addr_string_to_type(write_info.device_address.addr, conn_info->addr);

	if (event_data->conn_status.status != OAL_STATUS_SUCCESS) {
		result = BLUETOOTH_ERROR_INTERNAL;
		goto done;
	}

	/* Build event */
	/* SVC uuid */
	svc_uuid = g_memdup(&event_data->srvc_id.id.uuid.uuid[0], uuid_len);

	data_svc_uuid = g_variant_new_from_data(
			G_VARIANT_TYPE_BYTESTRING,
			svc_uuid,
			uuid_len,
			TRUE, NULL, NULL);

	/* Char uuid */
	char_uuid = g_memdup(&event_data->char_id.uuid.uuid[0], uuid_len);

	data_char_uuid = g_variant_new_from_data(
			G_VARIANT_TYPE_BYTESTRING,
			char_uuid,
			uuid_len,
			TRUE, NULL, NULL);

	param = g_variant_new("(isn@ayin@ayi)", result,
			conn_info->addr,
			16,
			data_svc_uuid,
			event_data->srvc_id.id.inst_id,
			16,
			data_char_uuid,
			event_data->char_id.inst_id);

	/* Send Event */
	_bt_send_event(BT_GATT_CLIENT_EVENT,
			BLUETOOTH_EVENT_GATT_WRITE_CHAR,
			param);

	/* Free data */
	if (svc_uuid)
		g_free(svc_uuid);
	if (char_uuid)
		g_free(char_uuid);
done:
	/* Send DBUS return */
	__bt_gatt_handle_pending_request_info(result,
			BT_GATT_WRITE_CHARACTERISTIC_VALUE_BY_TYPE,
			&write_info,
			sizeof(bluetooth_gatt_client_char_prop_info_t));
}


static void __bt_handle_client_descriptor_write_data(
		event_gattc_write_data *event_data)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_info_t *conn_info = NULL;
	bluetooth_gatt_client_desc_prop_info_t write_info;

	/* Write Information data structures */
	GVariant *param = NULL;
	GVariant *data_svc_uuid = NULL;
	GVariant *data_char_uuid = NULL;
	GVariant *data_desc_uuid = NULL;
	char *svc_uuid = NULL;
	char *char_uuid = NULL;
	char *desc_uuid = NULL;
	int uuid_len = 16;
	BT_INFO("+");

	memset(&write_info, 0x00, sizeof(bluetooth_gatt_client_desc_prop_info_t));

	/* Extract Address from conn_id of event data */
	conn_info = __bt_find_remote_gatt_server_info_from_conn_id(
			event_data->conn_status.conn_id);

	if (NULL == conn_info) {

		BT_INFO("Failed to get the conn info for conn_id [%d]", event_data->conn_status.conn_id);
		return;
	}

	BT_INFO("Descriptor Write callback from addr [%s] status [%d]",
			conn_info->addr, event_data->conn_status.status);

	/* Fill descriptor informations in buffer */
	memcpy(&write_info.descriptor.uuid,
			event_data->descr_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	write_info.descriptor.instance_id = event_data->descr_id.inst_id;

	/* Fill Characteristic informations in buffer */
	memcpy(&write_info.characteristic.uuid,
			event_data->char_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	write_info.characteristic.instance_id = event_data->char_id.inst_id;

	/* Fill Service informations in buffer */
	memcpy(&write_info.svc.uuid,
			event_data->srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	write_info.svc.instance_id = event_data->srvc_id.id.inst_id;

	/* Fill remote device address */
	_bt_convert_addr_string_to_type(write_info.device_address.addr, conn_info->addr);

	if (event_data->conn_status.status != OAL_STATUS_SUCCESS) {
		result = BLUETOOTH_ERROR_INTERNAL;
		goto done;
	}

	/* Build event */
	/* SVC uuid */
	svc_uuid = g_memdup(&event_data->srvc_id.id.uuid.uuid[0], uuid_len);

	data_svc_uuid = g_variant_new_from_data(
			G_VARIANT_TYPE_BYTESTRING,
			svc_uuid,
			uuid_len,
			TRUE, NULL, NULL);

	/* Char uuid */
	char_uuid = g_memdup(&event_data->char_id.uuid.uuid[0], uuid_len);

	data_char_uuid = g_variant_new_from_data(
			G_VARIANT_TYPE_BYTESTRING,
			char_uuid,
			uuid_len,
			TRUE, NULL, NULL);

	/* Desc uuid */
	desc_uuid = g_memdup(&event_data->descr_id.uuid.uuid[0], uuid_len);

	data_desc_uuid = g_variant_new_from_data(
			G_VARIANT_TYPE_BYTESTRING,
			desc_uuid,
			uuid_len,
			TRUE, NULL, NULL);

	param = g_variant_new("(isn@ayin@ayin@ayi)", result,
			conn_info->addr,
			16,
			data_svc_uuid,
			event_data->srvc_id.id.inst_id,
			16,
			data_char_uuid,
			event_data->char_id.inst_id,
			16,
			data_desc_uuid,
			event_data->descr_id.inst_id);

	/* Send Event */
	_bt_send_event(BT_GATT_CLIENT_EVENT,
			BLUETOOTH_EVENT_GATT_WRITE_DESC,
			param);

	/* Free data */
	if (svc_uuid)
		g_free(svc_uuid);
	if (char_uuid)
		g_free(char_uuid);
	if (desc_uuid)
		g_free(desc_uuid);
done:
	/* Send DBUS return */
	__bt_gatt_handle_pending_request_info(result,
			BT_GATT_WRITE_DESCRIPTOR_VALUE,
			&write_info,
			sizeof(bluetooth_gatt_client_desc_prop_info_t));
}

static void __bt_hanlde_le_device_disconnection(event_dev_conn_status_t *event_data)
{
	int result = BLUETOOTH_ERROR_INTERNAL;
	char *address = g_malloc0(BT_ADDRESS_STRING_SIZE);

	_bt_convert_addr_type_to_string(address, (unsigned char *)event_data->address.addr);

	/* DBUS Return with fail of pending BT_CONNECT_LE for all the apps */
	BT_INFO("Local GATT Client disconnected: Remote addr[%s] ", address);

	__bt_gatt_handle_pending_request_info(result, BT_CONNECT_LE, address,
						 BT_ADDRESS_STRING_SIZE);
	g_free(address);
}

static void __bt_handle_client_notification_registered(
		event_gattc_regdereg_notify_t *event_data,
		gboolean is_registered)
{
	int result = BLUETOOTH_ERROR_NONE;
	struct gatt_server_info_t *conn_info = NULL;
	bt_gatt_notif_reg_info_t notif_info;
	BT_INFO("+");

	memset(&notif_info, 0x00, sizeof(bt_gatt_notif_reg_info_t));

	BT_INFO("Client Interface [%d] status [%d]",
			event_data->client_if,
			event_data->status);

	/* Extract Address from conn_id of event data */
	conn_info = __bt_find_remote_gatt_server_info_from_client_if(
			event_data->client_if);

	if (!conn_info) {
		BT_INFO("Connection Info is not present, return");
		return;
	}
	BT_INFO("Notification Registered for addr [%s]",
			conn_info->addr);


	/* Fill svc informations in buffer */
	memcpy(&notif_info.svc_uuid,
			event_data->srvc_id.id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	notif_info.svc_inst = event_data->srvc_id.id.inst_id;

	/* Fill char in buffer */
	memcpy(&notif_info.char_uuid,
			event_data->char_id.uuid.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);
	notif_info.char_inst = event_data->char_id.inst_id;

	/* Fill remote device address */
	_bt_convert_addr_string_to_type(notif_info.addr.addr, conn_info->addr);

	notif_info.is_registered = is_registered;

	if (event_data->status != OAL_STATUS_SUCCESS)
		result = BLUETOOTH_ERROR_INTERNAL;

	/* Send DBUS Return for BT_GATT_WATCH_CHARACTERISTIC */
	__bt_gatt_handle_pending_request_info(result,
			BT_GATT_WATCH_CHARACTERISTIC,
			&notif_info,
			sizeof(bt_gatt_notif_reg_info_t));
}

static void __bt_handle_client_notification_data(event_gattc_notify_data *event_data)
{
	/* No status in this event from OAL */
	int result = BLUETOOTH_ERROR_NONE;

	/* Read Information data structures */
	GVariant *param = NULL;
	GVariant *data = NULL;
	GVariant *data_svc_uuid = NULL;
	GVariant *data_char_uuid = NULL;
	char *read_val = NULL;
	char *svc_uuid = NULL;
	char *char_uuid = NULL;
	char *addr = NULL;
	int i;
	int uuid_len = 16;
	BT_INFO("+");

	BT_INFO("Notifcation of charc data changed");

	if (event_data->data_len > 0) {
		/* DEBUG */
		for (i = 0; i < event_data->data_len; i++)
			BT_INFO("Data[%d] = [0x%x]", i, event_data->data[i]);

		/* Fill address */
		addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
		_bt_convert_addr_type_to_string(addr,
				(unsigned char *)&(event_data->address.addr));

		/* Read data */
		read_val = g_memdup(&event_data->data[0], event_data->data_len);

		data = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				read_val,
				event_data->data_len,
				TRUE, NULL, NULL);
		/* SVC uuid */
		svc_uuid = g_memdup(&event_data->srvc_id.id.uuid.uuid[0], uuid_len);

		data_svc_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				svc_uuid,
				uuid_len,
				TRUE, NULL, NULL);

		/* Char uuid */
		char_uuid = g_memdup(&event_data->char_id.uuid.uuid[0], uuid_len);

		data_char_uuid = g_variant_new_from_data(
				G_VARIANT_TYPE_BYTESTRING,
				char_uuid,
				uuid_len,
				TRUE, NULL, NULL);

		/* Build Param */
		param = g_variant_new("(isn@ayin@ayin@ay)", result,
				addr,
				16,
				data_svc_uuid,
				event_data->srvc_id.id.inst_id,
				16,
				data_char_uuid,
				event_data->char_id.inst_id,
				event_data->data_len,
				data);

		/* Send Event */
		_bt_send_event(BT_GATT_CLIENT_EVENT,
				BLUETOOTH_EVENT_GATT_CHAR_VAL_CHANGED,
				param);
	} else {
		BT_ERR("No Data!!");
	}
	/* Free data */
	if (read_val)
		g_free(read_val);
	if (svc_uuid)
		g_free(svc_uuid);
	if (char_uuid)
		g_free(char_uuid);
	if (addr)
		g_free(addr);
}

gboolean _bt_is_remote_gatt_device_connected(bluetooth_device_address_t *address)
{
	char *addr;
	struct gatt_server_info_t *conn_info =  NULL;
	gboolean connected = FALSE;
	BT_INFO("+");

	addr = g_malloc0(BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr,
			(unsigned char *)&(address->addr));

	BT_INFO("Check Connected or not for [%s]", addr);
	/* Check if device is already in connected list */
	conn_info = __bt_find_remote_gatt_server_info(addr);

	if (conn_info) {
		BT_INFO("Remote GATT Server device [%s] is Connected", conn_info->addr);
		connected = TRUE;
	} else
		BT_INFO("Remote GATT Server Device [%s] is not Connected", addr);
	g_free(addr);
	return connected;
}


int _bt_connect_le_device(bluetooth_device_address_t *address,
		int auto_connect, int client_id)
{
	struct gatt_server_info_t *conn_info = NULL;
	struct gatt_out_conn_info_t *out_conn_info = NULL;

	invocation_info_t *req_info = NULL;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;
	char *remote_address = NULL;

	BT_CHECK_PARAMETER(address, return);

	BT_INFO("+");

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, address->addr);
	BT_INFO("GATT Client connect request for address [%s] client instance [%d]",
			addr, client_id);


	/* Check if Remote Device is already under connection progress */
	req_info = _bt_get_request_info_data_from_function_name(BT_CONNECT_LE);
	if (req_info) {
		remote_address = (char*)req_info->user_data;
		if (remote_address && !strcasecmp(remote_address, addr)) {/* Address matched */
			BT_INFO("Already Connection ongoing for same remote GATT Server address [%s]", remote_address);
			/* Return and wait for events to be sent to all apps */
			g_free(addr);
			return BLUETOOTH_ERROR_IN_PROGRESS;
		}
	}

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_ERR("GATT Server is already connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_ALREADY_CONNECT;
	}

	/* TODO Check Requirement of holding Advertisement before initiating LE connect */

	/* Check if app sent 0 client id for connection, in such case, use default gatt client ID */
	if (client_id == 0) {
		BT_INFO("GATT CLient connect request sent by an app without any client instance [%d]",
				client_id);
		BT_INFO("Assign default GATT client id [%d]", gatt_default_client);
		client_id = gatt_default_client;
	}

	BT_INFO("Connect using CLient ID [%d]", client_id);
	ret = gattc_connect(client_id, (bt_address_t*)(address), auto_connect);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Mark this as outgoing connection */
	out_conn_info = g_new0(struct gatt_out_conn_info_t, 1);
	out_conn_info->addr = g_strdup(addr);
	out_conn_info->client_id = client_id;
	BT_INFO("Added outgoing connection info addr[%s]", out_conn_info->addr);
	outgoing_gatt_conn_list = g_slist_append(outgoing_gatt_conn_list, out_conn_info);

	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_get_primary_services(char *address)
{
	BT_CHECK_PARAMETER(address, return);
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	int ret = OAL_STATUS_SUCCESS;
	BT_INFO("+");

	/* Check if any app is already browsing primary services on the same remote GATT Server */
	req_info = _bt_get_request_info_data(BT_GATT_GET_PRIMARY_SERVICES, address);
	if (req_info) {
		BT_INFO("Already Primary Service Browsing ongoing for same rmeote GATT Server");
		/* Return and wait for events to be sent to all apps */
		return BLUETOOTH_ERROR_NONE;
	}

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(address);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	/* Send Primary Service Browsing request to stack */
	ret = gattc_search_service(conn_info->connection_id, NULL);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_get_all_characteristic(bluetooth_gatt_client_svc_prop_info_t *svc)
{
	BT_CHECK_PARAMETER(svc, return);
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	bluetooth_gatt_client_svc_prop_info_t *prop;
	oal_gatt_srvc_id_t srvc_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;
	BT_INFO("+");

	/* Check if any app is already browsing characteristics of the same service on the same remote GATT Server */
	req_info = _bt_get_request_info_data_from_function_name(BT_GATT_GET_SERVICE_PROPERTIES);
	if (req_info) {
		prop = (bluetooth_gatt_client_svc_prop_info_t*)req_info->user_data;
		if (prop && !memcmp(svc->device_address.addr, prop->device_address.addr, sizeof(bluetooth_device_address_t))
				&& memcmp(prop->svc.uuid, svc->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN)
				&& prop->svc.instance_id == svc->svc.instance_id) {
			BT_INFO("Already Properties browsing for Primary Service ongoing for same remote GATT Server");
			/* Return and wait for events to be sent to all apps */
			return BLUETOOTH_ERROR_NONE;
		}
	}

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, svc->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = svc->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, svc->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Search All Characteristic */
	ret = gattc_get_characteristic(conn_info->connection_id, &srvc_id, NULL);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_get_all_characteristic_properties(
		bluetooth_gatt_client_char_prop_info_t *chr)
{
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	bluetooth_gatt_client_char_prop_info_t *prop;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;

	BT_CHECK_PARAMETER(chr, return);

	BT_INFO("+");

	/* Check if any app is already browsing descriptors of the same char of
	   particular service on the same remote GATT Server */
	req_info = _bt_get_request_info_data_from_function_name(BT_GATT_GET_CHARACTERISTIC_PROPERTIES);
	if (req_info) {
		prop = (bluetooth_gatt_client_char_prop_info_t*)req_info->user_data;
		if (prop && !memcmp(chr->device_address.addr, prop->device_address.addr, sizeof(bluetooth_device_address_t)) /* Address matched */
				&& memcmp(chr->svc.uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Service UUID matched */
				&& chr->svc.instance_id == prop->svc.instance_id /* Service Instance ID matched */
				&& memcmp(chr->characteristic.uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Characteristic UUID matched */
				&& chr->characteristic.instance_id == prop->characteristic.instance_id) { /* Characteristic Instance ID matched */
			BT_INFO("Already Properties browsing for Characteristic ongoing for same remote GATT Server");
			/* Return and wait for events to be sent to all apps */
			return BLUETOOTH_ERROR_NONE;
		}
	}

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, chr->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = chr->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, chr->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = chr->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, chr->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Search All Descriptors */
	ret = gattc_get_descriptor(conn_info->connection_id, &srvc_id, &char_id, NULL);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_read_characteristic_value(
		bluetooth_gatt_client_char_prop_info_t *chr)
{
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	bluetooth_gatt_client_char_prop_info_t *prop;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;

	BT_CHECK_PARAMETER(chr, return);

	BT_INFO("+");

	/* Check if any app is already Reading characteristic of the same char of
	   particular service on the same remote GATT Server */
	req_info = _bt_get_request_info_data_from_function_name(BT_GATT_READ_CHARACTERISTIC);
	if (req_info) {
		prop = (bluetooth_gatt_client_char_prop_info_t*)req_info->user_data;
		if (prop && !memcmp(chr->device_address.addr, prop->device_address.addr, sizeof(bluetooth_device_address_t)) /* Address matched */
				&& memcmp(chr->svc.uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Service UUID matched */
				&& chr->svc.instance_id == prop->svc.instance_id /* Service Instance ID matched */
				&& memcmp(chr->characteristic.uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Characteristic UUID matched */
				&& chr->characteristic.instance_id == prop->characteristic.instance_id) { /* Characteristic Instance ID matched */
			BT_INFO("Already Characteristic value Read operation in progress for same remote GATT Server");
			/* Return and wait for events to be sent to all apps */
			return BLUETOOTH_ERROR_NONE;
		}
	}

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, chr->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = chr->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, chr->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = chr->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, chr->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Search All Descriptors */
	ret = gattc_read_characteristic(conn_info->connection_id, &srvc_id, &char_id, OAL_GATT_AUTH_REQ_NONE);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_read_descriptor_value(
		bluetooth_gatt_client_desc_prop_info_t *desc)
{
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	bluetooth_gatt_client_desc_prop_info_t *prop;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	oal_gatt_id_t desc_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;

	BT_CHECK_PARAMETER(desc, return);

	BT_INFO("+");

	/* Check if any app is already Reading descriptors of the same char of
	   particular service on the same remote GATT Server */
	req_info = _bt_get_request_info_data_from_function_name(BT_GATT_READ_DESCRIPTOR_VALUE);
	if (req_info) {
		prop = (bluetooth_gatt_client_desc_prop_info_t*)req_info->user_data;
		if (prop && !memcmp(desc->device_address.addr, prop->device_address.addr, sizeof(bluetooth_device_address_t)) /* Address matched */
				&& memcmp(desc->svc.uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Service UUID matched */
				&& desc->svc.instance_id == prop->svc.instance_id /* Service Instance ID matched */
				&& memcmp(desc->characteristic.uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Characteristic UUID matched */
				&& desc->characteristic.instance_id == prop->characteristic.instance_id /* Characteristic Instance ID matched */
				&& memcmp(desc->descriptor.uuid, prop->descriptor.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Descriptor UUID matched */
				&& desc->descriptor.instance_id == prop->descriptor.instance_id) { /* Descriptor Instance ID matched */
			BT_INFO("Already Descriptor value Read operation in progress for same remote GATT Server");
			/* Return and wait for events to be sent to all apps */
			return BLUETOOTH_ERROR_NONE;
		}
	}

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, desc->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = desc->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, desc->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = desc->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, desc->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	desc_id.inst_id = desc->descriptor.instance_id;
	memcpy(desc_id.uuid.uuid, desc->descriptor.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Search All Descriptors */
	ret = gattc_read_descriptor(conn_info->connection_id, &srvc_id, &char_id,
			&desc_id, OAL_GATT_AUTH_REQ_NONE);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

/*acquire Notify*/
int _bt_gatt_acquire_notify(bluetooth_gatt_client_char_prop_info_t *chr, int *fd, int *mtu)
{
	struct gatt_server_info_t *conn_info = NULL;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;

	BT_CHECK_PARAMETER(chr, return);

	BT_INFO("+");

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, chr->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = chr->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, chr->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = chr->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, chr->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

		ret = gattc_acquire_notify(conn_info->connection_id, &srvc_id, &char_id, fd, mtu);

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	BT_INFO("GATT characterstics FD [%d]  mtu[%d]", *fd, *mtu);
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

/*acquire Write*/
int _bt_gatt_acquire_write(bluetooth_gatt_client_char_prop_info_t *chr, int *fd, int *mtu)
{

	struct gatt_server_info_t *conn_info = NULL;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;

	BT_CHECK_PARAMETER(chr, return);

	BT_INFO("+");

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, chr->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = chr->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, chr->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = chr->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, chr->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

		ret = gattc_acquire_write(conn_info->connection_id, &srvc_id, &char_id,
					OAL_GATT_AUTH_REQ_NONE, fd, mtu);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	BT_INFO("GATT characterstics FD [%d]  mtu [%d]", *fd, *mtu);
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;

}


/* Write Characteristic */
int _bt_gatt_write_characteristic_value_by_type(
		bluetooth_gatt_client_char_prop_info_t *chr,
		bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_write_type_e write_type)
{
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	bluetooth_gatt_client_char_prop_info_t *prop;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;
	int k;

	BT_CHECK_PARAMETER(chr, return);
	BT_CHECK_PARAMETER(data, return);

	BT_INFO("+");

	/* Check if any app is already writing same char of
	   particular service on the same remote GATT Server */
	req_info = _bt_get_request_info_data_from_function_name(BT_GATT_WRITE_CHARACTERISTIC_VALUE_BY_TYPE);
	if (req_info) {
		prop = (bluetooth_gatt_client_char_prop_info_t*)req_info->user_data;
		if (prop && !memcmp(chr->device_address.addr, prop->device_address.addr, sizeof(bluetooth_device_address_t)) /* Address matched */
				&& memcmp(chr->svc.uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Service UUID matched */
				&& chr->svc.instance_id == prop->svc.instance_id /* Service Instance ID matched */
				&& memcmp(chr->characteristic.uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Characteristic UUID matched */
				&& chr->characteristic.instance_id == prop->characteristic.instance_id) { /* Characteristic Instance ID matched */
			BT_INFO("Already Characteristic Write Value operation in progress for same remote GATT Server");
			/* Return and wait for events to be sent to all apps */
			return BLUETOOTH_ERROR_NONE;
		}
	}

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, chr->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = chr->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, chr->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = chr->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, chr->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Write CHar value */
	/* DEBUG */
	BT_INFO("Connection ID [%d] write type [%d] data length [%d]", conn_info->connection_id, write_type, data->length);
	for (k = 0; k < data->length; k++)
		BT_INFO("Data[%d] [0x%x]", k, data->data[k]);

	ret = gattc_write_characteristic(conn_info->connection_id,
			&srvc_id, &char_id,
			(oal_gatt_write_type_t)write_type, data->length,
			OAL_GATT_AUTH_REQ_NONE, (char *)(&data->data[0]));
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

/* Write Descriptor */
int _bt_gatt_write_descriptor_value_by_type(
		bluetooth_gatt_client_desc_prop_info_t *desc,
		bluetooth_gatt_att_data_t *data,
		bluetooth_gatt_write_type_e write_type)
{
	struct gatt_server_info_t *conn_info = NULL;
	invocation_info_t *req_info = NULL;
	bluetooth_gatt_client_desc_prop_info_t *prop;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	oal_gatt_id_t desc_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;
	int k;

	BT_CHECK_PARAMETER(desc, return);
	BT_CHECK_PARAMETER(data, return);

	BT_INFO("+");

	/* Check if any app is already writing on same Descriptor of the same char of
	   particular service on the same remote GATT Server */
	req_info = _bt_get_request_info_data_from_function_name(BT_GATT_WRITE_DESCRIPTOR_VALUE);
	if (req_info) {
		prop = (bluetooth_gatt_client_desc_prop_info_t*)req_info->user_data;
		if (prop && !memcmp(desc->device_address.addr, prop->device_address.addr, sizeof(bluetooth_device_address_t)) /* Address matched */
				&& memcmp(desc->svc.uuid, prop->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Service UUID matched */
				&& desc->svc.instance_id == prop->svc.instance_id /* Service Instance ID matched */
				&& memcmp(desc->characteristic.uuid, prop->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Characteristic UUID matched */
				&& desc->characteristic.instance_id == prop->characteristic.instance_id /* Characteristic Instance ID matched */
				&& memcmp(desc->descriptor.uuid, prop->descriptor.uuid, BLUETOOTH_UUID_HEX_MAX_LEN) /* Descriptor UUID matched */
				&& desc->descriptor.instance_id == prop->descriptor.instance_id) { /* Descriptor Instance ID matched */
			BT_INFO("Already Descriptor value Write operation in progress for same remote GATT Server");
			/* Return and wait for events to be sent to all apps */
			return BLUETOOTH_ERROR_NONE;
		}
	}

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, desc->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}

	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = desc->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, desc->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = desc->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, desc->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	desc_id.inst_id = desc->descriptor.instance_id;
	memcpy(desc_id.uuid.uuid, desc->descriptor.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* DEBUG */
	BT_INFO("Connection ID [%d] write type [%d] data length [%d]", conn_info->connection_id, write_type, data->length);
	for (k = 0; k < data->length; k++)
		BT_INFO("Data[%d] [0x%x]", k, data->data[k]);

	ret = gattc_write_descriptor(conn_info->connection_id,
			&srvc_id, &char_id, &desc_id,
			(oal_gatt_write_type_t)write_type, data->length,
			OAL_GATT_AUTH_REQ_NONE, (char *)(&data->data[0]));
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_watch_characteristic(
		bluetooth_gatt_client_char_prop_info_t *chr,
		int client_id,
		gboolean is_notify)
{
	struct gatt_server_info_t *conn_info = NULL;
	oal_gatt_srvc_id_t srvc_id;
	oal_gatt_id_t char_id;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;

	BT_CHECK_PARAMETER(chr, return);

	BT_INFO("Client ID [%d] Is Notify [%d]", client_id, is_notify);

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, chr->device_address.addr);

	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		BT_INFO("GATT Server [%s] is connected, conn Id [%d]",
				conn_info->addr, conn_info->connection_id);
	} else {
		BT_ERR("GATT Server is not yet connected..");
		g_free(addr);
		return BLUETOOTH_ERROR_NOT_CONNECTED;
	}
	srvc_id.is_prmry = TRUE;
	srvc_id.id.inst_id = chr->svc.instance_id;
	memcpy(srvc_id.id.uuid.uuid, chr->svc.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	char_id.inst_id = chr->characteristic.instance_id;
	memcpy(char_id.uuid.uuid, chr->characteristic.uuid, BLUETOOTH_UUID_HEX_MAX_LEN);

	/* Register or unregister Notification characteristic */
	if (is_notify)
		ret = gattc_register_for_notification(conn_info->client_id,
				(bt_address_t*)&(chr->device_address),
				&srvc_id, &char_id);
	else
		ret = gattc_deregister_for_notification(conn_info->client_id,
				(bt_address_t*)&(chr->device_address),
				&srvc_id, &char_id);

	BT_INFO("Result[%d]", ret);
	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}


int _bt_disconnect_le_device(bluetooth_device_address_t *address,
		int client_id)
{
	struct gatt_server_info_t *conn_info = NULL;
	struct gatt_client_info_t *rem_client_conn_info = NULL;
	invocation_info_t *req_info = NULL;
	int ret = OAL_STATUS_SUCCESS;
	char *addr;
	char *remote_address = NULL;

	BT_CHECK_PARAMETER(address, return);

	BT_INFO("+");

	addr = g_malloc0(sizeof(char) * BT_ADDRESS_STRING_SIZE);
	_bt_convert_addr_type_to_string(addr, address->addr);
	BT_INFO("GATT Client Disconnect request for address [%s]", addr);

	/* Check if Remote Device is already under connection progress */
	req_info = _bt_get_request_info_data_from_function_name(BT_DISCONNECT_LE);
	if (req_info) {
		remote_address = (char*)req_info->user_data;
		if (remote_address && !strcasecmp(remote_address, addr)) {/* Address matched */
			BT_INFO("Already DisConnection ongoing for same remote GATT Server address [%s]", remote_address);
			/* Return success and wait for events to be sent to all apps */
			g_free(addr);
			return BLUETOOTH_ERROR_IN_PROGRESS;
		}
	}
	/* Check if remote GATT Server is connected or not */
	conn_info = __bt_find_remote_gatt_server_info(addr);
	if (conn_info) {
		/* Check if app sent 0 client id for Disconnection, in such case, use default gatt client ID */
		if (client_id == 0) {
			BT_INFO("GATT CLient Disconnect request sent by an app without any client instance [%d]",
					client_id);
			BT_INFO("Assign default GATT client id [%d]", gatt_default_client);
			client_id = gatt_default_client;
		}

		BT_INFO("Disconnect remote gatt server using CLient ID [%d] Connection ID [%d]", client_id, conn_info->connection_id);
		ret = gattc_disconnect(client_id, (bt_address_t*)(address),
				conn_info->connection_id);
	} else {
		/* check if remote client is connected */
		rem_client_conn_info = __bt_find_remote_gatt_client_info(addr);

		if (!rem_client_conn_info || client_id != 0) {
			BT_ERR("GATT device is not connected..");
			g_free(addr);
			return BLUETOOTH_ERROR_NOT_IN_OPERATION;
		}

		BT_INFO("Disconnect remote gatt client ");

		ret = gatts_disconnect(rem_client_conn_info->instance_id,
			(bt_address_t*)(address), rem_client_conn_info->connection_id);
	}

	if (ret != OAL_STATUS_SUCCESS) {
		BT_ERR("ret: %d", ret);
		g_free(addr);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_free(addr);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_gatt_watch_service_changed_indication(const char *sender,
		bluetooth_device_address_t *address,
		gboolean is_enabled)
{
	int k;
	bt_service_app_info_t *info = NULL;

	BT_INFO("Enable Servic changed Indication watcher [%d] for app [%s]",
			is_enabled, sender);

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		if (g_strcmp0(sender, info->sender) == 0 &&
				memcmp(info->address.addr, address->addr,
					sizeof(bluetooth_device_address_t)) == 0) {
			BT_INFO("Found GATT client App.. UUID [%s], sender [%s]", info->uuid, info->sender);
			info->is_watcher_enabled = is_enabled;
		}
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_unregister_gatt_client_instance(const char *sender, int client_id)
{
	BT_INFO("Unregister Allocated GATT CLient instance [%s] Client ID [%d]",
			sender, client_id);
	int result = BLUETOOTH_ERROR_NONE;
	int k;

	/* Unregister CLient instance associated with address X. It is possible that another app still
	   has client_id valid for same remote address */
	bt_service_app_info_t *info = NULL;

	for (k = 1; k < MAX_APPS_SUPPORTED; k++) {
		info = &numapps[k];

		/* Exact matching of sender */
		if (!g_strcmp0(info->sender, sender) && info->client_id == client_id) {  /* Check for only valid GATT client Instance */
			BT_INFO("Unregister GATT client instance [%d]", info->client_id);
			result = __bt_do_unregister_gatt_instance(info->client_id);
			if (result != BLUETOOTH_ERROR_NONE)
				BT_ERR("Error in unregistering GATT Client Interface");

			break;
		}
	}
	return result;
}

#endif
