/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <glib.h>
#include <string.h>
#include <dlog.h>
#include <vconf.h>
#include <vconf-internal-bt-keys.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-service-common.h"
#include "bt-service-event.h"
#include "bt-service-main.h"
#include "bt-service-core-adapter.h"
#include "bt-service-core-adapter-le.h"
#include "bt-service-device.h"

#ifdef TIZEN_FEATURE_BT_DPM
#include "bt-service-dpm.h"
#endif

#define BT_MEDIA_OBJECT_PATH "/Musicplayer"
static GDBusConnection *manager_conn;



static void __bt_ipsp_property_changed_event(GVariant *msg, const char *path);


static void __bt_device_property_changed_event(GVariant *msg, const char *path)
{
	BT_DBG("+");

	int event;
	int result = BLUETOOTH_ERROR_NONE;
	GVariantIter value_iter;
	GVariant *val;
	char *property = NULL;
	char *address;
	GVariant *param = NULL;
	g_variant_iter_init(&value_iter, msg);

	while ((g_variant_iter_loop(&value_iter, "{sv}", &property, &val))) {
		BT_DBG("Property %s", property);
		if (strcasecmp(property, "IpspConnected") == 0) {
			gboolean connected = FALSE;

			g_variant_get(val, "b", &connected);


			event = connected ? BLUETOOTH_EVENT_IPSP_CONNECTED :
							BLUETOOTH_EVENT_IPSP_DISCONNECTED;

			address = g_malloc0(BT_ADDRESS_STRING_SIZE);

			_bt_convert_device_path_to_address(path, address);

			BT_DBG("Ipspconnected: %d", connected);
			BT_DBG("address: %s", address);
			param = g_variant_new("(is)", result, address);

			/* Send event to application */
			_bt_send_event(BT_DEVICE_EVENT,
					event,
					param);
			g_free(address);
		} else if (strcasecmp(property, "IpspBtInterfaceInfo") == 0) {
			char *ifname = NULL;

			g_variant_get(val, "s", &ifname);

			address = g_malloc0(BT_ADDRESS_STRING_SIZE);

			_bt_convert_device_path_to_address(path, address);

			BT_DBG("Ipsp BT Interface Name: %s", ifname);
			BT_DBG("address: %s", address);
			param = g_variant_new("(iss)", result, address, ifname);

			/* Send event to application */
			_bt_send_event(BT_DEVICE_EVENT,
					BLUETOOTH_EVENT_IPSP_INTERFACE_INFO,
					param);
			g_free(address);
		}
	}
	BT_DBG("-");
}

void _bt_handle_ipsp_device_event(GVariant *msg, const char *member, const char *path)
{
	int event = 0;
	int result = BLUETOOTH_ERROR_NONE;
	char *address;
	GVariant *param = NULL;
	ret_if(path == NULL);

	if  (strcasecmp(member, "IpspStateChanged") == 0) {
		gboolean connected = FALSE;
		char *ifname = NULL;
		GVariant *ipsp_param = NULL;

		g_variant_get(msg, "(bs)", &connected, &ifname);

		event = connected ? BLUETOOTH_EVENT_IPSP_CONNECTED :
						BLUETOOTH_EVENT_IPSP_DISCONNECTED;

		address = g_malloc0(BT_ADDRESS_STRING_SIZE);
		_bt_convert_device_path_to_address(path, address);

		BT_DBG("Ipsp BT Interface Name: %s, address: %s", ifname, address);

		param = g_variant_new("(iss)", result, address, ifname);
		ipsp_param = g_variant_new("(ss)", ifname, address);

		g_free(ifname);

		/* Set Ipv6 Addr */
		GDBusProxy *ipsp_proxy;
		if (connected) {
			BT_DBG("IPSP connected, Set Ipv6 Addr");
			ipsp_proxy = _bt_get_ipsp_proxy();
			if (ipsp_proxy == NULL) {
				BT_ERR("can not get ipsp proxy");
				g_free(address);
				g_variant_unref(param);
				g_variant_unref(ipsp_param);
				return;
			}

			g_dbus_proxy_call(ipsp_proxy, "SetIpv6Addr",
					ipsp_param, G_DBUS_CALL_FLAGS_NONE,
					-1, NULL, NULL, NULL);
		} else {
			g_variant_unref(ipsp_param);
			BT_DBG("IPSP disconnected");
			ipsp_proxy = _bt_get_ipsp_proxy();
			if (ipsp_proxy == NULL) {
				BT_ERR("can not get ipsp proxy");
				g_free(address);
				g_variant_unref(param);
				return;
			}

			g_dbus_proxy_call(ipsp_proxy, "DisableIpsp",
					NULL, G_DBUS_CALL_FLAGS_NONE,
					-1, NULL, NULL, NULL);
		}

		/* Send event to application */
		_bt_send_event(BT_DEVICE_EVENT,	event, param);
		g_free(address);
	}
}


static  void __bt_manager_event_filter(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{
	const char *path = object_path;

	if (signal_name == NULL)
		return;

	if (g_strcmp0(interface_name, BT_PROPERTIES_INTERFACE) == 0) {
		char *inf_name = NULL;
		GVariant *val = NULL;

		if (strncmp(path, BT_MEDIA_OBJECT_PATH,
				strlen(BT_MEDIA_OBJECT_PATH)) == 0)
			return;

		g_variant_get(parameters, "(&s@a{sv}@as)", &inf_name, &val, NULL);

		if (strcasecmp(inf_name, BT_ADAPTER_INTERFACE) == 0)
			__bt_ipsp_property_changed_event(val, object_path);
		else if (strcasecmp(inf_name, BT_DEVICE_INTERFACE) == 0)
			__bt_device_property_changed_event(val, object_path);

		if (val)
			g_variant_unref(val);
	} else if (g_strcmp0(interface_name, BT_DEVICE_INTERFACE) == 0) {
			_bt_handle_ipsp_device_event(parameters, signal_name, object_path);
	}

	return;
}


int _bt_register_ipsp_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{

	if (conn == NULL)
		return -1;

	static int subs_interface_added_id = -1;
	static int subs_interface_removed_id = -1;
	static int subs_property_id = -1;
	static int subs_ipsp_state_id = -1;


	if (subscribe) {
		if (subs_interface_added_id == -1) {
			subs_interface_added_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_ADDED, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
		if (subs_interface_removed_id == -1) {
			subs_interface_removed_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_MANAGER_INTERFACE,
				BT_INTERFACES_REMOVED, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
		if (subs_property_id == -1) {
			subs_property_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_PROPERTIES_INTERFACE,
				BT_PROPERTIES_CHANGED, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
		if (subs_ipsp_state_id == -1) {
			subs_ipsp_state_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_DEVICE_INTERFACE,
				BT_IPSP_STATE_CHANGE, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_interface_added_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_interface_added_id);
			subs_interface_added_id = -1;
		}
		if (subs_interface_removed_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_interface_removed_id);
			subs_interface_removed_id = -1;
		}
		if (subs_property_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_property_id);
			subs_property_id = -1;
		}
		if (subs_ipsp_state_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_ipsp_state_id);
			subs_ipsp_state_id = -1;
		}
	}
	return 0;
}

int _bt_init_ipsp_receiver(void)
{
	BT_DBG("+");

	GError *error = NULL;

	if (manager_conn == NULL) {
		manager_conn =  g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
		if (error != NULL) {
			BT_ERR("ERROR: Can't get on system bus [%s]", error->message);
			g_clear_error(&error);
		}
		retv_if(manager_conn == NULL, BLUETOOTH_ERROR_INTERNAL);
	}

	if (_bt_register_ipsp_subscribe_signal(manager_conn, TRUE) != BLUETOOTH_ERROR_NONE)
		goto fail;

	return BLUETOOTH_ERROR_NONE;
fail:
	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	BT_DBG("-");

	return BLUETOOTH_ERROR_INTERNAL;
}

void _bt_deinit_service_event_receiver(void)
{
	BT_DBG("+");

	_bt_register_ipsp_subscribe_signal(manager_conn, FALSE);

	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	BT_DBG("-");
}

static void __bt_ipsp_property_changed_event(GVariant *msg, const char *path)
{
	GVariantIter value_iter;
	GVariant *val = NULL;
	char *property = NULL;
	GVariant *param = NULL;
	g_variant_iter_init(&value_iter, msg);
	while ((g_variant_iter_loop(&value_iter, "{sv}", &property, &val))) {
		BT_INFO("Property %s", property);

		if (strcasecmp(property, "IpspInitStateChanged") == 0) {
			gboolean ipsp_initialized = FALSE;

			g_variant_get(val, "b", &ipsp_initialized);
			BT_DBG("IPSP init state changed: %d", ipsp_initialized);
			param = g_variant_new("(b)", ipsp_initialized);

			/* Send event to application */
			_bt_send_event(BT_ADAPTER_EVENT,
					BLUETOOTH_EVENT_IPSP_INIT_STATE_CHANGED,
					param);
		}
	}
}
