/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <glib.h>
#include <gio/gio.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-service-common.h"
#include "bt-service-event.h"
#include "bt-service-util.h"
#include "bt-service-map-client.h"
#include "bt-service-obex-agent.h"
#include "bt-service-core-adapter.h"

#define DBUS_TIMEOUT 20 * 1000  /* 20 Seconds */

enum bt_map_transfer_type {
	BT_MAP_TRANSFER_GET_MESSAGE = 0,
	BT_MAP_TRANSFER_PUSH_MESSAGE
};

typedef struct {
	enum bt_map_transfer_type transfer_type;
	char *transfer_path;
	int request_id;
} bt_map_callback_data_t;

static GSList *transfer_list = NULL;

bt_session_info_t *session_info;

static void __bt_free_session_info(bt_session_info_t *info)
{
	ret_if(info == NULL);
	g_free(info->address);
	g_free(info);
}

void _bt_map_disconnected(const char *session_path)
{
	BT_DBG("+");
	GVariant *param = NULL;
	ret_if(session_info == NULL);

	if (g_strcmp0(session_info->session_path,
			session_path) != 0) {
		BT_INFO("Path mismatch, previous transfer failed! Returning");
		return;
	}

	param = g_variant_new("(isi)", session_info->result,
				session_info->address,
				session_info->request_id);
	_bt_send_event(BT_MAP_CLIENT_EVENT,
			BLUETOOTH_EVENT_MAP_DISCONNECTED,
			param);

	__bt_free_session_info(session_info);
	session_info = NULL;

	BT_DBG("-");
}

int _bt_create_session_sync(const char* address, char** session_id)
{
	BT_DBG("Entered SERVICE create session");
	GDBusConnection *g_conn;
	GDBusProxy *session_proxy;
	GError *err = NULL;

	retv_if(address == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	session_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE,
						NULL, BT_OBEX_SERVICE_NAME,
						BT_OBEX_CLIENT_PATH,
						BT_OBEX_CLIENT_INTERFACE,
						NULL, &err);
	if (err) {
		BT_ERR("Unable to create session_proxy: %s", err->message);
		g_clear_error(&err);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	retv_if(session_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariantBuilder* builder = g_variant_builder_new(G_VARIANT_TYPE_ARRAY);
	g_variant_builder_add(builder, "{sv}", "Target",
			g_variant_new("s", "map"));
	GVariant *args = g_variant_builder_end(builder);
	g_variant_builder_unref(builder);
	GVariant *param = g_variant_new("(s@a{sv})", address, args);

	GVariant *value = g_dbus_proxy_call_sync(session_proxy, "CreateSession", param,
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &err);
	if (err != NULL) {
		BT_ERR("Could not create session: %s\n", err->message);
		g_error_free(err);
		return BLUETOOTH_ERROR_INTERNAL;
	} else {
		if (NULL == value) {
			BT_ERR("create session returned value is null\n");
			return BLUETOOTH_ERROR_INTERNAL;
		} else {
			BT_DBG("create session succeed\n");
		}
	}

	g_variant_get(value, "(&o)", session_id);
	BT_DBG("session_id = \"%s\"\n", *session_id);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_destroy_session_sync(const char* session_id)
{
	BT_DBG("Entered SERVICE destroy session with id: \"%s\"", session_id);
	GDBusConnection *g_conn;
	GDBusProxy *session_proxy;
	GError *err = NULL;

	retv_if(session_id == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	session_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE,
						NULL, BT_OBEX_SERVICE_NAME,
						BT_OBEX_CLIENT_PATH,
						BT_OBEX_CLIENT_INTERFACE,
						NULL, &err);
	if (err) {
		BT_ERR("Unable to create session_proxy: %s", err->message);
		g_clear_error(&err);
		return BLUETOOTH_ERROR_INTERNAL;
	}
	retv_if(session_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *param = g_variant_new("(o)", session_id);

	g_dbus_proxy_call_sync(session_proxy, "RemoveSession", param,
			G_DBUS_CALL_FLAGS_NONE, -1, NULL, &err);
	if (err != NULL) {
		BT_ERR("Could not remove session: %s\n", err->message);
		g_error_free(err);
		return BLUETOOTH_ERROR_INTERNAL;
	} else {
		BT_DBG("remove session succeed\n");
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_map_client_set_folder(const char* session_id, const char* name)
{
	BT_DBG("+");

	GError *err = NULL;
	GDBusConnection *g_conn;
	GDBusProxy *message_proxy;
	GVariant *ret = NULL;

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	message_proxy = g_dbus_proxy_new_sync(g_conn,
		G_DBUS_PROXY_FLAGS_NONE, NULL,
		BT_OBEX_SERVICE_NAME, session_id,
		BT_OBEX_MESSAGE_INTERFACE, NULL, &err);

	retv_if(message_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	ret = g_dbus_proxy_call_sync(message_proxy,
				"SetFolder", g_variant_new("(s)", name),
				G_DBUS_CALL_FLAGS_NONE, DBUS_TIMEOUT,
				NULL, &err);

	if (ret == NULL) {
		if (err != NULL) {
			BT_ERR("SetFolder Error: %s\n", err->message);
			g_error_free(err);
		}
	} else {
		g_variant_unref(ret);
	}

	g_object_unref(message_proxy);

	BT_DBG("-");

	return BLUETOOTH_ERROR_NONE;
}

static void __bt_list_folders_cb(GDBusProxy *proxy,
				GAsyncResult *res, gpointer user_data)
{
	BT_DBG("+");

	GError *error = NULL;
	GVariant *value, *in_param, *param;

	int result = BLUETOOTH_ERROR_NONE;
	int request_id;

	in_param = (GVariant*) user_data;
	g_variant_get(in_param, "(i)", &request_id);

	value = g_dbus_proxy_call_finish(proxy, res, &error);

	if (error) {
		BT_ERR("%s", error->message);
		g_clear_error(&error);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	param = g_variant_new("(iiv)", result, request_id, value);
	BT_DBG("RequestID[%d]", request_id);
	result = _bt_send_event(BT_MAP_CLIENT_EVENT,
			BLUETOOTH_EVENT_MAP_LIST_FOLDERS_COMPLETE, param);

	g_variant_unref(value);
	g_variant_unref(in_param);

	BT_DBG("-");
}

int _bt_map_client_list_folders(
	int request_id,
	GDBusMethodInvocation *context,
	const char* session_id,
	const char* filter_serialized)
{
	BT_DBG("Entered _bt_map_list_folders with session id: \"%s\"", session_id);
	GDBusConnection *g_conn;
	GDBusProxy *message_access_proxy;
	GError *error = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	retv_if(session_id == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *out_param1 = g_variant_new_from_data((const GVariantType *)"ay",
			&request_id, sizeof(int),
			TRUE, NULL, NULL);
	g_dbus_method_invocation_return_value(context,
			g_variant_new("(iv)", result, out_param1));

	// create message access proxy
	g_clear_error(&error);
	message_access_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_OBEX_SERVICE_NAME, session_id, "org.bluez.obex.MessageAccess1",
			NULL, &error);
	if (error != NULL) {
		BT_ERR("Could not create message access proxy: %s\n", error->message);
		result = BLUETOOTH_ERROR_INTERNAL;
	} else {
		if (!message_access_proxy) {
			BT_ERR("message proxy handle is null\n");
			result = BLUETOOTH_ERROR_INTERNAL;
		} else {
			BT_DBG("message proxy set");

			GVariant *filter_variant = g_variant_parse(NULL, filter_serialized, NULL, NULL, NULL);

			GVariant *params = g_variant_new("(@a{sv})", filter_variant);
			GVariant *param = g_variant_new("(i)", request_id);

			g_dbus_proxy_call(message_access_proxy,
				"ListFolders", params,
				G_DBUS_CALL_FLAGS_NONE, -1,
				NULL,
				(GAsyncReadyCallback)__bt_list_folders_cb,
				(void*)param);
		}
	}

	if (message_access_proxy)
		g_object_unref(message_access_proxy);
	BT_DBG("-");

	return result;
}

static void __bt_list_filter_fields_cb(GDBusProxy *proxy,
				GAsyncResult *res, gpointer user_data)
{
	BT_DBG("+");

	GError *error = NULL;
	GVariant *value, *in_param, *param;

	int result = BLUETOOTH_ERROR_NONE;
	int request_id;

	in_param = (GVariant*) user_data;
	g_variant_get(in_param, "(i)", &request_id);

	value = g_dbus_proxy_call_finish(proxy, res, &error);

	if (error) {
		BT_ERR("%s", error->message);
		g_clear_error(&error);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	param = g_variant_new("(ivi)", result, value, request_id);

	_bt_send_event(BT_MAP_CLIENT_EVENT, BLUETOOTH_EVENT_MAP_LIST_FILTER_FIELD_COMPLETE,
			param);

	g_variant_unref(value);
	g_variant_unref(in_param);

	BT_DBG("-");
}

int _bt_map_client_list_filter_fields(int request_id, GDBusMethodInvocation *context, const char* session_id)
{
	BT_DBG("+");

	GError *err = NULL;
	GDBusConnection *g_conn;
	GDBusProxy *message_proxy;
	int result = BLUETOOTH_ERROR_NONE;

	retv_if(session_id == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *out_param = g_variant_new_from_data((const GVariantType *)"ay",
							&request_id, sizeof(int),
							TRUE, NULL, NULL);

	g_dbus_method_invocation_return_value(context,
			g_variant_new("(iv)", result, out_param));


	message_proxy = g_dbus_proxy_new_sync(g_conn,
		G_DBUS_PROXY_FLAGS_NONE, NULL,
		BT_OBEX_SERVICE_NAME, session_id,
		BT_OBEX_MESSAGE_INTERFACE, NULL, &err);

	retv_if(message_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *param = g_variant_new("(i)", request_id);

	g_dbus_proxy_call(message_proxy,
			"ListFilterFields", g_variant_new("()"),
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL,
			(GAsyncReadyCallback)__bt_list_filter_fields_cb,
			(void*)param);

	g_object_unref(message_proxy);

	BT_DBG("-");

	return BLUETOOTH_ERROR_NONE;
}

static void __bt_list_messages_cb(
	GDBusProxy *proxy, GAsyncResult *res, gpointer user_data)
{
	BT_DBG("+");

	GError *error = NULL;
	GVariant *value, *in_param, *param;

	int result = BLUETOOTH_ERROR_NONE;
	int request_id;

	in_param = (GVariant*) user_data;
	g_variant_get(in_param, "(i)", &request_id);

	value = g_dbus_proxy_call_finish(proxy, res, &error);

	if (error) {
		BT_ERR("%s", error->message);
		g_clear_error(&error);
		result = BLUETOOTH_ERROR_INTERNAL;
	}

	param = g_variant_new("(iiv)", result, request_id, value);
	BT_DBG("RequestID[%d]", request_id);
	result = _bt_send_event(BT_MAP_CLIENT_EVENT,
			BLUETOOTH_EVENT_MAP_LIST_MESSAGES_COMPLETE, param);

	g_variant_unref(value);
	g_variant_unref(in_param);

	BT_DBG("-");
}

int _bt_map_client_list_messages(
		int request_id,
		GDBusMethodInvocation *context,
		const char* session_id,
		const char* folder,
		const char* filter_serialized)
{
	BT_DBG("Entered _bt_map_client_list_messages with session id: \"%s\"", session_id);
	BT_DBG("Entered folder: %s", folder);
	GDBusConnection *g_conn;
	GDBusProxy *message_access_proxy;
	GError *error = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	retv_if(session_id == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *out_param1 = g_variant_new_from_data((const GVariantType *)"ay",
			&request_id, sizeof(int),
			TRUE, NULL, NULL);
	g_dbus_method_invocation_return_value(context,
			g_variant_new("(iv)", result, out_param1));

	// create message access proxy
	g_clear_error(&error);
	message_access_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_OBEX_SERVICE_NAME, session_id, "org.bluez.obex.MessageAccess1",
			NULL, &error);
	if (error != NULL) {
		BT_ERR("Could not create message access proxy: %s\n", error->message);
		result = BLUETOOTH_ERROR_INTERNAL;
	} else {
		if (!message_access_proxy) {
			BT_ERR("message proxy handle is null\n");
			result = BLUETOOTH_ERROR_INTERNAL;
		} else {
			BT_DBG("message proxy set");

			GVariant *filter_variant = g_variant_parse(NULL, filter_serialized, NULL, NULL, NULL);

			GVariant *params = g_variant_new("(s@a{sv})", folder, filter_variant);
			GVariant *param = g_variant_new("(i)", request_id);

			g_dbus_proxy_call(message_access_proxy,
				"ListMessages", params,
				G_DBUS_CALL_FLAGS_NONE, -1,
				NULL,
				(GAsyncReadyCallback)__bt_list_messages_cb,
				(void*)param);
		}
	}

	if (message_access_proxy)
		g_object_unref(message_access_proxy);
	BT_DBG("-");

	return result;
}

int _bt_map_client_update_inbox(const char* session_id)
{
	BT_DBG("+");

	GError *err = NULL;
	GDBusConnection *g_conn;
	GDBusProxy *message_proxy;
	GVariant *ret = NULL;

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	message_proxy = g_dbus_proxy_new_sync(g_conn,
		G_DBUS_PROXY_FLAGS_NONE, NULL,
		BT_OBEX_SERVICE_NAME, session_id,
		BT_OBEX_MESSAGE_INTERFACE, NULL, &err);

	retv_if(message_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	ret = g_dbus_proxy_call_sync(message_proxy,
				"UpdateInbox", g_variant_new("()"),
				G_DBUS_CALL_FLAGS_NONE, -1,
				NULL, &err);

	if (ret == NULL) {
		if (err != NULL) {
			BT_ERR("UpdateInbox Error: %s\n", err->message);
			g_error_free(err);
		}
	} else {
		g_variant_unref(ret);
	}

	g_object_unref(message_proxy);

	BT_DBG("-");

	return BLUETOOTH_ERROR_NONE;
}

void _bt_map_on_transfer_finished(const char *transfer_object_path, const int error)
{
	BT_DBG("Entered _bt_map_on_transfer_finished");
	BT_DBG("Looking for transfer %s id", transfer_object_path);

	bt_map_callback_data_t *callback_data = NULL;
	GSList* transfer = NULL;
	for (transfer = transfer_list; transfer != NULL; transfer = g_slist_next(transfer)) {
		callback_data = transfer->data;
		if (NULL == callback_data)
			continue;

		if (0 == strcmp(transfer_object_path, callback_data->transfer_path)) {
			BT_DBG("request id FOUND - triggering event");

			GVariant *param = g_variant_new("(ii)", error, callback_data->request_id);

			int event = -1;
			switch (callback_data->transfer_type) {
			case BT_MAP_TRANSFER_GET_MESSAGE:
				event = BLUETOOTH_EVENT_MAP_GET_MESSAGE_COMPLETE;
				break;
			case BT_MAP_TRANSFER_PUSH_MESSAGE:
				event = BLUETOOTH_EVENT_MAP_PUSH_MESSAGE_COMPLETE;
				break;
			}

			_bt_send_event(BT_MAP_CLIENT_EVENT, event, param);

			// remove callback data from list
			transfer_list = g_slist_remove(transfer_list, transfer);

			//free memory and break loop
			g_free(callback_data->transfer_path);
			callback_data->transfer_path = NULL;
			g_free(callback_data);
			break;
		}

	}
}

static void __bt_push_message_cb(GDBusProxy *proxy,
				GAsyncResult *res, gpointer user_data)
{
	BT_DBG("+");

	GError *error = NULL;
	GVariant *value, *in_param;

	char *transfer_object_path = NULL;
	GVariantIter *iter = NULL;

	int request_id;

	in_param = (GVariant*) user_data;
	g_variant_get(in_param, "(i)", &request_id);

	value = g_dbus_proxy_call_finish(proxy, res, &error);
	if (error) {
		BT_ERR("%s", error->message);
		g_clear_error(&error);
	}

	if (value) {
		g_variant_get(value, "(oa{sv})", &transfer_object_path, &iter);
		g_variant_unref(value);
	}

	BT_DBG("transfer object path: [%s]", transfer_object_path);

	BT_DBG("Adding request id %d - transfer [%s]", request_id, transfer_object_path);
	bt_map_callback_data_t* callback_data = g_malloc0(sizeof(bt_map_callback_data_t));
	callback_data->transfer_type = BT_MAP_TRANSFER_PUSH_MESSAGE;
	callback_data->request_id = request_id;
	callback_data->transfer_path = transfer_object_path;

	transfer_list = g_slist_append(transfer_list, callback_data);

	g_variant_unref(value);
	g_variant_unref(in_param);

	BT_DBG("-");
}

int _bt_map_client_push_message(
		int request_id,
		GDBusMethodInvocation *context,
		const char* session_id,
		const char* source_file,
		const char* folder,
		const char* args_serialized)
{
	BT_DBG("Entered _bt_map_client_push_message with session id: \"%s\"", session_id);
	BT_DBG("Entered source_file: %s", source_file);
	BT_DBG("Entered folder: %s", folder);

	GDBusConnection *g_conn;
	GDBusProxy *message_access_proxy;
	GError *error = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	retv_if(session_id == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	// TODO event listeners registration on first call, where should it be unregistered??
//	_bt_map_client_event_init();	--need to do

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *out_param1 = g_variant_new_from_data((const GVariantType *)"ay",
			&request_id, sizeof(int),
			TRUE, NULL, NULL);
	g_dbus_method_invocation_return_value(context, g_variant_new("(iv)", result, out_param1));

	// create message access proxy
	g_clear_error(&error);
	message_access_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_OBEX_SERVICE_NAME, session_id, BT_OBEX_MESSAGE_INTERFACE,
			NULL, &error);
	if (error != NULL) {
		BT_ERR("Could not create message access proxy: %s\n", error->message);
		result = BLUETOOTH_ERROR_INTERNAL;
	} else {
		if (!message_access_proxy) {
			BT_ERR("message proxy handle is null\n");
			result = BLUETOOTH_ERROR_INTERNAL;
		} else {
			BT_DBG("message proxy set");

			GVariant *args_variant = g_variant_parse(NULL, args_serialized, NULL, NULL, NULL);

			GVariant *params = g_variant_new("(ss@a{sv})", source_file, folder, args_variant);
			GVariant *req_id = g_variant_new("(i)", request_id);

			g_dbus_proxy_call(message_access_proxy,
				"PushMessage", params,
				G_DBUS_CALL_FLAGS_NONE, -1,
				NULL, (GAsyncReadyCallback)__bt_push_message_cb,
				(void*)req_id);
		}
	}

	if (message_access_proxy)
		g_object_unref(message_access_proxy);
	BT_DBG("-");

	return result;
}

static void __bt_get_message_cb(GDBusProxy *proxy,
				GAsyncResult *res, gpointer user_data)
{
	BT_DBG("+");

	GError *error = NULL;
	GVariant *value, *in_param;

	char *transfer_object_path = NULL;
	GVariantIter *iter = NULL;

	int request_id;

	in_param = (GVariant*) user_data;
	g_variant_get(in_param, "(i)", &request_id);

	value = g_dbus_proxy_call_finish(proxy, res, &error);
	if (error) {
		BT_ERR("%s", error->message);
		g_clear_error(&error);
	}

	if (value) {
		g_variant_get(value, "(oa{sv})", &transfer_object_path, &iter);
		g_variant_unref(value);
	}

	BT_DBG("transfer object path: [%s]", transfer_object_path);

	BT_DBG("Adding request id %d - transfer [%s]", request_id, transfer_object_path);
	bt_map_callback_data_t* callback_data = g_malloc0(sizeof(bt_map_callback_data_t));
	callback_data->transfer_type = BT_MAP_TRANSFER_GET_MESSAGE;
	callback_data->request_id = request_id;
	callback_data->transfer_path = transfer_object_path;

	transfer_list = g_slist_append(transfer_list, callback_data);

	g_variant_unref(value);
	g_variant_unref(in_param);

	BT_DBG("-");
}

int _bt_map_client_get_message(
		int request_id,
		GDBusMethodInvocation *context,
		const char* message_object,
		const char* target_file,
		bool attachment)
{
	BT_DBG("Entered _bt_map_client_get_message");

	GDBusConnection *g_conn;
	GDBusProxy *message_proxy;
	GError *error = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	retv_if(message_object == NULL, BLUETOOTH_ERROR_INVALID_PARAM);

	// TODO event listeners registration on first call, where should it be unregistered??
//	_bt_map_client_event_init();

	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	GVariant *out_param1 = g_variant_new_from_data((const GVariantType *)"ay",
			&request_id, sizeof(int),
			TRUE, NULL, NULL);
	g_dbus_method_invocation_return_value(context, g_variant_new("(iv)", result, out_param1));

	// create message proxy
	g_clear_error(&error);
	message_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_OBEX_SERVICE_NAME, message_object, "org.bluez.obex.Message1",
			NULL, &error);
	if (error != NULL) {
		BT_ERR("Could not create message proxy: %s\n", error->message);
		result = BLUETOOTH_ERROR_INTERNAL;
	} else {
		if (!message_proxy) {
			BT_ERR("message proxy handle is null\n");
			result = BLUETOOTH_ERROR_INTERNAL;
		} else {
			BT_DBG("message proxy set");
			GVariant *params = g_variant_new("(sb)", target_file, attachment);
			GVariant *req_id = g_variant_new("(i)", request_id);

			g_dbus_proxy_call(message_proxy, "Get", params, G_DBUS_CALL_FLAGS_NONE, -1,
				NULL, (GAsyncReadyCallback)__bt_get_message_cb, (void*)req_id);
		}
	}

	if (message_proxy)
		g_object_unref(message_proxy);
	BT_DBG("-");

	return result;
}
