/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <dirent.h>
#ifdef TIZEN_FEATURE_BT_DPM
#include "bt-service-dpm.h"
#endif

#include <vconf.h>

#include <gio/gio.h>

#include "bluetooth-api.h"
#include "bt-internal-types.h"

#include "bt-service-common.h"
#include "bt-service-event.h"
#include "bt-service-util.h"
#include "bt-service-obex-agent.h"
#include "bt-service-obex-server.h"

#include <oal-device-mgr.h>

#define DBUS_TIMEOUT 20 * 1000 /* 20 Seconds */
#define BT_OBEX_SERVER_AGENT_PATH "/org/obex/server_agent"

#define BT_OBEX_SERVICE "org.bluez.obex"
#define BT_OBEX_MANAGER "org.bluez.obex.AgentManager1"
#define BT_OBEX_PATH "/org/bluez/obex"

#define BT_OBEX_PATH_PREFIX "/opt/usr/media"
#define BT_OBEX_DEFAULT_PATH "/opt/usr/home/owner/media"
#define BT_OBEX_PATH_MAX_LENGTH 255


typedef struct {
	char *filename;
	char *file_path;
	char *path;
	char *type;
	char *device_name;
	int transfer_id;
	gint64 file_size;
	gint64 progress;
	char *address;
} bt_transfer_info_t;

typedef struct {
	GDBusMethodInvocation *reply_context;
	guint64 file_size;
	char *filename;
	char *file_path;
	char *device_name;
	char *transfer_path;
	char *address;
	unsigned char contact_auth_info[5];
} bt_auth_info_t;

typedef struct {
	char *dest_path;
	char *sender;
	int app_pid;
} bt_server_info_t;

typedef struct {
	GDBusProxy *proxy;
	int server_type;
	int accept_id;
	bt_auth_info_t *auth_info;
	bt_server_info_t *native_server;
	bt_server_info_t *custom_server;
} bt_obex_agent_info_t;

typedef struct {
	char *path;
	char *address;
	gboolean authorized;
} bt_session_info_t;

static GSList *transfers;
static bt_obex_agent_info_t agent_info;
static GSList *session_list = NULL;

static char *pending_auth_address = NULL;

static bt_session_info_t *__bt_find_session_by_path(char *transfer_path)
{
	GSList *l;
	bt_session_info_t *session;

	retv_if(transfer_path == NULL, NULL);

	for (l = session_list; l != NULL; l = l->next) {
		session = l->data;

		if (session == NULL)
			continue;

		if (g_strcmp0(session->path, transfer_path) == 0)
			return session;
	}

	return NULL;
}

static GQuark __bt_obex_error_quark(void)
{
	static GQuark quark = 0;
	if (!quark)
		quark = g_quark_from_static_string("agent");

	return quark;
}

static bt_transfer_info_t *__bt_find_transfer_by_id(int transfer_id)
{
	GSList *l;
	bt_transfer_info_t *transfer;

	for (l = transfers; l != NULL; l = l->next) {
		transfer = l->data;

		if (transfer == NULL)
			continue;

		if (transfer->transfer_id == transfer_id)
			return transfer;
	}

	return NULL;
}

bt_transfer_info_t *__bt_find_transfer_by_address(const char *address)
{
	BT_DBG("+");
	GSList *l;
	bt_transfer_info_t *transfer;

	retv_if(address == NULL, NULL);

	for (l = transfers; l != NULL; l = l->next) {
		transfer = l->data;

		if (transfer == NULL)
			continue;

		if (g_strcmp0(transfer->address, address) == 0)
			return transfer;
	}
	BT_DBG("-");
	return NULL;
}

static bt_transfer_info_t *__bt_find_transfer_by_path(const char *transfer_path)
{
	GSList *l;
	bt_transfer_info_t *transfer;

	retv_if(transfer_path == NULL, NULL);

	for (l = transfers; l != NULL; l = l->next) {
		transfer = l->data;

		if (transfer == NULL)
			continue;

		if (g_strcmp0(transfer->path, transfer_path) == 0)
			return transfer;
	}

	return NULL;
}

static void __bt_free_server_info(bt_server_info_t *server_info)
{
	ret_if(server_info == NULL);

	g_free(server_info->sender);
	g_free(server_info->dest_path);
	g_free(server_info);
}

static void __bt_free_auth_info(bt_auth_info_t *auto_info)
{
	ret_if(auto_info == NULL);

	g_free(auto_info->filename);
	g_free(auto_info->transfer_path);
	g_free(auto_info->device_name);
	g_free(auto_info->address);
	g_free(auto_info);
}

static void __bt_free_transfer_info(bt_transfer_info_t *transfer_info)
{
	ret_if(transfer_info == NULL);

	g_free(transfer_info->path);
	g_free(transfer_info->filename);
	g_free(transfer_info->file_path);
	g_free(transfer_info->type);
	g_free(transfer_info->device_name);
	g_free(transfer_info->address);
	g_free(transfer_info);
}

void _bt_obex_check_pending_transfer(const char *address)
{
	BT_DBG("+");
	GVariant *param = NULL;
	bt_transfer_info_t *transfer_info = __bt_find_transfer_by_address(address);
	if (transfer_info != NULL) {
		int result = BLUETOOTH_ERROR_CANCEL;
		param = g_variant_new("(issssstii)", result,
					transfer_info->filename,
					transfer_info->type,
					transfer_info->device_name,
					transfer_info->file_path,
					transfer_info->address,
					transfer_info->file_size,
					transfer_info->transfer_id,
					agent_info.server_type);
		_bt_send_event(BT_OPP_SERVER_EVENT,
			BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_COMPLETED,
			param);
		transfers = g_slist_remove(transfers, transfer_info);
		__bt_free_transfer_info(transfer_info);
	}
	BT_DBG("-");
}

static char *__bt_get_remote_device_name(const char *bdaddress)
{
	char *device_path = NULL;
	char *name = NULL;
	GVariant *value;
	GVariant *result = NULL;
	GError *err = NULL;
	GDBusProxy *device_proxy;
	GDBusConnection *conn;

	retv_if(bdaddress == NULL, NULL);

	device_path = _bt_get_device_object_path((char *)bdaddress);
	retv_if(device_path == NULL, NULL);

	conn = _bt_get_system_gconn();
	retv_if(conn == NULL, NULL);
	BT_INFO("Device_path %s", device_path);
	device_proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
						NULL, BT_BLUEZ_NAME,
						device_path,
						BT_PROPERTIES_INTERFACE,
						NULL, &err);

	g_free(device_path);
	retv_if(device_proxy == NULL, NULL);

	result = g_dbus_proxy_call_sync(device_proxy, "GetAll",
			g_variant_new("(s)", BT_DEVICE_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_TIMEOUT, NULL,
			&err);
	if (err) {
		BT_ERR("DBus Error : %s", err->message);
		g_clear_error(&err);
		return NULL;
	}
	if (result == NULL) {
		BT_ERR("g_dbus_proxy_call_sync function return NULL");
		return NULL;
	}
	g_variant_get(result, "(@a{sv})", &value);

	if (value) {
		GVariant *temp_value = g_variant_lookup_value(value, "Alias",
			G_VARIANT_TYPE_STRING);
		g_variant_get(temp_value, "s", &name);
		if (temp_value)
			g_variant_unref(temp_value);

		if (name != NULL)
			DBG_SECURE("Alias Name [%s]", name);
		else {
			temp_value = g_variant_lookup_value(value, "Name", G_VARIANT_TYPE_STRING);
			g_variant_get(temp_value, "s", &name);
			if (temp_value)
				g_variant_unref(temp_value);
			DBG_SECURE("Name = %s", name);
		}
	}
	g_variant_unref(result);
	g_object_unref(device_proxy);
	return name;
}

void __bt_get_auth_info(GVariant *reply, char *auth_info)
{
	int cursor;
	GVariant *tmp_value;
	char *manufacturer_data = NULL;
	int manufacturer_data_len;
	gboolean is_alias_set;
	GVariantIter *value_iter;
	guint8 m_value;
	int i = 0;

	tmp_value = g_variant_lookup_value(reply, "IsAliasSet",
								G_VARIANT_TYPE_BOOLEAN);
	if (tmp_value) {
		is_alias_set = g_variant_get_boolean(tmp_value);
		g_variant_unref(tmp_value);
	} else {
		is_alias_set = FALSE;
	}
	if (is_alias_set == FALSE) {
		tmp_value = g_variant_lookup_value(reply, "LegacyManufacturerDataLen",
								G_VARIANT_TYPE_UINT16);
		if (tmp_value) {
			manufacturer_data_len = g_variant_get_uint16(tmp_value);
			if (manufacturer_data_len >
					BLUETOOTH_MANUFACTURER_DATA_LENGTH_MAX) {
				BT_ERR("manufacturer_data_len is too long");
				manufacturer_data_len = BLUETOOTH_MANUFACTURER_DATA_LENGTH_MAX;
			}
			g_variant_unref(tmp_value);
		} else
			manufacturer_data_len = 0;

		tmp_value = g_variant_lookup_value(reply, "LegacyManufacturerData",
								G_VARIANT_TYPE_ARRAY);
		if (tmp_value) {
			if ((manufacturer_data_len == 0) ||
					manufacturer_data_len != g_variant_get_size(tmp_value)) {
				BT_ERR("manufacturer data length doesn't match");
				manufacturer_data_len = 0;
				manufacturer_data = NULL;
			} else {
				manufacturer_data = g_malloc0(manufacturer_data_len);
				g_variant_get(tmp_value, "ay", &value_iter);
				while (g_variant_iter_loop(value_iter, "y", &m_value))
					manufacturer_data[i++] = m_value;
			}
			g_variant_unref(tmp_value);
		} else {
			BT_INFO("manufacture data is not a G_VARIANT_TYPE_ARRAY ");
			manufacturer_data_len = 0;
			manufacturer_data = NULL;
		}
		/*minimum Size of the samsung specific manufacturer data is greater than 30 */
		if (manufacturer_data_len < 30) {
			g_free(manufacturer_data);
			return;
		}
		if (manufacturer_data[0] != 0x00 || manufacturer_data[1] != 0x75) {
			BT_DBG("This is not a samsung specific manufaturer data");
			g_free(manufacturer_data);
			return;
		}

		/* 2  samsung (0x00 0x75) + 1 (control and version) + 1 (service ID) +
		1 (discovery version) + 1 (associated service ID)
		2 (Proxamity and locality) + 2 (Device type and icon) */

		cursor = 10;

		memcpy(auth_info, &(manufacturer_data[cursor]), 5);
	}
	 g_free(manufacturer_data);
}

static void __bt_get_remote_device_name_authinfo(const char *bdaddress,
					char **device_name, unsigned char *auth_info)
{
	char *device_path = NULL;
	char *name = NULL;
	gboolean is_alias_set;
	GVariant *value;
	GVariant *result = NULL;
	GError *err = NULL;
	GDBusProxy *device_proxy;
	GDBusConnection *conn;

	ret_if(bdaddress == NULL);

	device_path = _bt_get_device_object_path((char *)bdaddress);
	ret_if(device_path == NULL);

	conn = _bt_get_system_gconn();
	ret_if(conn == NULL);
	BT_INFO("Device_path %s", device_path);
	device_proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
						NULL, BT_BLUEZ_NAME,
						device_path,
						BT_PROPERTIES_INTERFACE,
						NULL, &err);

	g_free(device_path);
	ret_if(device_proxy == NULL);

	result = g_dbus_proxy_call_sync(device_proxy, "GetAll",
			g_variant_new("(s)", BT_DEVICE_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_TIMEOUT, NULL,
			&err);
	if (err) {
		BT_ERR("DBus Error : %s", err->message);
		g_clear_error(&err);
		return;
	}
	if (result == NULL) {
		BT_ERR("g_dbus_proxy_call_sync function return NULL");
		return;
	}
	g_variant_get(result, "(@a{sv})", &value);

	if (value) {
		GVariant *temp_value = g_variant_lookup_value(value, "Alias",
			G_VARIANT_TYPE_STRING);
		g_variant_get(temp_value, "s", &name);
		if (temp_value)
			g_variant_unref(temp_value);

		if (name != NULL)
			DBG_SECURE("Alias Name [%s]", name);
		else {
			temp_value = g_variant_lookup_value(value, "Name", G_VARIANT_TYPE_STRING);
			g_variant_get(temp_value, "s", &name);
			if (temp_value)
				g_variant_unref(temp_value);
			DBG_SECURE("Name = %s", name);
		}
		temp_value = g_variant_lookup_value(value, "IsAliasSet", G_VARIANT_TYPE_BOOLEAN);
		if (temp_value) {
			is_alias_set = g_variant_get_boolean(temp_value);
			g_variant_unref(temp_value);
		} else {
			is_alias_set = FALSE;
		}

		if (is_alias_set == FALSE) {
			DBG_SECURE("Do Nothing");
			__bt_get_auth_info(value, (char *)auth_info);
		}
	}
	g_variant_unref(result);
	g_object_unref(device_proxy);

	*device_name = g_strdup(name);
	g_free(name);
	return;
}

static int __bt_get_transfer_id(const char *path)
{
	char *tmp = NULL;
	if (path == NULL)
		return -1;

	tmp = strrchr(path, 'r') + 1;
	retv_if(tmp == NULL, -1);

	return atoi(tmp);
}

static GDBusProxy *__bt_get_transfer_proxy(const char *transfer_path)
{
	GDBusConnection *conn;
	GDBusProxy *proxy;
	GError *err = NULL;

	conn = _bt_get_system_gconn();
	retv_if(conn == NULL, NULL);

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
					NULL, BT_OBEX_SERVICE_NAME,
					transfer_path,
					BT_OBEX_TRANSFER_INTERFACE,
					NULL, &err);

	if (err) {
		BT_ERR("Error : %s", err->message);
		g_clear_error(&err);
		return NULL;
	}

	return proxy;
}

static GDBusProxy *__bt_get_transfer_properties_proxy(const char *transfer_path)
{
	GDBusConnection *conn;
	GDBusProxy *proxy;
	GError *err = NULL;
	conn = _bt_get_system_gconn();
	retv_if(conn == NULL, NULL);

	proxy = g_dbus_proxy_new_sync(conn, G_DBUS_PROXY_FLAGS_NONE,
					NULL, BT_OBEX_SERVICE_NAME,
					transfer_path,
					BT_PROPERTIES_INTERFACE,
					NULL, &err);
	if (err) {
		BT_ERR("Error : %s", err->message);
		g_clear_error(&err);
		return NULL;
	}
	return proxy;
}

static int __bt_get_transfer_properties(bt_transfer_info_t *transfer_info,
					const char *transfer_path)
{
	GDBusProxy *transfer_proxy;
	GVariant *result = NULL;
	GError *err = NULL;
	GVariantIter *iter = NULL;
	BT_CHECK_PARAMETER(transfer_info, return);
	BT_CHECK_PARAMETER(transfer_path, return);

	transfer_proxy = __bt_get_transfer_properties_proxy(transfer_path);

	retv_if(transfer_proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(transfer_proxy, "GetAll",
			g_variant_new("(s)", BT_OBEX_TRANSFER_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_TIMEOUT, NULL,
			&err);

	if (err) {
		BT_ERR("DBus Error : %s", err->message);
		g_clear_error(&err);
		goto fail;
	}
	if (result == NULL) {
		BT_ERR("g_dbus_proxy_call_sync function return NULL");
		goto fail;
	}

	g_variant_get(result, "(a{sv})", &iter);
	g_variant_unref(result);
	if (iter) {
		const gchar *key;
		GVariant *val;
		gsize len = 0;
		while (g_variant_iter_loop(iter, "{sv}", &key, &val)) {
			if (g_strcmp0(key, "Operation") == 0) {
				transfer_info->type = g_variant_dup_string(val, &len);
			} else if (g_strcmp0(key, "Name") == 0) {
				transfer_info->filename = g_variant_dup_string(val, &len);
			} else if (g_strcmp0(key, "Size") == 0) {
				transfer_info->file_size = g_variant_get_uint64(val);
			} else if (g_strcmp0(key, "Address") == 0) {
				transfer_info->address = g_variant_dup_string(val, &len);
				BT_INFO("addressss %s", transfer_info->address);
			} else if (g_strcmp0(key, "Filename") == 0) {
				transfer_info->file_path = g_variant_dup_string(val, &len);
				if (!transfer_info->file_path)
					transfer_info->file_path = g_strdup(transfer_info->filename);
			}
		}
		g_variant_iter_free(iter);

		if (transfer_info->address == NULL)
			goto fail;
		transfer_info->device_name = __bt_get_remote_device_name(transfer_info->address);
		transfer_info->transfer_id = __bt_get_transfer_id(transfer_path);
		if (!transfer_info->device_name)
			transfer_info->device_name = g_strdup(transfer_info->address);

		if (transfer_info->type == NULL)
			goto fail;

		transfer_info->path = g_strdup(transfer_path);
	}

	g_object_unref(transfer_proxy);
	return BLUETOOTH_ERROR_NONE;

fail:
	g_object_unref(transfer_proxy);
	return BLUETOOTH_ERROR_INTERNAL;
}

static gboolean __bt_authorize_cb(GDBusMethodInvocation *context,
					const char *path,
					gpointer user_data)
{
	char *device_name = NULL;
	unsigned char auth_info[5] = {0, };
	int result = BLUETOOTH_ERROR_NONE;
	GDBusProxy *transfer_properties_proxy;
	char * bdaddress = NULL;
	GVariant *ret;
	GVariantIter *iter;
	GVariant *param = NULL;
	GError *err = NULL;
	bt_session_info_t *session_info = NULL;
#ifdef TIZEN_FEATURE_BT_DPM
	int value = DPM_BT_ERROR;
#endif

	BT_DBG(" path [%s] \n", path);

	transfer_properties_proxy = __bt_get_transfer_properties_proxy(path);

	retv_if(transfer_properties_proxy == NULL, FALSE);

	ret = g_dbus_proxy_call_sync(transfer_properties_proxy, "GetAll",
			g_variant_new("(s)", BT_OBEX_TRANSFER_INTERFACE),
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_TIMEOUT, NULL,
			&err);
	if (err) {
		BT_ERR("DBus Error : %s", err->message);
		g_clear_error(&err);
		return FALSE;
	}
	if (ret == NULL) {
		BT_ERR("g_dbus_proxy_call_sync function return NULL");
		return FALSE;
	}
	g_variant_get(ret, "(a{sv})", &iter);
	g_variant_unref(ret);
	if (iter == NULL) {
		g_object_unref(transfer_properties_proxy);
		return FALSE;
	}

	__bt_free_auth_info(agent_info.auth_info);

	agent_info.auth_info = g_malloc(sizeof(bt_auth_info_t));

	memset(agent_info.auth_info, 0, sizeof(bt_auth_info_t));

	agent_info.auth_info->reply_context = context;

	agent_info.auth_info->transfer_path = g_strdup(path);

#ifdef TIZEN_FEATURE_BT_DPM
	_bt_dpm_get_allow_bluetooth_mode(&value);
	if (value == DPM_BT_HANDSFREE_ONLY) {
		/* Free auth info in next function */
		_bt_obex_server_reject_authorize();
		return FALSE;
	}
#endif
	if (iter) {
		const gchar *key;
		GVariant *val;
		gsize len = 0;
		while (g_variant_iter_loop(iter, "{sv}", &key, &val)) {
			if (g_strcmp0(key, "Name") == 0)
				agent_info.auth_info->filename = g_variant_dup_string(val, &len);
			else if (g_strcmp0(key, "Address") == 0)
				bdaddress = g_variant_dup_string(val, &len);
			else if (g_strcmp0(key, "Size") == 0)
				agent_info.auth_info->file_size = g_variant_get_uint64(val);
		}
		g_variant_iter_free(iter);
	}

	__bt_get_remote_device_name_authinfo(bdaddress, &device_name, auth_info);

	if (!device_name)
		device_name = g_strdup(bdaddress);

	agent_info.auth_info->address = g_strdup(bdaddress);
	agent_info.auth_info->device_name = device_name;
	memcpy(agent_info.auth_info->contact_auth_info, auth_info, 5);

	session_info = __bt_find_session_by_path((char *)path);
	if (NULL == session_info) {
		session_info = g_malloc0(sizeof(bt_session_info_t));
		session_info->path = g_strdup(path);
		session_info->address = g_strdup(bdaddress);
		session_info->authorized = FALSE;
		session_list = g_slist_append(session_list, session_info);
	}

	g_object_unref(transfer_properties_proxy);
	g_free(bdaddress);

	if (agent_info.server_type == BT_CUSTOM_SERVER) {
		/* No need to send the event */
		_bt_obex_server_accept_authorize(agent_info.auth_info->filename, FALSE);
		return TRUE;
	}

	if (session_info->authorized == FALSE) {
		if (headed_plugin_info->plugin_headed_enabled)
			headed_plugin_info->headed_plugin->bt_launch_system_popup(BT_AGENT_EVENT_EXCHANGE_REQUEST, device_name,
					auth_info, NULL, NULL, BT_OBEX_SERVER_AGENT_PATH);
	} else {
/* TODO_40 : 4.0 merge */
		param = g_variant_new("(istss)", result,
				agent_info.auth_info->filename,
				agent_info.auth_info->file_size,
				agent_info.auth_info->address,
				agent_info.auth_info->device_name);
		_bt_send_event(BT_OPP_SERVER_EVENT,
				BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_AUTHORIZE, param);
	}

	return TRUE;
}

void _bt_obex_transfer_started(const char *transfer_path)
{
	bt_transfer_info_t *transfer_info;
	request_info_t *req_info;
	GVariant *out_param1 = NULL;
	GVariant *param = NULL;
	GVariantBuilder *builder = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	int i = 0;

	BT_DBG("%s", transfer_path);

	transfer_info = g_malloc0(sizeof(bt_transfer_info_t));

	if (agent_info.auth_info != NULL
	     && g_strcmp0(transfer_path, agent_info.auth_info->transfer_path) == 0) {
		transfer_info->filename = g_strdup(agent_info.auth_info->filename);
		transfer_info->file_size = agent_info.auth_info->file_size;
		transfer_info->type = g_strdup(TRANSFER_PUT);
		transfer_info->path = g_strdup(agent_info.auth_info->transfer_path);
		transfer_info->device_name = g_strdup(agent_info.auth_info->device_name);
		transfer_info->transfer_id = __bt_get_transfer_id(transfer_path);
		transfer_info->file_path = agent_info.auth_info->file_path;
		transfer_info->address = g_strdup(agent_info.auth_info->address);

	} else {
		if (__bt_get_transfer_properties(transfer_info, transfer_path) < 0) {
			BT_ERR("Get Properties failed");
			__bt_free_transfer_info(transfer_info);
			return;
		}
		agent_info.server_type = BT_FTP_SERVER;
	}

	if (agent_info.server_type == BT_CUSTOM_SERVER) {
		if (agent_info.custom_server == NULL) {
			__bt_free_transfer_info(transfer_info);
			__bt_free_auth_info(agent_info.auth_info);
			agent_info.auth_info = NULL;
			return;
		}

		req_info = _bt_get_request_info(agent_info.accept_id);
		if (req_info == NULL || req_info->context == NULL) {
			BT_ERR("info is NULL");
			goto done;
		}

		agent_info.accept_id = 0;
		result = BLUETOOTH_ERROR_NONE;
		GArray *g_out_param1 = NULL;
		g_out_param1 = g_array_new(FALSE, FALSE, sizeof(gchar));
		if (out_param1 == NULL) {
			out_param1 = g_variant_new_from_data((const GVariantType *)"ay",
					g_out_param1->data, g_out_param1->len,
					TRUE, NULL, NULL);
		}

		g_dbus_method_invocation_return_value(req_info->context,
				g_variant_new("(iv)", result, out_param1));
		g_array_free(g_out_param1, TRUE);
		_bt_delete_request_list(req_info->req_id);
	}
done:
	transfers = g_slist_append(transfers, transfer_info);

	BT_DBG("Transfer id %d", transfer_info->transfer_id);

	builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for (i = 0; i < 5; i++) {
		if (agent_info.auth_info)
			g_variant_builder_add(builder, "y", agent_info.auth_info->contact_auth_info[i]);
	}

	param = g_variant_new("(isssstii(ay))", result,
				transfer_info->device_name,
				transfer_info->filename,
				transfer_info->type,
				transfer_info->address,
				transfer_info->file_size,
				transfer_info->transfer_id,
				agent_info.server_type,
				builder);

	__bt_free_auth_info(agent_info.auth_info);
	agent_info.auth_info = NULL;

	g_variant_builder_unref(builder);

	_bt_send_event(BT_OPP_SERVER_EVENT,
		BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_STARTED,
		param);
}

void _bt_obex_transfer_progress(const char *transfer_path,
					guint64 transferred)
{
	BT_DBG("+");
	bt_transfer_info_t *transfer_info;
	int current_progress = 0;
	int previous_progress;
	GVariant *param = NULL;
	int result = BLUETOOTH_ERROR_NONE;

	transfer_info = __bt_find_transfer_by_path(transfer_path);
	ret_if(transfer_info == NULL);

	current_progress = (int)(((gdouble)transferred /
			(gdouble)transfer_info->file_size) * 100);

	previous_progress = (int)(((gdouble)transfer_info->progress /
			(gdouble)transfer_info->file_size) * 100);

	if (current_progress == previous_progress) {
		BT_DBG("Same Percentage Value: Do not emit Signal");
		return;
	}

	transfer_info->progress = transferred;
	param = g_variant_new("(isssstiii)", result,
				transfer_info->filename,
				transfer_info->type,
				transfer_info->device_name,
				transfer_info->address,
				transfer_info->file_size,
				transfer_info->transfer_id,
				current_progress,
				agent_info.server_type);
	_bt_send_event(BT_OPP_SERVER_EVENT,
		BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_PROGRESS,
		param);
	BT_DBG("-");
}

void _bt_obex_transfer_completed(const char *transfer_path, gboolean success)
{
	bt_transfer_info_t *transfer_info;
	GVariantBuilder *builder = NULL;
	GVariant *param = NULL;
	int result;
	int i = 0;
	BT_DBG("Transfer [%s] Success [%d] \n", transfer_path, success);

	result = (success == TRUE) ? BLUETOOTH_ERROR_NONE
				: BLUETOOTH_ERROR_CANCEL;

	transfer_info = __bt_find_transfer_by_path(transfer_path);

	if (transfer_info == NULL) {
		BT_DBG("Very small files receiving case, did not get Active status from obexd");
		if (agent_info.auth_info == NULL ||
				g_strcmp0(transfer_path,
				agent_info.auth_info->transfer_path) != 0) {
			BT_ERR("auth_info is NULL, returning");
			return;
		}

		transfer_info = g_new0(bt_transfer_info_t, 1);

		transfer_info->filename = g_strdup(agent_info.auth_info->filename);
		transfer_info->file_size = agent_info.auth_info->file_size;
		transfer_info->type = g_strdup(TRANSFER_PUT);
		transfer_info->path = g_strdup(agent_info.auth_info->transfer_path);
		transfer_info->device_name = g_strdup(agent_info.auth_info->device_name);
		transfer_info->transfer_id = __bt_get_transfer_id(transfer_path);
		transfer_info->file_path = agent_info.auth_info->file_path;
		transfer_info->address = g_strdup(agent_info.auth_info->address);

		builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
		for (i = 0; i < 5; i++)
			g_variant_builder_add(builder, "y", agent_info.auth_info->contact_auth_info[i]);

		param = g_variant_new("(isssstii(ay))", result,
					transfer_info->device_name,
					transfer_info->filename,
					transfer_info->type,
					transfer_info->address,
					transfer_info->file_size,
					transfer_info->transfer_id,
					agent_info.server_type,
					builder);
		_bt_send_event(BT_OPP_SERVER_EVENT,
			BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_STARTED,
			param);
		g_variant_builder_unref(builder);
	}
	param = g_variant_new("(issssstii)", result,
				transfer_info->filename,
				transfer_info->type,
				transfer_info->device_name,
				transfer_info->file_path,
				transfer_info->address,
				transfer_info->file_size,
				transfer_info->transfer_id,
				agent_info.server_type);
	_bt_send_event(BT_OPP_SERVER_EVENT,
		BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_COMPLETED,
		param);
	transfers = g_slist_remove(transfers, transfer_info);
	__bt_free_transfer_info(transfer_info);
}

void _bt_obex_transfer_connected(const char *obj_path)
{
	BT_DBG("+");

	int result = BLUETOOTH_ERROR_NONE;
	GVariant *param = NULL;
	bt_transfer_info_t *transfer_info = NULL;

	transfer_info = g_new0(bt_transfer_info_t, 1);
	__bt_get_transfer_properties(transfer_info, obj_path);
	DBG_SECURE("Address[%s] Name[%s] TransferID[%d] ", transfer_info->address,
			transfer_info->device_name, transfer_info->transfer_id);

	param = g_variant_new("(issi)", result, transfer_info->address,
			transfer_info->device_name, transfer_info->transfer_id);

	_bt_send_event(BT_OPP_SERVER_EVENT,
		BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_CONNECTED,
		param);

	__bt_free_transfer_info(transfer_info);
	BT_DBG("-");
}

void _bt_obex_transfer_disconnected(char * obj_path)
{
	BT_DBG("+");

	int result = BLUETOOTH_ERROR_NONE;
	GVariant *param = NULL;
	bt_session_info_t *session = NULL;
	int transfer_id = -1;

	session = __bt_find_session_by_path(obj_path);
	ret_if(session == NULL);

	transfer_id = __bt_get_transfer_id(obj_path);
	DBG_SECURE("transfer_id: [%d]", transfer_id);

	DBG_SECURE("%s", session->address);
	param = g_variant_new("(isi)", result, session->address, transfer_id);
	_bt_send_event(BT_OPP_SERVER_EVENT,
		BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_DISCONNECTED,
		param);
	session_list = g_slist_remove(session_list, session);
	g_free(session->address);
	g_free(session->path);
	g_free(session);
	BT_DBG("-");
}

int _bt_register_obex_server(void)
{
	GDBusConnection *g_conn;
	GDBusProxy *manager_proxy;
	GVariant *result = NULL;
	GError *g_error = NULL;

	/* Get the session bus. */
	g_conn = _bt_get_system_gconn();
	retv_if(g_conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_obex_agent_new(BT_OBEX_SERVER_AGENT_PATH);

	_bt_obex_setup(BT_OBEX_SERVER_AGENT_PATH);

	_bt_obex_set_authorize_cb(BT_OBEX_SERVER_AGENT_PATH,
					__bt_authorize_cb, NULL);

	manager_proxy = g_dbus_proxy_new_sync(g_conn, G_DBUS_PROXY_FLAGS_NONE,
						NULL, BT_OBEX_SERVICE,
						BT_OBEX_PATH,
						BT_OBEX_MANAGER,
						NULL, &g_error);

	if (manager_proxy == NULL)
		return BLUETOOTH_ERROR_INTERNAL;

	result = g_dbus_proxy_call_sync(manager_proxy, "RegisterAgent",
				g_variant_new("(o)", BT_OBEX_SERVER_AGENT_PATH),
				G_DBUS_CALL_FLAGS_NONE,
				DBUS_TIMEOUT, NULL,
				&g_error);

	if (g_error != NULL) {
		BT_ERR("Agent registration failed: %s\n", g_error->message);
		g_object_unref(manager_proxy);
		g_error_free(g_error);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	if (result)
		g_variant_unref(result);

	agent_info.proxy = manager_proxy;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_unregister_obex_server(void)
{
	GVariant *result = NULL;
	GError *g_error = NULL;

	retv_if(agent_info.proxy == NULL,
				BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(agent_info.proxy, "UnregisterAgent",
				g_variant_new("(o)", BT_OBEX_SERVER_AGENT_PATH),
				G_DBUS_CALL_FLAGS_NONE,
				DBUS_TIMEOUT, NULL,
				&g_error);
	if (g_error != NULL) {
		BT_ERR("Agent unregistration failed: %s", g_error->message);
		g_error_free(g_error);
	}

	if (result)
		g_variant_unref(result);

	_bt_obex_agent_destroy(BT_OBEX_SERVER_AGENT_PATH);
	g_object_unref(agent_info.proxy);
	agent_info.proxy = NULL;

	return BLUETOOTH_ERROR_NONE;
}

gboolean __bt_check_folder_path(const char *dest_path)
{
	DIR *dp;

	retv_if(dest_path == NULL, FALSE);

	dp = opendir(dest_path);

	if (dp == NULL) {
		BT_ERR("The directory does not exist");
		return FALSE;
	}

	closedir(dp);

	return TRUE;
}

char *__bt_transfer_folder_path(char *dest_path)
{
	char *dst_path = (char *)g_malloc0(BT_OBEX_PATH_MAX_LENGTH);
	if (g_str_has_prefix(dest_path, BT_OBEX_PATH_PREFIX))
		snprintf(dst_path, BT_OBEX_PATH_MAX_LENGTH, BT_OBEX_DEFAULT_PATH"%s", dest_path + strlen(BT_OBEX_PATH_PREFIX));
	else
		snprintf(dst_path, BT_OBEX_PATH_MAX_LENGTH, "%s", dest_path);

	BT_INFO("obex transfed path : %s", dst_path);
	return dst_path;
}

int _bt_obex_server_allocate(char *sender, const char *dest_path, int app_pid, gboolean is_native)
{
	BT_DBG("+");

	char *dst_path;
	dst_path = __bt_transfer_folder_path((char *)dest_path);

	if (__bt_check_folder_path(dst_path) == FALSE) {
		g_free(dst_path);
		return BLUETOOTH_ERROR_INVALID_PARAM;
	}

	if (is_native == TRUE) {
		if (agent_info.native_server) {
			BT_ERR("obex native server busy");
			g_free(dst_path);
			return BLUETOOTH_ERROR_DEVICE_BUSY;
		}

		/* Force to change the control to native */
		agent_info.native_server = g_malloc0(sizeof(bt_server_info_t));
		agent_info.native_server->dest_path = g_strdup(dst_path);
		agent_info.native_server->sender = g_strdup(sender);
		agent_info.native_server->app_pid = app_pid;
		agent_info.server_type = BT_NATIVE_SERVER;
		if (OAL_STATUS_SUCCESS !=
			device_set_osp_server(OAL_OSP_SERVER_OBEX, FALSE))
			BT_ERR("device_set_osp_server failed");
	} else {
		if (agent_info.custom_server) {
			BT_ERR("obex custom server busy");
			g_free(dst_path);
			return BLUETOOTH_ERROR_DEVICE_BUSY;
		}

		/* Force to change the control to custom */
		agent_info.custom_server = g_malloc0(sizeof(bt_server_info_t));
		agent_info.custom_server->dest_path = g_strdup(dst_path);
		agent_info.custom_server->sender = g_strdup(sender);
		agent_info.custom_server->app_pid = app_pid;
		agent_info.server_type = BT_CUSTOM_SERVER;
		if (OAL_STATUS_SUCCESS !=
			device_set_osp_server(OAL_OSP_SERVER_OBEX, TRUE))
			BT_ERR("device_set_osp_server failed");
	}

	g_free(dst_path);
	BT_DBG("-");
	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_deallocate(int app_pid, gboolean is_native)
{
	if (is_native == TRUE) {
		retv_if(agent_info.native_server == NULL,
				BLUETOOTH_ERROR_AGENT_DOES_NOT_EXIST);

		retv_if(agent_info.native_server->app_pid != app_pid,
				BLUETOOTH_ERROR_ACCESS_DENIED);

		__bt_free_server_info(agent_info.native_server);
		agent_info.native_server = NULL;

		/* Change the control to custom */
		if (agent_info.custom_server &&
				agent_info.server_type != BT_CUSTOM_SERVER) {

			agent_info.server_type = BT_CUSTOM_SERVER;

			if (OAL_STATUS_SUCCESS !=
				device_set_osp_server(OAL_OSP_SERVER_OBEX, TRUE))
				BT_ERR("device_set_osp_server failed");
		}
	} else {
		retv_if(agent_info.custom_server == NULL,
				BLUETOOTH_ERROR_AGENT_DOES_NOT_EXIST);

		retv_if(agent_info.custom_server->app_pid != app_pid,
				BLUETOOTH_ERROR_ACCESS_DENIED);


		__bt_free_server_info(agent_info.custom_server);
		agent_info.custom_server = NULL;

		/* Change the control to native */
		if (agent_info.native_server &&
				agent_info.server_type != BT_NATIVE_SERVER) {

			agent_info.server_type = BT_NATIVE_SERVER;

			if (OAL_STATUS_SUCCESS !=
				device_set_osp_server(OAL_OSP_SERVER_OBEX, FALSE))
				BT_ERR("device_set_osp_server failed");
		}
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_accept_authorize(const char *filename, gboolean is_native)
{
	char file_path[BT_FILE_PATH_MAX] = { 0 };
	bt_server_info_t *server_info;
	bt_auth_info_t *new_auth_info;

	BT_CHECK_PARAMETER(filename, return);

	retv_if(agent_info.auth_info == NULL, BLUETOOTH_ERROR_INTERNAL);

	retv_if(agent_info.auth_info->reply_context == NULL,
				BLUETOOTH_ERROR_INTERNAL);

	if (is_native == TRUE)
		server_info = agent_info.native_server;
	else
		server_info = agent_info.custom_server;

	retv_if(server_info == NULL, BLUETOOTH_ERROR_INTERNAL);

	if (server_info->dest_path != NULL)
		snprintf(file_path, sizeof(file_path), "%s/%s",
			server_info->dest_path, filename);
	else
		snprintf(file_path, sizeof(file_path), "%s", filename);

	g_dbus_method_invocation_return_value(agent_info.auth_info->reply_context,
		g_variant_new("(s)", &file_path));

	new_auth_info = g_malloc0(sizeof(bt_auth_info_t));

	new_auth_info->file_size = agent_info.auth_info->file_size;
	new_auth_info->file_path = g_strdup(file_path);
	new_auth_info->filename = g_strdup(filename);
	new_auth_info->device_name = g_strdup(agent_info.auth_info->device_name);
	new_auth_info->transfer_path = g_strdup(agent_info.auth_info->transfer_path);
	new_auth_info->address = g_strdup(agent_info.auth_info->address);
	memcpy(new_auth_info->contact_auth_info, agent_info.auth_info->contact_auth_info, 5);

	__bt_free_auth_info(agent_info.auth_info);

	agent_info.auth_info = new_auth_info;

	return BLUETOOTH_ERROR_NONE;
}

void _bt_obex_server_reply_accept(void)
{
	GVariant *param = NULL;
	bt_session_info_t *session_info = NULL;
	int result = BLUETOOTH_ERROR_NONE;
	param = g_variant_new("(istss)", result,
			agent_info.auth_info->filename,
			agent_info.auth_info->file_size,
			agent_info.auth_info->address,
			agent_info.auth_info->device_name);
	BT_INFO("Send Obex Authorize");
	_bt_send_event(BT_OPP_SERVER_EVENT, BLUETOOTH_EVENT_OBEX_SERVER_TRANSFER_AUTHORIZE, param);

	session_info = __bt_find_session_by_path(agent_info.auth_info->transfer_path);

	if (NULL == session_info)
		BT_ERR("Couldn't get the session info from the list");
	else
		session_info->authorized = TRUE;
}

int _bt_obex_server_reject_authorize(void)
{
	GError *g_error;

	retv_if(agent_info.auth_info->reply_context == NULL,
				BLUETOOTH_ERROR_INTERNAL);

	g_error = g_error_new(__bt_obex_error_quark(),
			BT_OBEX_AGENT_ERROR_CANCEL,
			"CancelledByUser");

	g_dbus_method_invocation_return_gerror(agent_info.auth_info->reply_context,
			g_error);
	g_error_free(g_error);

	__bt_free_auth_info(agent_info.auth_info);
	agent_info.auth_info = NULL;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_set_destination_path(const char *dest_path,
						gboolean is_native)
{
	bt_server_info_t *server_info;
	BT_CHECK_PARAMETER(dest_path, return);

	char *dst_path;
	dst_path = __bt_transfer_folder_path((char *)dest_path);

	DIR *dp = NULL;

	dp = opendir(dst_path);

	if (dp == NULL) {
		BT_ERR("The directory does not exist");
		g_free(dst_path);
		return BLUETOOTH_ERROR_INVALID_PARAM;
	}

	closedir(dp);

	if (is_native == TRUE)
		server_info = agent_info.native_server;
	else
		server_info = agent_info.custom_server;

	if (!server_info) {
		BT_ERR("obex server info is NULL");
		g_free(dst_path);
		return BLUETOOTH_ERROR_AGENT_DOES_NOT_EXIST;
	}

	g_free(server_info->dest_path);
	server_info->dest_path = g_strdup(dst_path);

	g_free(dst_path);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_set_root(const char *root)
{
	GVariant *result = NULL;
	GError *g_error = NULL;
	GVariant *folder = NULL;
	char *string = "Root";
	DIR *dp = NULL;

	BT_CHECK_PARAMETER(root, return);

	char *dst_root;
	dst_root = __bt_transfer_folder_path((char *)root);

	if (!agent_info.proxy) {
		BT_ERR("obex agent_info proxy error");
		g_free(dst_root);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	dp = opendir(dst_root);

	if (dp == NULL) {
		BT_ERR("The directory does not exist");
		g_free(dst_root);
		return BLUETOOTH_ERROR_INVALID_PARAM;
	}

	closedir(dp);

	folder = g_variant_new_string(dst_root);
	result = g_dbus_proxy_call_sync(agent_info.proxy, "SetProperty",
			g_variant_new("(sv)", string, folder),
			G_DBUS_CALL_FLAGS_NONE,
			DBUS_TIMEOUT, NULL,
			&g_error);

	if (g_error) {
		BT_ERR("SetProperty Fail: %s", g_error->message);
		g_error_free(g_error);
		g_free(dst_root);
		return BLUETOOTH_ERROR_INTERNAL;
	}

	if (result)
		g_variant_unref(result);

	g_free(dst_root);
	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_cancel_transfer(int transfer_id)
{
	bt_transfer_info_t *transfer = NULL;
	GDBusProxy *proxy;
	GVariant *result = NULL;
	GError *err = NULL;
	BT_DBG("+");
	transfer = __bt_find_transfer_by_id(transfer_id);

	retv_if(transfer == NULL, BLUETOOTH_ERROR_NOT_FOUND);
	proxy = __bt_get_transfer_proxy(transfer->path);

	retv_if(proxy == NULL, BLUETOOTH_ERROR_INTERNAL);

	result = g_dbus_proxy_call_sync(proxy, "Cancel", NULL,
		G_DBUS_CALL_FLAGS_NONE,
		DBUS_TIMEOUT, NULL, &err);
	if (err) {
		BT_ERR("Dbus Err: %s", err->message);
		g_clear_error(&err);
	}

	g_object_unref(proxy);

	if (result)
		g_variant_unref(result);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_cancel_all_transfers(void)
{
	GSList *l;
	bt_transfer_info_t *transfer;

	for (l = transfers; l != NULL; l = l->next) {
		transfer = l->data;

		if (transfer == NULL)
			continue;

		_bt_obex_server_cancel_transfer(transfer->transfer_id);
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_is_activated(gboolean *activated)
{
	BT_CHECK_PARAMETER(activated, return);

	if (agent_info.custom_server)
		*activated = TRUE;
	else
		*activated = FALSE;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_check_allocation(gboolean *allocation)
{
	BT_CHECK_PARAMETER(allocation, return);

	if (agent_info.native_server || agent_info.custom_server)
		*allocation = TRUE;
	else
		*allocation = FALSE;

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_check_termination(char *sender)
{
	BT_CHECK_PARAMETER(sender, return);

	if (agent_info.native_server) {
		if (g_strcmp0(sender, agent_info.native_server->sender) == 0) {
			_bt_obex_server_deallocate(agent_info.native_server->app_pid,
						TRUE);
		}
	}

	if (agent_info.custom_server) {
		if (g_strcmp0(sender, agent_info.custom_server->sender) == 0) {
			_bt_obex_server_deallocate(agent_info.custom_server->app_pid,
						FALSE);
		}
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_obex_server_is_receiving(gboolean *receiving)
{
	BT_CHECK_PARAMETER(receiving, return);

	if (transfers == NULL || g_slist_length(transfers) == 0)
		*receiving = FALSE;
	else
		*receiving = TRUE;

	return BLUETOOTH_ERROR_NONE;
}

void _bt_obex_server_set_pending_conn_auth_device_addr(const char *address)
{
	if (pending_auth_address)
		g_free(pending_auth_address);

	pending_auth_address = g_strdup(address);
}

/* To support the BOT  */
int _bt_obex_server_accept_connection(int request_id)
{
	int res;
	bt_address_t bd_addr;

	BT_INFO("address: %s", pending_auth_address);
	_bt_convert_addr_string_to_type(bd_addr.addr, pending_auth_address);
	res = device_reply_auth_request(&bd_addr, OPP_SERVICE_ID, TRUE, FALSE);
	g_free(pending_auth_address);
	pending_auth_address = NULL;
	if (res != OAL_STATUS_SUCCESS) {
		BT_ERR("device_reply_auth_request failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	agent_info.accept_id = request_id;

	return BLUETOOTH_ERROR_NONE;
}

/* To support the BOT  */
int _bt_obex_server_reject_connection(void)
{
	int res;
	bt_address_t bd_addr;

	BT_INFO("address: %s", pending_auth_address);
	_bt_convert_addr_string_to_type(bd_addr.addr, pending_auth_address);
	res = device_reply_auth_request(&bd_addr, OPP_SERVICE_ID, FALSE, FALSE);
	g_free(pending_auth_address);
	pending_auth_address = NULL;
	if (res != OAL_STATUS_SUCCESS) {
		BT_ERR("device_reply_auth_request failed");
		return BLUETOOTH_ERROR_INTERNAL;
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_opp_get_server_progress(int transfer_id, guint8 *progress)
{
	bt_transfer_info_t *requested_transfer = NULL;
	requested_transfer = __bt_find_transfer_by_id(transfer_id);
	if (requested_transfer == NULL) {
		BT_ERR("No Matching Inbound transfer");
		return BLUETOOTH_ERROR_NOT_FOUND;
	}

	*progress = (int)(((double)requested_transfer->progress /
			requested_transfer->file_size) * 100);

	BT_DBG("Percentage: %d", *progress);
	return BLUETOOTH_ERROR_NONE;
}

gboolean _bt_obex_server_is_custom(void)
{
	return (agent_info.server_type == BT_CUSTOM_SERVER) ? TRUE : FALSE;
}

