/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <string.h>
#include <dlog.h>
#include <vconf.h>
#include "bt-service-common.h"
#include <vconf-internal-bt-keys.h>
#include "bt-service-event.h"
//#include "bt-service-device.h"

guint nap_connected_device_count = 0;
static GDBusConnection *manager_conn;

void bt_set_device_values(gboolean connected, int state)
{
	int bt_device_state = VCONFKEY_BT_DEVICE_NONE;

	if (vconf_get_int(VCONFKEY_BT_DEVICE, &bt_device_state) != 0)
		BT_ERR("vconf_get_int failed");

	if (connected == TRUE)
		bt_device_state |= state;
	else if (bt_device_state & state)
		bt_device_state ^= state;

	if (vconf_set_int(VCONFKEY_BT_DEVICE, bt_device_state) != 0)
		BT_ERR("vconf_set_int failed");
}

void _bt_handle_network_server_event(GVariant *msg, const char *member)
{
	int result = BLUETOOTH_ERROR_NONE;
	char *address = NULL;
	char *device = NULL;
	GVariant *param = NULL;
	ret_if(member == NULL);

	BT_DBG("_bt_handle_network_server_event  Event handler called >>>>>>>>>>>>>");
	if (strcasecmp(member, "PeerConnected") == 0) {
		g_variant_get(msg, "(ss)", &device, &address);

		bt_set_device_values(TRUE,
				VCONFKEY_BT_DEVICE_PAN_CONNECTED);
		param = g_variant_new("(iss)", result, device, address);
		_bt_send_event(BT_NETWORK_EVENT, BLUETOOTH_EVENT_NETWORK_SERVER_CONNECTED,
			param);
		g_free(device);
		g_free(address);
		 nap_connected_device_count++;
	} else if (strcasecmp(member, "PeerDisconnected") == 0) {
		g_variant_get(msg, "(ss)", &device, &address);
		nap_connected_device_count--;
		if (nap_connected_device_count == 0)
			bt_set_device_values(FALSE,
				VCONFKEY_BT_DEVICE_PAN_CONNECTED);
		param = g_variant_new("(iss)", result, device, address);
		_bt_send_event(BT_NETWORK_EVENT, BLUETOOTH_EVENT_NETWORK_SERVER_DISCONNECTED,
			param);
		g_free(device);
		g_free(address);
	}
}

static  void __bt_manager_event_filter(GDBusConnection *connection,
					const gchar *sender_name,
					const gchar *object_path,
					const gchar *interface_name,
					const gchar *signal_name,
					GVariant *parameters,
					gpointer user_data)
{
	if (signal_name == NULL)
		return;

	if (g_strcmp0(interface_name, BT_NETWORK_SERVER_INTERFACE) == 0)
		_bt_handle_network_server_event(parameters, signal_name);
	return;
}

int _bt_register_network_subscribe_signal(GDBusConnection *conn,
		int subscribe)
{
	if (conn == NULL)
		return -1;

	static int subs_serv_id = -1;
	static int subs_client_id = -1;

	if (subscribe) {
		if (subs_serv_id == -1) {
				BT_DBG("_bt_register_network_subscribe_signal  SUBSCRIBE SERVER SUCCESS>>>>>>>>>>>>>>>>>>>>> ..");
			subs_serv_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_NETWORK_SERVER_INTERFACE,
				NULL, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
		if (subs_client_id == -1) {
				BT_DBG("_bt_register_network_subscribe_signal  SUBSCRIBE CLIENT SUCCESS>>>>>>>>>>>>>>>>>>>>> ..");
			subs_client_id = g_dbus_connection_signal_subscribe(conn,
				NULL, BT_NETWORK_CLIENT_INTERFACE,
				NULL, NULL, NULL, 0,
				__bt_manager_event_filter,
				NULL, NULL);
		}
	} else {
		if (subs_serv_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_serv_id);
			subs_serv_id = -1;
		}
		if (subs_client_id != -1) {
			g_dbus_connection_signal_unsubscribe(conn,
					subs_client_id);
			subs_client_id = -1;
		}
	}
	return 0;
}

static int __bt_init_manager_receiver(void)
{
	BT_DBG("+");

	GError *error = NULL;

	if (manager_conn == NULL) {
		manager_conn =  g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
		if (error != NULL) {
			BT_ERR("ERROR: Can't get on system bus [%s]", error->message);
			g_clear_error(&error);
		}

		retv_if(manager_conn == NULL, BLUETOOTH_ERROR_INTERNAL);
	}

	if (_bt_register_network_subscribe_signal(manager_conn, TRUE) == -1) {
		BT_DBG("_bt_register_network_subscribe_signal Failed..");
		goto fail;
	}


	return BLUETOOTH_ERROR_NONE;
fail:
	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	BT_DBG("-");

	return BLUETOOTH_ERROR_INTERNAL;
}

/* To receive the event from bluez */
int _bt_init_pan_nap_event_receiver(void)
{
	BT_DBG("+");

	int result;

	result = __bt_init_manager_receiver();
	retv_if(result != BLUETOOTH_ERROR_NONE, result);

	BT_DBG("-");

	return BLUETOOTH_ERROR_NONE;
}

void _bt_deinit_pan_nap_event_receiver(void)
{
	BT_DBG("+");

	_bt_register_network_subscribe_signal(manager_conn, FALSE);

	if (manager_conn) {
		g_object_unref(manager_conn);
		manager_conn = NULL;
	}

	BT_DBG("-");
}

