/*
 * Bluetooth-frwk
 *
 * Copyright (c) 2000 - 2011 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact:  Hocheol Seo <hocheol.seo@samsung.com>
 *		 Girishashok Joshi <girish.joshi@samsung.com>
 *		 Chanyeol Park <chanyeol.park@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <dlog.h>
#include <string.h>
#include <gio/gio.h>

#include "bluetooth-api.h"
#include "bt-service-common.h"
#include "bt-service-oob.h"
#include "bt-service-event.h"

static int __get_oob_data_from_variant(GVariant *variant,
		unsigned char *buf, int size)
{
	int i;
	unsigned char *v;

	i = g_variant_get_size(variant);
	v = (unsigned char *)g_variant_get_data(variant);
	if (i && v) {
		i = i > size ? size : i;
		memcpy(buf, v, i);
	}

	return i;
}

int _bt_oob_read_local_data(bt_oob_data_t *local_oob_data)
{
	GDBusProxy *proxy;
	GVariant *reply;
	GError *err = NULL;
	char *adapter_path;
	GDBusConnection *conn;
	GVariant *hash192 = NULL;
	GVariant *randomizer192 = NULL;
	GVariant *hash256 = NULL;
	GVariant *randomizer256 = NULL;

	BT_CHECK_PARAMETER(local_oob_data, return);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_path = _bt_get_adapter_path();
	retv_if(adapter_path == NULL, BLUETOOTH_ERROR_INTERNAL);

	proxy =  g_dbus_proxy_new_sync(conn,
			G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_BLUEZ_NAME, adapter_path,
			BT_OOB_INTERFACE, NULL, &err);
	g_free(adapter_path);
	if (!proxy) {
		BT_ERR("Unable to create proxy");
		if (err) {
			BT_ERR("Error: %s", err->message);
			g_clear_error(&err);
		}
		return BLUETOOTH_ERROR_INTERNAL;
	}

	reply = g_dbus_proxy_call_sync(proxy, "ReadLocalData",
			NULL,
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &err);
	g_object_unref(proxy);

	if (reply == NULL) {
		BT_ERR("ReadLocalData dBUS-RPC is failed");
		if (err != NULL) {
			BT_ERR("D-Bus API failure: errCode[%x], message[%s]",
					err->code, err->message);
			g_clear_error(&err);
		}
		return BLUETOOTH_ERROR_INTERNAL;
	}

	g_variant_get(reply, "(@ay@ay@ay@ay)", &hash192, &randomizer192,
			&hash256, &randomizer256);
	g_variant_unref(reply);

	if (hash192) {
		local_oob_data->hash_len = __get_oob_data_from_variant(hash192,
				local_oob_data->hash,
				sizeof(local_oob_data->hash));
		g_variant_unref(hash192);
	}

	if (randomizer192) {
		local_oob_data->randomizer_len = __get_oob_data_from_variant(randomizer192,
				local_oob_data->randomizer,
				sizeof(local_oob_data->randomizer));
		g_variant_unref(randomizer192);
	}

	if (hash256) {
		local_oob_data->hash256_len = __get_oob_data_from_variant(hash256,
				local_oob_data->hash256,
				sizeof(local_oob_data->hash256));
		g_variant_unref(hash256);
	}

	if (randomizer256) {
		local_oob_data->randomizer256_len = __get_oob_data_from_variant(randomizer256,
				local_oob_data->randomizer256,
				sizeof(local_oob_data->randomizer256));
		g_variant_unref(randomizer256);
	}

	return BLUETOOTH_ERROR_NONE;
}

int _bt_oob_add_remote_data(
			bluetooth_device_address_t *remote_device_address,
			bt_oob_data_t *remote_oob_data)
{
	int i;
	GDBusProxy *proxy;
	GVariant *reply;
	GError *err = NULL;
	char *dev_addr;
	char *adapter_path;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	unsigned char *remote_hash;
	unsigned char *remote_randomizer;
	GDBusConnection *conn;
	GVariantBuilder *builder;
	GVariantBuilder *builder2;
	GVariant *hash192;
	GVariant *randomizer192;
	GVariant *hash256;
	GVariant *randomizer256;

	BT_CHECK_PARAMETER(remote_device_address, return);
	BT_CHECK_PARAMETER(remote_oob_data, return);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_path = _bt_get_adapter_path();
	retv_if(adapter_path == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_type_to_string(address, remote_device_address->addr);

	proxy =  g_dbus_proxy_new_sync(conn,
			G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_BLUEZ_NAME, adapter_path,
			BT_OOB_INTERFACE, NULL, &err);
	g_free(adapter_path);
	if (!proxy) {
		BT_ERR("Unable to create proxy");
		if (err) {
			BT_ERR("Error: %s", err->message);
			g_clear_error(&err);
		}
		return BLUETOOTH_ERROR_INTERNAL;
	}

	/* Remote address */
	dev_addr = g_strdup(address);

	/* P-192 */
	remote_hash = remote_oob_data->hash;
	remote_randomizer = remote_oob_data->randomizer;

	builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	builder2 = g_variant_builder_new(G_VARIANT_TYPE("ay"));

	for (i = 0; i < remote_oob_data->hash_len; i++)
		g_variant_builder_add(builder, "y", remote_hash[i]);

	for (i = 0; i < remote_oob_data->randomizer_len; i++)
		g_variant_builder_add(builder2, "y", remote_randomizer[i]);

	hash192 = g_variant_new("ay", builder);
	randomizer192 = g_variant_new("ay", builder2);

	g_variant_builder_unref(builder2);
	g_variant_builder_unref(builder);

	/* P-256 */
	remote_hash = remote_oob_data->hash256;
	remote_randomizer = remote_oob_data->randomizer256;

	builder = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	builder2 = g_variant_builder_new(G_VARIANT_TYPE("ay"));

	for (i = 0; i < remote_oob_data->hash256_len; i++)
		g_variant_builder_add(builder, "y", remote_hash[i]);

	for (i = 0; i < remote_oob_data->randomizer256_len; i++)
		g_variant_builder_add(builder2, "y", remote_randomizer[i]);

	hash256 = g_variant_new("ay", builder);
	randomizer256 = g_variant_new("ay", builder2);

	g_variant_builder_unref(builder2);
	g_variant_builder_unref(builder);

	/* Call AddRemoteData Method */
	reply = g_dbus_proxy_call_sync(proxy, "AddRemoteData",
			g_variant_new("(s@ay@ay@ay@ay)", dev_addr,
				hash192, randomizer192, hash256, randomizer256),
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &err);
	g_object_unref(proxy);
	g_free(dev_addr);

	/* Check the reply */
	if (reply == NULL) {
		BT_ERR("AddRemoteData dBUS-RPC is failed");
		if (err != NULL) {
			BT_ERR("D-Bus API failure: errCode[%x], message[%s]",
					err->code, err->message);
			g_clear_error(&err);
		}
		return BLUETOOTH_ERROR_INTERNAL;
	}
	g_variant_unref(reply);

	return BLUETOOTH_ERROR_NONE;
}

int _bt_oob_remove_remote_data(
			bluetooth_device_address_t *remote_device_address)
{
	GDBusProxy *proxy;
	GVariant *reply;
	GError *err = NULL;
	char *dev_addr;
	char *adapter_path;
	char address[BT_ADDRESS_STRING_SIZE] = { 0 };
	GDBusConnection *conn;

	BT_CHECK_PARAMETER(remote_device_address, return);

	conn = _bt_gdbus_get_system_gconn();
	retv_if(conn == NULL, BLUETOOTH_ERROR_INTERNAL);

	adapter_path = _bt_get_adapter_path();
	retv_if(adapter_path == NULL, BLUETOOTH_ERROR_INTERNAL);

	_bt_convert_addr_type_to_string(address,
		remote_device_address->addr);

	proxy =  g_dbus_proxy_new_sync(conn,
			G_DBUS_PROXY_FLAGS_NONE, NULL,
			BT_BLUEZ_NAME, adapter_path,
			BT_OOB_INTERFACE, NULL, &err);
	g_free(adapter_path);
	if (!proxy) {
		BT_ERR("Unable to create proxy");
		if (err) {
			BT_ERR("Error: %s", err->message);
			g_clear_error(&err);
		}
		return BLUETOOTH_ERROR_INTERNAL;
	}

	dev_addr = g_strdup(address);

	/* Call RemoveRemoteData Method*/
	reply = g_dbus_proxy_call_sync(proxy, "RemoveRemoteData",
			g_variant_new("(s)", dev_addr),
			G_DBUS_CALL_FLAGS_NONE, -1,
			NULL, &err);
	g_object_unref(proxy);
	g_free(dev_addr);

	/* Check the reply*/
	if (reply == NULL) {
		BT_ERR("RemoveRemoteData dBUS-RPC is failed");
		if (err != NULL) {
			BT_ERR("D-Bus API failure: errCode[%x], message[%s]",
					err->code, err->message);
			g_clear_error(&err);
		}
		return BLUETOOTH_ERROR_INTERNAL;
	}

	g_variant_unref(reply);
	return BLUETOOTH_ERROR_NONE;
}
